// AfhEmitX86.c -- X86 code emitters.
//
//  Copyright 1997-2014 AFH Systems.
//
//  FilterMeister is free software. You can redistribute and modify it
//  under the terms of the MPL2 License.
//  See: https://www.mozilla.org/en-US/MPL/2.0/
//  Any changes to the FilterMeister files have to be published or sent
//  to harry@thepluginsite.com. The filter plugin code itself does not
//  need to be disclosed.
//
//	10JUL2013	A. Hunter	1. Add rip-relative addressing mode.
//							2. Fix some /Wp64 warnings.
//  17JUL2013   A. Hunter	1. Change CPUID code for Win64.
//							2. Try using IsProcessorFeaturePresent instead?
//	23JUL2013	A. Hunter	1. Modified exception handling code for Win64.
//  16AUG2013	A. Hunter	1. Added support for sse2 ops (CVTTSS2SI, etc.)
//  13SEP2013   A. Hunter	1. Treat FWAIT (9Bh) as an opcode prefix, so other prefixes
//							   such as FS: will follow, not precede, the FWAIT.
//	18Sep2013	A. Hunter	1. Add the cbLocals parameter to the X86_emit_*_epilog()
//							   routines.
//	20Sep2013	A. Hunter	1. Begin split-off of delayed prolog, and merger of
//							   function/handler prologs and epilogs.
//	30Sep2013	A. Hunter	1. Make func struct local to CB.
//	18Jan2014	A. Hunter	1. Fix X86_append_cb to update appendee cb with data
//							   from the appender cb.
//	20Jan2014	A. Hunter	1. Don't display code in the exception handler if not
//							   at a valid code address, to avoid recursive access
//							   violations.
//	20Jan2014	A. Hunter	1. Move formatted print routine macros to AfhFM.h.
//							2. Change format "+%d" to "%+d" to handle signed values.
//							3. Emit pseudocode for X86_emit_prolog, X86_emit_delayed_prolog,
//							   X86_emit_epilog, X86_emit_handler_prolog, 
//							   X86_emit_handler_epilog, X86_emit_DIVREM_TT,
//							   X86_emit_divCheck, X86_emit_defaultHandler.
//	22Jan2014	A. Hunter	1. Emit pseudocode for X86_emit_doOneRow_part1 and part2.
//						Later do:
//								X86_emit_compute_iuv, X86_emit_compute_d, X86_emit_compute_m
//	23Jan2014	A. Hunter	1. Fix bug in X86_emit_compute_d: truncate result to int32 using
//							   in-line SSE instruction cvttsd2si.
//	26Jan2014	A. Hunter	1. Fix X64 bug in X86_emit_op_Tm.
//	31Jan2014	A. Hunter   1. Abort X86_call_fmf1 if invalid code in buffer.
//  24Feb2014	A. Hunter	1. Expand some error messages using X86_abort2().
//	21Mar2014	A. Hunter	1. Add diagram of the new FM32/64 ABI stack frame layout.
//	27Mar2014	A. Hunter	1. For now, we *always* align the X64 stack to 16 bytes in the
//							   prolog; later, we will do this only for non-LEAF procedures.
//							   (Note that it is not sufficient to align the stack only
//							   if a static arglist is allocated, since calls to intrinsic
//							   functions do not currently allocate a static arglist -- though
//							   they *should* if the call is at level-1.)
//	26May2014	A. Hunter	1. Honor the XL macro in X86_emit_prefixes only when reg is SPL,
//							   BPL, SIL, or DIL. (The XL macro forces emission of a bare REX
//							   byte, which causes byte registers 4, 5, 6, and 7 to be interpreted
//							   as the X64 registers SPL, BPL, SIL, and DIL instead of AH, CH,
//							   DH, and BH, respectively.)
//	28May2014	A. Hunter	1. Add X86_emit_op_TiR.
//							2. Change X86_emit_op_MiR to use edx instead of eax.
//	26Jun2014	A. Hunter	1. Align cbLocals to a multiple of BPP (fixes a bug in X32 caused
//							   by a misaligned stack).
//
// TODO:
// - Optimize MOV reg,0 to XOR reg,reg.

#include <assert.h>
#include <stddef.h>     //offsetof
#include <stdio.h>
#include <stdlib.h>
#include "plmcvax.h"
#include "mda.h"
#include "AfhFM.h"

// We need both 32-bit and 64-bit versions of the FM context record...
#undef TARGET
#define TARGET 64
#include "AfhFMcontext.h"               /* 64-bit FM context record.          */
#undef TARGET
#define TARGET 32
#include "AfhFMcontext.h"               /* 32-bit FM context record.          */

#include "EmitX86.h"

#ifdef _WIN64
#include <intrin.h>		//for __cpuid
#endif

// Minimum valid executable address:
#define MIN_CODE_ADDRESS 0x10000U

#ifdef NDEBUG
#define X86_ASSERT(exp) (void)( (exp) || (X86_abort("Assertion failed:\r\n" #exp),0) )
#else
#define X86_ASSERT(exp) assert(exp)
#endif

#ifdef FML
#define PAUSE
#else
#define PAUSE (printf("pause..."),(getchar)())
#endif

#ifndef STANDALONE
static const char *reg64_name[] =
    { "RAX", "RCX", "RDX", "RBX", "RSP", "RBP", "RSI", "RDI",
      "R8",  "R9",  "R10", "R11", "R12", "R13", "R14", "R15", "RIP"};

static const char *reg32_name[] =
    { "EAX", "ECX", "EDX",  "EBX",  "ESP",  "EBP",  "ESI",  "EDI",
      "R8D", "R9D", "R10D", "R11D", "R12D", "R13D", "R14D", "R15D", "EIP"};

static const char *reg16_name[] =
    { "AX",  "CX",  "DX",   "BX",   "SP",   "BP",   "SI",   "DI",
      "R8W", "R9W", "R10W", "R11W", "R12W", "R13W", "R14W", "R15W", "IP"};

//non-orthogonal High/Low 8-bit register names (x86 or x64 noREX mode)
static const char *reg8H_name[] =
    { "AL",  "CL",  "DL",   "BL",   "AH",   "CH",   "DH",   "BH",
      "R8B", "R9B", "R10B", "R11B", "R12B", "R13B", "R14B", "R15B"};

//x64 orthogonal (Low byte only) 8-bit register names (X64 REX mode)
static const char *reg8L_name[] =
    { "AL",  "CL",  "DL",   "BL",   "SPL",  "BPL",  "SIL",  "DIL",
      "R8B", "R9B", "R10B", "R11B", "R12B", "R13B", "R14B", "R15B"};

static const char *mmxreg_name[] =
    { "MMX0", "MMX1", "MMX2", "MMX3", "MMX4", "MMX5", "MMX6", "MMX7" };

static const char *xmmreg_name[] =
    { "XMM0", "XMM1", "XMM2",  "XMM3",  "XMM4",  "XMM5",  "XMM6",  "XMM7",
	  "XMM8", "XMM9", "XMM10", "XMM11", "XMM12", "XMM13", "XMM14", "XMM15"};

static const char *ymmreg_name[] =
    { "YMM0", "YMM1", "YMM2",  "YMM3",  "YMM4",  "YMM5",  "YMM6",  "YMM7",
	  "YMM8", "YMM9", "YMM10", "YMM11", "YMM12", "YMM13", "YMM14", "YMM15"};

static const char *stnreg_name[] =
	{ "ST(0)", "ST(1)", "ST(2)", "ST(3)", "ST(4)", "ST(5)", "ST(6)", "ST(7)"};


static BOOL fInvalid;
static BOOL fPrivileged;
static BOOL fFpuCode;
static BOOL fIo;
static BOOL fTrap;
static BOOL fFarCode;
static BOOL fStringCode;
static BOOL fOpSizeOverride;
static BOOL fAddrSizeOverride;

#endif //ifndef STANDALONE

static BOOL cpuidDetected = FALSE;
static int cpuid_version;
static int cpuid_features;
static BOOL mmx_available;
static BOOL cmov_available;
static char vendor_id[13];

#define PLURAL(n) ((n)==1?"":"s")

#if 0 //moved to AfhFM.h
#define PRINTF(f) X86_print(f)

#define PRINTF2(f,a) \
    do {\
        char buff[128];\
        sprintf(buff, f, a);\
        X86_print(buff);\
    } while(0)

#define PRINTF3(f,a,b) \
    do {\
        char buff[128];\
        sprintf(buff, f, a, b);\
        X86_print(buff);\
    } while(0)

#define PRINTF4(f,a,b,c) \
    do {\
        char buff[128];\
        sprintf(buff, f, a, b, c);\
        X86_print(buff);\
    } while(0)
#endif

static int mmxsupport(void)
{
#ifndef MSVC_SUPPORTS_CPUID
    //BUG!!! It seems MSVC doesn't realize that CPUID clobbers ebx,
    //so ebx doesn't get preserved across calls to mmxsupport.
    //This causes problems when the calling code is saving a var
    //(e.g., hDlg) in ebx across calls to X86_MMX_available!!!
    //So we save ebx ourselves in the following code (or maybe
    //just make some visible change to ebx to clue MSVC in to
    //the fact that ebx needs to be preserved???)...
#endif

#ifdef _WIN64
// _asm is not supported in x64 MSVC, so we 
// use the __cpuid() intrinsic function:
//
//	void __cpuid(
//   int CPUInfo[4],
//   int InfoType
//  );

	int CPUInfo[4]; //receives eax, ebx, ecx, edx

#endif //_WIN64

	int mmx_supported = 0;

    cpuid_version = 0;
    cpuid_features = 0;
	memset(vendor_id, 0, sizeof(vendor_id));

#ifdef _WIN64
	// If we are running in x64 mode, we can safely
	// assume that the CPUID instruction is implemented;
	// no need to test for it.
	
	//set eax=0 for Vendor ID.
	__cpuid(CPUInfo, 0);
	if (CPUInfo[0] > 0) {
		//make sure eax returned a non-zero value
		*(int *)&vendor_id[0] = CPUInfo[1];	//ebx = chars 0-3
		*(int *)&vendor_id[4] = CPUInfo[3]; //edx = chars 4-7
		*(int *)&vendor_id[8] = CPUInfo[2];	//ecx = chars 8-11
		vendor_id[12] = 0;					//ensure NUL-terminated

		//set eax=1 for features request
		__cpuid(CPUInfo, 1);
		cpuid_version  = CPUInfo[0];	//eax returns version info
		cpuid_features = CPUInfo[3];	//edx returns feature flags

		//Test for MMX supported
		if (cpuid_features & 0x00800000) {
			// MMX is supported.
			mmx_supported = 1;
		}
	}

#else
	_asm {

#ifndef MSVC_SUPPORTS_CPUID
        xor ebx, ebx            //Make a dummy change to ebx
                                //to force MSVC to preserve it
                                //(see note above)
#endif

		pushfd					//Save Eflag to stack
		pop eax					//Get Eflag from stack into eax
		mov ecx, eax			//Make another copy of Eflag in ecx
		xor eax, 0x200000		//Toggle ID bit in Eflag [i.e. bit(21)] 
		push eax				//Save modified Eflag back to stack

		popfd					//Restore modified value back to Eflag reg 
		pushfd					//Save Eflag to stack
		pop eax					//Get Eflag from stack
		xor eax, ecx			//Compare the new Eflag with the original Eflag
		jz NOT_SUPPORTED		//If the same, CPUID instruction is not supported,
								//skip following instructions and jump to
								//NOT_SUPPORTED label

		xor eax, eax			//Set eax to zero
					
#ifndef MSVC_SUPPORTS_CPUID
		_asm _emit 0x0f			//CPUID instruction  (two byte opcode)
		_asm _emit 0xa2 
#else
        cpuid                   //CPUID instruction
#endif

        mov dword ptr vendor_id, ebx      //Save vendor ID string
        mov dword ptr vendor_id+4, edx
        mov dword ptr vendor_id+8, ecx

		cmp eax, 1				//make sure eax returned non-zero value
		jl NOT_SUPPORTED		//If eax is zero, mmx not supported

		xor eax, eax			//set eax to zero
		inc eax					//Now increment eax to 1.  This instruction is 
								//faster than the instruction "mov eax, 1"
		
#ifndef MSVC_SUPPORTS_CPUID
		_asm _emit 0x0f			//CPUID instruction
		_asm _emit 0xa2 
#else
        cpuid                   //CPUID instruction
#endif

        mov cpuid_version, eax  //save version info
        mov cpuid_features, edx //save feature flags

		and edx, 0x00800000		//mask out all bits but mmx bit(24)
		cmp edx, 0				// 0 = mmx not supported
		jz	NOT_SUPPORTED		// non-zero = Yes, mmx IS supported

		mov	mmx_supported, 1	//set return value to 1

NOT_SUPPORTED:
		mov	eax, mmx_supported	//move return value to eax	

	}//_asm
#endif //!_WIN64

	return mmx_supported;		
} /*mmxsupport*/

BOOL X86_MMX_available(void)
{
    mmxsupport();
    mmx_available  = (cpuid_features & 0x00800000) != 0;
    cmov_available = (cpuid_features & 0x00008000) != 0;
    return mmx_available;
} /*X86_MMX_available*/

BOOL X86_CMOV_available(void)
{
    mmxsupport();
    mmx_available  = (cpuid_features & 0x00800000) != 0;
    cmov_available = (cpuid_features & 0x00008000) != 0;
    return cmov_available;
} /*X86_CMOV_available*/

#ifndef STANDALONE

int X86_cpuid_version(void)
{
    mmxsupport();
    return cpuid_version;
} /*X86_cpuid_version*/

int X86_cpuid_features(void)
{
    mmxsupport();
    return cpuid_features;
} /*X86_cpuid_features*/

char *X86_cpuid_vendor_id(void)
{
    mmxsupport();
    return vendor_id;
} /*X86_cpuid_vendor_id*/

const char *X86_reg64_name(int reg)
{
    return reg64_name[reg & 15];
} /*X86_reg64_name*/

const char *X86_reg32_name(int reg)
{
    return reg32_name[reg & 15];
} /*X86_reg32_name*/

const char *X86_reg16_name(int reg)
{
    return reg16_name[reg & 15];
} /*X86_reg16_name*/

const char *X86_reg8H_name(int reg)
{
    return reg8H_name[reg & 15];
} /*X86_reg8H_name*/

const char *X86_reg8L_name(int reg)
{
    return reg8L_name[reg & 15];
} /*X86_reg8L_name*/

const char *X86_regv_name(int reg)
{
	if (X64gen) return X86_reg64_name(reg);	//X64
	if (X86gen) return X86_reg32_name(reg);	//X86
	return X86_reg16_name(reg);				//X16
} /*X86_regv_name*/

const char *X86_mmxreg_name(int reg)
{
    return mmxreg_name[reg & 7];
} /*X86_mmreg_name*/

const char *X86_xmmreg_name(int reg)
{
    return xmmreg_name[reg & 15];
} /*X86_xmmreg_name*/

const char *X86_ymmreg_name(int reg)
{
    return ymmreg_name[reg & 15];
} /*X86_ymmreg_name*/

const char *X86_stnreg_name(int reg)
{
	return stnreg_name[reg & 7];
} /*X86_stnreg_name*/

//////////////////////////
// Code Buffer Routines //
//////////////////////////

// Initialize a Code Buffer, assigning it a byte buffer and
// a display name. Initialize the array of CGLs for this
// CB to null.  Make this CB the current one, and return
// the address of the previous CB (if any).
CB* X86_emit_init(CB *this_cbp, BYTE cb[], int cb_size, const char *displayName)
{
    CB *prev_cbp = cbp;
    cbp = this_cbp;
    cbp->displayName = displayName;
#ifdef _WIN64
    cbp->pcStart = cb;			 //address of first code buffer byte
	cbp->pc = 0;				 //set pc to start of buffer
    cbp->pc_max = cb_size;		 //index of last buffer byte + 1
#else //Win32
    cbp->pc = cbp->pcStart = cb; //address of first buffer byte
    cbp->pc_max = &cb[cb_size];	 //address of last buffer byte + 1
#endif
    cbp->next_cgl = 0;
    cbp->max_cgl = 0;
    free(cbp->cgls);    //free any previous cgl info
    cbp->cgls = NULL;
	//Initialize various code buffer fields, in case of straight-line code
	//with no initializing prolog, or in case of appending code buffers.
	cbp->tosmax = -1;
	cbp->func.nTemps = 0;				//not yet known
	cbp->func.cbTemps = 0;				//not yet known
	cbp->func.nTempRegsUsed = 0;		//NOTE nTmps is max # temp regs; nTemps is # temps
	cbp->func.cbLocals = 0;				//none so far
	cbp->func.nArgs = 0;				//none so far
	cbp->func.cbStaticArglistSize = -1;	//none so far
	cbp->func.nargsNeeded = false;		//not so far as we know at this point...
	cbp->func.needFrameReg = true;		//so far as we know...
	cbp->func.homeArgsMask = 0;			//not yet known
	cbp->func.argTypeMask = 0;			//not yet known
	cbp->func.fpuUsed = false;			//not so far as we know yet...
	cbp->func.loadFMCP = false;			//not yet known
    return prev_cbp;
} /*X86_emit_init*/

void X86_record_tos(INT4 tos)
{
	if (tos > cbp->tosmax) {
		// record max value of tos
		cbp->tosmax = tos;
	}
} /*X86_record_tos*/


#if 0 //not used at present
// Special initializations for a code buffer intended to receive
// a function/handler prolog.
void X86_init_prolog_CB(CB *this_cbp, int max_prolog_size)
{
	int i;

	//must be at start of CB
	assert(this_cbp->pc == 0);

	//fill buffer prefix with INT3 opcode bytes.
	for (i = 0; i < max_prolog_size; ++i) {
		X86_emit_byte(INT3);
	}
} /*X86_init_prolog_CB*/
#endif

// Rotate the bytes in code buffer CB, swapping the initial (main body)
// code (of length middle) with the trailing (prolog) code.  Also fixes
// up any defined CGL labels in the CB.
static void X86_rotate_CB(CB *this_cbp, int middle)
{
	BYTE *cb = this_cbp->pcStart;
    int j;
	int first = 0;
	int mid = middle;
#ifdef _WIN64
	const int last = this_cbp->pc;
#else //Win32
	const int last = this_cbp->pc - this_cbp->pcStart;
#endif //Win32
	const int np = last - middle;	//length of prolog
	int next = mid;
	while (first != next) {
		int temp = cb[first];
		cb[first++] = cb[next];
		cb[next++] = temp;
		if (next == last) next = mid;
		else if (first == mid) mid = next;
	}

	// Now fix up any CGL labels defined in this CB
    for (j = 0; j < this_cbp->next_cgl; ++j) {
        if (this_cbp->cgls[j].defined &&
		   !this_cbp->cgls[j].isAbsVal) {
#ifdef _WIN64
            if (this_cbp->cgls[j].addr >= (ADDR)middle)
#else //Win32
            if (this_cbp->cgls[j].addr >= &cb[middle])
#endif //Win32
			{	//CGL is in the prolog...
				this_cbp->cgls[j].addr -= middle;	//shift left by size of main code
			}
			else {
				//CGL is in main body...
				this_cbp->cgls[j].addr += np;	//shift right by size of prolog
			}
        }//if defined label
    }//for j
} /*X86_rotate_CB*/

// Switch to another Code Buffer for code emission.
// Returns address of previous Code Buffer (if any).
CB* X86_switch_to(CB *this_cbp)
{
    CB *prev_cbp = cbp;
    if (this_cbp) cbp = this_cbp;
    return prev_cbp;
} /*X86_switch_to*/

// Finish emitting to the current Code Buffer.
// Returns number of bytes in buffer.
int X86_emit_finish(void)
{
    int i;
    for (i = 0; i < cbp->next_cgl; i++) {
        //check for undefined labels
        if (!cbp->cgls[i].defined) {
            PRINTF2("X86 EMIT: Warning -- cgl %d is undefined.\r\n", i);
        }
    }
    /* Don't free the cgls here ... disassemble needs them! */ //SLOW MEMORY LEAK?????
    //free(cbp->cgls);
    //cbp->cgls = NULL;
    //cbp->max_cgl = 0;
#ifdef _WIN64
    return cbp->pc;	//number of bytes emitted
#else //Win32
    return (int)(cbp->pc - cbp->pcStart);
#endif
} /*X86_emit_finish*/

// Free (remaining) resources for a Code Buffer
void X86_emit_free_resources(CB *this_cbp)
{
    // Free up resources that couldn't be freed up
    // by X86_emit_finish(), such as CGL's.
    free(this_cbp->cgls);
    this_cbp->cgls = NULL;
    this_cbp->max_cgl = 0;
    this_cbp->displayName = NULL;   //ptr might not survive re-invocation.
} /*X86_emit_free_resources*/

static int lines = 0;

#ifndef BETA

// Check to see if a CGL labels the specified set of 
// instruction bytes, and print a CGL label if so.
static void check_for_cgl(CB *this_cbp, int i, int nb)
{
    ///// Doesn't work for labels defined within an appended code buffer!!!
    ///// Maybe should take this out altogether???
    // Check for CGL at (or within) the instruction at offset i,
    // byte length nb, within code buffer this_cbp.  If so,
    // print it.
    BYTE *cb = this_cbp->pcStart;
    int j;

    for (j = 0; j < this_cbp->next_cgl; j++) {
        if (this_cbp->cgls[j].defined &&
            !this_cbp->cgls[j].isAbsVal &&
#ifdef _WIN64
            this_cbp->cgls[j].addr >= (ADDR)i &&
            this_cbp->cgls[j].addr < (ADDR)(i + nb)) {
            if (this_cbp->cgls[j].addr == i) {
#else
            this_cbp->cgls[j].addr >= &cb[i] &&
            this_cbp->cgls[j].addr < &cb[i]+nb) {
            if (this_cbp->cgls[j].addr == &cb[i]) {
#endif
                //cgl at start of instruction...
                PRINTF2("cgl%d:\r\n", j);
            }
            else {
                //cgl inside instruction...very strange!!!
#ifdef _WIN64
                PRINTF3("cgl%d = $%+d; INVALID LABEL POSITION\r\n", j,
                       this_cbp->cgls[j].addr);
#else
                PRINTF3("cgl%d = $%+d; INVALID LABEL POSITION\r\n", j,
                       this_cbp->cgls[j].addr - &cb[i]);
#endif
                fInvalid = TRUE;
            }
            ++lines;
        }
    }//for j
} /*check_for_cgl*/

// Disassemble the code in this Code Buffer
void X86_disassemble(CB *this_cbp)
{
    struct cpustate_struc state;
    int i, j, k;
    char oucs[128];
    UINT2 nb, nc;
    UINT4 flags;
    UINT4 iss;
    const BYTE *cb = this_cbp->pcStart;
#ifdef _WIN64
    const int nBytes = this_cbp->pc;
#else //Win32
    const int nBytes = (int)(this_cbp->pc - cb);
#endif
    int nBytesLeft;

    lines = 0;

    fInvalid = FALSE;
    fPrivileged = FALSE;
    fFpuCode = FALSE;
    fIo = FALSE;
    fTrap = FALSE;
    fFarCode = FALSE;
    fStringCode = FALSE;
    fOpSizeOverride = FALSE;
    fAddrSizeOverride = FALSE;

    for (i = 0; i < nBytes; i += nb) {
        nBytesLeft = nBytes - i;
#ifdef FML
		//Check for a switch-statement's jump table...
        if (nBytesLeft >= 4 &&
            cb[i] == 0x0f && cb[i+1] == 0x0b /*UD2*/) {
            // FM uses UD2 to mark the start of a table of dwords,
			// prefaced by a word containing the number of such
			// dwords to follow.
            int nDWords = (cb[i+3] << 8) + cb[i+2];	//little-endian word

            check_for_cgl(this_cbp, i, 2);	//possible jump-table label?
            PRINTF2("%8.8p: ", &cb[i]);
            for (j = 0; j < 2; j++) {
                PRINTF2("%2.2x", cb[i+j]);	//code bytes for the UD2 instruction
            }
            PRINTF("\r\n  UD2 ;dword table\r\n");

            check_for_cgl(this_cbp, i+2, 2);	//possible label?
            PRINTF2("%8.8p: ", &cb[i+2]);
            for (j = 0; j < 2; j++) {
                PRINTF2("%2.2x", cb[i+2+j]);	//2 data bytes for count word
            }
            PRINTF2("\r\n  DW %d ;number of dwords\r\n", nDWords);

            for (j = 4; j < nDWords*4 + 4; j += 4) {
                check_for_cgl(this_cbp, i+j, 4);	//possible label?
                if (j+3 >= nBytesLeft) X86_abort("Jump table runs past end of code buffer");
                PRINTF2("%8.8p: ", &cb[i+j]);
                for (k = 0; k < 4; k++) {
                    PRINTF2("%2.2x", cb[i+j+k]);	//4 data bytes per dword
                }
                PRINTF2("\r\n  DD %d\r\n", *(int *)&cb[i+j]);
            }

            nb = (nDWords << 2) + 4;	//number bytes in entire table
        }
        else {
#endif
            // Disassemble an ordinary instruction...
			UINT4 iopts = 8; //Use32
			if (X64GEN) iopts = 16; //Use64
            iss = Mda_disassemble(&cb[i], nBytesLeft, 16, iopts,
                                  &state, oucs, sizeof(oucs), &nb, &nc, &flags);
            if (!(iss & 1)) {
                fInvalid = TRUE;
                X86_abort2("Mda_disassemble failed (iss = 0x8.8x)", iss);
            }
            //check for CGL at (or within) this instruction:
            check_for_cgl(this_cbp, i, nb);
            ///PRINTF2("%8.8Ix: ", &cb[i]);	//prints 8 or more digits
            PRINTF2("%8.8p: ", &cb[i]);	//X64 always prints 16 digits for now
            for (j = 0; j < nb; j++) {
                PRINTF2("%2.2x", cb[i+j]);
            }
#ifdef FML
            PRINTF3("\r\n  %.*s", nc, oucs);
#else
            PRINTF3("\t%.*s", nc, oucs);
#endif
#if 1
            /* Output info from flags while debugging... */
            if (flags != 0) {
                if (flags & 0x00000001) {
                    PRINTF("; INVALID");
                    fInvalid = TRUE;
                }
                if (flags & 0x00000002) {
					// disable interrupts
                    PRINTF("; privileged");
                    fPrivileged = TRUE;
                }
                if (flags & 0x00000004) {
                    PRINTF("; seg override");
                    fFarCode = TRUE;
                }
#ifndef X64	//REP* may be sse prefix
                if (flags & 0x00000008) {
                    PRINTF("; repeat");
                    fStringCode = TRUE;
                }
#endif
                if (flags & 0x00000010) {
					// lock -- is it privileged???
                    PRINTF("; privileged");
                    fPrivileged = TRUE;
                }
                if (flags & 0x00000020) {
                    PRINTF("; inhibits single-step trap");
                }
                if (flags & 0x00000040) {
                    //PRINTF("; FPU instruction");
                    fFpuCode = TRUE;
                }
                if (flags & 0x00000080) {
                    //PRINTF("; wait");
                    fFpuCode = TRUE;
                }
                if (flags & 0x00000100) {
					// halt
                    PRINTF("; privileged");
                    fPrivileged = TRUE;
                }
                if (flags & 0x00000200) {
                    PRINTF("; privileged i/o");
                    fPrivileged = TRUE;
                    fIo = TRUE;
                }
                if (flags & 0x00000400) {
                    //PRINTF("; call");
                }
                if (flags & 0x00000800) {
                    //PRINTF("; return");
                }
                if (flags & 0x00001000) {
                    PRINTF("; interrupt 3");
                    fTrap = TRUE;
                }
                if (flags & 0x00002000) {
                    PRINTF("; interrupt n");
                    fTrap = TRUE;
                }
                if (flags & 0x00004000) {
                    PRINTF("; INTO/BOUND");
                    fTrap = TRUE;
                }
                if (flags & 0x00008000) {
					// interrupt return
                    PRINTF("; privileged");
                    fPrivileged = TRUE;
                }
                if (flags & 0x00010000) {
                    //PRINTF("; unconditional jump");
                }
                if (flags & 0x00020000) {
                    //PRINTF("; conditional jump");
                }
                if (flags & 0x00040000) {
                    PRINTF("; far");
                    fFarCode = TRUE;
                }
                if (flags & 0x00080000) {
					// pop flags -- privileged???
                    PRINTF("; privileged?");
                    fPrivileged = TRUE; //well,...
                }
#ifndef X64
                if (flags & 0x00100000) {
                    PRINTF("; operand size override");
                    fOpSizeOverride = TRUE;
                }
                if (flags & 0x00200000) {
                    PRINTF("; address size override");
                    fAddrSizeOverride = TRUE;
                }
#endif //X64
				if (flags & 0x00400000) {
					PRINTF("; redundant prefixes");
					fInvalid = TRUE;
				}
				if (flags & 0x00800000) {
					PRINTF("; misplaced REX prefix");
					fInvalid = TRUE;
				}
                if (flags & 0x0f000000) {
                    PRINTF("; unknown flags (MBZ)");
                    fInvalid = TRUE;
                }
				if (flags & 0xf0000000) {
					int prefixBytes = (flags >> 28) & 0xf;
					if (prefixBytes == 15) {
						PRINTF("; more than 14 prefix bytes!!");
						fInvalid = TRUE;
					}
#ifndef X64
					else {
						PRINTF3("; %u prefix byte%s", prefixBytes, PLURAL(prefixBytes));
					}
#endif //X64
				}
            }//if flags != 0
#endif //print flags
            PRINTF("\r\n");
#ifndef FML
            ++lines;
            if (lines > 40) {
                PAUSE;
                lines = 0;
            }
#endif
#ifdef FML
        }//if !UD2
#endif
    } /*for i*/
    //check for CGL at very end of buffer (or even up to 15
    //bytes past the end of the buffer):
    check_for_cgl(this_cbp, nBytes, 16);
    PAUSE;
    if (fPrivileged)        X86_warn("Privileged code was generated");
#if 0
    if (fFpuCode)           X86_warn("Floating point code was generated");
#endif
    if (fIo)                X86_warn("I/O code was generated");
    if (fTrap)              X86_warn("Trap code was generated");
    if (fFarCode)           X86_warn("FAR (non-FLAT) code was generated");
    if (fStringCode)        X86_warn("String code was generated");
    if (fOpSizeOverride)    X86_warn("16-bit operand code was generated");
    if (fAddrSizeOverride)  X86_warn("16-bit addressing was generated!!!");
    if (fInvalid)           X86_error("Invalid code was generated");
} /*X86_disassemble*/

#endif //ifndef BETA

void X86_emit_overflow(void)
{
    char msgBuffer[256];
    strcpy(msgBuffer, cbp->displayName ? cbp->displayName : "Unknown");
    strcat(msgBuffer, " code buffer overflow:\n\n"
           "Try to reduce the size of the source code.");
    X86_abort(msgBuffer);
} /*X86_emit_overflow*/

void X86_emit_byte(BYTE b)
{
    if (cbp->pc >= cbp->pc_max) X86_emit_overflow();
#ifdef _WIN64
    cbp->pcStart[cbp->pc++] = b;
#else //Win32
    *cbp->pc++ = b;
#endif
} /*X86_emit_byte*/

void X86_emit_word(WORD w)
{
    if (sizeof(WORD) > cbp->pc_max - cbp->pc) X86_emit_overflow();
#ifdef _WIN64
    *(WORD *)&cbp->pcStart[cbp->pc] = w;
#else //Win32
    *(WORD *)cbp->pc = w;
#endif
    cbp->pc += sizeof(WORD);
} /*X86_emit_word*/

void X86_emit_dword(DWORD dw)
{
    if (sizeof(DWORD) > cbp->pc_max - cbp->pc) X86_emit_overflow();
#ifdef _WIN64
    *(DWORD *)&cbp->pcStart[cbp->pc] = dw;
#else //Win32
    *(DWORD *)cbp->pc = dw;
#endif
    cbp->pc += sizeof(DWORD);
} /*X86_emit_dword*/

#ifdef X64
void X86_emit_qword(QWORD qw)
{
    if (sizeof(QWORD) > cbp->pc_max - cbp->pc) X86_emit_overflow();
#ifdef _WIN64
    *(QWORD *)&cbp->pcStart[cbp->pc] = qw;
#else //Win32
    *(QWORD *)cbp->pc = qw;
#endif
    cbp->pc += sizeof(QWORD);
} /*X86_emit_qword*/
#endif

void X86_emit_bytes(BYTE *cb, int nBytes)
{
#ifdef _WIN64
    if ((DWORD)nBytes > cbp->pc_max - cbp->pc) X86_emit_overflow();
    memmove(cbp->pcStart + cbp->pc, cb, nBytes);
#else //Win32
    if (nBytes > cbp->pc_max - cbp->pc) X86_emit_overflow();
    memmove(cbp->pc, cb, nBytes);
#endif
    cbp->pc += nBytes;
} /*X86_emit_bytes*/

void X86_append_cb(CB *that_cbp)
{
#ifdef _WIN64
    X86_emit_bytes(that_cbp->pcStart, that_cbp->pc);
#else //Win32
    X86_emit_bytes(that_cbp->pcStart, (int)(that_cbp->pc - that_cbp->pcStart));
#endif
	//merge function/handler info into current CB
	cbp->tosmax				= max(cbp->tosmax, that_cbp->tosmax);
	//cbp->next_cgl			= max(cbp->next_cgl, that_cbp->next_cgl);//no, next_cgl must be # used in cb!!
	cbp->func.nTemps		= max(cbp->func.nTemps, that_cbp->func.nTemps);
	cbp->func.cbTemps		= max(cbp->func.cbTemps, that_cbp->func.cbTemps);
	cbp->func.nTempRegsUsed	= max(cbp->func.nTempRegsUsed, that_cbp->func.nTempRegsUsed);
	cbp->func.cbLocals		= max(cbp->func.cbLocals, that_cbp->func.cbLocals);
	cbp->func.cbStaticArglistSize =
		max(cbp->func.cbStaticArglistSize, that_cbp->func.cbStaticArglistSize);
	///cbp->func.nArgs = max(cbp->func.nArgs, that_cbp->func.nArgs); NOT USED? min of negative???
	cbp->func.nargsNeeded	|= that_cbp->func.nargsNeeded;
	cbp->func.needFrameReg	|= that_cbp->func.needFrameReg;
	cbp->func.fpuUsed		|= that_cbp->func.fpuUsed;
	cbp->func.loadFMCP		|= that_cbp->func.loadFMCP;
	cbp->func.homeArgsMask	|= that_cbp->func.homeArgsMask;
	cbp->func.argTypeMask	|= that_cbp->func.argTypeMask;
/***
	int nTemps;					//number of temps used by this function
	int cbTemps;				//number of bytes to allocate for temps
	int nTempRegsUsed;			//number of temp registers used by this function (or a mask??)
	int nArgs;					//number of formal parameters (negated if VPL)
	int cbStaticArglistSize;	//number of bytes to allocate (-1 means none; minimum is 32 for X64)
	int nargsNeeded;			//true iff "nargs" is referenced by function
	int needFrameReg;			//true iff a stack frame register is needed
	int homeArgsMask;			//mask of args which must be homed (first 4 args only)
	int argTypeMask;			//mask of arg types; 0=integral, 1=floating point
	int fpuUsed;				//true only if 1 or more fpu intructions were emitted
	int loadFMCP;				//true iff the FM Context pointer needs to be loaded into rBX
	int cbLocals;				//number of bytes to allocate on stack for local vars
	CGL ref_cbLocals;			//fwd reference label to fixup cbLocals when known
/*****/
	//do we also need to keep track of cbLocals, tempResUsedMsk, etc.,
	//and merge them also????
	//should also merge the CGL info, but we'll save that for a rainy day...
} /*X86_append_cb*/

static void X86_emit_prefixes(int op, int reg, int base_reg, int index_reg)
{	//or make this a set of macros???
	//Output prefix bytes and/or a REX byte.
	//Note that legacy prefixes must follow a WAIT byte and precede the REX byte if present.
	//In general, we follow the ML64 assembler listing conventions,
	//and output the prefixes in the following order:
	//
	// WAIT(9B) REPE(F3/) REPNE(F2/) A32(67&) D16(66|) GS(65:) FS(64:) LOCK(F0/) REX(4X/)

	int REX_byte = 0;
	if (op & (WAIT_PB << 24)) {
		//FWAIT prefix for non-wait NDP op
		//output 9Bh fwait op before any other prefixes...
		X86_emit_byte(FWAIT);
	}
	// REP_E, REP_N, and LOCK are mutually exclusive
	if (op & (REP_E << 24)) {
		//REPE/REPZ/REP prefix (also used to extend SSE+ ops)
		//output 0F3h prefix byte
		X86_emit_byte(REPE);
	}
	if (op & (REP_N << 24)) {
		//REPNE/REPNZ prefix (also used to extend SSE+ ops)
		//output 0F2h prefix byte
		X86_emit_byte(REPNE);
	}
	if (op & (REX_A << 24)) {
		//address size = 32 bits instead of 64 bits
		//output 67h override byte
		X86_emit_byte(A32);
	}
	if (op & (REX_D << 24)) {
		//operand size = 16 bits
		//output 66h override byte
		X86_emit_byte(D16);
	}
#if 0	//Not yet implemented
	// FS: and GS: are mutually exclusive
	if (op & (GS_PB << 24)) {
		//segment override = GS:
		//output 65h segment override byte
		X86_emit_byte(GS_);
	}
	else if (op & (FS_PB << 24)) {
		//segment override = FS:
		//output 64h segment override byte
		X86_emit_byte(FS_);
	}
#endif //NYI
	if (op & (LOCK_PB << 24)) {
		//LOCK prefix
		//output 0F0h prefix byte
		X86_emit_byte(LOCK);
	}
	//construct a REX byte if needed...
#if 0	//old reference version
	if (op & (REX_W << 24)) {
		//operand size = 64 bits
		//output the REX.W bit flag
		REX_byte |= REX_W;
	}
	if ((op & (REX << 24)) && (reg >= spl) && (reg <= dil)) {
		//bare REX byte needed
		//for extended low byte regs
		//(SPL, BPL, SIL, and DIL only)
		REX_byte |= REX;
	}
	//set high order register extension bits
	if (reg & 8) REX_byte |= REX_R; 
	if (index_reg & 8) REX_byte |= REX_X;
	if (base_reg & 8) REX_byte |= REX_B;
#elif 0
	// Could use bit ops only, to set REX, REX_W, REX_R, 
	// REX_X, and REX_B bits...
	REX_byte |= (op >> 24) & REX_W;		//set REX_W = 8
	// Set REX bit if REX is set in op and bit 2 (0x04) is set in reg.
	// This sets REX if reg is 4,5,6,7,12,13,14,or 15, but REX will
	// be set anyway if reg is 12,13,14,or 15 because these regs
	// will always cause REX_R to be set, and hence REX. So the net
	// effect is to set REX specifically if reg is 4,5,6,or 7 --
	// which is exactly what we want. Phew.
	REX_byte |= (op >> 24) & REX & (reg << 4);
	REX_byte |= (reg & 8) >> 1;			//set REX_R = 4
	REX_byte |= (index_reg & 8) >> 2;	//set REX_X = 2
	REX_byte |= (base_reg & 8) >> 3;	//set REX_B = 1
	// N.B.  Is there a simple macro to compute log2(2**n)?
	// If so, can code e.g. 
	// REX_byte |= (reg & 8) >> (LOG2(8) - LOG2(REX_R));
#else
	// combining the above bitops...
	REX_byte |= (op >> 24) & REX & (reg << 4)|	//set REX (0x40) bit
				(op >> 24) & REX_W			 |	//set REX.W (0x08) bit
				(reg & 8) >> 1				 |	//set REX.R (0x04) bit
				(index_reg & 8) >> 2		 |	//set REX.X (0x02) bit
				(base_reg & 8) >> 3;			//set REX_B (0x01) bit
#endif
	if (REX_byte) X86_emit_byte(REX|REX_byte);
} /*X86_emit_prefixes*/

static void X86_emit_opc(int op)
{	// For internal use only!!
	// High-order byte of op has already been processed by X86_emit_prefixes
	// Or should X86_emit_prefixes be merged with this routine?
    if (op & 0xff0000) {
        X86_emit_byte((BYTE)(op >> 16));
        X86_emit_byte((BYTE)(op >> 8));
    }
    else if (op & 0xff00) {
        X86_emit_byte((BYTE)(op >> 8));
    }
    X86_emit_byte((BYTE)op);
} /*X86_emit_opc*/

void X86_emit_op(int op)
{
	X86_emit_prefixes(op, 0, 0, 0);	// Set/emit REX flags (not needed here???)

	X86_emit_opc(op);	// emit the opcode byte(s)
} /*X86_emit_op*/

void X86_emit_opreg(int op, int reg)
{
	//reg in low 3 bits of an opcode is extended by REX.b
	X86_emit_prefixes(op + (reg & 7), 0, reg, 0);	//or use macros???
	X86_emit_opc(op + (reg & 7));
} /*X86_emit_opreg*/

void X86_emit_mod_reg_rm(int mod, int reg, int rm)
{
    X86_emit_byte( (BYTE)(((mod & 3)<<6) | ((reg & 7)<<3) | (rm & 7)) );
} /*X86_emit_mod_reg_rm*/

void X86_emit_sib(int ss, int index, int base)
{
    X86_emit_byte( (BYTE)(((ss & 3)<<6) | ((index & 7)<<3) | (base & 7)) );
} /*X86_emit_sib*/

void X86_emit_RR(int reg1, int reg2)
{
    X86_emit_mod_reg_rm(3, reg1, reg2);
} /*X86_emit_RR*/

#if 0	// Old reference version of X86_emit_RM.  Delete when new x64 version is alive and well
void X86_emit_RM(int reg, int base_reg, int index_reg, int scaleFactor, int disp)
{
    int mod;
    int ss;
    if (scaleFactor != 0 && index_reg == esp) {
        //esp can't be used as scaled index reg
        X86_abort("Invalid use of ESP as index reg");
    }
    if (base_reg == esp) {
        //esp can't be used as base reg; must code as [ESP+ESP*1]
        X86_ASSERT(scaleFactor == 0);
        index_reg = esp;
        scaleFactor = 1;
    }
    if (base_reg == no_reg) {
        //no base reg ==> 4 byte displacement
        mod = 0;
        base_reg = ebp;     //[EBP] is actually [d32]
    }
    else if (disp == 0 && base_reg != ebp) {
        //no displacement and base reg is not EBP
        mod = 0;
    }
    else if (disp >= -128 && disp <= 127) {
        // 1 byte displacement (or no displacement and base reg is EBP)
        mod = 1;
    }
    else {
        // 4 byte displacement
        mod = 2;
    }
    if (scaleFactor == 0) {
        //no index reg => no sib byte
        X86_emit_mod_reg_rm(mod, reg, base_reg);
    }
    else {
        //need an sib byte
        if (scaleFactor == 1) ss = 0;
        else if (scaleFactor == 2) ss = 1;
        else if (scaleFactor == 4) ss = 2;
        else if (scaleFactor == 8) ss = 3;
        else X86_abort2("Invalid scale factor (%d.)", scaleFactor);
        X86_emit_mod_reg_rm(mod, reg, 4); //4 => sib is present
        X86_emit_sib(ss, index_reg, base_reg);
    }
    //emit displacement (if any)
    if (mod == 1) {
        //d8
        X86_emit_byte((BYTE)disp);
    }
    else if (mod == 2 || (mod == 0 && base_reg == ebp)) {
        //d32
        X86_emit_dword(disp);
    }
} /*X86_emit_RM*/
#endif // old reference version

#if 0
void X86_emit_RM(int reg, int base_reg, int index_reg, int scaleFactor, int disp)
{
    int mod;
    int ss;
    if (scaleFactor != 0 && index_reg == esp) {
		if (base_reg != esp && scaleFactor <= 1) {
			// 1*esp means NO index reg, so swap
			// [base+1*esp][+d] as [esp+1*base][+d]
			// (caller should probably have done that already!)
			index_reg = base_reg;
			if (index_reg == no_reg) index_reg = esp;
			base_reg = esp;
			scaleFactor = 1;
		}
		else {
			//esp can't be used as a scaled index reg (but r12 is allowed)
			X86_abort("Invalid use of ESP as an index reg");
		}
    }
#if 1	//this was wrong!
    if (base_reg == esp && scaleFactor == 0) {
        //esp can't be used as a simple base reg; must code as [ESP+ESP*1]
		//where ESP*1 indicates no index reg
        index_reg = esp;
        scaleFactor = 1;
    }
#endif
    if (base_reg == no_reg) {
        //no base reg ==> 4 byte displacement
        mod = 0;
        base_reg = ebp;     //[EBP] is actually [d32]
    }
    else if (disp == 0 && (base_reg & 7) != ebp) {
        //no displacement and base reg is neither EBP nor R13
        mod = 0;
    }
    else if (disp >= -128 && disp <= 127) {
        // 1 byte displacement (or no displacement and base reg is EBP or R13)
        mod = 1;
    }
    else {
        // 4 byte displacement
        mod = 2;
    }
    if (scaleFactor == 0 && (base_reg & 7) != esp) {
        //no index reg => no sib byte (unless base reg is esp or r12)
        X86_emit_mod_reg_rm(mod, reg, base_reg);
    }
    else {
        //need an sib byte
		if (scaleFactor == 0) ss = 0;  //base reg is esp or r12 
        else if (scaleFactor == 1) ss = 0;
        else if (scaleFactor == 2) ss = 1;
        else if (scaleFactor == 4) ss = 2;
        else if (scaleFactor == 8) ss = 3;
        else X86_abort2("Invalid scale factor (%d.)", scaleFactor);
        X86_emit_mod_reg_rm(mod, reg, 4); //4(esp) => sib is present
        X86_emit_sib(ss, index_reg, base_reg);
    }
    //emit displacement (if any)
    if (mod == 1) {
        //d8
        X86_emit_byte((BYTE)disp);
    }
    else if (mod == 2 || (mod == 0 && (base_reg & 7) == ebp)) {
        //d32 or [ebp]d32 or [r13]d32
        X86_emit_dword(disp);
    }
} /*X86_emit_RM*/
#else
/************************** X64 INSTRUCTION ENCODING ***************************/
//	
//	REX byte
//	--------
//	 
//	 7 6 5 4 3 2 1 0
//	+-+-+-+-+-+-+-+-+
//	|0 1 0 0|w r x b|
//	+-+-+-+-+-+-+-+-+
//	
//	w=1 => 64-bit operand size
//	r=1 => extends ModRM reg field
//	x=1 => extends SIB index field
//	b=1 => extends ModRM r/m field (unless mod<>3 and r/m=4, or mod=0 and r/m=5), 
//	       or SIB base field, or opcode reg field (lsb 3 bits of each?)
//	
//	
//	Reg in OPCODE byte:
//	-------------------
//	
//	 7 6 5 4 3 2 1 0
//	+-+-+-+-+-+-+-+-+
//	| opcode  | bbb | + REX.b * 8 (??)
//	+-+-+-+-+-+-+-+-+
//	Examples (all known cases? What about fpu, sse, etc.?):
//	BSWAP reg = 0F C8+reg
//	MOV reg,imm = B8+reg,imm
//	PUSH reg = 50+reg	(automatically promoted to 64bit, REX.w not needed, REX.b needed only for reg >= R8)
//	POP reg = 58+reg		"	"	    "
//	XCHG rAX,reg = 90+reg
//	(DEC reg = 48+reg and INC reg = 40+reg are usurped in 64-bit long)
//	
//	
//	ModRM byte
//	----------
//	
//	 7 6 5 4 3 2 1 0
//	+-+-+-+-+-+-+-+-+
//	|mod| reg | r/m |
//	+-+-+-+-+-+-+-+-+
//	
//	mod=0 => [r/m]
//	mod=1 => [r/m]disp8
//	mod=2 => [r/m]disp32
//	mod=3 => r/m
//	
//	except:
//	mod=0 and r/m=4(sp) => [SIB]								(REX.b ignored)
//	mod=1 and r/m=4(sp) => [SIB]disp8							(REX.b ignored)
//	mod=2 and r/m=4(sp) => [SIB]disp32							(REX.b ignored)
//	mod=0 and r/m=5(bp) => [rip]disp32 (X64) or [disp32] (x86)	(REX.b ignored)
//	
//	
//	SIB byte
//	--------
//	present if and only if r/m=4(sp)
//	
//	 7 6 5 4 3 2 1 0
//	+-+-+-+-+-+-+-+-+
//	| ss| xxx | bbb |
//	+-+-+-+-+-+-+-+-+
//	
//	ss = scale factor = 2**ss
//	     ss=0 => *1
//	     ss=1 => *2
//	     ss=2 => *4
//	     ss=3 => *8
//	
//	except:
//	REX.x:xxx=4(sp) => no index reg (and ss should be 0, at least for 386??)	(REX.x matters)
//	mod=0 and bbb=5(bp) => [xxx*scale]disp32 (but NO rip in this case!)			(REX.b ignored)
//	except:
//	mod=0 and bbb=5(bp) and REX.x:xxx=4(sp) => [disp32]			(REX.b ignored)	(REX.x matters)
//
//	References
//	----------
//	AMD 24594 Rev. 3.19
//		Tables A-34, A-35, A-36, A-37
//	
/*******************************************************************************/
void X86_emit_RM(int reg, int base_reg, int index_reg, int scaleFactor, int disp)
{
    int mod;
    int ss;
	// Test cases:
	//
	//	ok	[ecx]	==>	[ecx]			  m=00	r/m=ecx
	//	ok	[ecx+12]==>	[ecx]+12		  m=01	r/m=ecx	d8=12
	//	ok	[ecx+1234]	[ecx]+00001234	  m=10	r/m=ecx	d32=00001234
	//	ok	[r9]	==>	[r9]			  m=00	r/m=r9
	//	ok	[ebp]	==>	[ebp]+00		  m=01	r/m=5	d8=00
	//	ok	[r13]	==>	[r13]+00		  m=01	r/m=13	d8=00
	//	ok	[esp]	==>	[esp+1*esp]		  m=00	r/m=4	ss=00 iii=esp bbbb=esp
	//	ok	[r12]	==>	[r12+1*esp]		  m=00	r/m=4	ss=00 iii=esp bbbb=r12	???
	//if X86
	//	ok	[0]		==>	[ebp]+00000000    m=00	r/m=5	d32=00000000 (32-bit absolute address)
	//	ok	[1234]	==>	[ebp]+00001234    m=00	r/m=5	d32=00001234 (32-bit absolute address)
	//elseif X64
	//	??	[0]		==>	[rsp+1*rsp]+0000  m=00	r/m=4	d32=00000000 (64-bit sign-extended abs addr)
	//	??	[1234]	==>	[rsp+1*rsp]+1234  m=00	r/m=4	d32=00001234 (64-bit sign-extended abs addr)
	//	??	[rip]	==>	[rbp]+0000	      m=00	r/m=5	d32=00000000 (64-bit rip-relative address)
	//	??	[rip+1234]	[rbp]+00001234	  m=00	r/m=5	d32=00001234 (64-bit rip-relative)
	//endif
	//	ok	[2*ecx]	==>	[esp+2*ecx]	      m=00	r/m=4	ss=00 iii=ecx bbbb=esp d32=0000???
	//	ok	[1*ecx]	==>	[ecx]		      m=00	r/m=ecx
	//

#if 0	//Caller should optimize this (doesn't work with X86_emit_prefixes)
	// Optimize [1*xxx] to [xxx] if xxx <> esp/r12
	if (base_reg == no_reg && scaleFactor == 1 && (index_reg & 7) != esp) {
		base_reg = index_reg;
		index_reg = no_reg;
		scaleFactor = 0;
	}
	else
#endif
	// ESP can't be used as an index reg...
	// A possible workaround is to swap it with the base reg.
	// N.B. Caller should probably have done this already,
	// in which case we really don't need to waste cycles
	// checking for this workaround!
    if (index_reg == esp) {
		assert(scaleFactor != 0);
#if 0  //try a workaround (doesn't work with X86_emit_prefixes)
		// can't use esp as an index reg; see if we can use it
		// as a base reg by swapping it with the desired base reg (if any)
		if (base_reg != esp && scaleFactor == 1) {
			// 1*esp means NO index reg, so swap
			// [base+1*esp][+d] to [esp+1*base][+d]
			index_reg = base_reg;
			if (index_reg == no_reg) index_reg = esp;
			base_reg = esp;
			assert(scaleFactor == 1);
		}
		else
#endif //workaround
		{
			//esp can't be used as a scaled index reg (but r12 is allowed!)
			X86_abort("Invalid use of ESP as an index reg");
		}
    }
	else if ((base_reg & 7) == esp && scaleFactor == 0) {
        //esp or r12 can't be used as a simple base reg; must code as [esp+ESP*1]
		//where ESP*1 indicates no index reg
        index_reg = esp;
        scaleFactor = 1;	// ESP*1
    }

	//determine actual displacement byte size (mod=0, 1, or 2)
    if (base_reg == no_reg) {
        //no base reg ==> 4 byte displacement
        mod = 0;
        base_reg = ebp;     //[EBP] is actually [d32]	//X64!!
    }
    else if (disp == 0 && (base_reg & 7) != ebp) {
        //no displacement, and base reg is neither EBP nor R13
        mod = 0;
    }
    else if (disp >= -128 && disp <= 127) {
        // 1 byte displacement (or no displacement and base reg is EBP or R13)
        mod = 1;
    }
    else {
        // 4 byte displacement
        mod = 2;
    }

	//determine if SIB byte is needed				//X64 for abs addr!!
    if (scaleFactor == 0 && (base_reg & 7) != esp) {
        //no index reg => no sib byte (unless base reg is esp or r12)
        X86_emit_mod_reg_rm(mod, reg, base_reg);
    }
    else {
        //need SIB byte
		if (scaleFactor == 0) ss = 0;  //base reg is esp (or r12 ???)
        else if (scaleFactor == 1) ss = 0;
        else if (scaleFactor == 2) ss = 1;
        else if (scaleFactor == 4) ss = 2;
        else if (scaleFactor == 8) ss = 3;
        else X86_abort2("Invalid scale factor (%d.)", scaleFactor);
        X86_emit_mod_reg_rm(mod, reg, 4); //4(esp) => sib is present
        X86_emit_sib(ss, index_reg, base_reg);
    }

    //emit displacement (if any)
    if (mod == 1) {
        //d8
        X86_emit_byte((BYTE)disp);
    }
    else if (mod == 2 || (mod == 0 && (base_reg & 7) == ebp)) {	//last operand redundant???
        //d32 or [ebp]d32 or [r13]d32
        X86_emit_dword(disp);
    }
} /*X86_emit_RM*/
#endif

void X86_emit_op_RR(int op, int reg1, int reg2)
{
    switch (NP(op)) {
    case CMOVO:
    case CMOVNO:
    case CMOVB:
    case CMOVNB:
    case CMOVZ:
    case CMOVNZ:
    case CMOVBE:
    case CMOVNBE:
    case CMOVS:
    case CMOVNS:
    case CMOVP:
    case CMOVNP:
    case CMOVL:
    case CMOVNL:
    case CMOVLE:
    case CMOVNLE:
#ifndef NEVER_USE_CMOV
        if (gFmcp->use_CMOV) {
			X86_emit_prefixes(op, reg1, reg2, 0);
            X86_emit_opc(op);
            X86_emit_RR(reg1, reg2);
        } else
#endif //!NEVER_USE_CMOV
        {
            //Synthesize these...
            //  JNcc short L1
            //  MOV reg1, reg2
            //L1:
            CGL L1 = X86_genlab();
            //Change CMOVcc to Jcc (disp8 only), then toggle the cc to Ncc...
            X86_emit_op_lab_disp8((JO|(op & 0xf))^1, L1);	//X64???
            X86_emit_op_RR(MOV_V, reg1, reg2);
            X86_deflab(L1);
        }
        break;
#ifndef DONT_CHECK_FOR_ILLEGAL_INSTRUCTIONS
    case LEA:
    case LDS:
    case LES:
    case LSS:
    case LFS:
    case LGS:
        X86_abort2("No RR form of op (0x%8.8x)", op);
        break;
#endif
    default:
		X86_emit_prefixes(op, reg1, reg2, 0);
        X86_emit_opc(op);
        X86_emit_RR(reg1, reg2);
        break;
    } /*switch*/
} /*X86_emit_op_RR*/

void X86_emit_op_M(int op,
                   int base_reg, int index_reg, int scaleFactor, int disp)
{
    switch (NP(op)) {
    case SETO_D:
    case SETNO_D:
    case SETB_D:
    case SETNB_D:
    case SETZ_D:
    case SETNZ_D:
    case SETBE_D:
    case SETNBE_D:
    case SETS_D:
    case SETNS_D:
    case SETP_D:
    case SETNP_D:
    case SETL_D:
    case SETNL_D:
    case SETLE_D:
    case SETNLE_D:
        //Synthesize these...
#if 0
        // MOV m, 0
        // SETxx m
        X86_emit_op_M_imm32(MOVI_V, base_reg, index_reg, scaleFactor, disp, 0);	//X64??
        X86_emit_op_M(op - 1, base_reg, index_reg, scaleFactor, disp);
#else
        // SETxx al
        // AND eax, 1
        // MOV m, eax
        X86_ASSERT(base_reg != eax);
        X86_ASSERT(index_reg != eax);
        X86_emit_op_R(op - 1, al);
        X86_emit_op_R_simm32(ANDI_D, eax, 1);	//ANDI_V not needed, X64 zero-extends!!
        X86_emit_op_MR(MOV_D, base_reg, index_reg, scaleFactor, disp, eax);	//X64: op_store, assume int32 for now!!
#endif
        break;
    default:
		X86_emit_prefixes(OC2(op), 0, base_reg, index_reg);
		X86_emit_opc(OC2(op));
        ////X86_emit_op(op >> 8);	//X64???	//RESUME HERE TOO!!!!
        X86_emit_RM(op & 7, base_reg, index_reg, scaleFactor, disp);
        break;
    } /*switch*/
} /*X86_emit_op_M*/

void X86_emit_op_T(int op, int tmp)
{
    if (isReg(tmp)) {
        X86_emit_op_R(op, tmpReg(tmp));
    }
    else {
        X86_emit_op_M(op, TMP(tmp));
    }
} /*X86_emit_op_T*/

void X86_emit_op_Tm(int op, int tmp)
{   //Emit op against temp which MUST be in memory (e.g., an x87 op)
    switch (NP(op)) {
    case FST_I:
    case FSTP_I:
		// store 32-bit integer
        X86_emit_op_M(op, TMP(tmp));
        if (isReg(tmp)) {
            //Load reg from temp location -- should sign-extend to 64-bits if X64gen per op_load???
			if (X64gen) {
				//X64: sign-extend to 64 bits
				X86_emit_op_RM(MOVSXD_Q, tmpReg(tmp), TMP(tmp));
			}
			else {	
				//X32: load 32 bits
				X86_emit_op_RM(MOV_D, tmpReg(tmp), TMP(tmp));
			}
        }
        break;
    case FSTP_L:
		// store 64-bit integer
        X86_emit_op_M(op, TMP(tmp));
        if (isReg(tmp)) {
            //Load reg from temp location...
            X86_emit_op_RM(MOV_V, tmpReg(tmp), TMP(tmp));	//NB: Only use 32-bits if not X64???
        }
		break;
    case FST_F:
    case FST_D:
    case FST_S:
    case FSTP_F:
    case FSTP_D:
    case FSTP_S:
    case FSTP_T:
    case FSTP_B:
        X86_abort2("Unsupported opcode (0x%8.8x) for op_Tm", op);
        break;
    default:
        if (isReg(tmp)) {
            //Store reg in temp location...
            X86_emit_op_MR(MOV_V, TMP(tmp), tmpReg(tmp));	//X64 ???
        }
        X86_emit_op_M(op, TMP(tmp));
        break;
    } /*switch*/
} /*X86_emit_op_Tm*/

void X86_emit_op_R(int op, int reg)
{
    switch (NP(op)) {
    case PUSH_D:
        X86_emit_opreg(PUSH_REG, reg);	//X64??
        break;
    case POP_D:
        X86_emit_opreg(POP_REG, reg);	//X64??
        break;
    case SETO_D:
    case SETNO_D:
    case SETB_D:
    case SETNB_D:
    case SETZ_D:
    case SETNZ_D:
    case SETBE_D:
    case SETNBE_D:
    case SETS_D:
    case SETNS_D:
    case SETP_D:
    case SETNP_D:
    case SETL_D:
    case SETNL_D:
    case SETLE_D:
    case SETNLE_D:
        //Synthesize these...
        // can't always do a SETxx directly to reg, because
        // SETxx targets a byte reg (al,ah, etc.), not
        // a dword reg...
        switch (reg) {			// Could do better for X64, using orthogonal byte regs!!!
        case eax:
            X86_emit_op_R(op - 1, al);              //SETxx_B AL
            X86_emit_op_R_simm32(ANDI_D, reg, 1);   //zero-extend 1 bit to 32/64 bits [sic]
            break;
        case ebx:
            X86_emit_op_R(op - 1, bl);              //SETxx_B BL
            X86_emit_op_R_simm32(ANDI_D, reg, 1);   //zero-extend 1 bit to 32/64 bits
            break;
        case ecx:
            X86_emit_op_R(op - 1, cl);              //SETxx_B CL
            X86_emit_op_R_simm32(ANDI_D, reg, 1);   //zero-extend 1 bit to 32/64 bits
            break;
        case edx:
            X86_emit_op_R(op - 1, dl);              //SETxx_B DL
            X86_emit_op_R_simm32(ANDI_D, reg, 1);   //zero-extend 1 bit to 32/64 bits
            break;
        default:
            X86_emit_op_R(op - 1, al);              //SETxx_B AL
            X86_emit_op_RR(MOVZB_D, reg, al);       //MOVZX reg, AL (zero-extends to 32/64 bits)
            break;
        } /*switch reg*/
        break;
#ifndef X64ONLY
	//X64 has no INC_REG or DEC_REG opcodes
    case INC_D:
		if (X86GEN) {
	        X86_emit_opreg(INC_REG, reg);
		    break;
		}
		//fall through, ultimately to default (sneaky code!!)
    case DEC_D:
		if (X86GEN) {
	        X86_emit_opreg(DEC_REG, reg);
		    break;
		}
		//else fall through to default
#endif //!X64ONLY
    default:
		X86_emit_prefixes(OC2(op), 0, reg, 0);
        X86_emit_opc(OC2(op));
        X86_emit_RR(op & 7, reg);
        break;
    } /*switch op*/
} /*X86_emit_op_R*/

void X86_emit_op_RM(int op,
                    int reg,
                    int base_reg, int index_reg, int scaleFactor, int disp)
{
    switch (NP(op)) {
    case CMOVO:
    case CMOVNO:
    case CMOVB:
    case CMOVNB:
    case CMOVZ:
    case CMOVNZ:
    case CMOVBE:
    case CMOVNBE:
    case CMOVS:
    case CMOVNS:
    case CMOVP:
    case CMOVNP:
    case CMOVL:
    case CMOVNL:
    case CMOVLE:
    case CMOVNLE:
#ifndef NEVER_USE_CMOV
        if (gFmcp->use_CMOV) {
			X86_emit_prefixes(op, reg, base_reg, index_reg);
            X86_emit_opc(op);
            X86_emit_RM(reg, base_reg, index_reg, scaleFactor, disp);
        } else
#endif //!NEVER_USE_CMOV
        {
            //Synthesize these...
            //  JNcc short L1
            //  MOV reg, m
            //L1:
            CGL L1 = X86_genlab();
            //Change CMOVcc to Jcc (disp8 only), then toggle the cc to Ncc...
            X86_emit_op_lab_disp8((JO|(op & 0xf))^1, L1);	//X64???
            X86_emit_op_RM(MOV_D, reg, base_reg, index_reg, scaleFactor, disp);	//X64 op_load???
            X86_deflab(L1);
        }
        break;
    default:
		X86_emit_prefixes(op, reg, base_reg, index_reg);
        X86_emit_opc(op);
        X86_emit_RM(reg, base_reg, index_reg, scaleFactor, disp);
        break;
    } /*switch*/
} /*X86_emit_op_RM*/

void X86_emit_op_MR(int op,
                    int base_reg, int index_reg, int scaleFactor, int disp,
                    int reg)
{
    switch (NP(op)) {	//X64
    case TEST_B:
    case XCHG_B:
    case TEST_D:
    case XCHG_D:
        //operands are commutative, so don't need to invert direction
        X86_emit_op_RM(op, reg, base_reg, index_reg, scaleFactor, disp);
        break;

    case IMUL32:
        X86_ASSERT(base_reg != eax);
        X86_ASSERT(index_reg != eax);
        X86_ASSERT(reg != eax);
        X86_emit_op_RM(MOV_V, eax, base_reg, index_reg, scaleFactor, disp);
        X86_emit_op_RR(op, eax, reg);
        X86_emit_op_MR(MOV_V, base_reg, index_reg, scaleFactor, disp, eax);
        break;

	// Reverse direction bit is 0x01 for these SSE1 instructions
	#if (MOVSSR - MOVSS) != 0x01
	#error Incorrect toggle bit for MOVSS and MOVSDX
	#endif
	case NP(MOVUPS):	//SSE1	0x0F10
	case NP(MOVUPSR):	//SSE1	0x0F11
	case NP(MOVAPS):	//SSE1	0x0F28
	case NP(MOVAPSR):	//SSE1	0x0F29
	// remainder are redundant...
//	case NP(MOVUPD):	//SSE2	W(0x0F10)
//	case NP(MOVUPDR):	//SSE2	W(0x0F11)
//	case NP(MOVAPD):	//SSE2	W(0x0F28)
//	case NP(MOVAPDR):	//SSE2	W(0x0F29)
//	case NP(MOVSS):		//SSE1	F3(0x0F10)
//	case NP(MOVSSR):	//SSE1	F3(0x0F11)
//	case NP(MOVSDX):	//SSE1	F2(0x0F10)
//	case NP(MOVSDXR):	//SSE1	F2(0x0F11)
		op |= 0x01;			//reverse direction bit is 0x01
        X86_emit_op_RM(op,	//use the MR direction opcode
                       reg, base_reg, index_reg, scaleFactor, disp);
        break;

//	case NP(XMOVQX):	//SSE2	F3(MOVDR)	special case: no direction bit
	case NP(XMOVQXR):	//SSE2	W(0x0FD6)	special case: reverse of XMOVQX
		if (op == XMOVQX) {
			op = XMOVQXR;	//special case: doesn't use a "direction" bit
		}
        X86_emit_op_RM(op,	//use the MR direction opcode
                       reg, base_reg, index_reg, scaleFactor, disp);
        break;

	// Reverse direction bit is 0x10 for these MMX/SSE2 instructions
	#if (MOVDR - MOVD) != 0x10
	#error Incorrect toggle bit for MOVD et al.
	#endif
    case NP(MOVD):		//MMX	0x0F6E
    case NP(MOVDR):		//MMX	0x0F7E
	case NP(MOVQX):		//MMX	0x0F6F
	case NP(MOVQXR):	//MMX	0x0F7F
	// remainder are redundant...
//	case NP(MOVQ):		//MMX	Q(MOVD)
//	case NP(MOVQR):		//MMX	Q(MOVDR)
//	case NP(XMOVD):		//SSE2	W(MOVD)
//	case NP(XMOVDR):	//SSE2	W(MOVDR)
//	case NP(XMOVQ):		//SSE2	W(MOVQ) = W(Q(MOVD))
//	case NP(XMOVQR):	//SSE2	W(MOVQR)= W(Q(MOVDR))
		op |= 0x10;		//reverse direction bit is 0x10
        X86_emit_op_RM(op,	//use the MR direction opcode
                       reg, base_reg, index_reg, scaleFactor, disp);
        break;

	// Forward direction bit is 0x02 for these instructions
	#if (ADD_B - ADDR_B) != 0x02
	#error Incorrect toggle bit for OR et al.
	#endif
    case ADDR_B:
    case ADD_B:
    case ORR_B:
    case OR_B:
    case ADCR_B:
    case ADC_B:
    case SBBR_B:
    case SBB_B:
    case ANDR_B:
    case AND_B:
    case SUBR_B:
    case SUB_B:
    case XORR_B:
    case XOR_B:
    case CMPR_B:
    case CMP_B:
    case MOVR_B:
    case MOV_B:

    case ADDR_D:
    case ADD_D:
    case ORR_D:
    case OR_D:
    case ADCR_D:
    case ADC_D:
    case SBBR_D:
    case SBB_D:
    case ANDR_D:
    case AND_D:
    case SUBR_D:
    case SUB_D:
    case XORR_D:
    case XOR_D:
    case CMPR_D:
    case CMP_D:
    case MOVR_D:
    case MOV_D:
        X86_emit_op_RM(op & ~0x02, //Clear the forward direction bit
                       reg, base_reg, index_reg, scaleFactor, disp);
        break;
//
// Could synthesize MR form of MOVSXD_Q mem,reg as:
//    MOV RAX,reg / CDQE / MOV mem,RAX (if RAX is free)
// or: MOVSXD_Q RAX,reg / MOV mem,RAX (if RAX is free)
//
    default:
        //No MR form for these instructions...
        X86_abort2("No MR form of op (0x%8.8x)", op);
        break;
    }
} /*X86_emit_op_MR*/

void X86_emit_op_RT(int op, int reg, int tmp)
{
    if (isReg(tmp)) {
        X86_emit_op_RR(op, reg, tmpReg(tmp));
    }
    else {
        X86_emit_op_RM(op, reg, TMP(tmp));
    }
} /*X86_emit_op_RT*/

void X86_emit_op_TR(int op, int tmp, int reg)
{
    if (isReg(tmp)) {
        X86_emit_op_RR(op, tmpReg(tmp), reg);
    }
    else {
        X86_emit_op_MR(op, TMP(tmp), reg);
    }
} /*X86_emit_op_TR*/

void X86_emit_op_TT(int op, int tmp1, int tmp2)
{
    if (isReg(tmp1)) {
        X86_emit_op_RT(op, tmpReg(tmp1), tmp2);
    }
    else if (isReg(tmp2)) {
        X86_emit_op_TR(op, tmp1, tmpReg(tmp2));
    }
    else {
        X86_emit_op_MM(op, TMP(tmp1), TMP(tmp2));
    }
} /*X86_emit_op_TT*/

void X86_emit_op_MM(int op, MEM1, MEM2)		//what is this really used for?
											//if both mem operands are read-only,
											//as in TEST(mem1,mem2) or CMP(mem1,mem2),
											//then no op_store (and its intricacies)
											//need be involved
{
    switch (NP(op)) {	//X64

    case XCHG_B:
    case XCHG_D:
	case IMUL32:
        //no inverse direction (MR) forms for these...
        X86_ASSERT(base_reg1 != eax);
        X86_ASSERT(index_reg1 != eax);
        X86_ASSERT(base_reg2 != eax);
        X86_ASSERT(index_reg2 != eax);
        X86_emit_op_RM(MOV_V, eax, mem1);
        X86_emit_op_RM(op, eax, mem2);
        X86_emit_op_MR(MOV_V, mem1, eax);
        break;

#ifdef X64
	case MOVSD_D:	//For X86, this is ARPL!
#endif //X64
    case LEA:
        X86_ASSERT(base_reg1 != eax);
        X86_ASSERT(index_reg1 != eax);
        X86_emit_op_RM(op, eax, mem2);		//X64 op_load??
#if 0
        X86_emit_op_MR(MOV_V, mem1, eax);	//X64 op_store???
#else	
		// following is not totally correct!!!
#ifdef X64
		if (op != NP(op)) {
			//probably a QWORD (or WORD??) op
			X86_emit_op_MR(MOV_Q, mem1, eax);	//64-bit result
		}
		else 
#endif
		{ //must be a DWORD op???
			X86_emit_op_MR(MOV_D, mem1, eax);	//32-bit result
		}
#endif
        break;

    default:	//or swap MOV_V and op, and use common LEA code above??? no --NFG for e.g. ADD_D tmp33,tmp44
        X86_ASSERT(base_reg1 != eax);
        X86_ASSERT(index_reg1 != eax);
        X86_emit_op_RM(MOV_V, eax, mem2);	//X64 op_load???
        X86_emit_op_MR(op, mem1, eax);		//assumes MR form exists!!
        break;
    }
} /*X86_emit_op_MM*/

void X86_emit_op_TM(int op, int tmp,
                    int base_reg, int index_reg, int scaleFactor, int disp)
{
    if (isReg(tmp)) {
        X86_emit_op_RM(op, tmpReg(tmp), base_reg, index_reg, scaleFactor, disp);
    }
    else {
        X86_emit_op_MM(op, TMP(tmp), base_reg, index_reg, scaleFactor, disp);
    }
} /*X86_emit_op_TM*/

void X86_emit_op_MT(int op,
                    int base_reg, int index_reg, int scaleFactor, int disp,
                    int tmp)
{
    if (isReg(tmp)) {
        X86_emit_op_MR(op, base_reg, index_reg, scaleFactor, disp, tmpReg(tmp));
    }
    else {
        X86_emit_op_MM(op, base_reg, index_reg, scaleFactor, disp, TMP(tmp));
    }
} /*X86_emit_op_MT*/

//In following, 'Mi' means "memory indirect"...

void X86_emit_op_RMi(int op, int reg, MEM1)
{
    X86_ASSERT(reg != eax);
    X86_emit_op_RM(MOV_V, eax, mem1);	//x64 op_load???
    X86_emit_op_RM(op, reg, eax, no_reg, 0, 0);
} /*X86_emit_op_RMi*/

void X86_emit_op_Mi(int op, MEM1)
{
    X86_emit_op_RM(MOV_V, eax, mem1);	//X64 -- move ptr to eax
    X86_emit_op_M(op, eax, no_reg, 0, 0);	//op [eax]
} /*X86_emit_op_Mi*/

void X86_emit_op_MMi(int op, MEM1, MEM2)
{
    X86_ASSERT(base_reg1 != edx);
    X86_ASSERT(index_reg1 != edx);
    X86_emit_op_RMi(MOV_V, edx, mem2);	//X64 op_load???
    X86_emit_op_MR(op, mem1, edx);
#if 0
//or for movsxd_q:
	X86_emit_op_RMi(op, edx, mem2);
	X86_emit_op_MR(MOV_V, mem1, edx);
#endif
} /*X86_emit_op_MMi*/

void X86_emit_op_TMi(int op, int tmp, MEM1)
{
    if (isReg(tmp)) {
        X86_emit_op_RMi(op, tmpReg(tmp), mem1);
    }
    else {
        X86_emit_op_MMi(op, TMP(tmp), mem1);
    }
} /*X86_emit_op_TMi*/

void X86_emit_op_TTi(int op, int tmp1, int tmp2)
{
    if (isReg(tmp2)) {
        X86_emit_op_TM(op, tmp1, tmpReg(tmp2), no_reg, 0, 0);
    }
    else {
        X86_emit_op_TMi(op, tmp1, TMP(tmp2));
    }
} /*X86_emit_op_TTi*/

void X86_emit_op_Ti(int op, int tmp)
{
    if (isReg(tmp)) {
        X86_emit_op_M(op, tmpReg(tmp), no_reg, 0, 0);
    }
    else {
        X86_emit_op_Mi(op, TMP(tmp));
    }
} /*X86_emit_op_Ti*/

void X86_emit_op_MiR(int op, MEM1, int reg)
{
    X86_ASSERT(reg != edx);
    X86_emit_op_RM(MOV_V, edx, mem1);	//X64 -- move ptr to edx
    X86_emit_op_MR(op, edx, no_reg, 0, 0, reg);	// op [edx],reg
} /*X86_emit_op_MiR*/

void X86_emit_op_MiM(int op, MEM1, MEM2)
{
    X86_ASSERT(base_reg1 != edx);
    X86_ASSERT(index_reg1 != edx);
    X86_emit_op_RM(MOV_V, edx, mem2);	//X64 op_load???
    X86_emit_op_MiR(op, mem1, edx);
} /*X86_emit_op_MiM*/

void X86_emit_op_MiT(int op, MEM1, int tmp)
{
    if (isReg(tmp)) {
        X86_emit_op_MiR(op, mem1, tmpReg(tmp));
    }
    else {
        X86_emit_op_MiM(op, mem1, TMP(tmp));
    }
} /*X86_emit_op_MiT*/

void X86_emit_op_TiT(int op, int tmp1, int tmp2)
{
    if (isReg(tmp1)) {
        X86_emit_op_MT(op, tmpReg(tmp1), no_reg, 0, 0, tmp2);
    }
    else {
        X86_emit_op_MiT(op, TMP(tmp1), tmp2);
    }
} /*X86_emit_op_TiT*/

void X86_emit_op_TiR(int op, int tmp, int reg)
{
	if (isReg(tmp)) {
		X86_emit_op_MR(op, tmpReg(tmp), no_reg, 0, 0, reg);
	}
	else {
		X86_emit_op_MiR(op, TMP(tmp), reg);
	}
} /*X86_emit_op_TiR*/

void X86_emit_op_imm8(int op, BYTE immData)
{
	X86_emit_prefixes(op, 0, 0, 0);
    X86_emit_opc(op);
    X86_emit_byte(immData);
} /*X86_emit_op_imm8*/

void X86_emit_op_imm16(int op, WORD immData)
{
	X86_emit_prefixes(op, 0, 0, 0);
    X86_emit_opc(op);
    X86_emit_word(immData);
} /*X86_emit_op_imm16*/

void X86_emit_op_imm32(int op, DWORD immData)
{
	X86_emit_prefixes(op, 0, 0, 0);
    X86_emit_opc(op);
    X86_emit_dword(immData);
} /*X86_emit_op_imm32*/

void X86_emit_op_simm32(int op, SDWORD immData)
{
    if (immData >= -128 && immData <= 127) {
        //use sign-extended byte immediate data
#if 1
		X86_emit_prefixes(op | 2, 0, 0, 0);
        X86_emit_opc(op | 2);  //s=1
        X86_emit_byte((BYTE)immData);
#else
		X86_emit_op_imm8(op, immData);	//??? op | 2 ???
#endif
    }
    else {
        //use full 32-bit immediate data
#if 1
		X86_emit_prefixes(op, 0, 0, 0);				//16bit??
        X86_emit_opc(op);    //s=0
        X86_emit_dword(immData);				//16bit??
#else
		X86_emit_op_imm32(op, immData);	//???
#endif
    }
} /*X86_emit_op_simm32*/

void X86_emit_op_M_imm8(int op,
                        int base_reg, int index_reg, int scaleFactor, int disp,
                        BYTE immData)
{
	X86_emit_prefixes(OC2(op), 0, base_reg, index_reg);
    X86_emit_opc(OC2(op));
    X86_emit_RM(op & 7, base_reg, index_reg, scaleFactor, disp);
    X86_emit_byte(immData);
} /*X86_emit_op_M_imm8*/

void X86_emit_op_T_imm8(int op, int tmp, BYTE immData)
{
    if (isReg(tmp)) {
        X86_emit_op_R_imm8(op, tmpReg(tmp), immData);
    }
    else {
        X86_emit_op_M_imm8(op, TMP(tmp), immData);
    }
} /*X86_emit_op_T_imm8*/

void X86_emit_op_R_imm8(int op, int reg, BYTE immData)
{
    //doesn't (yet) optimize e.g. ADD AL,imm8
    switch (NP(op)) {
    case MOVI_B:
		X86_emit_opreg(MOVI_B_REG, reg);
		X86_emit_byte(immData);
        break;
    default:
		X86_emit_prefixes(OC2(op), 0, reg, 0);	//reg is extended by REX.b
        X86_emit_opc(OC2(op));
        X86_emit_RR(op & 7, reg);
        X86_emit_byte(immData);
        break;
    }
} /*X86_emit_op_R_imm8*/

void X86_emit_op_MR_imm8(int op,
                         int base_reg, int index_reg, int scaleFactor, int disp,
                         int reg,
                         BYTE immData)
{
	X86_emit_prefixes(op, reg, base_reg, index_reg);
    X86_emit_opc(op);
    X86_emit_RM(reg, base_reg, index_reg, scaleFactor, disp);	//should this be emit_MR???
    X86_emit_byte(immData);
} /*X86_emit_op_MR_imm8*/

void X86_emit_op_TR_imm8(int op, int tmp, int reg, BYTE immData)
{
    if (isReg(tmp)) {
        X86_emit_op_RR_imm8(op, tmpReg(tmp), reg, immData);
    }
    else {
        X86_emit_op_MR_imm8(op, TMP(tmp), reg, immData);
    }
} /*X86_emit_op_RT_imm8*/

void X86_emit_op_RR_imm8(int op, int reg1, int reg2, BYTE immData)
{
	X86_emit_prefixes(op, reg2, reg1, 0); //REX.b extends reg1
    X86_emit_opc(op);
    X86_emit_RR(reg2, reg1);    //[sic]
    X86_emit_byte(immData);
} /*X86_emit_op_RR_imm8*/

void X86_emit_op_RM_simm32(int op, int reg,
                           int base_reg, int index_reg, int scaleFactor, int disp,
                           SDWORD immData)
{
    if (immData >= -128 && immData <= 127) {
        //use sign-extended byte immediate data
		X86_emit_prefixes(op | 2, reg, base_reg, index_reg);
        X86_emit_opc(op | 2);  //s=1
        X86_emit_RM(reg, base_reg, index_reg, scaleFactor, disp);
        X86_emit_byte((BYTE)immData);
    }
    else {
        //use full 32-bit immediate data				//16bit??
		X86_emit_prefixes(op, reg, base_reg, index_reg);
        X86_emit_opc(op);    //s=0
        X86_emit_RM(reg, base_reg, index_reg, scaleFactor, disp);
        X86_emit_dword(immData);				//16bit??
    }
} /*X86_emit_op_RM_simm32*/

void X86_emit_op_RT_simm32(int op, int reg, int tmp, SDWORD immData)
{
    if (isReg(tmp)) {
        X86_emit_op_RR_simm32(op, reg, tmpReg(tmp), immData);
    }
    else {
        X86_emit_op_RM_simm32(op, reg, TMP(tmp), immData);
    }
} /*X86_emit_op_RT_simm32*/

void X86_emit_op_MR_simm32(int op,
                           int base_reg, int index_reg, int scaleFactor, int disp,
                           int reg,
                           SDWORD immData)
{
    X86_ASSERT(reg != eax);
    X86_ASSERT(base_reg != eax);
    X86_ASSERT(index_reg != eax);
    X86_emit_op_RR_simm32(op, eax, reg, immData);
    X86_emit_op_MR(MOV_V, base_reg, index_reg, scaleFactor, disp, eax);	//X64 op_store???
} /*X86_emit_op_MR_simm32*/

void X86_emit_op_TR_simm32(int op, int tmp, int reg, SDWORD immData)
{
    if (isReg(tmp)) {
        X86_emit_op_RR_simm32(op, tmpReg(tmp), reg, immData);
    }
    else {
        X86_emit_op_MR_simm32(op, TMP(tmp), reg, immData);
    }
} /*X86_emit_op_TR_simm32*/

void X86_emit_op_RR_simm32(int op, int reg1, int reg2, SDWORD immData)
{
    if (immData >= -128 && immData <= 127) {
        //use sign-extended byte immediate data
		X86_emit_prefixes(op | 2, reg1, reg2, 0);	//REX.b extends reg2
        X86_emit_opc(op | 2);  //s=1
        X86_emit_RR(reg1, reg2);
        X86_emit_byte((BYTE)immData);
    }
    else {
        //use full 32-bit immediate data				//16bit??
		X86_emit_prefixes(op, reg1, reg2, 0);	//REX.b extends reg2
        X86_emit_opc(op);    //s=0
        X86_emit_RR(reg1, reg2);
        X86_emit_dword(immData);				//16bit??
    }
} /*X86_emit_op_RR_simm32*/

void X86_emit_op_MM_simm32(int op,
                           int base_reg1, int index_reg1, int scaleFactor1, int disp1,
                           int base_reg2, int index_reg2, int scaleFactor2, int disp2,
                           SDWORD immData)
{
    X86_ASSERT(base_reg1 != eax);
    X86_ASSERT(index_reg1 != eax);
    X86_emit_op_RM_simm32(op, eax, base_reg2, index_reg2, scaleFactor2, disp2, immData);
    X86_emit_op_MR(MOV_V, base_reg1, index_reg1, scaleFactor1, disp1, eax);	//X64 op_store???
} /*X86_emit_op_MM_simm32*/

void X86_emit_op_TT_simm32(int op, int tmp1, int tmp2, SDWORD immData)
{
    if (isReg(tmp1)) {
        X86_emit_op_RT_simm32(op, tmpReg(tmp1), tmp2, immData);
    }
    else if (isReg(tmp2)) {
        X86_emit_op_TR_simm32(op, tmp1, tmpReg(tmp2), immData);
    }
    else {
        X86_emit_op_MM_simm32(op, TMP(tmp1), TMP(tmp2), immData);
    }
} /*X86_emit_op_TT_simm32*/

void X86_emit_op_M_imm32(int op,
                         int base_reg, int index_reg, int scaleFactor, int disp,
                         DWORD immData)
{
	X86_emit_prefixes(OC2(op), 0, base_reg, index_reg);
    X86_emit_opc(OC2(op));
    X86_emit_RM(op & 7, base_reg, index_reg, scaleFactor, disp);
    X86_emit_dword(immData);
} /*X86_emit_op_M_imm32*/

void X86_emit_op_T_imm32(int op, int tmp, DWORD immData)
{
    if (isReg(tmp)) {
        X86_emit_op_R_imm32(op, tmpReg(tmp), immData);
    }
    else {
        X86_emit_op_M_imm32(op, TMP(tmp), immData);
    }
} /*X86_emit_op_T_imm32*/

void X86_emit_op_R_imm32(int op, int reg, DWORD immData)
{
    switch (NP(op)) {
    case MOVI_D:
        ///X86_emit_op_imm32(MOVI_D_REG + reg, immData);
		X86_emit_opreg(MOVI_D_REG, reg);	//X64???
		X86_emit_dword(immData);
        break;
    default:
		X86_emit_prefixes(OC2(op), 0, reg, 0);	//REX.b extends reg
		X86_emit_opc(OC2(op));
        X86_emit_RR(op & 7, reg);
        X86_emit_dword(immData);
        break;
    }
} /*X86_emit_op_R_imm32*/

void X86_emit_op_M_simm32(int op,
                          int base_reg, int index_reg, int scaleFactor, int disp,
                          SDWORD immData)
{
    // op encoding: 000 ADD (low 3 bits)
    //              001 OR
    //              010 ADC
    //              011 SBB
    //              100 AND
    //              101 SUB
    //              110 XOR
    //              111 CMP
    if (NP(op) == IMULI_D) {			// CHECK X64!!!
        X86_emit_op_MM_simm32(op,
                              base_reg, index_reg, scaleFactor, disp,
                              base_reg, index_reg, scaleFactor, disp,
                              immData);
    }
    else if (immData >= -128 && immData <= 127) {
        //use sign-extended byte immediate data
		X86_emit_prefixes(OC2(op) | 2, 0, base_reg, index_reg);	//s=1
        X86_emit_opc(OC2(op) | 2);    //s=1
        X86_emit_RM(op & 7, base_reg, index_reg, scaleFactor, disp);
        X86_emit_byte((BYTE)immData);
    }
    else {
        //use full 32-bit immediate data								//16bit???
		X86_emit_prefixes(OC2(op), 0, base_reg, index_reg);	//s=0
        X86_emit_opc(OC2(op));    //s=0
        X86_emit_RM(op & 7, base_reg, index_reg, scaleFactor, disp);
        X86_emit_dword(immData);										//16bit??
    }
} /*X86_emit_op_M_simm32*/

void X86_emit_op_T_simm32(int op, int tmp, SDWORD immData)
{
    if (isReg(tmp)) {
        X86_emit_op_R_simm32(op, tmpReg(tmp), immData);
    }
    else {
        X86_emit_op_M_simm32(op, TMP(tmp), immData);
    }
} /*X86_emit_op_T_simm32*/

void X86_emit_op_Ti_simm32(int op, int tmp, SDWORD immData)
{
    if (isReg(tmp)) {
        X86_emit_op_M_simm32(op, tmpReg(tmp), no_reg, 0, 0, immData);
    }
    else {
        X86_emit_op_RM(MOV_D, eax, TMP(tmp));
        X86_emit_op_M_simm32(op, eax, no_reg, 0, 0, immData);
    }
} /*X86_emit_op_Ti_simm32*/

void X86_emit_op_R_simm32(int op, int reg, SDWORD immData)
{
    // op encoding: 000 ADD (low 3 bits)
    //              001 OR
    //              010 ADC
    //              011 SBB
    //              100 AND
    //              101 SUB
    //              110 XOR
    //              111 CMP
    if (NP(op) == IMULI_D) {
        X86_emit_op_RR_simm32(op, reg, reg, immData);
    }
    else if (immData >= -128 && immData <= 127) {
        //use sign-extended byte immediate data
		X86_emit_prefixes(OC2(op) | 2, 0, reg, 0);	//s=1	//REX.B extends reg
        X86_emit_opc(OC2(op) | 2);					//s=1
        X86_emit_RR(op & 7, reg);
        X86_emit_byte((BYTE)immData);
    }
    else if (reg == eax) {
        //use short form (immediate with accumulator)
        //saves 1 byte...
        ///X86_emit_byte( (BYTE)(((op & 7) << 3) | 0x04 | ((op >> 8) & 1)) );		//RESUME HERE!!!!
		X86_emit_prefixes(PF(op), 0, 0, 0);	//eax is implicit	//x64 RAX???
        X86_emit_opc( (BYTE)(((op & 7) << 3) | 0x04 | ((op >> 8) & 1)) );	//concoct eax short form
        X86_emit_dword(immData);								//should emit word16 for 16-bit op???
    }
    else {
        //use full 32-bit immediate data						//or 16-bit for 16-bit op???			
		X86_emit_prefixes(OC2(op), 0, reg, 0);	//s=0	//REX.B extends reg
        X86_emit_opc(OC2(op));					//s=0
        X86_emit_RR(op & 7, reg);
        X86_emit_dword(immData);					//should emit word16 for 16-bit op???
    }
} /*X86_emit_op_R_simm32*/

/*---------------------------------------*/
/*    LABELS / REFS / DEFS               */
/*---------------------------------------*/

CGL X86_genlab(void)
{
    if (cbp->next_cgl >= cbp->max_cgl) {
		//(re-)allocate the CGL_INFO array if necessary...
        CGL_INFO *temp;
        cbp->max_cgl += 128;
        temp = realloc(cbp->cgls, cbp->max_cgl * sizeof(*cbp->cgls));
        if (temp == NULL) {
#if 0       //unnecessary...
            free(cbp->cgls);
            cbp->cgls = NULL;
            cbp->max_cgl = 0;
#endif
            X86_abort2("Not enough memory for %d compiler-generated labels", cbp->max_cgl);
        }
        else {
            cbp->cgls = temp;
        }
    }
	//init a new GGL_INFO array element
    cbp->cgls[cbp->next_cgl].defined = FALSE;
    cbp->cgls[cbp->next_cgl].addr = 0; //NULL;
    cbp->cgls[cbp->next_cgl].fixup_list = 0; //NULL;
    cbp->cgls[cbp->next_cgl].isDisp8Fixup = FALSE;
    cbp->cgls[cbp->next_cgl].isAbsVal = FALSE;
    return cbp->next_cgl++;
} /*X86_genlab*/

void X86_deflab(CGL lab)
{
    // Define pc-relative displacement to this label.

    if (lab >= cbp->next_cgl) X86_abort2("Invalid CGL (%d.) in X86_deflab", lab);
    if (cbp->cgls[lab].defined)
		X86_abort2("Multiple definition of CGL (%d.) in X86_deflab", lab);
    // Process the fixup list...
    if (cbp->cgls[lab].isDisp8Fixup) {
        //Single disp8 fixup
#ifdef _WIN64
		int disp;
        ADDR pB = cbp->cgls[lab].fixup_list; //index of byte to be fixed up
        X86_ASSERT(pB != 0); // at least ONE fixup!
        disp = (cbp->pc - (pB + 1));	//rip-relative fixup
        if (disp < -128 || disp > 127)
            X86_abort2("Fixup target exceeds disp8 range (%d.)", disp);
        cbp->pcStart[pB] = disp;	//plant the fixup
#else //Win32
	    int disp;
        BYTE *pB = (BYTE *)cbp->cgls[lab].fixup_list; //ptr to byte to be fixed up
        X86_ASSERT(pB != NULL); // at least ONE fixup!
        disp = (int)(cbp->pc - (pB + 1));
        if (disp < -128 || disp > 127)
            X86_abort2("Fixup target exceeds disp8 range (%d.)", disp);
        *pB = disp;
#endif
    }
    else {
        //Chain of disp32 fixups
#ifdef _WIN64
		//X64 -- need to change ptr list to offset32 list
		ADDR p;
		ADDR q;
        for (p = cbp->cgls[lab].fixup_list; p != 0; ) {
            q = *(ADDR *)&cbp->pcStart[p];	//link to previous fixup
            *(ADDR *)&cbp->pcStart[p] = cbp->pc - (p + sizeof(ADDR)); //rip-relative fixup
            p = q;
        }
#else //Win32
		ADDR *p;
		ADDR *q;
        for (p = cbp->cgls[lab].fixup_list; p != NULL; ) {
            q = (ADDR *)*p;
            *p = (ADDR)(cbp->pc - (BYTE *)(p + 1));	//Note: scaled ptr math for (p+1)
            p = q;
        }
#endif
    }
    cbp->cgls[lab].addr = cbp->pc;	//save payload for later reference
    cbp->cgls[lab].defined = TRUE;
} /*X86_deflab*/

void X86_def_abs32(CGL lab, int ival32)
{   
    // Define absolute value for this label.

    if (lab >= cbp->next_cgl) X86_abort2("Invalid CGL (%d.) in X86_def_abs32", lab);
    if (cbp->cgls[lab].defined)
		X86_abort2("Multiple definition of CGL (%d.) in X86_def_abs32", lab);
	// Process the fixup list...
    if (cbp->cgls[lab].isDisp8Fixup) {
		//Single 8-bit fixup
#ifdef _WIN64
		ADDR pB;
        if (!cbp->cgls[lab].isAbsVal)
			X86_abort2("Relative disp8 fixup to absolute value for CGL (%d.)", lab);
        // Allow a single disp8 fixup for now..
        pB = cbp->cgls[lab].fixup_list; //index of byte to be fixed up
        X86_ASSERT(pB != 0); // at least ONE fixup!
        if (ival32 < -128 || ival32 > 255) {
            // a bit loosey-goosey at this point, since we don't know
            // whether ival32 should be signed or unsigned --
            // so caller should check for tighter bounds if it matters...
            X86_abort2("Fixup value exceeds 8-bit range (%d.)", ival32);
        }
        cbp->pcStart[pB] = ival32; // plant the payload (assume no further fixup chain)
#else //Win32
		BYTE *pB;
        if (!cbp->cgls[lab].isAbsVal)
			X86_abort2("Relative disp8 fixup to absolute value for CGL (%d.)", lab);
        // Allow a single disp8 fixup for now..
        pB = (BYTE *)cbp->cgls[lab].fixup_list; //ptr to byte to be fixed up
        X86_ASSERT(pB != NULL); // at least ONE fixup!
        if (ival32 < -128 || ival32 > 255) {
            // a bit loosey-goosey at this point, since we don't know
            // whether ival32 should be signed or unsigned --
            // so caller should check for tighter bounds if it matters...
            X86_abort2("Fixup value exceeds 8-bit range (%d.) in X86_def_abs32", ival32);
        }
        *pB = ival32;   // plant the payload (assume no further fixup chain)
#endif
    }
    else {
        //Chain of abs32 fixups
#ifdef _WIN64
					////// RESUME HERE!!!
	    ADDR p;
	    ADDR q;
        for (p = cbp->cgls[lab].fixup_list; p != 0; ) {
            q = *(ADDR *)&cbp->pcStart[p];	//link to previous fixup
            *(ADDR *)&cbp->pcStart[p] = ival32;	//abs32 fixup
            p = q;
        }
#else //Win32
	    ADDR *p;
	    ADDR *q;
        for (p = cbp->cgls[lab].fixup_list; p != NULL; ) {
            q = (ADDR *)*p;
            *p = (ADDR) ival32;
            p = q;
        }
#endif
    }
    cbp->cgls[lab].addr = (ADDR) ival32;  //save payload for later reference
    cbp->cgls[lab].defined = TRUE;
    cbp->cgls[lab].isAbsVal = TRUE;
} /*X86_def_abs32*/

ADDR X86_lab_offset(CGL lab)
{
    if (lab >= cbp->next_cgl) X86_abort2("Invalid CGL (%d.) in X86_lab_offset", lab);
    if (cbp->cgls[lab].defined) {
        return cbp->cgls[lab].addr;
    }
    else {
        X86_abort2("Offset of undefined CGL (%d.)", lab);
        return 0;
    }
} /*X86_lab_offset*/

int X86_lab_disp_size(CGL lab, int nOpcodeBytes)
{
    int disp;
    if (lab >= cbp->next_cgl) X86_abort2("Invalid CGL (%d.) in X86_lab_disp_size", lab);
    if (cbp->cgls[lab].defined) {
        //cgl already defined.  Is it within disp8 range?
        disp = (int)(cbp->cgls[lab].addr - (cbp->pc+nOpcodeBytes+1));	//+1 for the disp8 byte
        if (disp >= -128 && disp <= 127) {
            //can use disp8
            return 1;
        }
        else {
            //use disp32
            return 4;
        }
    }
    else {
        //cgl not yet defined -- use disp32.
        return 4;
    }
} /*X86_lab_disp_size*/

void X86_emit_lab_disp32(CGL lab)
{
    if (lab >= cbp->next_cgl) X86_abort2("Invalid CGL (%d.) in X86_emit_lab_disp32", lab);
    //if (X86_lab_disp_size(lab, 0) != 4) /*0 since already beyond code bytes*/
    //    X86_warn("Non-optimal displacement to CGL");
    if (cbp->cgls[lab].defined) {
        X86_emit_dword((DWORD)(cbp->cgls[lab].addr - (cbp->pc+4)));
    }
    else {
        //cgl not yet defined -- add to fixup list
        if (cbp->cgls[lab].isDisp8Fixup)
            X86_abort2("Disp32 fixup after disp8 fixup to CGL (0x%8.8x)", lab);
        X86_emit_dword((DWORD)cbp->cgls[lab].fixup_list);
#ifdef _WIN64
        cbp->cgls[lab].fixup_list = cbp->pc-4;	//new head of fixup list
#else //Win32
        cbp->cgls[lab].fixup_list = (ADDR *)(cbp->pc-4);	//new head of fixup list
#endif
    }
} /*X86_emit_lab_disp32*/

void X86_emit_lab_disp8(CGL lab)
{
    int disp;
    if (lab >= cbp->next_cgl) X86_abort2("Invalid CGL (%d.) in X86_emit_lab_disp8", lab);
    //X86_ASSERT(X86_lab_disp_size(lab, 0) == 1); //0 since beyond code bytes
    if (cbp->cgls[lab].defined) {
        disp = (int)(cbp->cgls[lab].addr - (cbp->pc+1)); //+1 for the disp8 byte
        if (disp < -128 || disp > 127)
            X86_abort2("Out of range of disp8 (%d.)", disp);
        X86_emit_byte((BYTE)disp);	//emit disp8 byte
    }
    else {
		//lab not yet defined -- emit a fixup
        if (cbp->cgls[lab].isDisp8Fixup)
            X86_abort2("More than one disp8 fixup for CGL (%d.) in X86_emit_lab_disp8", lab);
        if (cbp->cgls[lab].fixup_list != 0)
            X86_abort2("Disp8 fixup after disp32 fixup for CGL (%d.) in X86_emit_lab_disp8", lab);
        //Special 1-time fixup for disp8 forward reference.
        cbp->cgls[lab].isDisp8Fixup = TRUE;
#ifdef _WIN64
        cbp->cgls[lab].fixup_list = cbp->pc;
#else //Win32
        cbp->cgls[lab].fixup_list = (ADDR *)cbp->pc;
#endif
        X86_emit_byte(0);	//emit placeholder for disp8 byte
    }
} /*X86_emit_lab_disp8*/

void X86_emit_op_lab_disp32(int op, CGL lab)
{
    if (OC2M(op) == 0) X86_abort2("No disp32 form of op (0x%8.8x)", op);
    X86_emit_op(OC2(op));
    X86_emit_lab_disp32(lab);
} /*X86_emit_op_lab_disp32*/

void X86_emit_op_lab_disp8(int op, CGL lab)
{
    if (OC1M(op) == 0) X86_abort2("No disp8 form of op (0x%8.8x)", op);
    X86_emit_op(OC1(op));
    X86_emit_lab_disp8(lab);
} /*X86_emit_op_lab_disp8*/

void X86_emit_op_lab(int op, CGL lab)	//RESUME X64 VETTING
{
    if (OC1M(op) == 0) {
        //no disp8 form (e.g., CALL)
        X86_emit_op_lab_disp32(op, lab);
    }
    else if (OC2M(op) == 0) {
        //no disp32 form (e.g., JECXZ, LOOP, JMP_SHORT...)
        X86_emit_op_lab_disp8(op, lab);
    }
    else if (X86_lab_disp_size(lab, 1) == 1) {	//X64???
        //can reach with 8-bit displacement
        X86_emit_op_lab_disp8(op, lab);
    }
    else {
        //resort to disp32
        X86_emit_op_lab_disp32(op, lab);
    }
} /*X86_emit_op_lab*/

void X86_emit_op_R_ref_simm32(int op, int reg, CGL lab)
{
    // Emit an op_R_simm32 instruction where the simm32 value
    // is a (usually forward-referenced) label.
    if (lab >= cbp->next_cgl)
		X86_abort2("Invalid CGL (%d.) in X86_emit_op_R_ref_simm32", lab);
    if (cbp->cgls[lab].defined) {
        // Label is already defined...
        if (!cbp->cgls[lab].isAbsVal)
			X86_abort2("Wrong CGL reference (%d.) in X86_emit_op_R_ref_simm32", lab);
        // Optimize to instruction with 8- or 32-bit immediate value...				//16bit??
        X86_emit_op_R_simm32(op, reg, (SDWORD)cbp->cgls[lab].addr);
    }
    else {
        // Forward reference to label value...
        // must use imm32 form of instruction.				//16bit??
        if (cbp->cgls[lab].isDisp8Fixup)
			X86_abort2("Wrong fixup type for CGL (%d.) in X86_emit_op_R_ref_simm32", lab);
        cbp->cgls[lab].isAbsVal = TRUE;
        X86_emit_op_R(op, reg);
#if 1 //either method should work...
        X86_emit_dword((DWORD)cbp->cgls[lab].fixup_list);  //link to previous fixup if any
#ifdef _WIN64
        cbp->cgls[lab].fixup_list = (cbp->pc-4);	//-4 for dword just emitted
#else //Win32
        cbp->cgls[lab].fixup_list = (ADDR *)(cbp->pc-4);	//-4 for dword just emitted
#endif
#else //this will also work...
        { //alternate method doesn't depend on size of emitted fixup...
#ifdef _WIN64
			ADDR fix_prev = cbp->cgls[lab].fixup_list;
			cbp->cgls[lab].fixup_list = cbp->pc;
			X86_emit_dword((DWORD)fix_prev);
#else //Win32
			ADDR *fix_prev = cbp->cgls[lab].fixup_list;
			cbp->cgls[lab].fixup_list = (ADDR *)cbp->pc;
			X86_emit_dword((DWORD)fix_prev);
#endif
        }
#endif
    }
} /*X86_emit_op_R_ref_simm32*/

#if 0	//unused
void X86_emit_op_RM_ref_simm32(int op,
							   int reg,
							   int base_reg, int index_reg, int scaleFactor, CGL lab)
{
    X86_abort("X86_emit_op_RM_ref_simm32 is not yet implemented");
} /*X86_emit_op_RM_ref_simm32*/
#endif	//unused

#if 0	//unused
void X86_emit_op_R_ref_imm32(int op, int reg, CGL lab)
{
    X86_abort("X86_emit_op_R_ref_imm32 is not yet implemented");
} /*X86_emit_op_R_ref_imm32*/
#endif	//unused

#if 0	//unused
void X86_emit_op_R_ref_imm8(int op, int reg, CGL lab)
{
    X86_abort("X86_emit_op_R_ref_imm8 is not yet implemented");
} /*X86_emit_op_R_ref_imm8*/
#endif	//unused

// Align code buffer pc -- assumes alignment of pcStart >=
// requested alignment of pc.
// Notes on multibyte NOPs:
// Agner Fog says 0F 1F mod-000-rm
// is a variable length NOP available on Pentium Pro and up.

// http://www.asmpedia.org/index.php?title=NOP lists the following
// NOPs for P6+
//
// 90              nop
// 6690            xchg    ax,ax ; 66: switch to 16-bit operand 90: opcode 
// 0f1f00          nop     dword ptr [eax] ; 0f1f: 2-byte opcode 00: mod=00 reg=000 rm=000 [EAX]
// 0f1f4000        nop     dword ptr [eax] ; 0f1f: 2-byte opcode 40: mod=01 reg=000 rm=000 [EAX+0x00]
// 0f1f440000      nop     dword ptr [eax+eax] ; 0f1f: 2-byte opcode 44: mod=01 reg=000 rm=100 SIB + 0x00
// 660f1f440000    nop     word ptr [eax+eax] ; 66: switch to 16-bit operand 0f1f: 2-byte opcode 44: mod=01 reg=000 rm=100 SIB + 0x00
// 0f1f8000000000  nop     dword ptr [eax] ; 0f1f: 2-byte opcode 80: mod=10 reg=000 rm=000 [EAX+0x00000000]

// http://stackoverflow.com/questions/2052698/encode-data-in-a-multi-byte-noop
// lists the following 9-byte AT&T format NOP:
//
// 48 0f 1f 84 00 ef cd ab 12  nopq 0x12abcdef(%rax,%rax,1)

//  Intel Architecture Software developer's guide, volume 2B (instructions N-Z)
//  lists:
//
//  2 bytes   66 NOP                                     66 90H
//  3 bytes   NOP DWORD ptr [EAX]                        0F 1F 00H
//  4 bytes   NOP DWORD ptr [EAX + 00H]                  0F 1F 40 00H
//  5 bytes   NOP DWORD ptr [EAX + EAX*1 + 00H]          0F 1F 44 00 00H
//  6 bytes   66 NOP DWORD ptr [EAX + EAX*1 + 00H]       66 0F 1F 44 00 00H
//  7 bytes   NOP DWORD ptr [EAX + 00000000H]            0F 1F 80 00 00 00 00H
//  8 bytes   NOP DWORD ptr [EAX + EAX*1 + 00000000H]    0F 1F 84 00 00 00 00 00H
//  9 bytes   66 NOP DWORD ptr [EAX + EAX*1 + 00000000H] 66 0F 1F 84 00 00 00 00 00H
// 
/***
> The Intel Arch Optimization Guide recommends the following regarding NOPs:
>
> 3.5.1.8 Using NOPs
> Code generators generate a no-operation (NOP) to align instructions.
> Examples of NOPs of different lengths in 32-bit mode are shown below:
>
>    1-byte: XCHG EAX, EAX
>    2-byte: 66 NOP
>    3-byte: LEA REG, 0 (REG) (8-bit displacement)
>    4-byte: NOP DWORD PTR [EAX + 0] (8-bit displacement)
>    5-byte: NOP DWORD PTR [EAX + EAX*1 + 0] (8-bit displacement)
>    6-byte: LEA REG, 0 (REG) (32-bit displacement)
>    7-byte: NOP DWORD PTR [EAX + 0] (32-bit displacement)
>    8-byte: NOP DWORD PTR [EAX + EAX*1 + 0] (32-bit displacement)
>    9-byte: NOP WORD  PTR [EAX + EAX*1 + 0] (32-bit displacement)
***/


// From http://mail.openjdk.java.net/pipermail/hotspot-compiler-dev/2010-September/003882.html
//
// Using multi-bytes nops "0x0F 0x1F [address]" for Intel
//  1: 0x90
//  2: 0x66 0x90
//  3: 0x66 0x66 0x90 (don't use "0x0F 0x1F 0x00" - need patching safe padding)
//  4: 0x0F 0x1F 0x40 0x00
//  5: 0x0F 0x1F 0x44 0x00 0x00
//  6: 0x66 0x0F 0x1F 0x44 0x00 0x00
//  7: 0x0F 0x1F 0x80 0x00 0x00 0x00 0x00
//  8: 0x0F 0x1F 0x84 0x00 0x00 0x00 0x00 0x00
//  9: 0x66 0x0F 0x1F 0x84 0x00 0x00 0x00 0x00 0x00
// 10: 0x66 0x66 0x0F 0x1F 0x84 0x00 0x00 0x00 0x00 0x00
// 11: 0x66 0x66 0x66 0x0F 0x1F 0x84 0x00 0x00 0x00 0x00 0x00

// The rest coding is Intel specific - don't use consecutive address nops

// 12: 0x0F 0x1F 0x84 0x00 0x00 0x00 0x00 0x00 0x66 0x66 0x66 0x90
// 13: 0x66 0x0F 0x1F 0x84 0x00 0x00 0x00 0x00 0x00 0x66 0x66 0x66 0x90
// 14: 0x66 0x66 0x0F 0x1F 0x84 0x00 0x00 0x00 0x00 0x00 0x66 0x66 0x66 0x90
// 15: 0x66 0x66 0x66 0x0F 0x1F 0x84 0x00 0x00 0x00 0x00 0x00 0x66 0x66 0x66 0x90

/***
>      while(i>= 15) {
>        // For Intel don't generate consecutive address nops (mix with regular nops)
>        i -= 15;
>        emit_byte(0x66);   // size prefix
>        emit_byte(0x66);   // size prefix
>        emit_byte(0x66);   // size prefix
>        addr_nop_8();
>        emit_byte(0x66);   // size prefix
>        emit_byte(0x66);   // size prefix
>        emit_byte(0x66);   // size prefix
>        emit_byte(0x90);   // nop
>      }
>      switch (i) {
>        case 14:
>          emit_byte(0x66); // size prefix
>        case 13:
>          emit_byte(0x66); // size prefix
>        case 12:
>          addr_nop_8();
>          emit_byte(0x66); // size prefix
>          emit_byte(0x66); // size prefix
>          emit_byte(0x66); // size prefix
>          emit_byte(0x90); // nop
>          break;
>        case 11:
>          emit_byte(0x66); // size prefix
>        case 10:
>          emit_byte(0x66); // size prefix
>        case 9:
>          emit_byte(0x66); // size prefix
>        case 8:
>          addr_nop_8();
>          break;
>        case 7:
>          addr_nop_7();
>          break;
>        case 6:
>          emit_byte(0x66); // size prefix
>        case 5:
>          addr_nop_5();
>          break;
>        case 4:
>          addr_nop_4();
>          break;
>        case 3:
>          // Don't use "0x0F 0x1F 0x00" - need patching safe padding
>          emit_byte(0x66); // size prefix
>        case 2:
>          emit_byte(0x66); // size prefix
>        case 1:
>          emit_byte(0x90); // nop
>          break;
>        default:
>          assert(i == 0, " ");
>      }
>      return;
>    }
>
>
> void Assembler::addr_nop_4() {
>    // 4 bytes: NOP DWORD PTR [EAX+0]
>    emit_byte(0x0F);
>    emit_byte(0x1F);
>    emit_byte(0x40); // emit_rm(cbuf, 0x1, EAX_enc, EAX_enc);
>    emit_byte(0);    // 8-bits offset (1 byte)
> }
>
> void Assembler::addr_nop_5() {
>    // 5 bytes: NOP DWORD PTR [EAX+EAX*0+0] 8-bits offset
>    emit_byte(0x0F);
>    emit_byte(0x1F);
>    emit_byte(0x44); // emit_rm(cbuf, 0x1, EAX_enc, 0x4);
>    emit_byte(0x00); // emit_rm(cbuf, 0x0, EAX_enc, EAX_enc);
>    emit_byte(0);    // 8-bits offset (1 byte)
> }
>
> void Assembler::addr_nop_7() {
>    // 7 bytes: NOP DWORD PTR [EAX+0] 32-bits offset
>    emit_byte(0x0F);
>    emit_byte(0x1F);
>    emit_byte(0x80); // emit_rm(cbuf, 0x2, EAX_enc, EAX_enc);
>    emit_long(0);    // 32-bits offset (4 bytes)
> }
>
> void Assembler::addr_nop_8() {
>    // 8 bytes: NOP DWORD PTR [EAX+EAX*0+0] 32-bits offset
>    emit_byte(0x0F);
>    emit_byte(0x1F);
>    emit_byte(0x84); // emit_rm(cbuf, 0x2, EAX_enc, 0x4);
>    emit_byte(0x00); // emit_rm(cbuf, 0x0, EAX_enc, EAX_enc);
>    emit_long(0);    // 32-bits offset (4 bytes)
> }
>
>
***/

/***
From http://compgroups.net/comp.lang.asm.x86/multi-byte-nop-s/147243

90						NOP
89 C0					MOV EAX,EAX
8D 40 00				LEA EAX,[    EAX + (byte)0 ]
3E 8D 40 00				LEA EAX,[ DS:EAX + (byte)0 ]
3E 66 8D 40 00			LEA AX, [ DS:EAX + (byte)0 ]
8D 80 00 00 00 00		LEA EAX,[    EAX + (dword)0 ]
3E 8D 80 00 00 00 00	LEA EAX,[ DS:EAX + (dword)0 ]
3E 66 8D 80 00 00 00 00	LEA AX, [ DS:EAX + (dword)0 ]

***/

#ifdef _WIN64
// From http://mail.openjdk.java.net/pipermail/hotspot-compiler-dev/2010-September/003882.html
//
// Helper routines for X86_align
static void addr_nop_4() {
    // 4 bytes: NOP DWORD PTR [EAX+0]
    X86_emit_byte(0x0F);
    X86_emit_byte(0x1F);
    X86_emit_byte(0x40); // emit_rm(cbuf, 0x1, EAX_enc, EAX_enc);
    X86_emit_byte(0);    // 8-bits offset (1 byte)
}

static void addr_nop_5() {
    // 5 bytes: NOP DWORD PTR [EAX+EAX*0+0] 8-bits offset
    X86_emit_byte(0x0F);
    X86_emit_byte(0x1F);
    X86_emit_byte(0x44); // emit_rm(cbuf, 0x1, EAX_enc, 0x4);
    X86_emit_byte(0x00); // emit_rm(cbuf, 0x0, EAX_enc, EAX_enc);
    X86_emit_byte(0);    // 8-bits offset (1 byte)
}

static void addr_nop_7() {
    // 7 bytes: NOP DWORD PTR [EAX+0] 32-bits offset
    X86_emit_byte(0x0F);
    X86_emit_byte(0x1F);
    X86_emit_byte(0x80); // emit_rm(cbuf, 0x2, EAX_enc, EAX_enc);
    X86_emit_dword(0);   // 32-bits offset (4 bytes)
}

static void addr_nop_8() {
    // 8 bytes: NOP DWORD PTR [EAX+EAX*0+0] 32-bits offset
    X86_emit_byte(0x0F);
    X86_emit_byte(0x1F);
    X86_emit_byte(0x84); // emit_rm(cbuf, 0x2, EAX_enc, 0x4);
    X86_emit_byte(0x00); // emit_rm(cbuf, 0x0, EAX_enc, EAX_enc);
    X86_emit_dword(0);   // 32-bits offset (4 bytes)
}
#endif

void X86_align(int align)
{
#ifdef _WIN64
	// 64-bit cpus also support the multibyte NOP, which
	// we can make use of here...
	// From
	// http://mail.openjdk.java.net/pipermail/hotspot-compiler-dev/2010-September/003882.html

    int i = (align - (cbp->pc) % align) % align;
	while (i >= 15) {
		// For Intel don't generate consecutive address nops (mix with regular nops)
		i -= 15;
		X86_emit_byte(0x66);   // size prefix
		X86_emit_byte(0x66);   // size prefix
		X86_emit_byte(0x66);   // size prefix
		addr_nop_8();
		X86_emit_byte(0x66);   // size prefix
		X86_emit_byte(0x66);   // size prefix
		X86_emit_byte(0x66);   // size prefix
		X86_emit_byte(0x90);   // nop
	}
	switch (i) {
	case 14:
		X86_emit_byte(0x66);   // size prefix
	case 13:
		X86_emit_byte(0x66);   // size prefix
	case 12:
		addr_nop_8();
		X86_emit_byte(0x66);   // size prefix
		X86_emit_byte(0x66);   // size prefix
		X86_emit_byte(0x66);   // size prefix
		X86_emit_byte(0x90);   // nop
		break;
	case 11:
		X86_emit_byte(0x66);   // size prefix
	case 10:
		X86_emit_byte(0x66);   // size prefix
	case 9:
		X86_emit_byte(0x66);   // size prefix
	case 8:
		addr_nop_8();
		break;
	case 7:
		addr_nop_7();
		break;
	case 6:
		X86_emit_byte(0x66);   // size prefix
	case 5:
		addr_nop_5();
		break;
	case 4:
		addr_nop_4();
		break;
	case 3:
		// Don't use "0x0F 0x1F 0x00" - need patching safe padding
		X86_emit_byte(0x66);   // size prefix
	case 2:
		X86_emit_byte(0x66);   // size prefix
	case 1:
		X86_emit_byte(0x90);   // nop
		break;
	default:
		X86_ASSERT(i == 0);
	}

#else //Win32
	// FM32 must support older cpus that don't support multibyte NOP;
	// so just use a brute force series of NOP (90)'s.
	int i;
    int bytesNeeded = (align - (int)(cbp->pc - cbp->pcStart) % align) % align;
    switch (bytesNeeded) {
    case 1:
        X86_emit_op(NOP);
        break;
    // fill in other special cases (2-byte nops, etc)...
    default:
        for (i = 0; i < bytesNeeded; i++) {
            X86_emit_op(NOP);
        }
        break;
    } /*switch*/
#ifdef _WIN64
    X86_ASSERT((cbp->pc) % align == 0);
#else //Win32
    X86_ASSERT((cbp->pc - cbp->pcStart) % align == 0);
#endif
#endif //Win32
} /*X86_align*/

static void X86_disassemble_code(void *cbp)
{
    struct cpustate_struc state;
    int i, j;
    char oucs[128];
    UINT2 nb, nc;
    UINT4 flags;
    UINT4 iss;
    BYTE *cb = cbp;
    int nBytes = 16;
    int lines = 0;

    int fInvalid = FALSE;
    int fPrivileged = FALSE;
    int fFpuCode = FALSE;
    int fIo = FALSE;
    int fTrap = FALSE;
    int fFarCode = FALSE;
    int fStringCode = FALSE;
    int fOpSizeOverride = FALSE;
    int fAddrSizeOverride = FALSE;
    int reply;

	UINT4 iopts = 8; //Use32
	if (X64GEN) iopts = 16; //Use64

    for (i = 0; i < nBytes; i += nb) {
        iss = Mda_disassemble(&cb[i], nBytes - i, 16, iopts,
                              &state, oucs, sizeof(oucs), &nb, &nc, &flags);
        if (!(iss & 1)) {
            //partial instruction .. stop here
            break;
        }
        PRINTF2("%8.8p:", &cb[i]);
        for (j = 0; j < nb; j++) {
            PRINTF2("%3.2x", cb[i+j]);
        }
#ifdef FML
        PRINTF3("\r\n  %.*s", nc, oucs);
        reply = msgBox(MB_ICONINFORMATION|MB_OKCANCEL,
                       "FilterMeister Run-time Code Disassembly",
                       "%.*s", nc, oucs);
        if (reply == IDCANCEL) break;
#else
        PRINTF3("\t%.*s", nc, oucs);
#endif
#ifndef BETA
#if 1
        /* Output info from flags while debugging... */
        if (flags != 0) {
            if (flags & 0x00000001) {
                PRINTF("; INVALID");
                fInvalid = TRUE;
            }
            if (flags & 0x00000002) {
				// disable interrupts
                PRINTF("; privileged");
                fPrivileged = TRUE;
            }
            if (flags & 0x00000004) {
                PRINTF("; seg override");
                fFarCode = TRUE;
            }
            if (flags & 0x00000008) {
                PRINTF("; repeat");
                fStringCode = TRUE;
            }
            if (flags & 0x00000010) {
				// lock -- privileged???
                PRINTF("; privileged");
                fPrivileged = TRUE;
            }
            if (flags & 0x00000020) {
                PRINTF("; inhibits single-step trap");
            }
            if (flags & 0x00000040) {
                //PRINTF("; FPU instruction");
                fFpuCode = TRUE;
            }
            if (flags & 0x00000080) {
                //PRINTF("; wait");
                fFpuCode = TRUE;
            }
            if (flags & 0x00000100) {
				// halt
                PRINTF("; privileged");
                fPrivileged = TRUE;
            }
            if (flags & 0x00000200) {
                PRINTF("; privileged i/o");
                fPrivileged = TRUE;
                fIo = TRUE;
            }
            if (flags & 0x00000400) {
                //PRINTF("; call");
            }
            if (flags & 0x00000800) {
                //PRINTF("; return");
            }
            if (flags & 0x00001000) {
                PRINTF("; interrupt 3");
                fTrap = TRUE;
            }
            if (flags & 0x00002000) {
                PRINTF("; interrupt n");
                fTrap = TRUE;
            }
            if (flags & 0x00004000) {
                PRINTF("; INTO/BOUND");
                fTrap = TRUE;
            }
            if (flags & 0x00008000) {
				// interrupt return
                PRINTF("; privileged");
                fPrivileged = TRUE;
            }
            if (flags & 0x00010000) {
                //PRINTF("; unconditional jump");
            }
            if (flags & 0x00020000) {
                //PRINTF("; conditional jump");
            }
            if (flags & 0x00040000) {
                PRINTF("; far");
                fFarCode = TRUE;
            }
            if (flags & 0x00080000) {
				// pop flags -- privileged???
                PRINTF("; privileged?");
                fPrivileged = TRUE; //well,...
            }
            if (flags & 0x00100000) {
                PRINTF("; operand size override");
                fOpSizeOverride = TRUE;
            }
            if (flags & 0x00200000) {
                PRINTF("; address size override");
                fAddrSizeOverride = TRUE;
            }
			if (flags & 0x00400000) {
				PRINTF("; redundant prefixes");
				fInvalid = TRUE;
			}
			if (flags & 0x00800000) {
				PRINTF("; misplaced REX prefix");
				fInvalid = TRUE;
			}
            if (flags & 0x0f000000) {
                PRINTF("; unknown flags (MBZ)");
                fInvalid = TRUE;
            }
            if (flags & 0xf0000000) {
				int prefixBytes = (flags >> 28) & 0xf;
                if (prefixBytes == 15) {
                    PRINTF("; more than 14 prefix bytes!!");
                    fInvalid = TRUE;
                }
                else {
                    PRINTF3("; %u prefix byte%s", prefixBytes, PLURAL(prefixBytes));
                }
            }
        }//if flags != 0
#endif
        PRINTF("\r\n");
#ifndef FML
        ++lines;
        if (lines > 40) {
            PAUSE;
            lines = 0;
        }
#endif
#endif //ifndef BETA
    } /*for*/
#ifndef BETA
    if (fPrivileged)        PRINTF("Privileged code was detected\r\n");
    if (fFpuCode)           PRINTF("Floating point code was detected\r\n");
    if (fIo)                PRINTF("I/O code was detected\r\n");
    if (fTrap)              PRINTF("Trap code was detected\r\n");
    if (fFarCode)           PRINTF("FAR (non-FLAT) code was detected\r\n");
    if (fStringCode)        PRINTF("String code was detected\r\n");
    if (fOpSizeOverride)    PRINTF("16-bit operand code was detected\r\n");
    if (fAddrSizeOverride)  PRINTF("16-bit addressing was detected!!!\r\n");
    if (fInvalid)           PRINTF("Invalid code was detected\r\n");
    PAUSE;
    PRINTF("\r\n");
#endif //ifndef BETA
} /*X86_disassemble_code*/

#endif //ifndef STANDALONE

static int show_reg(const char *regname, INT_PTR regval)
{
    return msgBox(MB_ICONINFORMATION|MB_OKCANCEL,
               "FilterMeister Register Contents",
#ifdef _WIN64
               "%s = %#16.16I64X", regname, regval);
#else
               "%s = %#8.8X", regname, regval);
#endif
} /*show_reg*/

static int show_flags(const char *regname, int regval)
{
    return msgBox(MB_ICONINFORMATION|MB_OKCANCEL,
               "FilterMeister Flags Register Contents",
               "%s = %#8.8X", regname, (DWORD)regval);
} /*show_flags*/

static int show_sreg(const char *regname, int regval)
{
    return msgBox(MB_ICONINFORMATION|MB_OKCANCEL,
               "FilterMeister Segment Register Contents",
               "%s = %#4.4X", regname, (WORD)regval);
} /*show_sreg*/

static int Eval_Exception(LPEXCEPTION_POINTERS ep)
{
    int n_except;
    unsigned char *p;
    int reply;
#ifdef _WIN64
#define FloatSave FltSave
    XMM_SAVE_AREA32 env387;  //Win64 context area for FPU/MMX.
#else
    FLOATING_SAVE_AREA env387;  //don't actually need the Cr0NpxState field, but it's harmless
#endif
#ifndef STANDALONE
    int top; // 87sw.Top
    int tag, R;
#endif //ifndef STANDALONE
    static const char *tag_type[] = {"valid", "zero", "special", "EMPTY"};
    static const char *precision_control[] = {"24 bits", "RESERVED", "53 bits", "64 bits"};
    static const char *rounding_control[] = {"round to nearest", "round down", "round up", "chop toward 0"};
#if !defined(BETA) && !defined(STANDALONE)
    int i, n;
    int opcode;
#endif

    n_except = ep->ExceptionRecord->ExceptionCode;

#if 1
    if ( n_except != STATUS_INTEGER_DIVIDE_BY_ZERO && 
         n_except != STATUS_INTEGER_OVERFLOW &&
         n_except != STATUS_ACCESS_VIOLATION &&
         n_except != STATUS_ILLEGAL_INSTRUCTION &&
         n_except != STATUS_ARRAY_BOUNDS_EXCEEDED &&
         n_except != STATUS_PRIVILEGED_INSTRUCTION &&
         n_except != STATUS_STACK_OVERFLOW &&
         n_except != STATUS_CONTROL_C_EXIT &&
         //n_except != STATUS_BREAKPOINT &&
         //n_except != STATUS_SINGLE_STEP &&
         n_except != STATUS_FLOAT_INVALID_OPERATION &&
         n_except != STATUS_FLOAT_DENORMAL_OPERAND &&
         n_except != STATUS_FLOAT_DIVIDE_BY_ZERO &&
         n_except != STATUS_FLOAT_INEXACT_RESULT &&
         n_except != STATUS_FLOAT_STACK_CHECK &&
         n_except != STATUS_FLOAT_UNDERFLOW &&
         n_except != STATUS_FLOAT_OVERFLOW )		// Pass on most
        return EXCEPTION_CONTINUE_SEARCH;			//  exceptions
#endif

    // abort() raises STATUS_CONTROL_C_EXIT...
    if (n_except == STATUS_CONTROL_C_EXIT)
        return EXCEPTION_EXECUTE_HANDLER;

#ifdef _WIN64
	// save FPU/MMX context and mask exceptions (via implicit fninit) so we don't
	// recurse if this is a f.p. exception
	// In a later version, we should use fxsave to save entire FPU/MMX/SSE context.
	// N.B. The x64 VC++ compiler does not suppport in-line assembly!
	////X86_fninit();			//mask any exceptions -- is this needed???
	X86_fnsave(&env387);	//save FPU/MMX/SSE2 status
#else
    // save NDP context and mask exceptions (via implicit fninit) so we don't recurse
    // if this is a f.p. exception...
    __asm {
        fnsave env387
    }
#endif

#if !defined(BETA) && !defined(STANDALONE)
    PRINTF("\r\n\r\n");
#endif

    switch (n_except) {
    case STATUS_INTEGER_DIVIDE_BY_ZERO:
        p = "Integer (zero-)divide error!";
        break;
    case STATUS_INTEGER_OVERFLOW:
        p = "Integer overflow!";
        break;
    case STATUS_ACCESS_VIOLATION:
        p = "Memory access violation!";
        break;
    case STATUS_ILLEGAL_INSTRUCTION:
        p = "Illegal instruction!";
        break;
    case STATUS_ARRAY_BOUNDS_EXCEEDED:
        p = "Array bounds exceeded!";
        break;
    case STATUS_PRIVILEGED_INSTRUCTION:
        p = "Privileged instruction!";
        break;
    case STATUS_STACK_OVERFLOW:
        p = "Stack overflow!";
        break;
    case STATUS_CONTROL_C_EXIT:
        p = "Control-C exit!";
        break;
    case STATUS_BREAKPOINT:
        p = "Breakpoint trap!";
        break;
    case STATUS_SINGLE_STEP:
        p = "Single step trap or debug fault!";
        break;
    case STATUS_FLOAT_INVALID_OPERATION:
        p = "Floating-point invalid operation!";
        break;
    case STATUS_FLOAT_DENORMAL_OPERAND:
        p = "Floating-point denormal operand!";
        break;
    case STATUS_FLOAT_DIVIDE_BY_ZERO:
        p = "Floating-point divide by zero!";
        break;
    case STATUS_FLOAT_INEXACT_RESULT:
        p = "Floating-point inexact result!";
        break;
    case STATUS_FLOAT_STACK_CHECK:
        // FP stack check: C1 = 0 ==> overflow, else underflow
        if (env387.StatusWord & 0x0200)
            p = "Floating-point stack overflow!";
        else
            p = "Floating-point stack underflow!";
        break;
    case STATUS_FLOAT_UNDERFLOW:
        p = "Floating-point underflow!";
        break;
    case STATUS_FLOAT_OVERFLOW:
        p = "Floating-point overflow!";
        break;
    default:
        p = "Unrecognized exception.";
        break;
    }
#if !defined(BETA) && !defined(STANDALONE)
    PRINTF(p);
    PRINTF("\r\n\r\n");
#endif
  

#if STANDALONE
    reply = msgBox(MB_ICONERROR|MB_OK,		//SHOULD DISPLAY AT LEAST A
                   "Run-time Exception",	//LITTLE MORE INFO FOR DEBUG!!
                   "%s\n\n"
                   "",
                   p);

    return EXCEPTION_EXECUTE_HANDLER;
#endif
//----------------------------------- should have #else //!STANDALONE  ??? #endif
    reply = msgBox(MB_ICONERROR|MB_YESNOCANCEL,
                   "FilterMeister Run-time Exception",
                   "%s\n\n"
                   "Display exception details?",
                   p);
    if (reply == IDCANCEL)
        return EXCEPTION_EXECUTE_HANDLER;
    if (reply == IDNO)
#if 0
        return EXCEPTION_CONTINUE_EXECUTION;
#else
        return EXCEPTION_EXECUTE_HANDLER;
#endif

#if !defined(BETA) && !defined(STANDALONE)
    PRINTF2("Exception code  = %8.8x\r\n", n_except);
    PRINTF2("Exception flags = %8.8x\r\n", ep->ExceptionRecord->ExceptionFlags);
    PRINTF2("Exception rec_p = %8.8p\r\n", ep->ExceptionRecord->ExceptionRecord);
    PRINTF2("Exception addr  = %8.8p\r\n", ep->ExceptionRecord->ExceptionAddress);
    PRINTF2("Exception nparm = %8.8x\r\n", n = ep->ExceptionRecord->NumberParameters);
    for (i = 0; i < n; i++) {
		ULONG_PTR parm = ep->ExceptionRecord->ExceptionInformation[i];
#ifdef _WIN64
        PRINTF3("Exception param %d = %16.16Ix\r\n", i, parm);
#else //Win32
        PRINTF3("Exception param %d = %8.8x\r\n", i, parm);
#endif
        if (n_except == STATUS_ACCESS_VIOLATION) {
            if (i == 0) {
                //first param is read/write flag
                //Win95 *always* sets this to 0 for vaddr in [0,0xa820]???
                if (parm == 0) {
                    PRINTF("  Read access violation\r\n");
                    //Info("Read access violation.");
                }
                else if (parm == 1) {
                    PRINTF("  Write access violation\r\n");
                    Info("Write access violation.");
                }
            }
            else if (i == 1) {
                //2nd param is virtual address that caused access violation
                //Win95 *always* sets this to -1 for vaddr in [0,0xa820]???
#ifdef _WIN64
                PRINTF2("  Virtual address = %16.16Ix\r\n", parm);
                if (parm != -1) Info("Virtual address = %#16.16IX.", parm);
#else //Win32
                PRINTF2("  Virtual address = %8.8x\r\n", parm);
                if (parm != -1) Info("Virtual address = %#8.8X.", parm);
#endif
            }
        }
    }

	if ((UINT_PTR)ep->ExceptionRecord->ExceptionAddress < MIN_CODE_ADDRESS) {
		//oops -- this would cause another GP, resulting in reentrance!!!
		PRINTF("\r\nException address is not a valid code address!\r\n");
		opcode = 0;
	}
	else {
		PRINTF("\r\nCode bytes at exception address:\r\n");
		p = (unsigned char *)ep->ExceptionRecord->ExceptionAddress;
		opcode = *p;	//save for use later
		for (i=0; i<8; i++) {
			PRINTF2("%3.2x", *p++);
		}
	}
    PRINTF("\r\n\r\n");
#endif //if !defined(BETA) && !defined(STANDALONE)


#ifndef STANDALONE
	if ((UINT_PTR)ep->ExceptionRecord->ExceptionAddress >= MIN_CODE_ADDRESS) {
	    X86_disassemble_code(ep->ExceptionRecord->ExceptionAddress);
	}
#endif //ifndef STANDALONE

#if !defined(BETA) && !defined(STANDALONE)
    PRINTF2("Context record address = %8.8p\r\n", ep->ContextRecord);
    PRINTF2("Context flags = %8.8x\r\n", ep->ContextRecord->ContextFlags);
    if ((ep->ContextRecord->ContextFlags & CONTEXT_DEBUG_REGISTERS) == CONTEXT_DEBUG_REGISTERS) {
        PRINTF2("dr0    = %8.8x\r\n", ep->ContextRecord->Dr0);
        PRINTF2("dr1    = %8.8x\r\n", ep->ContextRecord->Dr1);
        PRINTF2("dr2    = %8.8x\r\n", ep->ContextRecord->Dr2);
        PRINTF2("dr3    = %8.8x\r\n", ep->ContextRecord->Dr3);
        PRINTF2("dr6    = %8.8x\r\n", ep->ContextRecord->Dr6);
        PRINTF2("dr7    = %8.8x\r\n", ep->ContextRecord->Dr7);
    }
    if ((ep->ContextRecord->ContextFlags & CONTEXT_FLOATING_POINT) == CONTEXT_FLOATING_POINT) {

        /**** 87 Control Word ****/
        PRINTF2("87cw    = %4.4x\r\n", (WORD)ep->ContextRecord->FloatSave.ControlWord);
        if ((WORD)ep->ContextRecord->FloatSave.ControlWord != (WORD)env387.ControlWord) {
            PRINTF2("...but env387.cw = %4.4x!!!\r\n", (WORD)env387.ControlWord);
        }
        /**** individual bit fields thereof... ****/
        PRINTF2("87cw.RC = %d\r\n", (env387.ControlWord >> 10) & 3);
        PRINTF2("87cw.PC = %d\r\n", (env387.ControlWord >>  8) & 3);
        PRINTF2("87cw.PM = %d\r\n", (env387.ControlWord >>  5) & 1);
        PRINTF2("87cw.UM = %d\r\n", (env387.ControlWord >>  4) & 1);
        PRINTF2("87cw.OM = %d\r\n", (env387.ControlWord >>  3) & 1);
        PRINTF2("87cw.ZM = %d\r\n", (env387.ControlWord >>  2) & 1);
        PRINTF2("87cw.DM = %d\r\n", (env387.ControlWord >>  1) & 1);
        PRINTF2("87cw.IM = %d\r\n", (env387.ControlWord      ) & 1);

        /**** 87 Status Word ****/
        PRINTF2("87sw    = %4.4x\r\n", (WORD)ep->ContextRecord->FloatSave.StatusWord);
        if ((WORD)ep->ContextRecord->FloatSave.StatusWord != (WORD)env387.StatusWord) {
            PRINTF2("...but env387.sw = %4.4x!!!\r\n", (WORD)env387.StatusWord);
        }
        /**** individual bit fields thereof... ****/
        PRINTF2("87sw.Busy = %d\r\n", (env387.StatusWord >> 15) & 1);
        PRINTF2("87sw.Top  = %d\r\n", top = (env387.StatusWord >> 11) & 7);
        PRINTF2("87sw.C3   = %d\r\n", (env387.StatusWord >> 14) & 1);
        PRINTF2("87sw.C2   = %d\r\n", (env387.StatusWord >> 10) & 1);
        PRINTF2("87sw.C1   = %d\r\n", (env387.StatusWord >>  9) & 1);
        PRINTF2("87sw.C0   = %d\r\n", (env387.StatusWord >>  8) & 1);
        PRINTF2("87sw.ES   = %d\r\n", (env387.StatusWord >>  7) & 1);
        PRINTF2("87sw.SF   = %d\r\n", (env387.StatusWord >>  6) & 1);
        PRINTF2("87sw.PE   = %d\r\n", (env387.StatusWord >>  5) & 1);
        PRINTF2("87sw.UE   = %d\r\n", (env387.StatusWord >>  4) & 1);
        PRINTF2("87sw.OE   = %d\r\n", (env387.StatusWord >>  3) & 1);
        PRINTF2("87sw.ZE   = %d\r\n", (env387.StatusWord >>  2) & 1);
        PRINTF2("87sw.DE   = %d\r\n", (env387.StatusWord >>  1) & 1);
        PRINTF2("87sw.IE   = %d\r\n", (env387.StatusWord      ) & 1);

        /**** 87 Tag Word ****/
        PRINTF2("87tw    = %4.4x\r\n", (WORD)ep->ContextRecord->FloatSave.TagWord);
        if ((WORD)ep->ContextRecord->FloatSave.TagWord != (WORD)env387.TagWord) {
            PRINTF2("...but env387.tw = %4.4x!!!\r\n", (WORD)env387.TagWord);
        }

        PRINTF3("87cs:eip= %4.4x:%8.8x\r\n", (WORD)ep->ContextRecord->FloatSave.ErrorSelector,
                                             ep->ContextRecord->FloatSave.ErrorOffset);
        if ((WORD)ep->ContextRecord->FloatSave.ErrorSelector != (WORD)env387.ErrorSelector ||
            ep->ContextRecord->FloatSave.ErrorOffset != env387.ErrorOffset) {
            PRINTF3("...but env387.cs:eip= %4.4x:%8.8x!!!\r\n", (WORD)env387.ErrorSelector,
                                             env387.ErrorOffset);
        }

        PRINTF3("87ds:edo= %4.4x:%8.8x\r\n", (WORD)ep->ContextRecord->FloatSave.DataSelector,	//edo???
                                             ep->ContextRecord->FloatSave.DataOffset);
        if ((WORD)ep->ContextRecord->FloatSave.DataSelector != (WORD)env387.DataSelector ||
            ep->ContextRecord->FloatSave.DataOffset != env387.DataOffset) {
            PRINTF3("...but env387.ds:edo= %4.4x:%8.8x!!!\r\n", (WORD)env387.DataSelector,		//edo??
                                             env387.DataOffset);
        }

#ifdef _WIN64
		// What is the WIN64 equivalent of Cr0NpxState???
#else
		// FIXME: Cr0NpxState renamed to Spare0 in Win8.1 SDK, prevents compiling on VS2015
		//PRINTF2("87cr0   = %8.8x\r\n", ep->ContextRecord->FloatSave.Cr0NpxState);
#endif

#ifdef _WIN64
        for (i = 0; i < 8; i++) {
            PRINTF2("S%d=", i);
            PRINTF3("%4.4I64x%16.16I64x",
				ep->ContextRecord->FltSave.FloatRegisters[i].High & 0xFFFF, //hi 16 bits
				ep->ContextRecord->FltSave.FloatRegisters[i].Low);	//low 64 bits of tempreal
            PRINTF("\r\n");
            R = (i + top) & 7;
            tag = (env387.TagWord >> (2*R)) & 3;
            PRINTF3("   tag=%d (%s)\r\n", tag, tag_type[tag]);
        }
        PRINTF("\r\n");
#else
        for (i = 0; i < 8; i++) {
            PRINTF2("S%d=", i);
            PRINTF3("%2.2x%2.2x", ep->ContextRecord->FloatSave.RegisterArea[i*10 + 9],
                                  ep->ContextRecord->FloatSave.RegisterArea[i*10 + 8]);
            PRINTF3("%2.2x%2.2x", ep->ContextRecord->FloatSave.RegisterArea[i*10 + 7],
                                  ep->ContextRecord->FloatSave.RegisterArea[i*10 + 6]);
            PRINTF3("%2.2x%2.2x", ep->ContextRecord->FloatSave.RegisterArea[i*10 + 5],
                                  ep->ContextRecord->FloatSave.RegisterArea[i*10 + 4]);
            PRINTF3("%2.2x%2.2x", ep->ContextRecord->FloatSave.RegisterArea[i*10 + 3],
                                  ep->ContextRecord->FloatSave.RegisterArea[i*10 + 2]);
            PRINTF3("%2.2x%2.2x", ep->ContextRecord->FloatSave.RegisterArea[i*10 + 1],
                                  ep->ContextRecord->FloatSave.RegisterArea[i*10 + 0]);
            PRINTF("\r\n");
            R = (i + top) & 7;
            tag = (env387.TagWord >> (2*R)) & 3;
            PRINTF3("   tag=%d (%s)\r\n", tag, tag_type[tag]);
        }
        PRINTF("\r\n");
#endif

#if 0	//print the hardware version of the fpu regs
#ifdef _WIN64
        for (i = 0; i < 8; i++) {
            PRINTF2("r%d=", (i + top) & 7);
            PRINTF3("%4.4I64x%16.16I64x",
				env387.FloatRegisters[i].High & 0xFFFF, //hi 16 bits
				env387.FloatRegisters[i].Low);	//low 64 bits of tempreal
            PRINTF("\r\n");
        }
        PRINTF("\r\n");
#else
        for (i = 0; i < 8; i++) {
            PRINTF2("r%d=", (i + top) & 7);
            PRINTF3("%2.2x%2.2x", env387.RegisterArea[i*10 + 9],
                                  env387.RegisterArea[i*10 + 8]);
            PRINTF3("%2.2x%2.2x", env387.RegisterArea[i*10 + 7],
                                  env387.RegisterArea[i*10 + 6]);
            PRINTF3("%2.2x%2.2x", env387.RegisterArea[i*10 + 5],
                                  env387.RegisterArea[i*10 + 4]);
            PRINTF3("%2.2x%2.2x", env387.RegisterArea[i*10 + 3],
                                  env387.RegisterArea[i*10 + 2]);
            PRINTF3("%2.2x%2.2x", env387.RegisterArea[i*10 + 1],
                                  env387.RegisterArea[i*10 + 0]);
            PRINTF("\r\n");
        }
        PRINTF("\r\n");
#endif //Win32
#endif
    }//CONTEXT_FLOATING_POINT
#endif  //!defined(BETA) && !defined(STANDALONE)

    if ((ep->ContextRecord->ContextFlags & CONTEXT_SEGMENTS) == CONTEXT_SEGMENTS) {
#if !defined(BETA) && !defined(STANDALONE)
        PRINTF2("gs     = %4.4x\r\n", (WORD)ep->ContextRecord->SegGs);
        PRINTF2("fs     = %4.4x\r\n", (WORD)ep->ContextRecord->SegFs);
        PRINTF2("es     = %4.4x\r\n", (WORD)ep->ContextRecord->SegEs);
        PRINTF2("ds     = %4.4x\r\n", (WORD)ep->ContextRecord->SegDs);
#endif

        if (show_sreg("ds", ep->ContextRecord->SegDs) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_sreg("es", ep->ContextRecord->SegEs) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_sreg("fs", ep->ContextRecord->SegFs) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_sreg("gs", ep->ContextRecord->SegGs) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
    }
    if ((ep->ContextRecord->ContextFlags & CONTEXT_CONTROL) == CONTEXT_CONTROL) {
#if !defined(BETA) && !defined(STANDALONE)
#ifdef _WIN64
        PRINTF2("rbp    = %16.16I64x\r\n", ep->ContextRecord->Rbp);
        PRINTF2("rip    = %16.16I64x\r\n", ep->ContextRecord->Rip);
        PRINTF2("cs     = %4.4x\r\n",	   ep->ContextRecord->SegCs);
        PRINTF2("eflags = %8.8x\r\n",      ep->ContextRecord->EFlags);
        PRINTF2("rsp    = %16.16I64x\r\n", ep->ContextRecord->Rsp);
        PRINTF2("ss     = %4.4x\r\n",      ep->ContextRecord->SegSs);
#else	//Win32
        PRINTF2("ebp    = %8.8x\r\n", ep->ContextRecord->Ebp);
        PRINTF2("eip    = %8.8x\r\n", ep->ContextRecord->Eip);
        PRINTF2("cs     = %4.4x\r\n", (WORD)ep->ContextRecord->SegCs);
        PRINTF2("eflags = %8.8x\r\n", ep->ContextRecord->EFlags);
        PRINTF2("esp    = %8.8x\r\n", ep->ContextRecord->Esp);
        PRINTF2("ss     = %4.4x\r\n", (WORD)ep->ContextRecord->SegSs);
#endif	//Win32
#endif
    
#ifdef _WIN64
        if (show_sreg("ss", ep->ContextRecord->SegSs) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_sreg("cs", ep->ContextRecord->SegCs) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("rip", ep->ContextRecord->Rip) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_flags("eflags", ep->ContextRecord->EFlags) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("rsp", ep->ContextRecord->Rsp) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("rbp", ep->ContextRecord->Rbp) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
#else	//Win32
        if (show_sreg("ss", ep->ContextRecord->SegSs) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_sreg("cs", ep->ContextRecord->SegCs) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("eip", ep->ContextRecord->Eip) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_flags("eflags", ep->ContextRecord->EFlags) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("esp", ep->ContextRecord->Esp) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("ebp", ep->ContextRecord->Ebp) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
#endif	//Win32
    }
    if ((ep->ContextRecord->ContextFlags & CONTEXT_INTEGER) == CONTEXT_INTEGER) {
#if !defined(BETA) && !defined(STANDALONE)
#ifdef _WIN64
        PRINTF2("rdi    = %16.16I64x\r\n", ep->ContextRecord->Rdi);
        PRINTF2("rsi    = %16.16I64x\r\n", ep->ContextRecord->Rsi);
        PRINTF2("rbx    = %16.16I64x\r\n", ep->ContextRecord->Rbx);
        PRINTF2("rdx    = %16.16I64x\r\n", ep->ContextRecord->Rdx);
        PRINTF2("rcx    = %16.16I64x\r\n", ep->ContextRecord->Rcx);
        PRINTF2("rax    = %16.16I64x\r\n", ep->ContextRecord->Rax);
        PRINTF2("r8     = %16.16I64x\r\n", ep->ContextRecord->R8);
        PRINTF2("r9     = %16.16I64x\r\n", ep->ContextRecord->R9);
        PRINTF2("r10    = %16.16I64x\r\n", ep->ContextRecord->R10);
        PRINTF2("r11    = %16.16I64x\r\n", ep->ContextRecord->R11);
        PRINTF2("r12    = %16.16I64x\r\n", ep->ContextRecord->R12);
        PRINTF2("r13    = %16.16I64x\r\n", ep->ContextRecord->R13);
        PRINTF2("r14    = %16.16I64x\r\n", ep->ContextRecord->R14);
        PRINTF2("r15    = %16.16I64x\r\n", ep->ContextRecord->R15);
#else	//Win32
        PRINTF2("edi    = %8.8x\r\n", ep->ContextRecord->Edi);
        PRINTF2("esi    = %8.8x\r\n", ep->ContextRecord->Esi);
        PRINTF2("ebx    = %8.8x\r\n", ep->ContextRecord->Ebx);
        PRINTF2("edx    = %8.8x\r\n", ep->ContextRecord->Edx);
        PRINTF2("ecx    = %8.8x\r\n", ep->ContextRecord->Ecx);
        PRINTF2("eax    = %8.8x\r\n", ep->ContextRecord->Eax);
#endif	//Win32
#endif

#ifdef _WIN64
        if (show_reg("rax", ep->ContextRecord->Rax) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("rbx", ep->ContextRecord->Rbx) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("rcx", ep->ContextRecord->Rcx) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("rdx", ep->ContextRecord->Rdx) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("rdi", ep->ContextRecord->Rdi) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("rsi", ep->ContextRecord->Rsi) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("r8",  ep->ContextRecord->R8 ) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("r9",  ep->ContextRecord->R9 ) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("r10", ep->ContextRecord->R10) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("r11", ep->ContextRecord->R11) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("r12", ep->ContextRecord->R12) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("r13", ep->ContextRecord->R13) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("r14", ep->ContextRecord->R14) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("r15", ep->ContextRecord->R15) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
#else	//Win32
        if (show_reg("eax", ep->ContextRecord->Eax) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("ebx", ep->ContextRecord->Ebx) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("ecx", ep->ContextRecord->Ecx) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("edx", ep->ContextRecord->Edx) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("edi", ep->ContextRecord->Edi) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
        if (show_reg("esi", ep->ContextRecord->Esi) == IDCANCEL) return EXCEPTION_EXECUTE_HANDLER;
#endif	//Win32

#if !defined(BETA) && !defined(STANDALONE)
        if (opcode == 0xf6) {					//X64???
#ifdef _WIN64
            PRINTF2("dividend = %8.8x (EAX)\r\n", (DWORD)ep->ContextRecord->Rax);	//X64???
#else	//Win32
            PRINTF2("dividend = %4.4x (AX)\r\n", ep->ContextRecord->Eax & 0xffff);
#endif	//Win32
        }
        else if (opcode == 0xf7) {			//X64???
#ifdef _WIN64
			PRINTF3("dividend = %16.16I64x:%16.16I64x (RDX:RAX)\r\n",
						ep->ContextRecord->Rdx,
                        ep->ContextRecord->Rax);
#else	//Win32
			PRINTF3("dividend = %8.8x:%8.8x (EDX:EAX)\r\n", ep->ContextRecord->Edx,
                                                           ep->ContextRecord->Eax);
#endif	//Win32
        }
        else {
            PRINTF("not a recognized divide opcode\r\n");
        }
#endif
    }

#if 0
    for (;;) {} //pause for debugging
#endif

#if 0
    //see if we can modify fiber context... Yes, we can!!!
    if (n_except == STATUS_INTEGER_OVERFLOW) {
        ep->ContextRecord->Eax = 0x7FFFFFFF;
        return EXCEPTION_CONTINUE_EXECUTION;
    }
#endif
    return EXCEPTION_EXECUTE_HANDLER;		//EH???
} /*Eval_Exception*/

INT_PTR X86_call_fmf1(BYTE *code, INT_PTR p1)
{
	// Note that the rbx/ebx context ptr (fmcp) gets passed
	// on to the callee as p1...

	// Is nargs actually used by any called code buffers?
	// If not, eliminating setting up nargs would greatly
	// simplify the implementation of X86_call_fmf1 !!

#if 0	//debugging
	PRINTF2("Calling fmf1 code at 0x%p...\r\n", code);
#endif

    if (!gParams->fValidCompile){
		//Code not valid -- return immediately!
#if defined(_DEBUG) || defined(STANDALONE) || 1 ///always warn for now!!!!
		Warn("Aborting X86_call_fmf1 -- code not valid.");
#endif
		return 0;
	}//!fValidCompile

    __try {
#ifdef _WIN64
		if (!X64GEN) {
			//Code not X64 -- return immediately!
#ifdef _DEBUG
			///Don't display repeatedly e.g. when called from doForEveryRow!
			///Warn("Aborting X86_call_fmf1 -- not X64 code.");
#endif
			return 0;
		}
#if 1  //use X64_call_fmf1(BYTE *code, INT_PTR p1);
		return X64_call_fmf1(code, p1);
#else //use old method
		// set nargs = 1
		X64_set_nargs(1);	//rax = 1; but will it be preserved???
        return ((fmf1)code)(p1);
#endif
#else //Win32
		if (X64GEN) {
			//Code not X32 -- return immediately!
#ifdef _DEBUG
			///Don't display repeatedly e.g. when called from doForEveryRow!
			///Warn("Aborting X86_call_fmf1 -- not X32 code.");
#endif
			return 0;
		}
        // set ecx = 1 (nargs)
        _asm { mov ecx, 1 }
        return ((fmf1)code)(p1);
#endif  //Win32
    }
    __except (Eval_Exception(GetExceptionInformation( ))) {
        if (!gFmcp->doingProxy) {
            //if running main image, set gResult to abort ALL tiles...
            gResult = userCanceledErr;
        }
        return userCanceledErr;
    }
} /*X86_call_fmf1*/

/*-----------------------------------------------------*/
/*  FUNCTION Prolog/Epilog                             */
/*-----------------------------------------------------*/

/*****************************************************
Here's the new stack frame layout for the FM32/64 ABI:
******************************************************
      ...
{    arg 6    }
+-------------+
{    arg 5    }
+-------------+
| arg 4 home  |
| arg 3 home  | <- rBP + #cbTemps + #cbNdpRegs + nargsNeeded*BPP + (1+argn+nTempRegsUsed+loadFMCP)*BPP
| arg 2 home  |
| arg 1 home  | <- rBP + #cbTemps + #cbNdpRegs + nargsNeeded*BPP + (2+nTempRegsUsed+loadFMCP)*BPP
+-------------+ <- rSP before call (could save as rAP argument pointer?)
| return rIP  |
+-------------+ <- rBP + #cbTemps + #cbNdpRegs + nargsNeeded*BPP + (1+nTempRegsUsed+loadFMCP)*BPP
{ saved rBX   } - if loadFMCP
+-------------+
{ temp regs   } - saved rDI at top
{ saved in    } - nTempRegsUsed
{ order used  }
+-------------+
{ other saved }
{ nonvolatile } - none?
{ registers   }
+-------------+
| saved rBP   | <- rBP + #cbTemps + #cbNdpRegs + nargsNeeded*BPP
+-------------+
{    nargs    } - if nargsNeeded
+-------------+ <- rBP + #cbTemps + #cbNdpRegs
{    ndp      } #cbNdpRegs
{   backing   } <- rBP + #cbTemps + (stn-1)*8
{   storage   }
+-------------+ <- rBP + #cbTemps
{  tmp reg    } #cbTemps = nTemps*BPP
{  backing    } <- rBP + reg*BPP
{  storage    }
+-------------+ <- rBP (frame pointer)
{   local     } #cbLocals
{  variable   }
{  storage    }
+-------------+
{  alignment  }
{  padding    }
+-------------+ <- rBP - #cbLocals
{  16-byte    }
{  alignment  }
{  padding    }
+-------------+
{   static    } #cbStaticArglistSize (0, else min 32 bytes for X64)
{   arglist   } <- rSP + (argn-1)*BPP
+-------------+ <- rSP before level-1 call (at a 16-byte boundary)
{  alignment  }
{  padding    }
+-------------+
{   dynamic   }
{   arglist   } <- rSP + (argn-1)*BPP
+-------------+ <- rSP before call at level > 1 (at a 16-byte boundary)
     ...  last two areas can be repeated for nested calls
     
 Key:
  { } = optional or variable length field
  # = size of field in bytes
  BPP = bytes per pointer (4 for x32, 8 for X64)
  argn = argument number (1, 2, 3, ..)
  nargs = total # actual args passed by caller
  nargsNeeded = 1 if nargs saved on stack, else 0
  loadFMCP = 1 if rBX saved on stack, else 0
  call level = 1 at statement level, 2 at first inner nested call, ...
  reg = pseudoregister # (r0, r1, r2, ...)
  cbLocals = total # bytes allocated to local (auto) variables (padded to multiple of BPP)
  nTempRegsUsed = # of nonvolatile registers used to hold pseudoregisters
  nTemps = total # of pseudoregisters allocated
  cbTemps = total # bytes allocated for psesudoregisters = nTemps*BPP
  stn = ndp slot number (1 through 8), 0 means empty
  cbNdpRegs = total # bytes allocated for saving ndp registers (#regs * 8)
/***************************************************************************/

#ifndef STANDALONE

// This prolog flavor explicitly loads rBX from arg1 (which
// presumably contains the address of the FMC structure).

void X86_emit_prolog(int nTempsXXX, int cbLocalsXXX)
{
	assert(cbp != NULL);
    EMIT2("<prolog to load rBX for \"%s\">", cbp->displayName);
	// initialize variables at start of function/handler...
	cbp->tosmax = -1;					//no temps used so far
	cbp->func.nTemps = 0;				//not yet known
	cbp->func.cbTemps = 0;				//not yet known
	cbp->func.nTempRegsUsed = 0;		//NOTE nTmps is max # temp regs; nTemps is # temps
	cbp->func.cbLocals = 0;				//none so far
	cbp->func.nArgs = 0;				//none so far
	cbp->func.cbStaticArglistSize = -1;	//none so far
	cbp->func.nargsNeeded = false;		//not so far as we know at this point...
	cbp->func.needFrameReg = true;		//so far as we know...
	cbp->func.homeArgsMask = 0;			//not yet known
	cbp->func.argTypeMask = 0;			//not yet known
	cbp->func.fpuUsed = false;			//not so far as we know yet...
	cbp->func.loadFMCP = true;			//function prolog should load FMCP
} /*X86_emit_prolog*/

// Remainder of prolog is delayed until function body has been generated,
// so we know precisely what needs to be done in the prolog...

void X86_emit_delayed_prolog(int nTemps, int nTempRegsUsed, int cbLocals,
							 int nArgs, int cbStaticArglistSize,
							 int nargsNeeded, int needFrameReg,
							 int homeArgsMask, int argTypeMask,
							 int loadFMCP, int fpuUsed)
{
	int BPP = (X64gen ? 8 : 4);		//Bytes Per Pointer (4 or 8, depends on gen mode)
	int cbTemps = BPP*nTemps;		//size of temps in bytes
#ifdef _WIN64
	int cbMainBody = cbp->pc;		//size of main body code
#else //Win32
	int cbMainBody = cbp->pc - cbp->pcStart;
#endif //Winn32
	CGL entrypoint = X86_genlab();
	CGL endprolog = X86_genlab();

#ifdef _DEBUG	//or make this conditional on EMIT flags ????
	// Emit some function/handler stats during debug...
	PRINTF("\r\n");
	PRINTF2("<Function/Handler stats for cb \"%s\">\r\n", cbp->displayName);
	PRINTF2("pcStart    : %p\r\n", cbp->pcStart);
	PRINTF2("pc         : %x\r\n", cbp->pc);
	PRINTF2("pc_max     : %x\r\n", cbp->pc_max);
	PRINTF2("next_cgl   : %d.\r\n", cbp->next_cgl);
	PRINTF2("max_cgl    : %d.\r\n", cbp->max_cgl);
	PRINTF2("tosmax     : %d.\r\n", cbp->tosmax);
	// cardinals
	PRINTF2("func.nTemps              : %d.\r\n", cbp->func.nTemps);
	PRINTF2("func.cbTemps             : %d.\r\n", cbp->func.cbTemps);
	PRINTF2("func.cbLocals            : %d.\r\n", cbp->func.cbLocals);
	PRINTF2("func.nTempRegsUsed       : %d.\r\n", cbp->func.nTempRegsUsed);
	PRINTF2("func.nArgs               : %d.\r\n", cbp->func.nArgs);
	PRINTF2("func.cbStaticArglistSize : %d.\r\n", cbp->func.cbStaticArglistSize);
	// Booleans
	PRINTF2("func.nargsNeeded         : %d.\r\n", cbp->func.nargsNeeded);
	PRINTF2("func.needFrameReg        : %d.\r\n", cbp->func.needFrameReg);
	PRINTF2("func.fpuUsed             : %d.\r\n", cbp->func.fpuUsed);
	PRINTF2("func.loadFMCP            : %d.\r\n", cbp->func.loadFMCP);
	// bit masks
	PRINTF2("func.homeArgsMask        : %x\r\n", cbp->func.homeArgsMask);
	PRINTF2("func.argTypeMask         : %x\r\n", cbp->func.argTypeMask);
	/********************* temporary - delete this later... *****************
	// FUNC
	int nTemps;					//number of temps used by this function
	int cbTemps;				//number of bytes to allocate for temps
	int nTempRegsUsed;			//number of temp registers used by this function (or a mask??)
	int nArgs;					//number of formal parameters (negated if VPL)	NOT USED???
	int cbStaticArglistSize;	//number of bytes to allocate (-1 means none; minimum is 32 for X64)
	int nargsNeeded;			//true iff "nargs" is referenced by function
	int needFrameReg;			//true iff a stack frame register is needed
	int homeArgsMask;			//mask of args which must be homed (first 4 args only)
	int argTypeMask;			//mask of arg types; 0=integral, 1=floating point
	int fpuUsed;				//true only if 1 or more fpu intructions were emitted
	int loadFMCP;				//true iff the FM Context pointer needs to be loaded into rBX
	int cbLocals;				//number of bytes to allocate on stack for local vars
	CGL ref_cbLocals;			//fwd reference label to fixup cbLocals when known
	// CB32
    const char *displayName;//display name of this CB, for error messages
    BYTE *pc;				//ptr to next byte to be emitted in buffer
    BYTE *pcStart;			//ptr to start of the code buffer
    BYTE *pc_max;			//ptr to end of code buffer + 1
    CGL next_cgl;			//next CGL to be used for this CB
    CGL max_cgl;			//# of CGLs that will fit in current CGL_INFO array
    CGL_INFO32 *cgls;		//ptr to array of CGLs for this Code Buffer
	int tosmax;				//max # of temps used
	FUNC func;				//function-related info
	---------------------------------------------
	// CB64
    const char *displayName;//display name of this CB, for error messages
    BYTE *pcStart;			//ptr to start of the code buffer
    ADDR pc;				//index of next byte to be emitted in buffer
    ADDR pc_max;			//index of end of code buffer + 1 (= sizeof(cb))
    CGL next_cgl;			//next CGL to be used for this CB
    CGL max_cgl;			//# of CGLs that will fit in current CGL_INFO array
    CGL_INFO64 *cgls;		//ptr to array of CGLs for this Code Buffer
	int tosmax;				//max # of temps used
	FUNC func;				//function-related info
	---
	/*****************************/
#endif

	EMIT2("<Delayed prolog for \"%s\">", cbp->displayName);

#ifdef HOTPATCH //hot-patching not needed here!!
	// Precede the prolog entry point with 8 NOP's for a hot-patch jump
	// (really needs only 5 bytes max for a near (disp32) JMP),
	// leaving the prolog 8-byte aligned (assuming the CB is aligned
	// mod 8 or 16).
	//
	// So should calls to the prolog then be to CB+8???
	//<prolog entry point>:
	//	DB		0x40		;REX to make 2-byte opcode if needed for hot-patching (short jump)
#endif //HOTPATCH

    // Prolog template:
    //                      ;arg4=[esp+4*BPP]
    //                      ;arg3=[esp+3*BPP]
    //                      ;arg2=[esp+2*BPP]
    //                      ;arg1=[esp+1*BPP]	//arg1 has fmcp
    //                      ;return ip =[esp]
	//home first 4 args if needed
	//(store in caller's arg list)
	//if X64gen
	//if homeArgsMask
	//bit set means double, else int
	//	if argTypeMask & (1 << 0)
	//		xmovq	[rsp+8],xmm0
	//	else
	//		mov		[rsp+8],rcx
	//	if argTypeMask & (1 << 1)
	//		xmovq	[rsp+16],xmm1
	//	else
	//		mov		[rsp+16],rdx
	//	if argTypeMask & (1 << 2)
	//		xmovq	[rsp+24],xmm2
	//	else
	//		mov		[rsp+24],r8
	//	if argTypeMask & (1 << 3)
	//		xmovq	[rsp+32],xmm3
	//	else
	//		mov		[rsp+32],r9
	//endif homeArgsMask
	//endif X64gen
	//if loadFMCP
    //  push    ebx				;saved ebx =[esp]		//rBX will be loaded from arg1
	//if X64gen
	//	mov		rbx,rcx			;X64: load FM context ptr from arg1 register
	//else
    //  mov     ebx,[esp+2*BPP] ;X86: load FM context ptr from arg1 on stack
	//endif x86
	//endif loadFMCP
	//	//save following regs if used...
	//	//save temp regs in order allocated (yeah, I know edi and esi are
	//	//backwards -- it's a historical thing, and not worth fixing...)
	//if nTempRegsUsed > 0
    //  push    edi
	//if nTempRegsUsed > 1
    //  push    esi
	//if X64gen
	//if nTempRegsUsed > 2
	//	push	r12
	//if nTempRegsUsed > 3
	//	push	r13
	//if nTempRegsUsed > 4
	//	push	r14
	//if nTempRegsUsed > 5
	//	push	r15
	//endif X64gen
	//	//save frame ptr register
    //  push    ebp         ;[ebp+cbTemps+1*BPP]
	//	//save nargs (if needed)
	//if nargsNeeded
	//if X64gen
    //  push	rax         ;nargs=[ebp+cbTemps+0*BPP]	//X64:save nargs from rax
	//else
    //  push	ecx         ;nargs=[ebp+cbTemps+0*BPP]	//X86:save nargs from ecx
	//endif !X64gen
	//endif nargsNeeded
	//if cbTemps > 0
    //  sub     esp,cbTemps ;temp0=[ebp]
	//endif
    //  mov     ebp,esp     ;establish frame ptr (may want to test probe here?)
	//if cbLocals > 4000
	//	;need to probe stack
	//	mov		eax,cbLocals
	//	call	fm_chkstk[ebx]
	//if X64gen
	//	sub		rsp, rax	//x86 chkstk subtracts eax from esp before returning; X64 doesn't
	//endif X64gen
	//else cbLocals <= 4000
    //											//Why not just one sub esp,BPP*(nTemps+1)+cbLocals
    //											//and then point ebp to some optimal midpoint???
    //  sub     esp,cbLocals ;allocate locals	//Because for now, we always want ebp to point to
	//											//the first temp (r0) to make it easier to compute
	//											//the address of temps and stack variables. Later,
	//											//we can try to optimize the ebp pointer to cover
	//											//the most useful range of disp8-addressable vars.
	//endif cbLocals
	//if cbStaticArglistSize >= 0
	//if X64gen
	//-- actually can do better than this if odd # of args (see code elsewhere)
	//	and		rsp,-16		;align stack to 16 bytes
	//	if cbStaticArglistSize < 32 then cbStaticArglistSize = 32; min size required
	//	cbStaticArglistSize = (cbStaticArglistSize + 15) & -16; round up to 0 mod 16
	//endif X64gen
	//	sub		rsp,cbStaticArglistSize		;allocate static arglist frame (use chkstk if >= 4KB)
	//endif cbStaticArglistSize
	//if fpuUsed
	//	[finit--too expensive!!!]			;if body code uses fpu???
	//endif fpuUsed
	//	align 8 or 16 ??

#ifdef HOTPATCH
	//emit 8 NOPs to allow room for a near JMP when hot-patching
	{	int i;
		for i = 0; i < 8; ++i) {
			EMIT("nop ;hot patch");
			X86_emit_op(NOP);
		}
	}
#endif //HOTPATCH
	EMIT2L("cgl%d: ;entrypoint", entrypoint);
    X86_deflab(entrypoint);		//entrypoint label
#ifdef HOTPATCH
	//pad byte for hot patching
	EMIT("db 40h ;REX for hot patch");
	X86_emit_byte(REX);		// REX prefix to make two-byte PUSH_REG	//OOOPS following op is not necessrily a PUSH!!
#endif //HOTPATCH
#ifdef X64
	if (X64gen && homeArgsMask) {
		//home first 4 args if necessary
		if (homeArgsMask & (1 << 0)) {
			EMIT3("mov [RSP%+d], RCX ;home arg%d", 1*BPP, 1);
			X86_emit_op_MR(MOV_V, rsp, no_reg, 0, 1*BPP, rcx);	//arg1
		}
		if (homeArgsMask & (1 << 1)) {
			EMIT3("mov [RSP%+d], RDX ;home arg%d", 2*BPP, 2);
			X86_emit_op_MR(MOV_V, rsp, no_reg, 0, 2*BPP, rdx);	//arg2
		}
		if (homeArgsMask & (1 << 2)) {
			EMIT3("mov [RSP%+d], R8 ;home arg%d", 3*BPP, 3);
			X86_emit_op_MR(MOV_V, rsp, no_reg, 0, 3*BPP, r8);	//arg3
		}
		if (homeArgsMask & (1 << 3)) {
			EMIT3("mov [RSP%+d], R9 ;home arg%d", 4*BPP, 4);
			X86_emit_op_MR(MOV_V, rsp, no_reg, 0, 4*BPP, r9);	//arg4
		}
		//home first 4 float args if needed (using XMOVQ)
		if (argTypeMask) {		//NYI
			EMIT4("xmovq [RSP%+d], XMM%d ;home xarg%d", 1*BPP, 0, 1);
			EMIT4("xmovq [RSP%+d], XMM%d ;home xarg%d", 2*BPP, 1, 2);
			EMIT4("xmovq [RSP%+d], XMM%d ;home xarg%d", 3*BPP, 2, 3);
			EMIT4("xmovq [RSP%+d], XMM%d ;home xarg%d", 4*BPP, 3, 4);
			X86_emit_op_MR(XMOVQ, rsp, no_reg, 0, 1*BPP, xmm0);	//xarg1
			X86_emit_op_MR(XMOVQ, rsp, no_reg, 0, 2*BPP, xmm1);	//xarg2
			X86_emit_op_MR(XMOVQ, rsp, no_reg, 0, 3*BPP, xmm2);	//xarg3
			X86_emit_op_MR(XMOVQ, rsp, no_reg, 0, 4*BPP, xmm3);	//xarg4
		}
#if 0 //Disassembler chokes on MOVSDX for now...
		//home first 4 float args if needed (using MOVSDX)
		EMIT4("movsdx [RSP%+d], XMM%d ;home xarg%d", 1*BPP, 0, 1);
		EMIT4("movsdx [RSP%+d], XMM%d ;home xarg%d", 2*BPP, 1, 2);
		EMIT4("movsdx [RSP%+d], XMM%d ;home xarg%d", 3*BPP, 2, 3);
		EMIT4("movsdx [RSP%+d], XMM%d ;home xarg%d", 4*BPP, 3, 4);
		X86_emit_op_MR(MOVSDX, rsp, no_reg, 0, 1*BPP, xmm0);	//xarg1
		X86_emit_op_MR(MOVSDX, rsp, no_reg, 0, 2*BPP, xmm1);	//xarg2
		X86_emit_op_MR(MOVSDX, rsp, no_reg, 0, 3*BPP, xmm2);	//xarg3
		X86_emit_op_MR(MOVSDX, rsp, no_reg, 0, 4*BPP, xmm3);	//xarg4
#endif //disasm
	}
#endif //X64
	//save and load the FM Context register if necessary
	if (loadFMCP) {
		EMIT("push rBX ;save fmcp reg");
		X86_emit_opreg(PUSH_REG, ebx);	//save the fmcp reg.
#ifdef X64
		if (X64gen) {
			//X64: load FM context ptr from arg1 register
			EMIT("mov RBX, RCX ;load fmcp from arg1");
			X86_emit_op_RR(MOV_V, rbx, rcx);
		} else
#endif //X64
		{
			//X86: load FM context ptr from arg1 on stack
			EMIT2("mov EBX, [ESP%+d] ;load fmcp from arg1", 2*BPP);
			X86_emit_op_RM(MOV_V, ebx, esp, no_reg, 0, 2*BPP);	//esp was decremented by push ebx
		}
	}//loadFMCP
	//save X86 temp regs if used...
	if (nTempRegsUsed > 0) {
		EMIT("push rDI ;save tempreg r0");
		X86_emit_opreg(PUSH_REG, edi);
	}
	if (nTempRegsUsed > 1) {
		EMIT("push rSI ;save tempreg r1");
		X86_emit_opreg(PUSH_REG, esi);
	}
#ifdef X64
	if (X64gen) {
		//save additional X64 temp regs only if actually used!!
		if (nTempRegsUsed > 2) {
			EMIT("push R12 ;save tempreg r2");
			X86_emit_opreg(PUSH_REG, r12);
		}
		if (nTempRegsUsed > 3) {
			EMIT("push R13 ;save tempreg r3");
			X86_emit_opreg(PUSH_REG, r13);
		}
		if (nTempRegsUsed > 4) {
			EMIT("push R14 ;save tempreg r4");
			X86_emit_opreg(PUSH_REG, r14);
		}
		if (nTempRegsUsed > 5) {
			EMIT("push R15 ;save tempreg r5");
			X86_emit_opreg(PUSH_REG, r15);
		}
	}
#endif //X64
	//save the old frame ptr register
	EMIT("push rBP ;save old frame ptr");
    X86_emit_opreg(PUSH_REG, ebp);	//[ebp+cbTemps+1*BPP]

	//save nargs (if needed)
#ifdef X64
	if (X64gen) {
		//X64: nargs is passed in rax
		EMIT("push RAX ;save nargs");
		X86_emit_opreg(PUSH_REG, rax);	//nargs=[rbp+cbTemps+0*BPP]
	} else
#endif //X64
	{
		//X86: nargs is passed in ecx
		EMIT("push ECX ;save nargs");
		X86_emit_opreg(PUSH_REG, ecx);	//nargs=[ebp+cbTemps+0*BPP]
	}
	//allocate room for temp registers on stack
	if (cbTemps > 0) {
		EMIT2("sub rSP, %d ;allocate stack temps", cbTemps);
		X86_emit_op_R_simm32(SUBI_V, esp, cbTemps);
	}

	//establish frame ptr at base of temps
	EMIT("mov rBP, rSP ;set frame ptr to base of temps");
	X86_emit_op_RR(MOV_V, ebp, esp);	//do a test probe here??

	//allocate room for local (auto) variables on the stack
	if (cbLocals > 4000) {	//somewhat less than a 4KB page.
		//need to probe the stack for possible extension
		//need to have loaded ebp and ebx before we can call chkstk!!
		EMIT2("mov EAX, %d ;#bytes for auto vars", cbLocals);
		X86_emit_op_R_imm32(MOVI_D, eax, cbLocals);	//X64 zero-extends
		EMIT("call [rBX].chkstk ;extend stack if needed");
		X86_emit_op_M(CALLI_N, FMC(fm_chkstk));
		if (X64gen) {
			//X64 must subtract allocation from stack *after* chkstk returns...
			EMIT("sub RSP, RAX ;allocate auto vars");
			X86_emit_op_RR(SUB_V, rsp, rax);
		}
	}
	else if (cbLocals > 0) {
		//no chkstk needed
		EMIT2("sub rSP, %d ;allocate auto vars", cbLocals);
		X86_emit_op_R_simm32(SUBI_V, esp, cbLocals);
	} //no chkstk

	//allocate a static arglist if needed... COMBINE WITH ABOVE CBLOCALS (AND MAYBE CBTEMPS/NARGS)
#if 0
	if (cbStaticArglistSize > 4000) {
		//should use _alloca or _chkstk to allocate large arglist...
		NYI;
	} else
#endif
	if (cbStaticArglistSize >= 0) {
		//allocate static arglist frame
#ifdef X64
		if (X64gen) {
			//X64 requires a minimum arglist size of 32 bytes (first 4 args)
			if (cbStaticArglistSize < 32) {
				cbStaticArglistSize = 4*BPP;				//EH??
			}
			//allocate static arglist frame
			EMIT2("sub RSP, %d ;allocate static arglist", cbStaticArglistSize);
			X86_emit_op_R_simm32(SUBI_V, rsp, cbStaticArglistSize);
		} else
#endif //X64
		if (cbStaticArglistSize > 0) {
			//X86: allocate static arglist frame if more than 0 bytes
			EMIT2("sub ESP, %d ;allocate static arglist", cbStaticArglistSize);
			X86_emit_op_R_simm32(SUBI_V, rsp, cbStaticArglistSize);
		}
	}//cbStaticArglistSize
#ifdef X64
	if (X64gen) {
		//for now, we always align the X64 stack to 16 bytes;
		//later, we will do this only for non-LEAF procedures
		//align stack to 16 bytes
		EMIT("and RSP, -16 ;align 16");	//Should do this even if no static arglist; may need alignment for intrinsic func calls!!!
		X86_emit_op_R_simm32(ANDI_V, rsp, -16);	//round down to 0 mod 16.
	}
#endif //X64
	if (fpuUsed) {			//only in init code!!!
		EMIT("finit ;init fpu");
		X86_emit_op(FINIT);		//in case the body code uses the fpu? too expensive!!!
	}//fpuUsed

	//label to mark end of prolog
	EMIT2L("cgl%d: ;end of prolog", endprolog);
    X86_deflab(endprolog);

	// Now rotate the Code Buffer left by cbMainBody bytes,
	// so the Prolog will precede the Main Body (and epilog).
	X86_rotate_CB(cbp, cbMainBody);

#if 0
    // Patch test area...
    {//new scope
        CGL lab_skip_around = X86_genlab();
        X86_emit_op_lab(JMP, lab_skip_around);

        // following code will never be executed so can be used
        // as a test bed to test and disassemble new code generators;
        {//new scope
            CGL ref_cbLocals = X86_genlab();
            X86_emit_op_R_ref_simm32(SUBI_V, esp, ref_cbLocals);
            // just some dummy code...
            X86_emit_opreg(PUSH_REG, ebx);
            X86_emit_opreg(PUSH_REG, esi);
			X86_emit_opreg(PUSH_REG, r12d);
			X86_emit_opreg(PUSH_REG, r12);

        X86_emit_op_lab(JMP, lab_skip_around);  //double-chain of disp32

            X86_emit_opreg(PUSH_REG, edi);
            X86_emit_opreg(PUSH_REG, ebp);
			X86_emit_opreg(PUSH_REG, rbp);
			X86_emit_opreg(PUSH_REG, r15);

            X86_def_abs32(ref_cbLocals, cbLocals);

			// Check disassembly of some instructions:

			X86_emit_op(WAIT);
			X86_emit_op(HLT);
			X86_emit_op(WAIT);
			X86_emit_op(WAIT);
			X86_emit_op(HLT);
			X86_emit_op(FLDZ);
			X86_emit_op(HLT);
			X86_emit_op(WAIT);
			X86_emit_op(FLDZ);
			X86_emit_op(HLT);
			X86_emit_op(FNINIT);
			X86_emit_op(HLT);
			X86_emit_op(FINIT);
			X86_emit_op(HLT);
			X86_emit_op(WAIT);
			X86_emit_op(FNINIT);
			X86_emit_op(HLT);

			X86_emit_op_R(INC_V, eax);
			X86_emit_op_RR(MOV_V, eax, ebx);
			X86_emit_op_R(INC_V, eax);
			X86_emit_op_RR(ADD_V, eax, ebx);
			X86_emit_op_R(DEC_V, ecx);
			X86_emit_op_RR(SUB_V, eax, ebx);



        }//end scope

        X86_deflab(lab_skip_around);
    }//end patch area scope
#endif//patch test area

} /*X86_emit_delayed_prolog*/

void X86_emit_epilog(int nTempsXXX, int cbLocalsXXX, int nArgs)
{
	int BPP = (X64gen ? 8 : 4);		//Bytes Per Pointer (4 or 8, depends on gen mode)
	CGL epilog = X86_genlab();

	assert(cbp != NULL);
    EMIT2("<epilog for \"%s\">", cbp->displayName);
	cbp->func.nTemps = cbp->tosmax + 1;	//+1 because tos is 0-based
	cbp->func.cbTemps = BPP*(cbp->func.nTemps);		//size of temps in bytes
	cbp->func.nTempRegsUsed = min(cbp->func.nTemps, nTmps);	//not actually necessary to minimize??

    //nArgs <= 0 ==> vpl, else fpl
    //Epilog template:
    //////  mov     esp,ebp
    //////  add     esp,cbTemps+1*BPP ;remove "nargs" and Temps
    //  lea     esp,[ebp+cbTemps+1*BPP] ;remove "nargs" and Temps
    //  pop     ebp
	//switch (nTempRegsUsed)
	//default:
	//case 6:	pop		r15
	//case 5:	pop		r14
	//case 4:	pop		r13
	//case 3:	pop		r12
    //case 2:	pop     esi
    //case 1:	pop     edi
	//case 0:	break;
	//end switch
	//if loadFMCP
    //  pop     ebx
	//endif
	//if X64gen or nargs <= 0
	//	ret					//if X64 or vpl (nargs <= 0), caller cleans stack args
	//else
    //  ret     [nArgs*BPP]	//if X86 and fpl (nargs > 0), callee cleans stack args
	//endif X64gen

	//Align cbLocals to a multiple of BPP
	cbp->func.cbLocals = (cbp->func.cbLocals + BPP - 1) & ~(BPP - 1);

    // Fixup forward refs to cbLocals, now that it's known...
    ///X86_def_abs32(ref_cbLocals, cbLocals);

	//epilog label for debugging
	EMIT2L("cgl%d: ;epilog", epilog);
    X86_deflab(epilog);

	//X86_emit_op_RR(MOV_V, esp, ebp);
    //X86_emit_op_R_simm32(ADDI_V, esp, cbTemps+BPP);	//BPP is for nargs
	EMIT2("lea rSP, [rBP%+d] ;restore rSP", cbp->func.cbTemps+BPP);
    X86_emit_op_RM(LEA_V, esp, ebp, no_reg, 0, cbp->func.cbTemps+BPP); //BPP is for nargs
	EMIT("pop rBP ;restore rBP");
    X86_emit_opreg(POP_REG, ebp);
	switch (cbp->func.nTempRegsUsed) {
	default:
#ifdef X64
	//restore the X64 temp regs that were used
	// (fall-through is deliberate!)
	case 6:
		EMIT("pop R15 ;restore");
		X86_emit_opreg(POP_REG, r15);
	case 5:
		EMIT("pop R14 ;restore");
		X86_emit_opreg(POP_REG, r14);
	case 4:
		EMIT("pop R13 ;restore");
		X86_emit_opreg(POP_REG, r13);
	case 3:
		EMIT("pop R12 ;restore");
		X86_emit_opreg(POP_REG, r12);
#endif
	//restore the X86 temp regs that were used
	// (fall-through is deliberate!)
	case 2:
		EMIT("pop rSI ;restore");
		X86_emit_opreg(POP_REG, esi);
	case 1:
		EMIT("pop rDI ;restore");
		X86_emit_opreg(POP_REG, edi);
	case 0:
		break;
	}//switch
	if (cbp->func.loadFMCP) {
		//restore rBX if it was loaded with fmcp
		EMIT("pop rBX ;restore");
		X86_emit_opreg(POP_REG, ebx);
	}
    if (X64gen || nArgs <= 0) {	//X64 is always VPL
        //X64 or vpl
		//if X64 or vpl (nargs <= 0), caller cleans stack args
		EMIT("ret");
        X86_emit_op(RET0_N);
    }
    else {
        //fpl
		//if X86 and fpl (nargs > 0), callee cleans stack args
#if 0
		//SHOULD NEVER HAPPEN now that we use UNIVERSAL calling sequence??
		//But an external caller might still expect us to clean the stack??
		ErrorOk("Invalid FPL return, bug #9001");
#endif
		EMIT2("ret %d ;FPL return", (WORD)(4*nArgs));
        X86_emit_op_imm16(RET_N, (WORD)(4*nArgs));
    }

	// Now emit the (delayed) prolog...
	X86_emit_delayed_prolog(cbp->func.nTemps, cbp->func.nTempRegsUsed,
							cbp->func.cbLocals,
							nArgs, cbp->func.cbStaticArglistSize,
							cbp->func.nargsNeeded, cbp->func.needFrameReg,
							cbp->func.homeArgsMask, cbp->func.argTypeMask,
							cbp->func.loadFMCP, cbp->func.fpuUsed);

} /*X86_emit_epilog*/

/*-----------------------------------------------------*/
/*  EVENT HANDLER Prolog/Epilog                        */
/*-----------------------------------------------------*/

void X86_emit_handler_prolog(int nTempsXXX, int cbLocalsXXX)
{
    EMIT2("<prolog without loading rBX, for \"%s\">", cbp->displayName);
	//same as X86_emit_prolog, except we don't need to
	//load the FM Context pointer reg for a handler,
	//since rBX is already assumed to contain the FMCP.
	X86_emit_prolog(nTempsXXX, cbLocalsXXX);
	cbp->func.loadFMCP = false;	//don't load rBX
} /*X86_emit_handler_prolog*/

void X86_emit_handler_epilog(int nTempsXXX, int cbLocalsXXX, int nArgs)
{
	//same as X86_emit_epilog, except we don't
	//restore register rBX, since it wasn't saved by
	//the prolog for this handler.
	X86_emit_epilog(nTempsXXX, cbLocalsXXX, nArgs);
} /*X86_emit_handler_epilog*/

/*-----------------------------------------------------*/
/*  Synthetics                                         */
/*-----------------------------------------------------*/

void X86_emit_DIVREM_TT(int op, int tmp1, int tmp2, BOOL check)
{
    //op=0 => div
    //op=1 => rem
	//[tmp1] = dividend
	//[tmp2] = divisor
	//check = 1 to check for divide error, 0 if not
    if (check) {
        //run-time check version
        CGL okdiv = X86_genlab();
        //rAX <= dividend
		EMIT2("mov rAX, r%d ;dividend", tmp1);
        X86_emit_op_RT(MOV_V, eax, tmp1);
        //rCX <= divisor+1
        if (isReg(tmp2)) {
			EMIT2("lea rCX, [r%d+1] ;rCX=divisor+1", tmp2);
            X86_emit_op_RM(LEA_V, ecx, tmpReg(tmp2), no_reg, 0, 1);
        }
        else {
			EMIT2("mov rCX, r%d ;divisor", tmp2);
			EMIT("inc rCX ;divisor+1");
            X86_emit_op_RT(MOV_V, ecx, tmp2);
            X86_emit_op_R(INC_V, ecx);
        }
        //restore divisor, setting CF and ZF
        //(can't use DEC -- it doesn't set CF)
		EMIT("sub eCX, 1 ;set CF and ZF");
		EMIT2("ja cgl%d ;okdiv", okdiv);
        X86_emit_op_R_simm32(SUBI_V, ecx, 1);
        X86_emit_op_lab_disp8(JA, okdiv);
		//need to check further when divisor=0 (ZF=1) or -1 (CF=1)
		//call divCheck(eAX=dividend, eCX=divisor)
		EMIT("call [rBX].divCheck ;divisor=0 or -1");
        X86_emit_op_M(CALLI_N, FMC(divCheck));
      X86_deflab(okdiv);
		EMIT2L("cgl%d: ;okdiv", okdiv);
	    EMIT("cdq ;sign-extend dividend");
        X86_emit_op(CDQ_V);	//sign-extend dividend to rDX:rAX
#if 1
        //No mods...
#elif 0	//Various testbeds...
        //NT SEH reports STATUS_ACCESS_VIOLATION??? for these (nparams=0);
        //Win95 apparently ignores HLT/CLI and continues execution!!!
        //(Maybe to try to support some old legacy real-mode code?),
        //For LTR, Win95 generates STATUS_ACCESS_VIOLATION, with the
        //usual 2 bogus params (0, -1).  NT correctly generates 
        //STATUS_ACCESS_VIOLATION???.  Both point excep addr to the LTR.
        X86_emit_op_R(LTR, eax); //DEBUG: cause a privileged opcode GP
        X86_emit_op(HLT);   //DEBUG: cause a privileged opcode GP
        X86_emit_op(CLI);   //DEBUG: cause a privileged opcode GP
#elif 1
        //test MDA's setting of IFLAGS...
        X86_emit_op(D16);      //operand size override
        X86_emit_op(A32);      //address size override
        X86_emit_op(D16);      //operand size override
#elif 0
        //Win95 reports STATUS_BREAKPOINT (exception addr = next instruction,parm0=0).
        X86_emit_op(INT3);  //DEBUG: cause a breakpont trap
#elif 1
        //Win95 reports STATUS_INTEGER_OVERFLOW (nparm=0, addr=>past INTO)
#ifdef X64
		if (X64gen) {
#if 0
			X86_emit_op_RR(XOR_D, eax, eax);	//rax=0
			X86_emit_op_R(DEC_Q, rax);			//rax=-1
			X86_emit_op_R(Q(SHR1_D), rax);		//rax=0x7FFFFFFFFFFFFFFF
#elif 1
			X86_emit_op_R_simm32(OR_Q, rax, -1);//rax=-1
			X86_emit_op_R(Q(SHR1_D), rax);		//rax=0x7FFFFFFFFFFFFFFF
#else
			X86_emit_op_R_imm64(MOVI_Q, rax, 0x7FFFFFFFFFFFFFFF);	//NYI
#endif
		} else
#endif //X64
		{
			X86_emit_op_R_imm32(MOVI_D, eax, 0x7FFFFFFF);
		}
        X86_emit_op_R_imm32(ADDI_V, eax, 1);	//rAX=0x80000000[00000000] (overflow into sign)
        X86_emit_op(INTO);  //DEBUG: cause an integer overflow trap, or INVALID in X64
        X86_emit_op(INT3);  //DEBUG: then break so we can examine regs...
#elif 1
        //Win95 reports STATUS_ILLEGAL_INSTRUCTION (nparm=0, addr=>illegal instr.)
        X86_emit_op_RR(LES, edi, eax);   //Illegal intruction (also need to defeat
                                         //the check in emit_op_RR)
#elif 1
        //Win95 reports STATUS_ARRAY_BOUNDS_EXCEEDED (nparm=0, addr=after INT5;
        //what about after a true BOUND opcode?)
        X86_emit_op_imm8(INTI_B, 5);   //DEBUG: fake a bounds exception
#endif //various testbeds
		//
		// Now do the actual division...
		//
		EMIT("idiv rCX ;divide rDX:rAX by rCX");
        X86_emit_op_R(IDIV_V, ecx);	//divide rDX:rAX by rCX
    } //if check
    else {
        //no run-time checking
		EMIT("<no run-time checking>");
		EMIT2("mov rAX, r%d ;dividend", tmp1);
		EMIT("cdq ;sign-extend to rDX:rAX");
		EMIT2("idiv r%d ;divide by divisor", tmp2);
        X86_emit_op_RT(MOV_V, eax, tmp1);	//move dividend to rAX
        X86_emit_op(CDQ_V);					//sign-extend dividend to rDX:rAX
        X86_emit_op_T(IDIV_V, tmp2);		//divide rDX:rAX by divisor
    } //if !check
	//
	// Now select the result as either the quotient (in eAX)
	// or the remainder (in eDX)...
	//
    if (op == 0) {
		//return quotient in tmp1
		EMIT2("mov r%d, rAX ;return quotient", tmp1);
        X86_emit_op_TR(MOV_V, tmp1, eax);
    }
    else {
		//return remainder in tmp1
		EMIT2("mov r%d, rDX ;return remainder", tmp1);
        X86_emit_op_TR(MOV_V, tmp1, edx);
    }
} /*X86_emit_DIVREM_TT*/

void X86_emit_divCheck(void)
{
    CGL entry = X86_genlab();
	CGL divok = X86_genlab();
    CGL zdiv = X86_genlab();
    CGL handled = X86_genlab();
    CGL defzdh = X86_genlab();
    CGL divov = X86_genlab();

	EMIT0("<divCheck handler>");

  X86_deflab(entry);	//for debug
  EMIT2L("cgl%d: ;entry", entry);
	//On entry, rAX=dividend and rCX=divisor
    //CC on entry reflects state of divisor:
    //ZF=1 if divisor was 0; CF=1 if divisor was -1
	EMIT2("jz cgl%d ;zdiv error", zdiv);
    X86_emit_op_lab_disp8(JZ, zdiv);		//zero-divide err if divisor=0
	// else divisor must have been -1;
	// overflow will occur iff dividend is 0x80000000[00000000]
#ifdef X64
	//Assumes intermediate type = int64, not int32!!!
	//Also assumes dividend and divisor are int64 in 64-bit mode.
	if (X64gen) {
		//64-bit code: test for dividend==0x8000000000000000
		//First, test if lo dword of dividend==0x00000000
		//Note that we can't use OR EAX,EAX here,
		//since it would zero out the hi dword of the dividend!!
		//However, TEST EAX,EAX works just fine here, since it
		//does not modify any operands(including hi dwords),
		//and it generates 2 code bytes, just as OR EAX,EAX would...
#if 1
		EMIT("test EAX, EAX ;test lo dword");
		X86_emit_op_RR(TEST_D, eax, eax);		//(2 bytes) Does NOT clear hi(rax) as OR EAX,EAX would.
#elif 1
		EMIT("cmp EAX, 0 ;test lo dword");
		X86_emit_op_R_simm32(CMPI_D, eax, 0);	//(3 bytes) Also works, but uses 1 more code byte.
#elif 1
		EMIT("test EAX, -1 ;test lo dword");
		X86_emit_op_R_imm32(TESTI_D, eax, -1);	//(6 bytes) Also works, but must use full 32-bit immediate with TEST.
#endif
		EMIT2("jne cgl%d ;divok", divok);
		EMIT("ror RAX, 32 ;swap dwords");
		EMIT("cmp EAX, 80000000H ;test hi word");
		EMIT("ror RAX, 32 ;restore dividend");
		EMIT2("je cgl%d ;divov", divov);
		X86_emit_op_lab_disp8(JNE, divok);		//ok to divide if not 0
		X86_emit_op_R_imm8(Q(RORI_D), rax, 32);	//swap hi and lo dwords
		X86_emit_op_R_simm32(CMPI_D, eax, 0x80000000);	//set ZF if dividend's hi dword==0x80000000
		X86_emit_op_R_imm8(Q(RORI_D), rax, 32);	//restore original rax (ZF not changed)
		X86_emit_op_lab_disp8(JE, divov);		//divide oflo if rax==0x8000000000000000
												//else no oflo possible
	} else
#endif X64
	{	//32-bit code: test for dividend==0x80000000
		EMIT("cmp EAX, 80000000H ;test dividend");
		EMIT2("je cgl%d ;divov", divov);
		X86_emit_op_R_simm32(CMPI_D, eax, 0x80000000);
		X86_emit_op_lab_disp8(JE, divov);		//divide oflo if eax==0x80000000
	}
    //quick return to perform the divide if no overflow possible
  X86_deflab(divok);
	EMIT("<okay to divide>");
	EMIT2L("cgl%d: ;divok", divok);
	EMIT("ret ;ok to divide");
    X86_emit_op(RET0_N);

  X86_deflab(zdiv);
	EMIT("<zero divide error>");
	EMIT2L("cgl%d: ;zdiv", zdiv);
    //zero divide: call the user's onZeroDivide handler
	//dividend & divisor should be int64 if calc64...
	EMIT("mov [rBX].dividend, rAX ;store dividend");
	EMIT("mov [rBX].divisor, rCX ;store divisor");
	EMIT("call [rBX].onZeroDivide ;notify user");
	EMIT("test rAX, rAX ;test return value");
	X86_emit_op_MR(MOV_V, FMC(dividend), eax);	//X64 op_store??
    X86_emit_op_MR(MOV_V, FMC(divisor), ecx);	//X64 op_store??
    X86_emit_op_M(CALLI_N, FMC(onZeroDivide));	//notify the user
    X86_emit_op_RR(TEST_V, eax, eax);			//user returns TRUE if error was handled
    //if user didn't handle error, go to default zero-divide handler
	EMIT2L("jz cgl%d ;defzdh", defzdh);
    X86_emit_op_lab_disp8(JZ, defzdh);

  X86_deflab(handled);
	EMIT("<user handled the divide error>");
	EMIT2L("cgl%d: ;handled", handled);
    //user handled the error; restore the (modified?) operands
    //and try the operation...
	EMIT("mov rAX, [rBX].dividend");
	EMIT("mov rCX, [rBX].divisor");
	EMIT("ret ;try operation again");
    X86_emit_op_RM(MOV_V, eax, FMC(dividend));	//X64 op_load??
    X86_emit_op_RM(MOV_V, ecx, FMC(divisor));	//X64 op_load??
    X86_emit_op(RET0_N);

  X86_deflab(defzdh);
	EMIT("<default zero divide handler>");
	EMIT2L("cgl%d: ;defzdh", defzdh);
    //default zero divide handler
    //(we force the result to 0 for FF compatibility)
	EMIT("xor EAX, EAX ;dividend=0");
	EMIT("mov ECX, 1 ;divisor=1");
	EMIT("ret ;divide 0/1=0");
    X86_emit_op_RR(XOR_D, eax, eax);			//dividend <- zext(0)
    X86_emit_op_R_imm32(MOVI_D, ecx, 1);		//divisor <- zext(1)
    X86_emit_op(RET0_N);						//go do the divide, result = 0 op 1 = 0

  X86_deflab(divov);
	EMIT("<divide overflow error>");
	EMIT2L("cgl%d: ;divov", divov);
    //divide overflow: call the user's onDivideOverflow handler
	EMIT("mov [rBX].dividend, rAX ;store dividend");
	EMIT("mov [rBX].divisor, rCX ;store divisor");
	EMIT("call [rBX].onDivideOverflow ;notify user");
	EMIT("test rAX, rAX ;test return value");
	EMIT2L("jnz cgl%d ;handled", handled);
    X86_emit_op_MR(MOV_V, FMC(dividend), eax);	//X64 op_store??
    X86_emit_op_MR(MOV_V, FMC(divisor), ecx);	//X64 op_store??
    X86_emit_op_M(CALLI_N, FMC(onDivideOverflow)); //notify the user
    X86_emit_op_RR(TEST_V, eax, eax);			//user returns TRUE if errorr was handled
    X86_emit_op_lab_disp8(JNZ, handled);
    //fall thru to the default divide overflow handler:
	EMIT("<default divide overflow handler>");
    //Result is forced to 0x80000000[00000000] (approximation to indefinite-Nan)
	EMIT("mov EAX, 80000000H ;dividend=80000000H");
    X86_emit_op_R_imm32(MOVI_D, eax, 0x80000000);	//dividend <- zext(0x80000000) (calc32)
#ifdef X64
	if (X64gen) {
		EMIT("ror RAX, 32 ;dividend=8000000000000000H");
		X86_emit_op_R_imm8(Q(RORI_D), rax, 32);		//calc64: dividend <- 0x8000000000000000
					 // or Q(SHLI_D) ?? any faster???
	}
#endif X64
	EMIT("mov ECX, 1 ;divisor=1");
	EMIT("ret ;go divide, res=int NaN");
    X86_emit_op_R_imm32(MOVI_D, ecx, 1);            //divisor <- zext(1) (X64 zero-extends)
    X86_emit_op(RET0_N);							//go do the divide, res = 0x80000000[00000000]
} /*X86_emit_divCheck*/

void X86_emit_defaultHandler(void)
{
	EMIT("<default handler>");
	EMIT("xor EAX, EAX; clear rAX");
	EMIT("ret ;return FALSE");
    X86_emit_op_RR(XOR_D, eax, eax);    // return FALSE;
    X86_emit_op(RET0_N);
} /*X86_emit_defaultHandler*/

#if 0 //not used
void X86_emit_defaultOnZeroDivide(void)
{	//assumes calc64
#if 0
    X86_emit_op_M_imm32(MOVI_V, FMC(dividend), 0);  //dividend=0;
    X86_emit_op_M_imm32(MOVI_V, FMC(divisor), 1);   //divisor=1;
#else
    X86_emit_op_RR(XOR_D, eax, eax);				//rAX=0;
    X86_emit_op_MR(MOV_V, FMC(dividend), eax);      //dividend=0;
    X86_emit_op_R(INC_D, eax);						//rAX=1;
    X86_emit_op_MR(MOV_V, FMC(divisor), eax);       //divisor=1;
#endif
    X86_emit_op(RET0_N);                            //return TRUE (rAX=1);
} /*X86_emit_defaultOnZeroDivide*/
#endif

#if 0 //not used
void X86_emit_onDivideOverflow(void)
{
    X86_emit_op_RR(XOR_D, eax, eax);                // rAX=0;
    X86_emit_op(RET0_N);							// return FALSE;
} /*X86_emit_onDivideOverflow*/
#endif

#if 0 //not used
void X86_emit_defaultOnDivideOverflow(void)
{
#if 0	//calc32
    X86_emit_op_M_imm32(MOVI_D, FMC(dividend), 0x80000000);  //dividend;
    X86_emit_op_M_imm32(MOVI_D, FMC(divisor), 1);   //divisor=1;
#else	//calc64
    X86_emit_op_RR(XOR_D, eax, eax);				//rAX=0;
    X86_emit_op_R(INC_D, eax);						//rAX=1;
    X86_emit_op_MR(MOV_V, FMC(divisor), eax);       //divisor=1;
    X86_emit_op_R(ROR1_V, eax);						//rAX=0x80000000[00000000]
    X86_emit_op_MR(MOV_V, FMC(dividend), eax);      //dividend=0x8000000[00000000];
	X86_emit_op_R(ROL1_V, eax);						//rAX=1 for return TRUE;
#endif
    X86_emit_op(RET0_N);                            //return TRUE;
} /*X86_emit_defaultOnDivideOverflow*/
#endif

    /***********************/
    /*  FM Inner Loop Code */
    /***********************/

void X86_emit_compute_iuv(void)							//X64 VETTING DONE
{
	EMIT("<compute i, u, v>");
    //  i = i0 = i1 = (( 76*r)+(150*g)+( 29*b))/256;
    //  u = u0 = u1 = ((-19*r)+(-37*g)+( 56*b))/256;
    //  v = v0 = v1 = (( 78*r)+(-65*g)+(-13*b))/256;

	// N.B. i,u,v,r,g,b, etc. are int32 regardless of 32/64-bit mode.
	// We use 32-bit math even for X64 code, for compatibility and
	// perhaps some speed advantage over 64-bit math.
	// N.B.2. Could split out calc of i with a separate need_i?
    X86_emit_op_RM_simm32(IMULI_D, eax, FMC(r), 76);
    X86_emit_op_RM_simm32(IMULI_D, edx, FMC(g), 150);
    X86_emit_op_RR(ADD_D, eax, edx);
    X86_emit_op_RM_simm32(IMULI_D, edx, FMC(b), 29);
    X86_emit_op_RR(ADD_D, eax, edx);
    X86_emit_op_R_imm8(SHRI_D, eax, 8); //should never be negative!
    X86_emit_op_MR(MOV_D, FMC(i), eax);
    //if (gFmcp->needPremiere) {
        //i0 and i1 don't set needPremiere for now...
        X86_emit_op_MR(MOV_D, FMC(globvars.i0), eax);
        X86_emit_op_MR(MOV_D, FMC(globvars.i1), eax);
    //}

    X86_emit_op_RM_simm32(IMULI_D, eax, FMC(r), -19);
    X86_emit_op_RM_simm32(IMULI_D, edx, FMC(g), -37);
    X86_emit_op_RR(ADD_D, eax, edx);
    X86_emit_op_RM_simm32(IMULI_D, edx, FMC(b), 56);
    X86_emit_op_R_imm32(MOVI_D, ecx, 256);
    X86_emit_op_RR(ADD_D, eax, edx);
    X86_emit_op(CDQ);
    X86_emit_op_R(IDIV_D, ecx); //can be negative!
    X86_emit_op_MR(MOV_D, FMC(u), eax);
    if (gFmcp->needPremiere) {
        X86_emit_op_MR(MOV_D, FMC(u0), eax);
        X86_emit_op_MR(MOV_D, FMC(u1), eax);
    }

    X86_emit_op_RM_simm32(IMULI_D, eax, FMC(r), 78);
    X86_emit_op_RM_simm32(IMULI_D, edx, FMC(g), -65);
    X86_emit_op_RR(ADD_D, eax, edx);
    X86_emit_op_RM_simm32(IMULI_D, edx, FMC(b), -13);
    //X86_emit_op_R_imm32(MOVI_D, ecx, 256);    //still live
    X86_emit_op_RR(ADD_D, eax, edx);
    X86_emit_op(CDQ);
    X86_emit_op_R(IDIV_D, ecx); //can be negative!
    X86_emit_op_MR(MOV_D, FMC(v), eax);
    if (gFmcp->needPremiere) {
        X86_emit_op_MR(MOV_D, FMC(v0), eax);
        X86_emit_op_MR(MOV_D, FMC(v1), eax);
    }
} /*X86_emit_compute_iuv*/

void X86_emit_compute_d(void)										//X64 VETTING DONE
{	//assumes x, X, y, Y, d, d0, d1 are int32 regardless of 32/64-bit mode

	EMIT("<compute d>");
	// Based on new implementation of c2d...
    //  d = d0 = d1 = c2d(x-(X>>1), y-(Y>>1));
    //where:
    //  c2d(x,y) = (int)( (256.0 * 0.63661977236758134308) * atan2((double)y, (double)x) );
    //
    X86_emit_op_RM(MOV_D, edx, FMC(y));
    X86_emit_op_RM(MOV_D, eax, FMC(Y));
    X86_emit_op_R(SHR1_D, eax);
    X86_emit_op_RR(SUB_D, edx, eax);    //edx = y-(Y>>1)
    X86_emit_op_MR(MOV_D, FMC(temp), edx);
    X86_emit_op_M(FLD_I, FMC(temp));    //(double)(y-(Y>>1))

    X86_emit_op_RM(MOV_D, edx, FMC(x));
    X86_emit_op_RM(MOV_D, eax, FMC(X));
    X86_emit_op_R(SHR1_D, eax);
    X86_emit_op_RR(SUB_D, edx, eax);    //edx = x-(X>>1)
    X86_emit_op_MR(MOV_D, FMC(temp), edx);
    X86_emit_op_M(FLD_I, FMC(temp));    //(double)(x-(X>>1)))

    X86_emit_op(FPATAN);                    // atan2()
    X86_emit_op_M(FMUL_D, FMC(d512byPi));   // * 512.0/Pi

	// truncate to integer for FF compatibilty
	if (X64gen) {
		//X64: convert double to int32 with truncation
		X86_emit_op_M(FSTP_D, FMC(temp));	//next op needs double in XMM reg or memory
		X86_emit_op_RM(CVTTSD2SI, eax, FMC(temp));
	}
	else
	{	//X86: call MSVC "chop to int" intrinsic
		X86_emit_op_M(CALLI_N, FMC(_ftol));	
	}
    X86_emit_op_MR(MOV_D, FMC(d), eax);
    if (gFmcp->needPremiere) {
        X86_emit_op_MR(MOV_D, FMC(d0), eax);
        X86_emit_op_MR(MOV_D, FMC(d1), eax);
    }
} /*X86_emit_compute_d*/

void X86_emit_compute_m(void)									//X64 VETTING DONE
{	//assumes x, X, y, Y, m, m0, m1 are int32 regardless of 32/64-bit mode

	EMIT("<compute m>");
    //  m = m0 = m1 = c2m(x-(X>>1), y-(Y>>1));
    //where:
    //  c2m(x,y) = (int)sqrt((double)(x*x+y*y));
    X86_emit_op_RM(MOV_D, eax, FMC(X));
    X86_emit_op_R(SHR1_D, eax);
    X86_emit_op_RM(SUB_D, eax, FMC(x)); //neg, but we're going
    X86_emit_op_RR(IMUL32, eax, eax);   //to square it anyway
    X86_emit_op_RM(MOV_D, ecx, FMC(Y));
    X86_emit_op_R(SHR1_D, ecx);
    X86_emit_op_RM(SUB_D, ecx, FMC(y)); //neg, but we're going
    X86_emit_op_RR(IMUL32, ecx, ecx);   //to square it anyway
    X86_emit_op_RR(ADD_D, eax, ecx);
    X86_emit_op_MR(MOV_D, FMC(temp), eax);
    X86_emit_op_M(FLD_I, FMC(temp));
    X86_emit_op(FSQRT);
    X86_emit_op_M(FSTP_I, FMC(m));		//should we call _ftol here???
    if (gFmcp->needPremiere) {
        X86_emit_op_RM(MOV_D, eax, FMC(m));
        X86_emit_op_MR(MOV_D, FMC(m0), eax);
        X86_emit_op_MR(MOV_D, FMC(m1), eax);
    }
} /*X86_emit_compute_m*/

#if 0									//RESUME X64 FIXUPS HERE...
//not used???
void X86_emit_doOneRow(int nTempsXXX, int cbLocalsXXX)
{
    X86_emit_doOneRow_part1(nTempsXXX, cbLocalsXXX);
    X86_emit_doOneRow_part2(nTempsXXX, cbLocalsXXX);
} /*X86_emit_doOneRow*/
#endif

static CGL L1;

void X86_emit_doOneRow_part1(int nTempsXXX, int cbLocalsXXX)
{	//assumes all FMC variables (except ptrs) are int32 regardless of 32/64-bit mode
    CGL L1a;
#if 0
    CGL L2, L3, L4;
#endif

    //  <prolog>
    //  if (alphaPresent) alphaMask = 0;
    //  else alphaMask = 0xff;
    //  col = 0;
    //L1:
    //  <copy over extra planes?>
    //  R = r = r0 = r1 = srcPtr[0];
    //  G = g = g0 = g1 = srcPtr[1];
    //  B = b = b0 = b1 = srcPtr[2];
    //  if (alphaPresent) A = a = a0 = a1 = srcPtr[3];
    //  if (maskPresent) S = s = s0 = s1 = mskPtr[0];
    //  if (need_iuv) {
    //      i = i0 = i1 = ( 76*r + 150*g + 29*b)/256;
    //      u = u0 = u1 = (-19*r -  37*g + 56*b)/256;
    //      v = v0 = v1 = ( 78*r -  65*g - 13*b)/256;
    //  }
    //  if (need_d) {
    //      d = d0 = d1 = c2d(x-(X >> 1), y-(Y >> 1));
    //  }
    //  if (need_m) {
    //      m = m0 = m1 = c2m(x-(X >> 1), y-(Y >> 1));
    //  }
    //  <RGBA channel code>

	EMIT("<prolog for doOneRow>");
    X86_emit_prolog(nTempsXXX, cbLocalsXXX);            // <prolog>
    //
    // For now, we assume alphaPresent == (Z >= 4), which is correct
    // for RGB, but not for CMYK, GrayScale, etc.
    //
    //  alphaMask = (Z >= 4) ? 0 : 0xff;
    //
	EMIT("xor EAX, EAX");
	EMIT("cmp [rBX].Z, 4");
	EMIT("setl AL ;EAX=1 if Z<4");
	EMIT("neg AL ;EAX=255 if Z<4");
	EMIT("mov [rBX].alphaMask, EAX");
    X86_emit_op_RR(XOR_D, eax, eax);                // xor eax, eax             ; eax <- 0
    X86_emit_op_M_simm32(CMPI_D, FMC(Z), 4);        // cmp fmc.Z, 4             ; compare Z to 4
    X86_emit_op_R(SETL_B, al);                      // setl al                  ; eax <- (Z < 4) ? 1 : 0
    X86_emit_op_R(NEG_B, al);                       // neg al                   ; eax <- (Z < 4) ? 0xff : 0
    X86_emit_op_MR(MOV_D, FMC(alphaMask), eax);     // mov fmc.alphaMask, eax   ; alphaMask = (Z >= 4) ? 0 : 0xff

	EMIT("mov [rBX].col, 0");
    X86_emit_op_M_imm32(MOVI_D, FMC(col), 0);       // mov fmc.col, 0           ; col = 0

  L1 = X86_genlab();
  X86_deflab(L1);                                   //L1:
	EMIT2L("cgl%d: ;L1", L1);

#if 0
	EMIT("mov rAX, [rBX].srcPtr");
	EMIT("mov EAX, [rAX]  ;EAX=[a,b,g,r]");
    X86_emit_op_RM(MOV_V, eax, FMC(srcPtr));        // mov rAX, fmcp->srcPtr
    X86_emit_op_RM(MOV_D, eax, eax, no_reg, 0, 0);  // mov eax, [rAX]   ; eax=[a, b, g, r]

#elif 0	//could revert to 1 in FM64??
    // Quick fix to avoid buffer overrun...
	L1a = X86_genlab();
	EMIT("mov rCX, [rBX].srcPtr");
	EMIT("mov AL, [rBX].alphaMask");
	EMIT("or AL, AL");
	EMIT2("jnz cgl%d ;use 255 if no alpha", L1a);
	EMIT("mov AL, [rCX+3] ;AL=alpha");
    X86_emit_op_RM(MOV_V, ecx, FMC(srcPtr));        // mov rCX, fmcp->srcptr
    X86_emit_op_RM(MOV_B, al, FMC(alphaMask));      // mov al, fmcp->alphaMask   ; 0 [alpha present] or 0xff
    X86_emit_op_RR(OR_B, al, al);                   // or al, al
    X86_emit_op_lab_disp8(JNZ, L1a);                // jnz L1a          ; br if no alpha present (alphaMask = 0xff)
    X86_emit_op_RM(MOV_B, al, ecx, no_reg, 0, 3);   // mov al, 3[rCX]   ; eax=[a] (use actual alpha channel pixel val)
  X86_deflab(L1a);                                  //L1a:
	EMIT2L("cgl%d: ;L1a", L1a);
	EMIT("shl EAX, 8 ;AX=[a,0]");
	EMIT("mov AL, [rCX+2] ;AX=[a,b]");
	EMIT("shl EAX, 16 ;EAX=[a,b,0,0]");
	EMIT("mov AX, [rCX] ;EAX=[a,b,g,r]");
    X86_emit_op_R_imm8(SHLI_D, eax, 8);             // shl eax, 8       ; eax=[a, 0]
    X86_emit_op_RM(MOV_B, al, ecx, no_reg, 0, 2);   // mov al, 2[rCX]   ; eax=[a, b]
    X86_emit_op_R_imm8(SHLI_D, eax, 16);            // shl eax, 16      ; eax=[a, b, 0, 0]
    X86_emit_op_RM(MOV_W, ax, ecx, no_reg, 0, 0);   // mov ax, 0[rCX]   ; eax=[a, b, g, r]

#else
    // Quick fix to avoid buffer overrun...
    //working it the other way around...
    //(so can use MOVZX_W instead of MOV_W, which generates a compile-time warning [fixed in FM64!!])
	//Would it be simpler/faster to just copy 1 byte at a time?? Don't really need packed [a, b, r, g], do we???
	L1a = X86_genlab();
	EMIT("mov rCX, [rBX].srcPtr");
	EMIT("movzx EAX, word ptr[rCX] ;EAX=[0,0,g,r]");
	EMIT("ror EAX, 16 ;EAX=[g,r,0,0]");
	EMIT("mov AL, [rCX+2] ;EAX=[g,r,0,b]");
	EMIT("ror EAX, 8 ;EAX=[b,g,r,0]");
	EMIT("mov AL, [rBX].alphaMask");
	EMIT("or AL, AL");
	EMIT2("jnz cgl%d ;use 255 if no alpha", L1a);
	EMIT("mov AL, [rCX+3] ;EAX=[b,g,r,a]");
    X86_emit_op_RM(MOV_V, ecx, FMC(srcPtr));        // mov rCX, fmcp->srcptr
    X86_emit_op_RM(MOVZW_D, eax, ecx, no_reg, 0, 0);// movzx eax, 0[rCX]; eax=[0, 0, g, r]
    X86_emit_op_R_imm8(RORI_D, eax, 16);            // ror eax, 16      ; eax=[g, r, 0, 0]
    X86_emit_op_RM(MOV_B, al, ecx, no_reg, 0, 2);   // mov al, 2[rCX]   ; eax=[g, r, 0, b]
    X86_emit_op_R_imm8(RORI_D, eax, 8);             // ror eax, 8       ; eax=[b, g, r, 0]
    X86_emit_op_RM(MOV_B, al, FMC(alphaMask));      // mov al, fmcp->alphaMask
    X86_emit_op_RR(OR_B, al, al);                   // or al, al        ; 0 [if alpha present] or 0xff [default alpha]
    X86_emit_op_lab_disp8(JNZ, L1a);                // jnz L1a          ; br if no alpha present (alphaMask = 0xff)
    X86_emit_op_RM(MOV_B, al, ecx, no_reg, 0, 3);   // mov al, 3[rCX]   ; al=[a] (use actual alpha channel pixel val)
  X86_deflab(L1a);                                  //L1a:              ; eax=[b, g, r, a]
	EMIT2L("cgl%d: ;L1a", L1a);
	EMIT("ror EAX, 8 ;EAX=[a,b,g,r]");
    X86_emit_op_R_imm8(RORI_D, eax, 8);             // ror eax, 8       ; eax=[a, b, g, r]
#endif
	//to avoid jump above:
	//	mov al, 3[rCX]
	//	or al, fmcp->alphaMask
	//but it doesn't avoid possible buffer overrun!!! NFG
	//for better speed, use dword load on all except last col of buffer??? or last line or pixel of entire buffer???
	//How about just changing the MOV_B's below to MOVZB_D's from memory, and delete other unnecessary byte shuffling??

	EMIT("xor ECX, ECX ;for zero-extension");
	EMIT("mov CL, AL ;zext(r)");
	EMIT("mov [rBX].r, ECX");
    X86_emit_op_RR(XOR_D, ecx, ecx);                // xor ecx, ecx     ; for zero-extension
    X86_emit_op_RR(MOV_B, cl, al);                  // mov cl, al       ; r
    X86_emit_op_MR(MOV_D, FMC(r), ecx);             // mov fmcp->r, ecx
    if (gFmcp->needPremiere) {
		EMIT("mov [rBX].r0, ECX ;Premiere");
		EMIT("mov [rBX].r1, ECX ;Premiere");
        X86_emit_op_MR(MOV_D, FMC(r0), ecx);        // mov fmcp->r0, ecx
        X86_emit_op_MR(MOV_D, FMC(r1), ecx);        // mov fmcp->r1, ecx
    }
	EMIT("mov [rBX].R, ECX");
    X86_emit_op_MR(MOV_D, FMC(R), ecx);             // mov fmcp->R, ecx

	EMIT("mov CL, AH ;zext(g)");
	EMIT("mov [rBX].g, ECX");
    X86_emit_op_RR(MOV_B, cl, ah);                  // mov cl, ah       ; g [note hi bytes still 0]	//X64 ok???
    X86_emit_op_MR(MOV_D, FMC(g), ecx);             // mov fmcp->g, ecx
    if (gFmcp->needPremiere) {
		EMIT("mov [rBX].g0, ECX ;Premiere");
		EMIT("mov [rBX].g1, ECX ;Premiere");
        X86_emit_op_MR(MOV_D, FMC(g0), ecx);        // mov fmcp->g0, ecx
        X86_emit_op_MR(MOV_D, FMC(g1), ecx);        // mov fmcp->g1, ecx
    }
	EMIT("mov [rBX].G, ECX");
    X86_emit_op_MR(MOV_D, FMC(G), ecx);             // mov fmcp->G, ecx

	EMIT("shr EAX, 16 ;AX=[a,b]");
	EMIT("mov CL, AL ;zext(b)");
	EMIT("mov [rBX].b, ECX");
    X86_emit_op_R_imm8(SHRI_D, eax, 16);            // shr eax, 16      ; eax=[0, 0, a, b]
    X86_emit_op_RR(MOV_B, cl, al);                  // mov cl, al       ; b [hi bytes still 0]
    X86_emit_op_MR(MOV_D, FMC(b), ecx);             // mov fmcp->b, ecx
    if (gFmcp->needPremiere) {
		EMIT("mov [rBX].b0, ECX ;Premiere");
		EMIT("mov [rBX].b1, ECX ;Premiere");
        X86_emit_op_MR(MOV_D, FMC(b0), ecx);        // mov fmcp->b0, ecx
        X86_emit_op_MR(MOV_D, FMC(b1), ecx);        // mov fmcp->b1, ecx
    }
    X86_emit_op_MR(MOV_D, FMC(B), ecx);             // mov fmcp->B, ecx

	EMIT("mov CL, AH ;zext(a)");
	EMIT("mov [rBX].a, ECX");
    X86_emit_op_RR(MOV_B, cl, ah);                  // mov cl, ah       ; a [hi bytes still 0]	//X64 ok???
    X86_emit_op_MR(MOV_D, FMC(a), ecx);             // mov fmcp->a, ecx
    if (gFmcp->needPremiere) {
		EMIT("mov [rBX].a0, ECX ;Premiere");
		EMIT("mov [rBX].a1, ECX ;Premiere");
        X86_emit_op_MR(MOV_D, FMC(a0), ecx);        // mov fmcp->a0, ecx
        X86_emit_op_MR(MOV_D, FMC(a1), ecx);        // mov fmcp->a1, ecx
    }
	EMIT("mov [rBX].A, ECX");
    X86_emit_op_MR(MOV_D, FMC(A), ecx);             // mov fmcp->A, ecx

#if 0
  L2 = X86_genlab();
    X86_emit_op_M_simm32(CMPI_D, FMC(need_iuv), 0); // cmp fmcp->need_iuv, 0
    X86_emit_op_lab_disp8(JZ, L2);                  // jz L2
        X86_emit_compute_iuv();                     // <compute i, u, v>
  X86_deflab(L2);                                   //L2:
  L3 = X86_genlab();
    X86_emit_op_M_simm32(CMPI_D, FMC(need_d), 0);   // cmp fmcp->need_d, 0
    X86_emit_op_lab_disp8(JZ, L3);                  // jz L3
        X86_emit_compute_d();                       // <compute d>
  X86_deflab(L3);                                   //L3:
  L4 = X86_genlab();
    X86_emit_op_M_simm32(CMPI_D, FMC(need_m), 0);   // cmp fmcp->need_m, 0
    X86_emit_op_lab_disp8(JZ, L4);                  // JZ L4
        X86_emit_compute_m();                       // <compute m>
  X86_deflab(L4);                                   //L4:
#else
    if (gFmcp->need_iuv) {
        X86_emit_compute_iuv();                     // <compute i, u, v>
    }
    if (gFmcp->need_d) {
        X86_emit_compute_d();                       // <compute d>
    }
    if (gFmcp->need_m) {
        X86_emit_compute_m();                       // <compute m>
    }
#endif
    //channel code goes here...
	EMIT("<RGBA channel code goes here>");
} /*X86_emit_doOneRow_part1*/

void X86_emit_doOneRow_part2(int nTempsXXX, int cbLocalsXXX)	//X64 RESUME HERE!!
{
    CGL L2, L3, L4, L5, L6, L7, L8, L9, L10;

	EMIT("<RGBA channel code goes above>");
    //  <RGBA channel code>
    //  dstPtr[0] = (R<0)?0:(R>255)?255:R;
    //  dstPtr[1] = (G<0)?0:(G>255)?255:G;
    //  dstPtr[2] = (B<0)?0:(B>255)?255:B;
    //  if (writeAlpha) dstPtr[3] = (A<0)?0:(A>255)?255:A;
    //  srcPtr += srcColDelta;
    //  dstPtr += dstColDelta;
    //  mskPtr += mskColDelta;
    //  ++x;
    //  ++col;
    //  if (col < columns) goto L1;
    //  <epilog>

    if (gFmcp->use_MMX) {	//X64 -- REVIEW THIS!!!
		EMIT("<using MMX extensions...>");
        // Following uses MMX extensions to pack and clip 4 ints into 4 bytes, faster than x86 ops.
        // But could cause GP at end of buffer if Z < 4 !!!
		EMIT("mov rCX, [rBX].dstPtr");
		EMIT("movq MMX0,[rBX].R ;MMX0=[G,R]");
		EMIT("packssdw MMX0, [rBX].B ;[A,B,G,R] shorts");
		EMIT("packuswb MMX0,MMX0 ;[A,B,G,R] bytes");
		EMIT("movd [rCX], MMX0 ;store in dest");
		EMIT("emms ;empty MMX state");
        X86_emit_op_RM(MOV_V, ecx, FMC(dstPtr));      // mov rCX, fmcp->dstPtr
        X86_emit_op_RM(MOVQ, mmx0, FMC(R));           // movq mmx0, fmcp->R	    ; mmx0=[G, R]
        X86_emit_op_RM(PACKSSDW, mmx0, FMC(B));       // packssdw mmx0, fmcp->B ; pack 4 ints [A, B, G, R] into 4 shorts
        X86_emit_op_RR(PACKUSWB, mmx0, mmx0);         // packuswb mmx0, mmx0    ; pack 4 signed shorts into 4 unsigned bytes
                                                      //                        ; (actually 4 lower bytes are replicated into
													  //						; 4 higher bytes: [A, B, G, R, A, B, G, R])
        X86_emit_op_MR(MOVD, ecx, no_reg, 0, 0, mmx0);//movd [rCX], mmx0        ; store R, G, B, A bytes in destination buffer
        X86_emit_op(EMMS);							  // emms                   ; empty MMX state
		EMIT("<end of MMX code>");
    }
    else {

        // Could simplify/optimize if a compile option is set to specify that no clipping is needed??

		EMIT("mov rCX, [rBX].dstPtr");
        X86_emit_op_RM(MOV_V, ecx, FMC(dstPtr));    // mov rCX, fmcp->dstPtr

        // Could combine following sections using macro or subr (for easier maintenance)??
        //
        // dstPtr[0] = (R<0)?0:(R>255)?255:R;
        //
        L2 = X86_genlab();
        L3 = X86_genlab();                          // (2 labels are necessary until disp8 fixup chaining is supported)
		EMIT("mov EAX, [rBX].R");
		EMIT("cmp EAX, 255");
		EMIT2("jbe cgl%d ;ok if <=255", L2);
		EMIT("mov AL, 255 ;clamp to 255");
		EMIT2("jnl cgl%d ;ok if not neg", L3);
		EMIT("inc EAX ;clamp to 0");
		X86_emit_op_RM(MOV_D, eax, FMC(R));         // mov eax, fmcp->R     ; eax=[R]
        X86_emit_op_R_simm32(CMPI_D, eax, 255);     // cmp eax, 255         ; clip to [0,255]
        X86_emit_op_lab_disp8(JBE, L2);             // jbe L2               ; okay if not above 255 (unsigned comparison)
        X86_emit_op_R_imm8(MOVI_B, al, 255);        // mov al, 255          ; tentatively clamp to 255
        X86_emit_op_lab_disp8(JNL, L3);             // jnl L3               ; flags still set from cmp above
        X86_emit_op_R(INC_D, eax);                  // inc eax              ; clamp to 0 if negative (treating as signed)
      X86_deflab(L2);                               //L2:
      X86_deflab(L3);                               //L3:
		EMIT2L("cgl%d: ;L2", L2);
		EMIT2L("cgl%d: ;L3", L3);
		EMIT("mov [rCX], AL ;store clamped R");
        X86_emit_op_MR(MOV_B, ecx, no_reg, 0, 0, al);//mov 0[rCX], al       ; store (clamped) R in destination byte
        //
        //  dstPtr[1] = (G<0)?0:(G>255)?255:G;
        //
        L4 = X86_genlab();
        L5 = X86_genlab();                          // (2 labels are necessary until disp8 fixup chaining is supported)
		EMIT("mov EAX, [rBX].G");
		EMIT("cmp EAX, 255");
		EMIT2("jbe cgl%d ;ok if <=255", L4);
		EMIT("mov AL, 255 ;clamp to 255");
		EMIT2("jnl cgl%d ;ok if not neg", L5);
		EMIT("inc EAX ;clamp to 0");
        X86_emit_op_RM(MOV_D, eax, FMC(G));         // mov eax, fmcp->G     ; eax=[G]
        X86_emit_op_R_simm32(CMPI_D, eax, 255);     // cmp eax, 255         ; clip to [0,255]
        X86_emit_op_lab_disp8(JBE, L4);             // jbe L4               ; okay if not above 255 (unsigned comparison)
        X86_emit_op_R_imm8(MOVI_B, al, 255);        // mov al, 255          ; tentatively clamp to 255
        X86_emit_op_lab_disp8(JNL, L5);             // jnl L5               ; flags still set from cmp above
        X86_emit_op_R(INC_D, eax);                  // inc eax              ; clamp to 0 if negative (treating as signed)
      X86_deflab(L4);                               //L4:
      X86_deflab(L5);                               //L5:
		EMIT2L("cgl%d: ;L4", L4);
		EMIT2L("cgl%d: ;L5", L5);
		EMIT("mov [rCX+1], AL ;store clamped G");
        X86_emit_op_MR(MOV_B, ecx, no_reg, 0, 1, al);//mov 1[rCX], al       ; store (clamped) G in destination byte
        //
        //  dstPtr[2] = (B<0)?0:(B>255)?255:B;
        //
        L6 = X86_genlab();
        L7 = X86_genlab();                          // (2 labels are necessary until disp8 fixup chaining is supported)
		EMIT("mov EAX, [rBX].B");
		EMIT("cmp EAX, 255");
		EMIT2("jbe cgl%d ;ok if <=255", L6);
		EMIT("mov AL, 255 ;clamp to 255");
		EMIT2("jnl cgl%d ;ok if not neg", L7);
		EMIT("inc EAX ;clamp to 0");
        X86_emit_op_RM(MOV_D, eax, FMC(B));         // mov eax, fmcp->B     ; eax=[B]
        X86_emit_op_R_simm32(CMPI_D, eax, 255);     // cmp eax, 255         ; clip to [0,255]
        X86_emit_op_lab_disp8(JBE, L6);             // jbe L6               ; okay if not above 255 (unsigned comparison)
        X86_emit_op_R_imm8(MOVI_B, al, 255);        // mov al, 255          ; tentatively clamp to 255
        X86_emit_op_lab_disp8(JNL, L7);             // jnl L7               ; flags still set from cmp above
        X86_emit_op_R(INC_D, eax);                  // inc eax              ; clamp to 0 if negative (treating as signed)
      X86_deflab(L6);                               //L6:
      X86_deflab(L7);                               //L7:
		EMIT2L("cgl%d: ;L6", L6);
		EMIT2L("cgl%d: ;L7", L7);
		EMIT("mov [rCX+2], AL ;store clamped B");
        X86_emit_op_MR(MOV_B, ecx, no_reg, 0, 2, al);//mov 2[rCX], al       ; store (clamped) B in destination byte
        //
        //  if (writeAlpha) dstPtr[3] = (A<0)?0:(A>255)?255:A;
        //
        L8 = X86_genlab();
        L9 = X86_genlab();                          // (2 labels are necessary until disp8 fixup chaining is supported)
        L10 = X86_genlab();
		EMIT("mov EAX, [rBX].writeAlpha");
		EMIT2("jz cgl%d ;br if not writing alpha", L10);
		EMIT("mov EAX, [rBX].A");
		EMIT("cmp EAX, 255");
		EMIT2("jbe cgl%d ;ok if <=255", L8);
		EMIT("mov AL, 255 ;clamp to 255");
		EMIT2("jnl cgl%d ;ok if not neg", L9);
		EMIT("inc EAX ;clamp to 0");
        X86_emit_op_RM(MOV_D, eax, FMC(writeAlpha));// mov eax, fmcp->writeAlpha ; writing alpha channel?
        X86_emit_op_RR(OR_D, eax, eax);             // or eax, eax
        X86_emit_op_lab_disp8(JZ, L10);             // jz L10               ; skip unless writing alpha channel to dest
        X86_emit_op_RM(MOV_D, eax, FMC(A));         // mov eax, fmcp->A     ; eax=[A]
        X86_emit_op_R_simm32(CMPI_D, eax, 255);     // cmp eax, 255         ; clip to [0,255]
        X86_emit_op_lab_disp8(JBE, L8);             // jbe L8               ; okay if not above 255 (unsigned comparison)
        X86_emit_op_R_imm8(MOVI_B, al, 255);        // mov al, 255          ; tentatively clamp to 255
        X86_emit_op_lab_disp8(JNL, L9);             // jnl L9               ; flags still set from cmp above
        X86_emit_op_R(INC_D, eax);                  // inc eax              ; clamp to 0 if negative (treating as signed)
      X86_deflab(L8);                               //L8:
      X86_deflab(L9);                               //L9:
		EMIT2L("cgl%d: ;L8", L8);
		EMIT2L("cgl%d: ;L9", L9);
		EMIT("mov [rCX+3], AL ;store clamped A");
        X86_emit_op_MR(MOV_B, ecx, no_reg, 0, 3, al);//mov 3[rCX], al       ; store (clamped) A in destination byte
      X86_deflab(L10);                              //L10:
		EMIT2L("cgl%d: ;L10", L10);
    } //if not use_MMX
    //
    // srcPtr += srcColDelta;
    //
	EMIT("mov EAX, [rBX].srcColDelta");
	EMIT("add [rBX].srcPtr, rAX ;increment");
    X86_emit_op_RM(MOV_D, eax, FMC(srcColDelta));   // mov eax, fmcp->srcColDelta (zero-extends to 64-bits in X64)	//X64 RESUME!!!
    X86_emit_op_MR(ADD_V, FMC(srcPtr), eax);        // add fmcp->srcPtr, rAX
    //
    // dstPtr += dstColDelta;
    //
	EMIT("mov EAX, [rBX].dstColDelta");
	EMIT("add [rBX].dstPtr, rAX ;increment");
    X86_emit_op_RM(MOV_D, eax, FMC(dstColDelta));   // mov eax, fmcp->dstColDelta (zero-extends to 64-bits in X64)
    X86_emit_op_MR(ADD_V, FMC(dstPtr), eax);        // add fmc->dstPtr, rAX
    //
    // mskPtr += mskColDelta;
    //
	EMIT("mov EAX, [rBX].mskColDelta");
	EMIT("add [rBX].mskPtr, rAX ;increment");
    X86_emit_op_RM(MOV_D, eax, FMC(mskColDelta));   // mov eax, fmcp->mskColDelta (zero-extends to 64-bits in X64)
    X86_emit_op_MR(ADD_V, FMC(mskPtr), eax);        // add fmc->mskPtr, rAX

	EMIT("inc dword ptr[rBX].x ;++x");
	EMIT("mov EAX, [rBX].col");
	EMIT("inc EAX ;++col");
	EMIT("mov [rBX].col, EAX");
	EMIT("cmp EAX, [rBX].columns");
	EMIT2("jl cgl%d ;L1 if more cols", L1);
    X86_emit_op_M(INC_D, FMC(x));                   // inc fmcp->x           ; ++x
    X86_emit_op_RM(MOV_D, eax, FMC(col));           // mov eax, fmcp->col
    X86_emit_op_R(INC_D, eax);                      // inc eax
    X86_emit_op_MR(MOV_D, FMC(col), eax);           // mov fmcp->col, eax    ; ++col
    X86_emit_op_RM(CMP_D, eax, FMC(columns));       // cmp eax, fmcp->columns
    X86_emit_op_lab(JL, L1);                        // jl L1				 ; if (col < columns) goto L1

    X86_emit_epilog(nTempsXXX, cbLocalsXXX, -1);			//nArgs = 1 vpl
} /*X86_emit_doOneRow_part2*/

#endif //ifndef STANDALONE

#if 0
////////////////// PATCH/TEST AREA /////////////////////

/// Test x87 fpu code generation by VC++

float funcx (float aa, float bb, float cc, float dd, float ee);

float func1 (float a, float b, double c, float d, float e, float f) {
	float zz;
	double dd;
	int ii;
	long long jj;
	zz = a+b+c+d+e+f;
	zz = funcx(a,b,c,d,zz);
	zz = (a+b)*funcx(a,b,c,d,e)*funcx(e,a,b,c,d)*(c-d);
	dd = zz;
	jj = dd;
	ii = zz;
	jj = zz;
	ii = dd;
	return zz*a;
}//func1

///////////////// END OF PATCH/TEST AREA ///////////////
#endif

#if 0
////////////////// PATCH/TEST AREA /////////////////////

// Test whether 32-byte static shadow arglist is generated
// by VC++ even if only function calls are to parameterless
// functions.

void v0(void);
int i0(void);
int i1(int i);

void test0(void) {
	//int i, j;
	v0();
	//i = i0();
	////i += i1(i);
}//test0

///////////////// END OF PATCH/TEST AREA ///////////////
#endif

#if 0
////////////////// PATCH/TEST AREA /////////////////////

// Test whether 32-byte static shadow arglist is generated
// by VC++ even if only function calls are to parameterless
// functions.

// Test whether structs of size 1, 2, 3, 4, 5, 6, 7, 8, 9,
// and 10 bytes are passed by value (in a reg or stack slot)
// or by reference.  The MSDN says by 1, 2, 4, and 8-byte
// structs ar passed by value, while > 8 byte structs are
// passed by reference. It is somewhat vague about 3, 5, 6,
// and 7-byte structs, but seems to imply they are passed
// buy reference.  However, I think this is wrong in principle,
// and that any struct <= 8 bytes should be passed by value.
// Also, check how a struct consisting or a single f.p.
// value is passed.

void ps();

struct {char b1;} sb1;
struct {char b1,b2;} sb2;
struct {char b1,b2,b3;} sb3;
struct {char b1,b2,b3,b4;} sb4;
struct {char b1,b2,b3,b4,b5;} sb5;
struct {char b1,b2,b3,b4,b5,b6;} sb6;
struct {char b1,b2,b3,b4,b5,b6,b7;} sb7;
struct {char b1,b2,b3,b4,b5,b6,b7,b8;} sb8;
struct {char b1,b2,b3,b4,b5,b6,b7,b8,b9;} sb9;
struct {char b1,b2,b3,b4,b5,b6,b7,b8,b9,b10;} sb10;

struct {short s1;} ss1;
struct {int i4;} si4;
struct {float f4;} sf4;
struct {__int64 i8;} si8;
struct {double d8;} sd8;
struct {int i4, j4;} si4j4;
struct {int i4; char b1;} si4b1;

void test1(void) {
	printf("%d %d %d %d %d %d %d %d %d %d",
		sizeof(sb1),
		sizeof(sb2),
		sizeof(sb3),
		sizeof(sb4),
		sizeof(sb5),
		sizeof(sb6),
		sizeof(sb7),
		sizeof(sb8),
		sizeof(sb9),
		sizeof(sb10) );
	ps(sb1,sb2,sb3,sb4,sb5,sb6,sb7,sb8,sb9,sb10);
	//sb3,5,6,7,9, and 10 are passed by reference
	//sb1,2,4, and 8 are passed by value
	//so MSDN is correct!!

	//using individual calls, so params are passed in RCX...
	ps(sb1); ps(sb2); ps(sb3); ps(sb4); ps(sb5); ps(sb6); ps(sb7); ps(sb8); ps(sb9); ps(sb10); 

	//some other tests...
	ps(ss1);
	ps(si4);
	ps(sf4);
	ps(si8);
	ps(sd8);
	
	//handling of scalars in a non-prototyped call...
	{//scope
		char bnoproto;
		short snoproto;
		unsigned short usnoproto;
		int inoproto;
		float fnoproto;
		double dnoproto;

		//integral types are zero/sign-extended to 32-bits, then zero-extended to 64-bits
		//reals are *promoted* to double in xmm0 if necessary, then copied to rcx
		ps();
		ps(bnoproto);
		ps(snoproto);
		ps(usnoproto);
		ps(inoproto);
		ps(fnoproto); //cvtss2sd xmm0, DWORD PTR fnoproto; movd rcx, xmm0
		ps(dnoproto); //movsdx   xmm0, QWORD PTR dnoproto; movd rcx, xmm0
	}//scope


}//test1

///////////////// END OF PATCH/TEST AREA ///////////////
#endif

#if 0
////////////////// PATCH/TEST AREA /////////////////////

// Test how VC++ passes real args to vararg and unprototyped
// functions.

int noproto();

float x, y, z;
double xx, yy, zz;

void func9(float a, float b, char *fmt) {
	//calls to unprototyped function
	noproto(1.0, 2.0, 3.0, 4.0, 5.0);
	noproto(x, xx, y, yy, z);
	noproto(a, b, x, y, z);

	//calls to (well-known) vararg function
	printf(fmt, 1.1, x, a, xx, b);

	//will VC++ make use of knowledge from a known format
	//string to optimize the arg passsing??
	//No, same arg passing code, at least for Debug
	//Does optimize level make a difference??
	//No, apparently not (tried /O2)
	printf("%f %f %f %f %f", 1.1, x, a, xx, b);
}//func9

///////////////// END OF PATCH/TEST AREA ///////////////
#endif

#if 0
////////////////// PATCH/TEST AREA /////////////////////

// Test how VC++ passes real args to vararg and unprototyped
// functions.

// Test whether VC++ (and/or VC++) allows arbitrary permutation
// of storage-specifiers, type-specifiers, and type-qualifiers.

//name scope
int f32(void) {
#if 0
	signed long int i1;
	int long signed i2;

	static int i3;
	int static i4;

	static i5;

	const int i6;
	int const i7;

	double long i8;	//=long double = double

	unsigned float x1;	//warning C4076: 'unsigned' : can not be used with type 'float'

	typedef long long long longinteger;	//error C2632: '__int64' followed by 'long' is illegal

	long typedef long longinteger;
	longinteger z1;

	long char c1;				//error C2632: 'long' followed by 'char' is illegal
	signed int unsigned i11;	//error C2042: signed/unsigned keywords mutually exclusive

	long float z2;			//warning C4215: nonstandard extension used : long float
	long long double z3;	//error C2632: '__int64' followed by 'double' is illegal
	long double long z4;	//error C2632: 'long double' followed by 'long' is illegal

	int int int i12;	//error C2632: 'int' followed by 'int' is illegal
	signed signed i13;	//warning C4114: same type qualifier used more than once
	unsigned long;		//warning C4091: ' ' : ignored on left of 'unsigned long' when no variable is declared
#endif
	///new///
	//auto register rax;	//error C2159: more than one storage class specified
	volatile register db01;
	register struct {char a,b;} xyz;
	//signed bool b1;
	//unsigned bool b2;
	signed long l1;
	unsigned long l2;
	//tempreal xx;
	enum {q,w,e,r,t} qwerty;
	short enum {f,g,h} fgh;	//error C2236: unexpected 'enum' '<unnamed-tag>'. Did you forget a ';'?
	signed enum {v,b,n} vbn;//error C2236: unexpected 'enum' '<unnamed-tag>'. Did you forget a ';'?
	unsigned enum {a,s,d} asd;//error C2236: unexpected 'enum' '<unnamed-tag>'. Did you forget a ';'?
	signed float xx1;		//warning C4076: 'signed' : can not be used with type 'float'
	signed double xxxx1;	//warning C4076: 'signed' : can not be used with type 'double'
	signed _int64 ll1;
	unsigned float zz1;		//warning C4076: 'unsigned' : can not be used with type 'float'
	unsigned double zzzz2;	//warning C4076: 'unsigned' : can not be used with type 'double'
	unsigned _int64 uin1;
	signed short ss01;
	unsigned short us02;

	//i14;				//taken as expressions, not empty declarations
	//i15, i16, i17;



	return db01 + xyz.b;

}//scope

///////////////// END OF PATCH/TEST AREA ///////////////
#endif

#if 0
////////////////// PATCH/TEST AREA /////////////////////

/// Test various parse-level error messages in MSVC

void parsetestxyzzy(void)
{
	{//unsigned integer constant overflow
	unsigned int dw1, dw2, dw3;
	unsigned __int64 qw1, qw2, qw3;
	dw1 = 0x12345678;
	dw2 = 0x123456789;		// warning C4305: '=' : truncation from '__int64' to 'unsigned int'
	dw3 = 0x123456789LL;				// warning C4305: '=' : truncation from '__int64' to 'unsigned int'
	qw1 = 0x1234567812345678;		//63 bits
	//qw2 = 0x12345678123456789;		// error C2177: constant too big
	//qw3 = 0x12345678123456789LL;		// error C2177: constant too big
	qw2 = 0x8765432187654321;		//64 bits
	qw3 = 0x8765432187654321L;		//no warn for 64-bit 'L' constant!!
	}

	{//signed integer constant overflow
		int i1, i2, i3, i4;
		__int64 ii1, ii2, ii3, ii4, ii5;
		i1 = INT_MAX;
		i2 = 2147483648;	//INT_MAX+1 (large neg #)
		i3 = LLONG_MAX;					// warning C4305: '=' : truncation from '__int64' to 'unsigned int'
		i4 = 9223372036854775808i64;	//LLONG_MAX+1 // warning C4305: '=' : truncation from '__int64' to 'unsigned int'
		ii1 = ULLONG_MAX;	//treated as -1
		ii2 = 2147483648;	//INT_MAX+1
		//ii3 = 18446744073709551616;		// error C2177: constant too big
		ii4 =  9223372036854775808i64;	//LLONG_MAX+1 (no error,large negv #?)
		ii5 = 0;
	}

	{//Character constants
		__int64 j0, j1, j2, j3, j4, j5, j6, j7,j8, j9;
		__int64 w0, w1, w2, w3, w4, w5;
		__int64 wx, wy, wz;
		//j0 = '';	//error C2137: empty character constant
		j1 = 'a';
		j2 = 'ab';
		j3 = 'abc';
		j4 = 'abcd';		//=61626364H
		//j5 = 'abcde';		//error C2015: too many characters in constant
		//j6 = 'abcdef';	//error C2015: too many characters in constant
		//j7 = 'abcdefg';	//error C2015: too many characters in constant
		//j8 = 'abcdefgh';	//error C2015: too many characters in constant
		//j9 = 'abcdefghi';	//error C2015: too many characters in constant
		w0 = L'';
		w1 = L'a';
		w2 = L'ab';		//warning C4066: characters beyond first in wide-character constant ignored
		w3 = L'abc';	//ditto
		w4 = L'abcd';	//ditto
		w5 = L'abcde';	//ditto
		//wx = l'a';	//error C2065: 'l' : undeclared identifier
		wx = L'';
		wy = L'A';
		wz = L'AB';		//warning C4066: characters beyond first in wide-character constant ignored
////i1 = "
////error C2001: newline in constant";
	}

} /*parsetestxyzzy*/
/** unterminated block style comment
fatal error C1071: unexpected end of file found in comment **/
///////////////// END OF PATCH/TEST AREA ///////////////
#endif

#if 0
////////////////// PATCH/TEST AREA /////////////////////

// Test how VC++ x32 handles save/restore of ndp regs across
// function calls, esp. calls to real functions.
// Bah. It evaluates calls first, if necessary.

int ifunc(void);
double xfunc(void);

void test0(void) {
	double xx, yy, zz;
	int i, j, k;
	i = ifunc();
	xx = xfunc();
	xx = xx*yy + zz*xfunc();
	xx = 1.0 + (xx + (yy + (zz * xfunc())));
}//test0

///////////////// END OF PATCH/TEST AREA ///////////////
#endif

#if 0
////////////////// PATCH/TEST AREA /////////////////////

// Try to find out *exactly* when VC++ x64 replicates
// float args into int regs.

int ifunc();	//unprototyped
double xfunc();	//unprototyped
int vfunc(double x, double y,...);	//varargs

void test0(void) {
	double xx, yy, zz;
	double x1, x2, x3, x4, x5, x6;
	int i, j, k;
	i = ifunc(x1, x2, x3, x4, x5, x6);
	xx = xfunc(x1, x2, x3, x4, x5, x6);
	j = vfunc(x1, x2, x3, x4, x5, x6);
}//test0

///////////////// END OF PATCH/TEST AREA ///////////////
#endif

#if 0
////////////////// PATCH/TEST AREA /////////////////////

// See how VC++ implements op_load/op_store of different
// sizes and signed-ness.
// Does VC++ use int32 as the default calculation type,
// or does it use minimujm calc size necessary.  Does
// it ever use int64 without being forced?

void test99(void) {
	float x, y, z;
	double xx, yy, zz;
	char c1, c2, c3;
	unsigned char b1, b2, b3;
	short s1, s2, s3;
	unsigned short w1, w2, w3;
	int i1, i2, i3;
	unsigned int u1, u2, u3;
	__int64 l1, l2, l3;
	unsigned __int64 q1, q2, q3;

	c1 = c2;
	s1 = c1;
	i1 = c3;
	l1 = c1;

	s1 = s2;
	i2 = s3;
	l2 = s1;

	i3 = i2;
	l3 = i3;

	l1 = l2;

	l1 = c1;
	l2 = s1;
	l3 = i1;
	l2 = l1;

	q1 = b1;
	q2 = w2;
	q3 = u1;
	q1 = q2;

	c1 = c2 + c3;
	s1 = s2 - s3;
	i2 = i1 + i3;
	l1 = l2 - l3;

	// intermediate computation when final result is int64

	l1 = i1 * i2;
	l2 = (__int64)i1 * i3;
	l3 = i1 * (__int64)i2;

	l1 = s1 + s2;
	l2 = s2 * s3;

	l2 = c1 - c2;
	l3 = c2 * c3;

	q1 = b1 + b2;
	q2 = b1 - b2;
	l1 = b1 + b2;
	q3 = w1 - w2;
	l3 = w1 - w2;
	q2 = u1 - u2;
	l2 = u1 + u3;
}//test99

void test100(void) {

	unsigned char	 b1 = 1, b2 = 2;
	unsigned short	 w1 = 1, w2 = 2;
	unsigned int	 u1 = 1, u2 = 2;
	unsigned __int64 q1, q2, q3;

	q1 = b1 - b2;	//int difference is sign-extended to uint64 result (0xffffffffffffffff)
	q2 = w1 - w2;	//int difference is sign-extended to uint64 result (0xffffffffffffffff)
	q3 = u1 - u2;	//int difference is zero-extended to uint64 result (0x00000000ffffffff)
}//test100

///////////////// END OF PATCH/TEST AREA ///////////////
#endif

#if 0
////////////////// PATCH/TEST AREA /////////////////////

// How does VC++ generate op_preinc (++i) and op_postinc (i++),
// with or without final rvalue?

void test0(void) {
	float x, y, z;
	double xx, yy, zz;
	char c1, c2, c3;
	unsigned char b1, b2, b3;
	short s1, s2, s3;
	unsigned short w1, w2, w3;
	int i1, i2, i3;
	unsigned int u1, u2, u3;
	__int64 l1, l2, l3;
	unsigned __int64 q1, q2, q3;

	//op_preinc (++i)
	++i1;
	i2 = ++i3;
	++xx;
	yy = ++zz;

	//op_postinc (i++)
	i1++;
	i2 = i3++;
	xx++;
	yy = zz++;

}//test0

///////////////// END OF PATCH/TEST AREA ///////////////
#endif

#if 1
////////////////// PATCH/TEST AREA /////////////////////

// How does VC++ handle forward references in decl-inits?
// Forms of static local inits?
// Forms of global inits?
// HANDLE?
//

#include <math.h>

void test3(int i, int j, int k) {
	int f;
	f = i+j+k;
}

void *test0(void) {
//	/*static*/ int foo = foo + 3; //VC++ doesn't allow non-constant static local inits?
	static void *pv = &pv;	//constant init is okay, even with self-ref
//	HANDLE h;
	double x = 2.1;
	double y = 3.3;
	double x2y;
	x2y = pow(x,y);
	test3(1,2,3);
	x = pow(y, 3);
	return pv;
}//test0

//void *pv2 = test0();

///////////////// END OF PATCH/TEST AREA ///////////////
#endif
