/*
  File: AfhFMUI.c

  C source file for Windows specific code for the
  AFH Filter Meister Photoshop Plug-in Filter.

    Copyright 1997-2014 AFH Systems.

  Derived from Photoshop SDK source files which are:

    Copyright 1991-95 by Adobe Systems, Inc.

  FilterMeister is free software. You can redistribute and modify it
  under the terms of the MPL2 License.
  See: https://www.mozilla.org/en-US/MPL/2.0/
  Any changes to the FilterMeister files have to be published or sent
  to harry@thepluginsite.com. The filter plugin code itself does not
  need to be disclosed.
*/

/****************************************************************
		REVISION HISTORY

	14JUL2013	A. Hunter	1. Changes for WIN64.
	30DEC2013	A. Hunter	1. Change format discriptor for zoom% from %d to %g.
	10JAN2014	A. Hunter	1. Change NOOPTIMZE_QUICK_TOGGLE to NOOPTIMIZE_QUICK_TOGGLE.
							2. Make changes for 32-bit coordinates.
	30JAN2014	A. Hunter	1. Reset filterTitleText, filterCategoryText, and
							   filterFilename strings at start of parse (since parser
							   may now set values for these strings).
	19FEB2014	A. Hunter	1. Set gParams->fValidCompile = TRUE when LoadFmCode succeeds.
	21Feb2014	A. Hunter	1. _MT mods (for statically linking the CRT) kindly provided 
							   by Harald Heim.
    27Feb2014	A. Hunter	1. Don't assert() in FMprint if no edit control -- but I still
							   need to determine who is calling FMprint from a standalone!
	27Feb2014	A. Hunter	1. Update the "welcome" message. Delete line breaks but retain
							   paragraph breaks per suggestion from Kohan.
	05MAR2014   A. Hunter	1. Guard all calls to X86_call_fmf1 by a call to setjmp,
							   while X64 exception handling is not yet working.

        TO DO

  - F12 no longer toggles editor?
  - use F11 to enter advanced mode?
  - editor controls are disabled every other time FM invoked - FIXED
  - preview cursor not restored on right button up?
  - globals param is obsolete in many cases, but fix setZoom
    et al anyway
  - local imageOffset in setZoom?
  - IDC_HELP --> IDC_HELP1 - DONE
  - Should setZoom skip 13, 15?
  - Is fDragging logic needed in setZoom?
  - Should setZoom be doing SetFocus???
  - Why did HH comment out MOUSEOUT logic in WM_MOUSEMOVE?
  - Fix &rcCtl != NULL logic
  - Check ClickDrag logic
  - Add pixelDepth, simplify logic
  - Make CTL_PREVIEW more like a user control (set inuse, etc.) and
    thereby simplify logic elsewhere
  - Should welcome be displayed only first time FM installed, or
    each time upgraded/re-installed?  (E.g., record current version
    in welcome field in registry, display welcome if different?)
  - Timer events use same ctl numbers as user controls?  Maybe they
    should be considered user controls in their own right, with own
    unique ctl numbers?
  - Add IDC_ARROW etc. as manifest constants??
  - Allow setPreviewCursor to accept a resource name?
  - Locate help.htm via various paths???
  - Add file name and dirty indicator to editor dialog title
  - Need to free the RichEdit library in AfhFMeditor.c???
  - Correctly restore minimized editor dialog window on subsequent
    invocations.
  - replace hard-coded control ids with symbolic names from resource.h
  - Init kills default region with upper rounded corners around the
    default filter dialog on XP (whether or not Visual Theme are enabled)
    -need to save/restore current dialog region (if present) instead of
    setting region to null in fm_resetAllCtls?
  - add itoa, ltoa, ultoa (non-ANSI)
  - add atof, atoi, atol (ANSI)
  - implement errno?

  - allow empty list of class-specific control attributes
  - ability to specify width of buddy1 NUMEDIT in STANDARD control
  - move Font... control back to position after Save...
  - bug when standard control val < -32k or > 32k?
  - createCtl() needs to reset divisor=1
  - setCtlDivisor() needs to update NUMEDIT buddy

  - cleanup and simplify the .pre_ctl logic:
  --- make sure .pre_ctl is set whenever .val is set
  --- better yet, replace .val everywhere with fmc.pre_ctl
      (except still use .val to save .pre_ctl across invocations)

  - restrict STANDARD/SCROLLBAR range to (-32k, +32k-1) ?
  - what is range for TRACKBAR?

  - interpolate filterFilenameText before using it in AfhFMmake
    so Filename:"!T" will work?

  ========================================================
  From the documentation for SetWindowRgn:

      If the bRedraw parameter is TRUE, the system sends the 
      WM_WINDOWPOSCHANGING and WM_WINDOWPOSCHANGED messages 
      to the window. 

  Would that be a better way to force our own redraws when needed?

  Also note:

      After a successful call to SetWindowRgn, the operating 
      system owns the region specified by the region handle hRgn. 
      The operating system does not make a copy of the region. 
      Thus, you should not make any further function calls with 
      this region handle. In particular, do not close this region 
      handle. 




  ======================= B U G S ========================

  - Correctly abort if SetupFilterRecordForProxy() gets an error,
    so host app won't crash. (Kohan)

  - fm_srcp() should mask off a if Z < 4 ???

  =======

    I'm currently looking at a problem in the 1.0 beta
    which crashes PSP and Photoshop, but doesn't crash in earlier
    versions, or in the latest build from my own sources.

    To reproduce the crash:

    1. Open a new instance of FM 1.0 beta in PSP or PS.
    2. Open the editor with blank source.
    3. Enter a single blank (or some other trivial FFP program) in
        the editor.
    4. Press the OK button.
    5. When asked if you want to recompile, click YES.
    6. The host app crashes.

   - above crash no longer occurs in 0.4.21 !???,
   - whoops, yes it does!!  PS disappears altogether now (no GP message)!

   - But a syntax error in 0.4.21 now crashes PS 6.0!!!
   --- could it be related to the NT/XP tooltip bug?
   --- seems to crash in SetFocus().  Problem with FM intercepting
       exceptions?

  **** AH! -- It's a simple stack overflow in FilterDialogProc(),
       caused by szCode[] growing too large.
       - make szCode static for now
       - eventually, eliminate use of szCode altogether

/***************************************************************/

#include <windows.h>
#include <commdlg.h>
#include <commctrl.h>
#include <windowsx.h>       // For SetDlgMsgResult()
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stddef.h>         /* offsetof */
#include <process.h>        // _beginthread
#include <time.h>
#include <assert.h>
#include <direct.h>
#include "PITypes.h"
#include "PIGeneral.h"
#include "PIFilter.h"
#include "PIAbout.h"
#include "PIProperties.h"
#include "PIUtilities.h"

#include "plmcvax.h"
#include "AfhFM.h"
#include "AfhFMcontext.h"
#include "EmitX86.h"
#include "mdp.h"            //parser flags

#include "md386def.hc"      //MD386__SYNTAX

#if defined(FM_LIFTCOV)
#include "liftcov/resliftcov.h"
#elif defined(FM_RADSINAL)
#include "RadSinal/resRadSinal.h"
#elif defined(FM_MONOMET)
#include "monomet/resmonomet.h"
#else
#include "resource.h"
#endif

#if 1	//deprecated for WIN64
#include "WinUtil.h"    // Message crackers and stuff
#endif

#ifndef USE_EDIT_CONTROL
#include "RichEdit.h"   //Rich Edit Control definitions
#endif //USE_EDIT_CONTROL

HANDLE hDllInstance=NULL;  /* DLL instance handle */

HWND ghTT;  //tooltip control handle
HWND hEditorDlg = NULL;

#ifndef STATIC_FILTER
HWND hEdit1 = NULL;
HWND hEdit2 = NULL;
#endif

HHOOK KeyHook = NULL;	//Not thread-safe???

int cancelEventPerformed=0;

//Moved to main level of AfhFMUI.c, Harald Heim. Dec 14, 2002
//Moved to Global Level, because of restoreImageData()
POINT imageOffset = {0, 0}; //Static

//Move up to top level of AfhFMUI.c, Harald Heim. January, 2003
static HCURSOR hcHand;
INT_PTR cursorResource;

//Declare
BOOL PerformControlAction(int n, FME e, int previous, int ctlMsg, int ctlCode,
                          BOOL fActionable, BOOL fProcessed, int overrideBusy);


#define IDC_PROXY   101

/* externals */

//extern UINT4 SDL[];        /* The semantic descriptor list (sized elsewhere). */

void Mdc_dump_estack(void *sdlp);
void Mdc_execute_init_code(void);

/*****************************************************************************/

/* Calculates the rectangle we will want to use for the proxy for an image. */

void GetProxyItemRect (HWND hDlg, Rect *proxyItemRect)
    {

    RECT    wRect;

    GetWindowRect(GetDlgItem(hDlg, IDC_PROXY), &wRect); 
    ScreenToClient (hDlg, (LPPOINT)&wRect);
    ScreenToClient (hDlg, (LPPOINT)&(wRect.right));

    proxyItemRect->left = (int16)wRect.left;
    proxyItemRect->top  = (int16)wRect.top;
    proxyItemRect->right = (int16)wRect.right;
    proxyItemRect->bottom = (int16)wRect.bottom;
    
    }

/*****************************************************************************/

static unsigned8 *pRGBbuffer = NULL;    //remember to deallocate at exit!!!!
static int pRGBbufferSize = 0;

void ShowOutputBuffer( GPtr globals, HWND hDlg, BOOL fDragging)
{
    PSPixelMap pixels;
    PSPixelMask mask;
    PAINTSTRUCT  ps;
    RECT  itemBounds;
    RECT  wRect;
    int proxyWidth;
    int proxyHeight;
#if 1
    RECT  rcIntersect;
#endif
    POINT   mapOrigin; 
    HDC     hDC;

	long  scaleFactor = gStuff->inputRate >> 16;


    if (gResult != noErr) {
        // Got an error somewhere along the way...
        // Get out quick in case the error was returned by
        // AdvanceState(), in which case the input/output data
        // pointers are bogus and can cause severe dyspepsia.
        return;
    }

    // If AFH_DRAG_FILTERED_OUTPUT is set, turn off fDragging locally so we
    // can display the filtered output instead of the raw input...
    if (gParams->flags & AFH_DRAG_FILTERED_OUTPUT)
    {
        fDragging = FALSE;
    }

    GetWindowRect(GetDlgItem(hDlg, IDC_PROXY), &wRect);
    proxyWidth = wRect.right - wRect.left;
    proxyHeight = wRect.bottom - wRect.top;
    mapOrigin.x = 0;
    mapOrigin.y = 0;
    ClientToScreen(hDlg, &mapOrigin);

    // Following rounded in different directions if Proxy was shifted off top or left of screen,
    // since the screen coords went negative, and on x86 architecture, integer division always truncates
    // towards zero.  This resulted in some strange off-by-one-pixel artifacts if the proxy
    // got shifted offscreen and then back onscreen.  Fix is to convert wRect to client coords
    // *before* division by 2, so the dividend will always be positive and truncate consistently
    // *down* towards zero.
    //itemBounds.left = ((wRect.right + wRect.left - gStuff->outRect.right + gStuff->outRect.left) / 2) - mapOrigin.x;
    //itemBounds.top = ((wRect.bottom + wRect.top - gStuff->outRect.bottom + gStuff->outRect.top) / 2) - mapOrigin.y;

    /*itemBounds.left = (wRect.right - mapOrigin.x + wRect.left - mapOrigin.x - (gStuff->outRect.right - gStuff->outRect.left)) / 2;
    itemBounds.top = (wRect.bottom - mapOrigin.y + wRect.top - mapOrigin.y - (gStuff->outRect.bottom - gStuff->outRect.top)) / 2;
    itemBounds.right = itemBounds.left + (gStuff->outRect.right - gStuff->outRect.left);
    itemBounds.bottom = itemBounds.top + (gStuff->outRect.bottom - gStuff->outRect.top);*/

    //Temporary Deactivated
	gFmcp->outRect32 = GetOutRect();	//needed??
    if (gFmcp->outRect32.bottom - gFmcp->outRect32.top <= wRect.bottom - wRect.top)
    {    
        itemBounds.top = (wRect.bottom - mapOrigin.y + wRect.top - mapOrigin.y - (gFmcp->outRect32.bottom - gFmcp->outRect32.top)) / 2;
        itemBounds.bottom = itemBounds.top + (gFmcp->outRect32.bottom - gFmcp->outRect32.top);
    } else { // in case preview is sized down
        itemBounds.top = (wRect.bottom - mapOrigin.y + wRect.top - mapOrigin.y - (wRect.bottom - wRect.top))/2;
        itemBounds.bottom = itemBounds.top + (wRect.bottom - wRect.top);
    }

    if (gFmcp->outRect32.right - gFmcp->outRect32.left <= wRect.right - wRect.left)
    {
        itemBounds.left = (wRect.right - mapOrigin.x + wRect.left - mapOrigin.x - (gFmcp->outRect32.right - gFmcp->outRect32.left)) / 2;
        itemBounds.right = itemBounds.left + (gFmcp->outRect32.right - gFmcp->outRect32.left);
    } else { // in case preview is sized down
        itemBounds.left = (wRect.right - mapOrigin.x + wRect.left - mapOrigin.x - (wRect.right - wRect.left))/2;
        itemBounds.right = itemBounds.left + (wRect.right - wRect.left);
    }
    


    hDC = BeginPaint(hDlg, &ps);

#if 1
    if (IntersectRect(&rcIntersect, &itemBounds, &ps.rcPaint)) {
#endif

    wRect = itemBounds;
    InflateRect(&wRect, 1, 1);
    FrameRect(hDC, &wRect, GetStockObject(BLACK_BRUSH));

    pixels.version = 1;
    pixels.bounds.top = gFmcp->outRect32.top;
    pixels.bounds.left = gFmcp->outRect32.left;

    /*pixels.bounds.bottom = gFmcp->outRect32.bottom;
    pixels.bounds.right = gFmcp->outRect32.right;*/

    //Temporary Deactivated
    if (gFmcp->outRect32.bottom - gFmcp->outRect32.top <= itemBounds.bottom - itemBounds.top)
    {
        pixels.bounds.bottom = gFmcp->outRect32.bottom;
    } else { // in case preview is sized down
        pixels.bounds.bottom = gFmcp->outRect32.top + (itemBounds.bottom - itemBounds.top);
    }

    if (gFmcp->outRect32.right - gFmcp->outRect32.left <= itemBounds.right - itemBounds.left)
    {
        pixels.bounds.right = gFmcp->outRect32.right;
    } else { // in case preview is sized down
        pixels.bounds.right = gFmcp->outRect32.left  + (itemBounds.right - itemBounds.left);
    }



    pixels.imageMode = gStuff->imageMode;
    pixels.rowBytes = fDragging ? gStuff->inRowBytes : gStuff->outRowBytes;
    pixels.colBytes = fDragging ? gStuff->inHiPlane - gStuff->inLoPlane + 1
                                : gStuff->outHiPlane - gStuff->outLoPlane + 1;
    pixels.planeBytes = 1;
    pixels.baseAddr = fDragging ? MyAdjustedInData : gStuff->outData;


#ifdef HIGHZOOM
	// >100% Zoom
	//if (pixels.imageMode == plugInModeRGB48 || pixels.imageMode == plugInModeLab48 || pixels.imageMode == plugInModeDeepMultichannel) 
	if (gFmcp->enlargeFactor != 1)
	{
        unsigned8 *inPtr1 = pixels.baseAddr;
		unsigned8 *inBase = pixels.baseAddr;
        unsigned8 *outPtr;
        int iRows = pixels.bounds.bottom - pixels.bounds.top;
        int iCols = pixels.bounds.right - pixels.bounds.left;
		int maxRC;

        if (pRGBbufferSize != proxyWidth*proxyHeight*(gStuff->inHiPlane+1)){ //3 //Only reallocate if size has to be changed
            
            if (pRGBbuffer) {
                free(pRGBbuffer);
                pRGBbuffer = NULL;
            }
            
            // Allocate an RGB buffer large enough to cover the entire proxy window.
            pRGBbuffer = calloc(proxyWidth*proxyHeight, (gStuff->inHiPlane+1)); //3
            
            if (!pRGBbuffer) {
#if 1
                ErrorOk("Can't allocate proxy preview buffer for 48-bit image!");
#endif
                pRGBbufferSize = 0;
                goto egress;
            
            } else {

                pRGBbufferSize = proxyWidth*proxyHeight*(gStuff->inHiPlane+1); //3
            }
        }

		maxRC = (iRows-1)/gFmcp->enlargeFactor*pixels.rowBytes + (iCols-1)/gFmcp->enlargeFactor*pixels.colBytes;
        outPtr = pRGBbuffer;
        while (iRows--) {
            //int16 *inPtr = (int16 *)inPtr1;
            int j = iCols;
            while (j--) {
                // Unroll 3 planes...
                int iPos = maxRC - (iRows/gFmcp->enlargeFactor*pixels.rowBytes + j/gFmcp->enlargeFactor*pixels.colBytes);
				int r = inBase[iPos + 0];
                int g = inBase[iPos + 1];
                int b = inBase[iPos + 2];
                outPtr[0] = r;
                outPtr[1] = g;
                outPtr[2] = b;

                if (gStuff->inHiPlane==3) {
					int a = inBase[iPos + 3];
					outPtr[3] = a;
				}
                //inPtr += pixels.colBytes;
                outPtr += (gStuff->inHiPlane+1); //3
            }//while j
            //inPtr1 += pixels.rowBytes;
        }//while iRows

        // Point pixel map to the RGB buffer...
        pixels.baseAddr = pRGBbuffer;
        pixels.rowBytes = iCols*(gStuff->inHiPlane+1); //3
        pixels.colBytes = (gStuff->inHiPlane+1); //3
        pixels.planeBytes = 1;
    }
#endif




#if 1


    //Added by Harald Heim, December 6, 2002
    if (pixels.imageMode == plugInModeRGB48 || pixels.imageMode == plugInModeLab48 || pixels.imageMode == plugInModeDeepMultichannel) {

        // Convert outData from 48-bit to 24-bit for display...
        unsigned8 *inPtr1 = pixels.baseAddr;
        unsigned8 *outPtr;
        int iRows = pixels.bounds.bottom - pixels.bounds.top;
        int iCols = pixels.bounds.right - pixels.bounds.left;

        //We need to reallocate the pRGBbuffer if the preview has been resized in the meantime.
        //This was the problem that caused the crash. The buffer wasn't reallocated, so sometimes it was too small.
        //if (!pRGBbuffer) {
        if (pRGBbufferSize != proxyWidth*proxyHeight*(gStuff->inHiPlane+1)){ //3 //Only reallocate if size has to be changed
            
            if (pRGBbuffer) {
                free(pRGBbuffer);
                pRGBbuffer = NULL;
            }
            
            // Allocate an RGB buffer large enough to cover the entire proxy window.
            pRGBbuffer = calloc(proxyWidth*proxyHeight, (gStuff->inHiPlane+1)); //3
            
            if (!pRGBbuffer) {
#if 1
                ErrorOk("Can't allocate proxy preview buffer for 48-bit image!");
#endif
                pRGBbufferSize = 0;
                goto egress;
            
            } else {

                pRGBbufferSize = proxyWidth*proxyHeight*(gStuff->inHiPlane+1); //3
            }
        }

        outPtr = pRGBbuffer;
        while (iRows--) {
            int16 *inPtr = (int16 *)inPtr1;
            int j = iCols;
            while (j--) {
                // Unroll 3 planes...
                int r = inPtr[0];
                int g = inPtr[1];
                int b = inPtr[2];
                int a = inPtr[3];	//not needed in non-alpha loop!!
                outPtr[0] = (r >> 7) | (r >> 15);	//Note that the >> 15 must be a SIGNED shift!!!
                outPtr[1] = (g >> 7) | (g >> 15);
                outPtr[2] = (b >> 7) | (b >> 15);
				//this loop needs to be tight! Use a separate loop if alpha is present??
                if (gStuff->inHiPlane==3) outPtr[3] = (a >> 7) | (a >> 15); //or inHiPlane >= 3 ???
                inPtr += pixels.colBytes;
                outPtr += (gStuff->inHiPlane+1); //3
            }//while j
            inPtr1 += pixels.rowBytes;
        }//while iRows
        // Point pixel map to the RGB buffer...
        pixels.baseAddr = pRGBbuffer;
        pixels.rowBytes = iCols*(gStuff->inHiPlane+1); //3
        pixels.colBytes = (gStuff->inHiPlane+1); //3
        pixels.planeBytes = 1;

        if (pixels.imageMode == plugInModeRGB48)
            pixels.imageMode = plugInModeRGBColor;      //data is RGB
        else if (pixels.imageMode == plugInModeLab48)
            pixels.imageMode = plugInModeLabColor;		//data is Lab
        ///else if (pixels.imageMode == plugInModeDeepMultichannel)
		else //plugInModeDeepMultichannel is the only possibility left!
            pixels.imageMode = plugInModeMultichannel;	//data is Multichannel
    }


    //Added by Harald Heim, December 6, 2002
    else if (pixels.imageMode == plugInModeCMYK64) {

        // Convert outData from CMYK64 to CMYK for display...
        unsigned8 *inPtr1 = pixels.baseAddr;
        unsigned8 *outPtr;
        int iRows = pixels.bounds.bottom - pixels.bounds.top;
        int iCols = pixels.bounds.right - pixels.bounds.left;
        
        //if (!pRGBbuffer) {
        if (pRGBbufferSize != proxyWidth*proxyHeight*(gStuff->inHiPlane+1)){ //4 //Only reallocate if size has to be changed

            if (pRGBbuffer) {
                free(pRGBbuffer);
                pRGBbuffer = NULL;
            }

            // Allocate a CMYK buffer large enough to cover the entire proxy window.
            pRGBbuffer = calloc(proxyWidth*proxyHeight, (gStuff->inHiPlane+1) ); //4
            if (!pRGBbuffer) {
#if 1
                ErrorOk("Can't allocate proxy preview buffer for CMYK64 image!");
#endif
                goto egress;

            } else {

                pRGBbufferSize = proxyWidth*proxyHeight*(gStuff->inHiPlane+1);//4
            }
        }
        outPtr = pRGBbuffer;
        while (iRows--) {
            int16 *inPtr = (int16 *)inPtr1;
            int j = iCols;
            while (j--) {
                // Unroll 4 planes...
                int c = inPtr[0];
                int m = inPtr[1];
                int y = inPtr[2];
                int k = inPtr[3];
                int a = inPtr[4];
                outPtr[0] = (c >> 7) | (c >> 15);
                outPtr[1] = (m >> 7) | (m >> 15);
                outPtr[2] = (y >> 7) | (y >> 15);
                outPtr[3] = (k >> 7) | (k >> 15);
				//this loop needs to be tight! Use a separate loop if alpha is present??
                if (gStuff->inHiPlane==4) outPtr[4] = (a >> 7) | (a >> 15); // or >= 4 (see rgb64 above)
                inPtr += pixels.colBytes;
                outPtr += (gStuff->inHiPlane+1);//4;
            }//while j
            inPtr1 += pixels.rowBytes;
        }//while iRows
        // Point pixel map to the CMYK buffer...
        pixels.baseAddr = pRGBbuffer;
        pixels.rowBytes = iCols*(gStuff->inHiPlane+1);//4;
        pixels.colBytes = (gStuff->inHiPlane+1);//4;
        pixels.planeBytes = 1;
        pixels.imageMode = plugInModeCMYKColor;      //data is CMYK
    }

    
    //Added by Harald Heim, December 6, 2002
    else if (pixels.imageMode == plugInModeDuotone16) {

        // Convert outData from Duotone16 to Duotone for display...
        unsigned8 *inPtr1 = pixels.baseAddr;
        unsigned8 *outPtr;
        int iRows = pixels.bounds.bottom - pixels.bounds.top;
        int iCols = pixels.bounds.right - pixels.bounds.left;
        
        //if (!pRGBbuffer) {
        if (pRGBbufferSize != proxyWidth*proxyHeight*(gStuff->inHiPlane+1)){ //2 //Only reallocate if size has to be changed

            if (pRGBbuffer) {
                free(pRGBbuffer);
                pRGBbuffer = NULL;
            }

            // Allocate a Duotone buffer large enough to cover the entire proxy window.
            pRGBbuffer = calloc(proxyWidth*proxyHeight, (gStuff->inHiPlane+1)); //2
            if (!pRGBbuffer) {
#if 1
                ErrorOk("Can't allocate proxy preview buffer for Duotone16 image!");
#endif
                goto egress;

            } else {

                pRGBbufferSize = proxyWidth*proxyHeight*(gStuff->inHiPlane+1);//2;
            }
        }
        outPtr = pRGBbuffer;
        while (iRows--) {
            int16 *inPtr = (int16 *)inPtr1;
            int j = iCols;
            while (j--) {
                // Unroll 2 planes...
                int d = inPtr[0];
                int u = inPtr[1];
                int a = inPtr[3];
                outPtr[0] = (d >> 7) | (d >> 15);
                outPtr[1] = (u >> 7) | (u >> 15);
				//this loop needs to be tight! Use a separate loop if alpha is present??
                if (gStuff->inHiPlane==2) outPtr[2] = (a >> 7) | (a >> 15);	// or >=2 ??
                inPtr += pixels.colBytes;
                outPtr += (gStuff->inHiPlane+1); //2;
            }//while j
            inPtr1 += pixels.rowBytes;
        }//while iRows
        // Point pixel map to the Duotone buffer...
        pixels.baseAddr = pRGBbuffer;
        pixels.rowBytes = iCols*(gStuff->inHiPlane+1);//2;
        pixels.colBytes = (gStuff->inHiPlane+1);//2;
        pixels.planeBytes = 1;
        pixels.imageMode = plugInModeDuotone;      //data is Duotone
    }
    

    else if (pixels.imageMode == plugInModeGray16) {
        // Convert outData from Gray16 to Grayscale for display...
        unsigned8 *inPtr1 = pixels.baseAddr;
        unsigned8 *outPtr;
        int iRows = pixels.bounds.bottom - pixels.bounds.top;
        int iCols = pixels.bounds.right - pixels.bounds.left;
        
        //if (!pRGBbuffer) {
        if (pRGBbufferSize != proxyWidth*proxyHeight*(gStuff->inHiPlane+1)){ //Only reallocate if size has to be changed

            if (pRGBbuffer) {
                free(pRGBbuffer);
                pRGBbuffer = NULL;
            }

            // Allocate a grayscale buffer large enough to cover the entire proxy window.
            pRGBbuffer = calloc(proxyWidth*proxyHeight, (gStuff->inHiPlane+1)); //1
            if (!pRGBbuffer) {
#if 1
                ErrorOk("Can't allocate proxy preview buffer for Gray16 image!");
#endif
                goto egress;
            
            } else {

                pRGBbufferSize = proxyWidth*proxyHeight;
            }
        }
        outPtr = pRGBbuffer;
        while (iRows--) {
            int16 *inPtr = (int16 *)inPtr1;
            int j = iCols;
            while (j--) {
                //int gray = *inPtr;
                int gray = inPtr[0];
                int a = inPtr[1];
                outPtr[0] = (gray >> 7) | (gray >> 15);
				//this loop needs to be tight! Use a separate loop if alpha is present??
				//and move invariants out of loop; e.g. (gStuff->inHiPlane+1)
                if (gStuff->inHiPlane==1) outPtr[1] = (a >> 7) | (a >> 15);	//or >=1 ??
                //*outPtr++ = (gray >> 7) | (gray >> 15);
                inPtr += pixels.colBytes;
                outPtr += (gStuff->inHiPlane+1); //2;
            }//while j
            inPtr1 += pixels.rowBytes;
        }//while iRows
        // Point pixel map to the grayscale buffer...
        pixels.baseAddr = pRGBbuffer;
        pixels.rowBytes = iCols*(gStuff->inHiPlane+1);
        pixels.colBytes = (gStuff->inHiPlane+1); //1;
        pixels.planeBytes = 1;
        pixels.imageMode = plugInModeGrayScale;      //data is GrayScale
    }
#endif

    pixels.mat = NULL;
    pixels.masks = NULL;
    pixels.maskPhaseRow = 0;
    pixels.maskPhaseCol = 0;

    if (gStuff->isFloating || //gStuff->depth == 16 || //Temp preview fix on 16bit layer -> but no transparency chess pattern
        (gStuff->haveMask && gStuff->filterCase == filterCaseProtectedTransparencyWithSelection)) {
        mask.next = NULL;
        mask.maskData = MyAdjustedMaskData;
        mask.rowBytes = gStuff->maskRowBytes;
        mask.colBytes = 1;
        mask.maskDescription = kSimplePSMask;
    
        pixels.masks = &mask;
    } else if ((gStuff->inLayerPlanes != 0) && (gStuff->inTransparencyMask != 0) &&
               (gStuff->filterCase != filterCaseProtectedTransparencyNoSelection) &&
               (gStuff->filterCase != filterCaseProtectedTransparencyWithSelection)) {
        // Was wrong for PS Protected Transparency cases, in which
        // case inLayerPlanes = 3 and inTransparencyMask=1, but inColumnBytes = 3,
        // not 4, and the transparency mask is NOT interleaved with the RGB data.
        // In fact, I don't know where the transparency mask data is for this case!
        
        //int multiply=1;
        //if (gStuff->depth == 16) multiply=2
        
        mask.next = NULL;
        
        if (gStuff->depth == 16){
            
            if (gStuff->inHiPlane == 3)
                mask.maskData = pRGBbuffer+3;
            else if (gStuff->inHiPlane == 4)
                mask.maskData = pRGBbuffer+4;
            else if (gStuff->inHiPlane == 1)
                mask.maskData = pRGBbuffer+1;

            mask.rowBytes = pixels.rowBytes;
            mask.colBytes = pixels.colBytes;

        } else {
            mask.maskData = fDragging ?
                        ((char *) MyAdjustedInData) + (gStuff->inHiPlane - gStuff->inLoPlane) :
                        ((char *) gStuff->outData) + (gStuff->outHiPlane - gStuff->outLoPlane);
        
            mask.rowBytes = fDragging ? gStuff->inRowBytes : gStuff->outRowBytes;
            mask.colBytes = fDragging ? gStuff->inHiPlane - gStuff->inLoPlane + 1
                                  : gStuff->outHiPlane - gStuff->outLoPlane + 1;
        }
        mask.maskDescription = kSimplePSMask;
    
        pixels.masks = &mask;
    }

	//if ( !fDragging) Info ("%d,%d - %d,%d     %d, %d", pixels.bounds.left, pixels.bounds.top, pixels.bounds.right, pixels.bounds.bottom,   itemBounds.left, itemBounds.top);

    (gStuff->displayPixels)(&pixels, &pixels.bounds, itemBounds.top, itemBounds.left, (void *)hDC);


#if 1
    }//if intersects
#endif
egress:
    EndPaint(hDlg, (LPPAINTSTRUCT) &ps);

	
	//DRAWITEM event for preview
	if (gParams->ctl[CTL_PREVIEW].properties & CTP_DRAWITEM){            
        PerformControlAction(
                CTL_PREVIEW,             //control index
                FME_DRAWITEM,           //FM event code
                0,                      //previous value
                WM_DRAWITEM,            //message type
                0,                      //notification code
                FALSE,                  //fActionable
                TRUE,0);                  //fProcessed
    }


} /*ShowOutputBuffer*/

////////////////////////////////////////////////////////////////////////////////////
void ShowFilterParameterBlock( GPtr globals, HWND hDlg)
{
#ifndef STATIC_FILTER
#if !defined(BETA) && !defined(STANDALONE)
    char buff[4096];
    int n;

    //Added by Harald Heim, December 6, 2002
    static const char *modeName[16] =
    {
        "Bitmap",
        "GrayScale",
        "IndexedColor",
        "RGBColor",
        "CMYKColor",
        "HSLColor",
        "HSBColor",
        "Multichannel",
        "Duotone",
        "LabColor",
        "Gray16",
        "RGB48",
        "Lab48",
        "CMYK64",
        "DeepMultichannel",
        "Duotone16"
    };


    static const char *layoutName[7] =
    {
        "Traditional -- Rows, columns, planes with colbytes = # planes",
        "RowsColumnsPlanes",
        "RowsPlanesColumns",
        "ColumnsRowsPlanes",
        "ColumnsPlanesRows",
        "PlanesRowsColumns",
        "PlanesColumnsRows"
    };
    static const char *filterCaseName[] =
    {
        "Unsupported",                              // -1
        "Not set -- use haveMask and isFloating",   //  0
        "FlatImageNoSelection",                     //  1
        "FlatImageWithSelection",                   //  2
        "FloatingSelection",                        //  3
        "EditableTransparencyNoSelection",          //  4
        "EditableTransparencyWithSelection",        //  5
        "ProtectedTransparencyNoSelection",         //  6
        "ProtectedTransparencyWithSelection"        //  7
    };
    static const char *paddingName[] =
    {
        "plugInWantsErrorOnBoundsException",    // -3
        "plugInDoesNotWantPadding",             // -2
        "plugInWantsEdgeReplication",           // -1
    };
    static const char *samplingName[] =
    {
        "hostDoesNotSupportSampling",           // 0
        "hostSupportsIntegralSampling",         // 1
        "hostSupportsFractionalSampling"        // 2
    };

if ((gParams->flags & (AFH_EDITMODE|AFH_ADVANCED)) == (AFH_EDITMODE|AFH_ADVANCED)) {
    // Apparently wsprintf writes a max of 1024 bytes (including final
    // NUL) to the output buffer, so we need to break up this call
    // into more than one piece...
    n = wsprintf(buff,
                   "------ Version 2.5 and later ------\r\n"
                   "serialNumber = %d\r\n"
                   "abortProc = %#x\r\n"
                   "progressProc = %#x\r\n"
                   "parameters = %#x\r\n"
                   "imageSize = (%d, %d)\r\n"
                   "planes = %d\r\n"
                   "filterRect = ((%d, %d), (%d, %d))\r\n"
                   "background (obs) \t= (%d, %d, %d)\r\n"
                   "foreground (obs) \t= (%d, %d, %d)\r\n"
                   "maxSpace = %d\r\n"
                   "bufferSpace = %d\r\n"
                   "inRect = ((%d, %d), (%d, %d))\r\n"
                   "inLoPlane = %d\r\n"
                   "inHiPlane = %d\r\n"
                   "outRect = ((%d, %d), (%d, %d))\r\n"
                   "outLoPlane = %d\r\n"
                   "outHiPlane = %d\r\n"
                   "inData = %#x\r\n"
                   "MyAdjustedInData = %#x\r\n"
                   "inRowBytes = %d\r\n"
                   "outData = %#x\r\n"
                   "outRowBytes = %d\r\n"
                   "isFloating = %d\r\n"
                   "haveMask = %d\r\n"
                   "autoMask = %d\r\n"
                   "maskRect = ((%d, %d), (%d, %d))\r\n"
                   "maskData = %#x\r\n"
                   "MyAdjustedMaskData = %#x\r\n"
                   "maskRowBytes = %d\r\n"

                   ,gStuff->serialNumber
                   ,gStuff->abortProc
                   ,gStuff->progressProc
                   ,gStuff->parameters
                   ,gStuff->imageSize.h, gStuff->imageSize.v
                   ,gStuff->planes
                   ,gStuff->filterRect.left, gStuff->filterRect.top
                   ,gStuff->filterRect.right, gStuff->filterRect.bottom
#if 0
                   ,gStuff->background.red & 0xff, 
                        gStuff->background.green & 0xff, 
                        gStuff->background.blue & 0xff
                   ,gStuff->foreground.red & 0xff, 
                        gStuff->foreground.green & 0xff, 
                        gStuff->foreground.blue & 0xff
#else
                   //these are actually 16-bit vals (8 bit * 257)...
                   ,gStuff->background.red, 
                        gStuff->background.green, 
                        gStuff->background.blue
                   ,gStuff->foreground.red, 
                        gStuff->foreground.green, 
                        gStuff->foreground.blue
#endif
                   ,gStuff->maxSpace
                   ,gStuff->bufferSpace
                   ,gStuff->inRect.left, gStuff->inRect.top,
                        gStuff->inRect.right, gStuff->inRect.bottom
                   ,gStuff->inLoPlane
                   ,gStuff->inHiPlane
                   ,gStuff->outRect.left, gStuff->outRect.top,
                        gStuff->outRect.right, gStuff->outRect.bottom
                   ,gStuff->outLoPlane
                   ,gStuff->outHiPlane
                   ,gStuff->inData
                   ,MyAdjustedInData
                   ,gStuff->inRowBytes
                   ,gStuff->outData
                   ,gStuff->outRowBytes
                   ,gStuff->isFloating
                   ,gStuff->haveMask
                   ,gStuff->autoMask
                   ,gStuff->maskRect.left, gStuff->maskRect.top,
                        gStuff->maskRect.right, gStuff->maskRect.bottom
                   ,gStuff->maskData
                   ,MyAdjustedMaskData
                   ,gStuff->maskRowBytes
                   );
#if 0
    //test grabbing entire rect, not just filterRect
    gStuff->filterRect.top = 0;
    gStuff->filterRect.left = 0;
    gStuff->filterRect.bottom = gStuff->wholeSize.v;	//update for BigDoc??
    gStuff->filterRect.right = gStuff->wholeSize.h;
#endif
    n += wsprintf(&buff[n],
                   "backColor = (%d, %d, %d, %d)\r\n"
                   "foreColor = (%d, %d, %d, %d)\r\n"
                   "hostSig = %#x (\"%.4s\")\r\n"
                   "hostProc = %#x\r\n"
                   "imageMode = %d (%s)\r\n"
                   "imageHRes = %d.%d\r\n"
                   "imageVRes = %d.%d\r\n"
                   "floatCoord = (%d, %d)\r\n"
                   "wholeSize = (%d, %d)\r\n"
                   "[monitor setup info not displayed]\r\n"
                   "platformData = %#x\r\n"
                   "bufferProcs = %#x\r\n"
                   "resourceProcs = %#x\r\n"
                   "processEvent = %#x\r\n"
                   "displayPixels = %#x\r\n"
                   "handleProcs = %#x\r\n"

                   ,gStuff->backColor[0], gStuff->backColor[1], gStuff->backColor[2],
                        gStuff->backColor[3]
                   ,gStuff->foreColor[0], gStuff->foreColor[1], gStuff->foreColor[2],
                        gStuff->foreColor[3]
                   ,gStuff->hostSig, &gStuff->hostSig
                   ,gStuff->hostProc
                   ,gStuff->imageMode,
                        (gStuff->imageMode >= 0 && gStuff->imageMode < 12) ?
                        modeName[gStuff->imageMode] : "unknown mode"
                   ,gStuff->imageHRes >> 16, gStuff->imageHRes & 0xffff
                   ,gStuff->imageVRes >> 16, gStuff->imageVRes & 0xffff
                   ,gStuff->floatCoord.h, gStuff->floatCoord.v
                   ,gStuff->wholeSize.h, gStuff->wholeSize.v
                   //,gStuff->monitor
                   ,gStuff->platformData	//X64??
                   ,gStuff->bufferProcs
                   ,gStuff->resourceProcs
                   ,gStuff->processEvent
                   ,gStuff->displayPixels
                   ,gStuff->handleProcs
                   );
    n += wsprintf(&buff[n],
                   "\r\n"
                   "------ Version 3.0 and later ------\r\n"
                   "supportsDummyPlanes = %d\r\n"
                   "supportsAlternateLayouts = %d\r\n"
                   "wantLayout = %d (%s)\r\n"
                   "filterCase = %d (%s)\r\n"
                   "dummyPlaneValue = %d (-1 means leave undefined)\r\n"
                   "premiereHook (obs) = %#x\r\n"
                   "advanceState = %#x\r\n"
                   "supportsAbsolute = %d\r\n"
                   "wantsAbsolute = %d\r\n"
                   "getProperty (obs) = %#x\r\n"
                   "cannotUndo = %d\r\n"
                   "supportsPadding = %d\r\n"
                   "inputPadding = %d (%s)\r\n"
                   "outputPadding = %d (%s)\r\n"
                   "maskPadding = %d (%s)\r\n"

                   //------ Version 3.0 and later ------
                   ,gStuff->supportsDummyChannels
                   ,gStuff->supportsAlternateLayouts
                   ,gStuff->wantLayout,
                        (gStuff->wantLayout >= 0 && gStuff->wantLayout < 7) ?
                        layoutName[gStuff->wantLayout] : "unknown layout"
                   ,gStuff->filterCase,
                        (gStuff->filterCase >= -1 && gStuff->filterCase <= 7) ?
                        filterCaseName[gStuff->filterCase + 1] : "unknown case"
                   ,gStuff->dummyPlaneValue
                   ,gStuff->premiereHook
                   ,gStuff->advanceState
                   ,gStuff->supportsAbsolute
                   ,gStuff->wantsAbsolute
                   ,gStuff->getPropertyObsolete
                   ,gStuff->cannotUndo
                   ,gStuff->supportsPadding
                   ,gStuff->inputPadding,
                        (gStuff->inputPadding >= 0 && gStuff->inputPadding <= 255) ?
                        "set to this value" :
                        (gStuff->inputPadding >= -3 && gStuff->inputPadding <= -1) ?
                        paddingName[gStuff->inputPadding + 3] : "unknown case"
                   ,gStuff->outputPadding,
                        (gStuff->outputPadding >= 0 && gStuff->outputPadding <= 255) ?
                        "set to this value" :
                        (gStuff->outputPadding >= -3 && gStuff->outputPadding <= -1) ?
                        paddingName[gStuff->outputPadding + 3] : "unknown case"
                   ,gStuff->maskPadding,
                        (gStuff->maskPadding >= 0 && gStuff->maskPadding <= 255) ?
                        "set to this value" :
                        (gStuff->maskPadding >= -3 && gStuff->maskPadding <= -1) ?
                        paddingName[gStuff->maskPadding + 3] : "unknown case"
                   );
    n += wsprintf(&buff[n],
                   "samplingSupport = %d (%s)\r\n"
                   "reservedByte = %d\r\n"
                   "inputRate = %d.%d\r\n"
                   "maskRate = %d.%d\r\n"
                   "colorServices = %#x\r\n"
                   "inLayerPlanes = %d\r\n"
                   "inTransparencyMask = %d\r\n"
                   "inLayerMasks = %d\r\n"
                   "inInvertedLayerMasks = %d\r\n"
                   "inNonLayerPlanes = %d\r\n"
                   "outLayerPlanes = %d\r\n"
                   "outTransparencyMask = %d\r\n"
                   "outLayerMasks = %d\r\n"
                   "outInvertedLayerMasks = %d\r\n"
                   "outNonLayerPlanes = %d\r\n"
                   "absLayerPlanes = %d\r\n"
                   "absTransparencyMask = %d\r\n"
                   "absLayerMasks = %d\r\n"
                   "absInvertedLayerMasks = %d\r\n"
                   "absNonLayerPlanes = %d\r\n"
                   "inPreDummyPlanes = %d\r\n"
                   "inPostDummyPlanes = %d\r\n"
                   "outPreDummyPlanes = %d\r\n"
                   "outPostDummyPlanes = %d\r\n"
                   "inColumnBytes = %d\r\n"
                   "inPlaneBytes = %d\r\n"
                   "outColumnBytes = %d\r\n"
                   "outPlaneBytes = %d\r\n"

                   ,gStuff->samplingSupport,
                        (gStuff->samplingSupport >= 0 && gStuff->samplingSupport <= 2) ?
                        samplingName[gStuff->samplingSupport] : "unknown case"
                   ,gStuff->reservedByte
                   ,gStuff->inputRate >> 16, gStuff->inputRate & 0xffff
                   ,gStuff->maskRate >> 16, gStuff->maskRate & 0xffff
                   ,gStuff->colorServices
                   ,gStuff->inLayerPlanes
                   ,gStuff->inTransparencyMask
                   ,gStuff->inLayerMasks
                   ,gStuff->inInvertedLayerMasks
                   ,gStuff->inNonLayerPlanes
                   ,gStuff->outLayerPlanes
                   ,gStuff->outTransparencyMask
                   ,gStuff->outLayerMasks
                   ,gStuff->outInvertedLayerMasks
                   ,gStuff->outNonLayerPlanes
                   ,gStuff->absLayerPlanes
                   ,gStuff->absTransparencyMask
                   ,gStuff->absLayerMasks
                   ,gStuff->absInvertedLayerMasks
                   ,gStuff->absNonLayerPlanes
                   ,gStuff->inPreDummyPlanes
                   ,gStuff->inPostDummyPlanes
                   ,gStuff->outPreDummyPlanes
                   ,gStuff->outPostDummyPlanes
                   ,gStuff->inColumnBytes
                   ,gStuff->inPlaneBytes
                   ,gStuff->outColumnBytes
                   ,gStuff->outPlaneBytes
                   );
    n += wsprintf(&buff[n],
                   "\r\n"
                   "------ Version 3.0.4 and later ------\r\n"
                   "imageServicesProcs = %#x\r\n"
                   "propertyProcs = %#x\r\n"
                   "inTileHeight = %d\r\n"
                   "inTileWidth = %d\r\n"
                   "inTileOrigin = (%d, %d)\r\n"
                   "absTileHeight = %d\r\n"
                   "absTileWidth = %d\r\n"
                   "absTileOrigin = (%d, %d)\r\n"
                   "outTileHeight = %d\r\n"
                   "outTileWidth = %d\r\n"
                   "outTileOrigin = (%d, %d)\r\n"
                   "maskTileHeight = %d\r\n"
                   "maskTileWidth = %d\r\n"
                   "maskTileOrigin = (%d, %d)\r\n"

                   //------ Version 3.0.4 and later ------
                   ,gStuff->imageServicesProcs
                   ,gStuff->propertyProcs
                   ,gStuff->inTileHeight
                   ,gStuff->inTileWidth
                   ,gStuff->inTileOrigin.h, gStuff->inTileOrigin.v
                   ,gStuff->absTileHeight
                   ,gStuff->absTileWidth
                   ,gStuff->absTileOrigin.h, gStuff->absTileOrigin.v
                   ,gStuff->outTileHeight
                   ,gStuff->outTileWidth
                   ,gStuff->outTileOrigin.h, gStuff->outTileOrigin.v
                   ,gStuff->maskTileHeight
                   ,gStuff->maskTileWidth
                   ,gStuff->maskTileOrigin.h, gStuff->maskTileOrigin.v
                   );
    n += wsprintf(&buff[n],
                   "\r\n"
                   "------ Version 4.0 and later ------\r\n"
                   "descriptorParameters = %#x\r\n"
                   "errorString = \"%.255s\"\r\n"
                   "channelPortProcs = %#x\r\n"
                   "documentInfo = %#x\r\n"

                   //------ Version 4.0 and later ------
                   ,gStuff->descriptorParameters
                   // following still wrong -- it's a Counted nul-terminated string!!!
                   ,gStuff->errorString ? gStuff->errorString : (Str255 *)"(null ptr)"
                   ,gStuff->channelPortProcs
                   ,gStuff->documentInfo
                    );

    n += wsprintf(&buff[n],
                    "\r\n"
                    "------ Version 5.0 and later ------\r\n"
                   "sSPBasic = %#x\r\n"
                   "plugInRef = %#x\r\n"
                   "depth = %d (1, 8, or 16 bit)\r\n "
                   
                   //------ Version 5.0 and later ------
                   ,gStuff->sSPBasic
                   ,gStuff->plugInRef
                   ,gStuff->depth
                   );

    n += wsprintf(&buff[n],
                   "\r\n" 
                   "------ Version 6.0 and later ------\r\n"
                   "iCCprofileData = %#x\r\n"
                   "iCCprofileSize = %d\r\n"
                   "canUseICCProfiles = %d\r\n "
                   
                   //------ Version 6.0 and later ------
                   ,gStuff->iCCprofileData
                   ,gStuff->iCCprofileSize
                   ,gStuff->canUseICCProfiles
                   );

    n += wsprintf(&buff[n],
                   "\r\n"
                   "------ Version 7.0 and later ------\r\n"
                   "hasImageScrap = %d\r\n"
                   
                   //------ Version 7.0 and later ------
                   ,gStuff->hasImageScrap
                   );

    n += wsprintf(&buff[n],
                   "\r\n" 
                   "------ Version CS (8.0) and later ------\r\n"
                   "bigDocumentData = %#x\r\n"
                   
                   //------ Version 8.0 and later ------
                   ,gStuff->bigDocumentData
                   );

    if (gStuff->bigDocumentData){
        n += wsprintf(&buff[n],
                   "PluginUsing32BitCoordinates = %d\r\n"
                   "imageSize32 = (%d, %d)\r\n"
                   "filterRect32 = ((%d, %d), (%d, %d))\r\n"
                   "inRect32 = ((%d, %d), (%d, %d))\r\n"
                   "outRect32 = ((%d, %d), (%d, %d))\r\n"
                   "maskRect32 = ((%d, %d), (%d, %d))\r\n"
                   "floatCoord32 = (%d, %d)\r\n"
                   "wholeSize32 = (%d, %d)\r\n"
                   
                   //------ Version 8.0 and later ------
                   ,gStuff->bigDocumentData->PluginUsing32BitCoordinates
                   ,gStuff->bigDocumentData->imageSize32.h,gStuff->bigDocumentData->imageSize32.v
                   ,gStuff->bigDocumentData->filterRect32.left, gStuff->bigDocumentData->filterRect32.top,
                        gStuff->bigDocumentData->filterRect32.right, gStuff->bigDocumentData->filterRect32.bottom
                   ,gStuff->bigDocumentData->inRect32.left, gStuff->bigDocumentData->inRect32.top,
                        gStuff->bigDocumentData->inRect32.right, gStuff->bigDocumentData->inRect32.bottom
                   ,gStuff->bigDocumentData->outRect32.left, gStuff->bigDocumentData->outRect32.top,
                        gStuff->bigDocumentData->outRect32.right, gStuff->bigDocumentData->outRect32.bottom
                   ,gStuff->bigDocumentData->maskRect32.left, gStuff->bigDocumentData->maskRect32.top,
                        gStuff->bigDocumentData->maskRect32.right, gStuff->bigDocumentData->maskRect32.bottom
                   ,gStuff->bigDocumentData->floatCoord32.h,gStuff->bigDocumentData->floatCoord32.v
                   ,gStuff->bigDocumentData->wholeSize32.h,gStuff->bigDocumentData->wholeSize32.v

                   );
    }


    // Property suite stuff
    {//scope
        GetPropertyProc pGPP;   //pointer to getPropertyProc
        OSErr err;
        int32 simpleProperty;
        Handle complexProperty;

        n += wsprintf(&buff[n],
                   "\r\n"
                   "------ Property Suite ------\r\n"
                   "getProperty (obs) = %#x\r\n"
                   "propertyProcs = %#x\r\n"
                   ,gStuff->getPropertyObsolete
                   ,gStuff->propertyProcs
                   );
        pGPP = gStuff->getPropertyObsolete;
        if (gStuff->propertyProcs) {
            n += wsprintf(&buff[n],
                   "propertyProcsVersion = %d\r\n"
                   "numPropertyProcs = %d\r\n"
                   "getPropertyProc = %#x\r\n"
                   "setPropertyProc = %#x\r\n"
                   ,gStuff->propertyProcs->propertyProcsVersion
                   ,gStuff->propertyProcs->numPropertyProcs
                   ,gStuff->propertyProcs->getPropertyProc
                   ,gStuff->propertyProcs->setPropertyProc
                   );
            if (gStuff->propertyProcs->propertyProcsVersion >= kCurrentPropertyProcsVersion &&
                gStuff->propertyProcs->numPropertyProcs >= 1 &&
                gStuff->propertyProcs->getPropertyProc) {
                //prefer this version of getPropertyProc over the obsolete one
                pGPP = gStuff->propertyProcs->getPropertyProc;
            }
        } //propertyProcs != 0
        if (pGPP) {
            //display some of the available properties

            // 'nuch'
            err = pGPP('8BIM', 'nuch', 0, &simpleProperty, &complexProperty);
            if (err == noErr) {
                n += wsprintf(&buff[n],
                   "propNumberOfChannels (nuch) = %d\r\n", simpleProperty
                   );
            }
            else if (err == errPlugInPropertyUndefined) {
                //not defined for this host
                n += wsprintf(&buff[n],
                   "propNumberOfChannels (nuch) = undefined\r\n"
                   );
            }
            else {
                n += wsprintf(&buff[n],
                   "propNumberOfChannels (nuch) = ERROR(%d)\r\n", err
                   );
            }

            // 'mode'
            err = pGPP('8BIM', 'mode', 0, &simpleProperty, &complexProperty);
            if (err == noErr) {
                n += wsprintf(&buff[n],
                   "propImageMode (mode) = %d\r\n", simpleProperty
                   );
            }
            else if (err == errPlugInPropertyUndefined) {
                //not defined for this host
                n += wsprintf(&buff[n],
                   "propImageMode (mode) = undefined\r\n"
                   );
            }
            else {
                n += wsprintf(&buff[n],
                   "propImageMode (mode) = ERROR(%d)\r\n", err
                   );
            }

            // 'cpyr'
            err = pGPP('8BIM', 'cpyr', 0, &simpleProperty, &complexProperty);
            if (err == noErr) {
                n += wsprintf(&buff[n],
                   "propCopyright (cpyr) = %d\r\n", simpleProperty
                   );
            }
            else if (err == errPlugInPropertyUndefined) {
                //not defined for this host
                n += wsprintf(&buff[n],
                   "propCopyright (cpyr) = undefined\r\n"
                   );
            }
            else {
                n += wsprintf(&buff[n],
                   "propCopyright (cpyr) = ERROR(%d)\r\n", err
                   );
            }

            // 'intp'
            err = pGPP('8BIM', propInterpolationMethod, 0, &simpleProperty, &complexProperty);
            if (err == noErr) {
                n += wsprintf(&buff[n],
                   "propInterpolationMethod (intp) = %d\r\n", simpleProperty
                   );
            }
            else if (err == errPlugInPropertyUndefined) {
                //not defined for this host
                n += wsprintf(&buff[n],
                   "propInterpolationMethod (intp) = undefined\r\n"
                   );
            }
            else {
                n += wsprintf(&buff[n],
                   "propInterpolationMethod (intp) = ERROR(%d)\r\n", err
                   );
            }

            // 'titl'
            err = pGPP('8BIM', 'titl', 0, &simpleProperty, &complexProperty);
            if (err == noErr) {
                int size;
                LPSTR ptr;
                size = HostGetHandleSize(gStuff->handleProcs, complexProperty);
                size = min(size, 256);  //limit to avoid buff overflow
                ptr = HostLockHandle(gStuff->handleProcs, complexProperty, FALSE);
                n += sprintf(&buff[n],  //use sprintf(); wsprintf() doesn't support '*'
                   "propTitle (titl) = \"%.*s\"\r\n", size, ptr
                   );
                HostUnlockHandle(gStuff->handleProcs, complexProperty);
                HostDisposeHandle(gStuff->handleProcs, complexProperty);
            }
            else if (err == errPlugInPropertyUndefined) {
                //not defined for this host
                n += wsprintf(&buff[n],
                   "propTitle (titl) = undefined\r\n"
                   );
            }
            else {
                n += wsprintf(&buff[n],
                   "propTitle (titl) = ERROR(%d)\r\n", err
                   );
            }

#if 0
            // NFG -- Caption is in IPTC-NAA format
            // 'capt'
            err = pGPP('8BIM', propCaption, 0, &simpleProperty, &complexProperty);
            //Info("err = %d", err);
            if (err == noErr) {
                int size;
                LPSTR ptr;
                size = HostGetHandleSize(gStuff->handleProcs, complexProperty);
                //Info("size = %d", size);
                if (size == 0) {
                    //HostLockHandle GPs if size=0
                    n += wsprintf(&buff[n],
                       "propCaption (capt) = \"\"\r\n"
                       );
                }
                else {
                    size = min(size, 256);  //limit to avoid buff overflow
                    ptr = HostLockHandle(gStuff->handleProcs, complexProperty, FALSE);
                    //Info("ptr = %#x", ptr);
                    n += sprintf(&buff[n],  //wsprintf() doesn't support '*'
                       "propCaption (capt) = \"%.*s\"\r\n", size, ptr
                       );
                    HostUnlockHandle(gStuff->handleProcs, complexProperty);
                }
                HostDisposeHandle(gStuff->handleProcs, complexProperty);
            }
            else if (err == errPlugInPropertyUndefined) {
                //not defined for this host
                n += wsprintf(&buff[n],
                   "propCaption (capt) = undefined\r\n"
                   );
            }
            else {
                n += wsprintf(&buff[n],
                   "propCaption (capt) = ERROR(%d)\r\n", err
                   );
            }
#endif
            // 'URL '
            err = pGPP('8BIM', propURL, 0, &simpleProperty, &complexProperty);
            //Info("err = %d", err);
            if (err == noErr) {
                int size;
                LPSTR ptr;
                size = HostGetHandleSize(gStuff->handleProcs, complexProperty);
                //Info("size = %d", size);
                if (size == 0) {
                    //HostLockHandle GPs if size=0
                    n += wsprintf(&buff[n],
                       "propURL (URL ) = \"\"\r\n"
                       );
                }
                else {
                    size = min(size, 256);  //limit to avoid buff overflow
                    ptr = HostLockHandle(gStuff->handleProcs, complexProperty, FALSE);
                    //Info("ptr = %#x", ptr);
                    n += sprintf(&buff[n],  //wsprintf() doesn't support '*'
                       "propURL (URL ) = \"%.*s\"\r\n", size, ptr
                       );
                    HostUnlockHandle(gStuff->handleProcs, complexProperty);
                }
                HostDisposeHandle(gStuff->handleProcs, complexProperty);
            }
            else if (err == errPlugInPropertyUndefined) {
                //not defined for this host
                n += wsprintf(&buff[n],
                   "propURL (URL ) = undefined\r\n"
                   );
            }
            else {
                n += wsprintf(&buff[n],
                   "propURL (URL ) = ERROR(%d)\r\n", err
                   );
            }

            // 'sstr' -- N/A in Paint Shop Pro
            err = pGPP('8BIM', 'sstr', 0, &simpleProperty, &complexProperty);
            if (err == noErr) {
                int size;
                LPSTR ptr;
                size = HostGetHandleSize(gStuff->handleProcs, complexProperty);
                size = min(size, 256);  //limit to avoid buff overflow
                ptr = HostLockHandle(gStuff->handleProcs, complexProperty, FALSE);
                n += sprintf(&buff[n],  //wsprintf() doesn't support '*'
                   "propSerialString (sstr) = \"%.*s\"\r\n", size, ptr
                   );
                HostUnlockHandle(gStuff->handleProcs, complexProperty);
                HostDisposeHandle(gStuff->handleProcs, complexProperty);
            }
            else if (err == errPlugInPropertyUndefined) {
                //not defined for this host
                n += wsprintf(&buff[n],
                   "propSerialString (sstr) = undefined\r\n"
                   );
            }
            else {
                n += wsprintf(&buff[n],
                   "propSerialString (sstr) = ERROR(%d)\r\n", err
                   );
            }


            // 'EXIF'
            err = pGPP('8BIM', 'EXIF', 0, &simpleProperty, &complexProperty);
            //Info("err = %d", err);
            if (err == noErr) {
                int size;
                LPSTR ptr;
                size = HostGetHandleSize(gStuff->handleProcs, complexProperty);
                //Info("size = %d", size);
                if (size == 0) {
                    //HostLockHandle GPs if size=0
                    n += wsprintf(&buff[n],
                       "propEXIFData (EXIF) = \"\"\r\n"
                       );
                }
                else {
                    size = min(size, 256);  //limit to avoid buff overflow
                    ptr = HostLockHandle(gStuff->handleProcs, complexProperty, FALSE);
                    //Info("ptr = %#x", ptr);
                    n += sprintf(&buff[n],  //wsprintf() doesn't support '*'
                       "propEXIFData (EXIF) = \"%.*s\"\r\n", size, ptr
                       );
                    HostUnlockHandle(gStuff->handleProcs, complexProperty);
                }
                HostDisposeHandle(gStuff->handleProcs, complexProperty);
            }
            else if (err == errPlugInPropertyUndefined) {
                //not defined for this host
                n += wsprintf(&buff[n],
                   "propEXIFData (EXIF) = undefined\r\n"
                   );
            }
            else {
                n += wsprintf(&buff[n],
                   "propEXIFData (EXIF) = ERROR(%d)\r\n", err
                   );
            }



        }//pGPP
    }//scope
#if 1
    if (n >= sizeof(buff)) {
#else
    if (TRUE) { //debugging
#endif
        ErrorOk("Internal error: FPB buffer overflow (%d bytes)", n);
        buff[sizeof(buff)-1] = '\0';
    }
    SetDlgItemText(hEditorDlg?hEditorDlg:hDlg, 999, buff);
} //AFH_EDITMODE|AFH_ADVANCED
#endif //if !defined(BETA) && !defined(STANDALONE)
#endif //ifndef STATIC_FILTER

    // Display proxy scale factor
    {
        char buff[20];
        long  scaleFactor = gStuff->inputRate >> 16;
        //HWND hCtl = GetDlgItem(hDlg, 498);
        //RECT rect;

        sprintf(buff, "%.3g%%", (100.0*gFmcp->enlargeFactor)/scaleFactor);
        SetDlgItemText(hDlg, 498, buff);
#if 0
        GetWindowRect(hCtl, &rect);
        MapWindowPoints(NULL, hDlg, (LPPOINT)&rect, 2);  // map screen coords to dialog client coords
        //causes bad flashing.  Are rect coords correct???
        InvalidateRect(hDlg, &rect, TRUE);  //erase background
#endif
#if 0
                Trace (globals, "scale factor\n"
                                " left = %d\n"
                                " top = %d\n"
                                " right = %d\n"
                                " bottom = %d\n",
                                rect.left,
                                rect.top,
                                rect.right,
                                rect.bottom);
#endif
    }

} /*ShowFilterParameterBlock*/

//Do any of these need to be dynamically issued at init time???
static BOOL fDragging = FALSE;
static BOOL fSizing = FALSE;
static BOOL fPaint= FALSE;
static BOOL fMaximized= FALSE;
BOOL fScaling = FALSE;			//Should be dynamically initialized at init time!?

#ifdef START_SCALEFACTOR_1
static long scaleFactor = 1;
#else
static long scaleFactor = -888; // -888 means compute optimal scale factor
#endif

void UpdateProxy (HWND hDlg, BOOL bEraseBackground)
{
    HWND hProxy;
    RECT imageRect;
    static long previousScaleFactor = -1; //force initial background erase. //Thread-Safe????

    // Question: Does this automatically take care of all
    // needed background erases?  If so, we can delete the
    // logic that does this everywhere else...
    // To test this, uncomment the following line:
    bEraseBackground = FALSE;
    if (scaleFactor > previousScaleFactor) {
        // If scale factor has increased, proxy image will
        // be smaller, so we need to erase the background.
        // (Actually, don't need to erase background if image
        // is still clipped to preview window at the new scale
        // factor -- we can do this optimization later if
        // unneeded flashing is a problem.)
        bEraseBackground = TRUE;
    }
    previousScaleFactor = scaleFactor;

    // Invalidate Proxy Item
    hProxy = GetDlgItem(hDlg, IDC_PROXY);
    GetWindowRect(hProxy, &imageRect);
    ScreenToClient (hDlg, (LPPOINT)&imageRect);
    ScreenToClient (hDlg, (LPPOINT)&(imageRect.right));
    InvalidateRect (hDlg, &imageRect, bEraseBackground);

} /*UpdateProxy*/

/*****************************************************************************/

int DoAction(CTLACTION action)
{
    //Perform the specified action for this control,,,
    switch(action) {
    
    case CA_SIZE: //Invoke FME_SIZE && FME_EXITSIZE events

        gFmcp->n = 0;
        gFmcp->e = FME_SIZE;
        gFmcp->previous = 0;
        gFmcp->ctlMsg = WM_SIZE;
        gFmcp->ctlCode = 0;
        gFmcp->ctlClass = gParams->ctl[0].class;
		//Catch any longjmp's back...
		if (setjmp(Mde_error_exit_jb)) {
			//someone longjmp'ed back to us...
			return FALSE;   //abort
		}
        X86_call_fmf1(gParams->CB_onCtl, (INT_PTR)gFmcp);
                
        //Make sure that the preview is displayed correctly
        GetProxyItemRect (MyHDlg, &gProxyRect); 
        SetupFilterRecordForProxy (MyGlobals, &scaleFactor, scaleFactor, &imageOffset);

        gFmcp->n = 0;
        gFmcp->e = FME_EXITSIZE;
        gFmcp->previous = 0;
        gFmcp->ctlMsg = WM_EXITSIZEMOVE;
        gFmcp->ctlCode = 0;
        gFmcp->ctlClass = gParams->ctl[0].class;
		//Catch any longjmp's back...
		if (setjmp(Mde_error_exit_jb)) {
			//someone longjmp'ed back to us...
			return FALSE;   //abort
		}
        X86_call_fmf1(gParams->CB_onCtl, (INT_PTR)gFmcp);

        break;

    case CA_PREVIEW:
        // Update the preview...
        if (!fDragging) {
            // Update Filter parameter block display
            ShowFilterParameterBlock(MyGlobals, MyHDlg);
            if (gResult == noErr) {
                // Do Filtering operation
                DoFilterRect (MyGlobals);
                // Invalidate Proxy Item
                UpdateProxy(MyHDlg, FALSE);
            }//noErr
        }
        break;
    case CA_APPLY:
        // Apply filter to original source image...
        SendMessage(MyHDlg, WM_COMMAND, IDOK, 0);
        break;
    case CA_CANCEL:
        cancelEventPerformed = 1;//Cancel Button was pressed
		// Exit without doing anything...
        SendMessage(MyHDlg, WM_COMMAND, IDCANCEL, 0);
        break;
#ifndef STANDALONE
    case CA_EDIT:
        // Enter/exit Edit Mode...
        SendMessage(MyHDlg, WM_COMMAND, IDC_EDIT, 0);
        break;
#endif
    case CA_ABOUT:
        // Display the ABOUT dialog box...
        SendMessage(MyHDlg, WM_COMMAND, IDC_HELP2, 0);
        break;
    case CA_RESET:
#if 0
        PlaySound("Goodbye", NULL, SND_ASYNC | SND_ALIAS);
#elif 0
        PlaySound(MAKEINTRESOURCE(IDR_WAVE1), hDllInstance, SND_ASYNC | SND_RESOURCE);
#elif 0
        PlaySound("Asterisk", NULL, SND_ASYNC | SND_ALIAS);
#elif 0
        PlaySound((const char *)SND_ALIAS_SYSTEMWELCOME, NULL, SND_ASYNC | SND_ALIAS);
#elif 0
        PlaySound("c:\\windows\\media\\chimes.wav", NULL, SND_ASYNC | SND_FILENAME);
#elif 0
        PlaySound("c:\\Program Files\\Microsoft Office\\Clipart\\MMedia\\rooster.wav", NULL, SND_ASYNC | SND_FILENAME);
#elif 0
        PlaySound("c:\\download\\thunder.wav", NULL, SND_ASYNC | SND_FILENAME);
#else
        Info("action=RESET is not yet implemented.");
#endif
        break;
    case CA_NONE:
    default:
        // No action...
        break;
    }//switch action
    return TRUE;
} /*DoAction*/


BOOL PerformControlAction(int n, FME e, int previous, int ctlMsg, int ctlCode,
                          BOOL fActionable, BOOL fProcessed, int overrideBusy)
{
    static int busy = 0;

    // Prevent reentry if reentry count != 0
    if (busy && overrideBusy==0) {
#if defined(_DEBUG) && 0
        OutputDebugString("PerformControlAction is BUSY!\n");
#endif
        return FALSE;
    }
    busy++;

#ifndef STANDALONE
    //make sure CB_onCtl is valid first!!!
    if (gParams->fValidCompile)
#endif
    {
        INT_PTR res;
        gFmcp->n = n;
        gFmcp->e = e;
        //Added by Harald Heim, May 22, 2003
        gFmcp->previous = previous;

        gFmcp->ctlMsg = ctlMsg;
        gFmcp->ctlCode = ctlCode;
        gFmcp->ctlClass = gParams->ctl[n].class;
		//Catch any longjmp's back...
		if (setjmp(Mde_error_exit_jb)) {
			//someone longjmp'ed back to us...
            busy--; //decrement reentry count
            return fProcessed;   //processed or not
		}
        res = X86_call_fmf1(gParams->CB_onCtl, (INT_PTR)gFmcp);
        if (res) {
            busy--; //decrement reentry count
            return TRUE;   //fully processed, skip default action
        }
    }//if
    //Perform the default action for this control...
    if (fActionable) {
        DoAction(gParams->ctl[n].action);
    }

    busy--; //decrement reentry count
    return fProcessed;   //processed or not
} /*PerformControlAction*/

/*****************************************************************************/

static int xFrom, xTo;
static int yFrom, yTo;

static void scaleControl (HWND hDlg, int nID)
{
    RECT r;
    HWND hCtl = GetDlgItem(hDlg, nID);

    if (hCtl) {
        GetWindowRect(hCtl, &r);                        //screen coords
        MapWindowPoints(NULL, hDlg, (LPPOINT)&r, 2);    //...to client coords
        // Undo the scaling...
        r.left   = (r.left * xTo)/xFrom;    // Order of multiply/divide is important!!
        r.right  = (r.right * xTo)/xFrom;
        r.top    = (r.top * yTo)/yFrom;
        r.bottom = (r.bottom * yTo)/yFrom;
        // Resize and reposition the window.
        SetWindowPos(hCtl, NULL,
                     r.left, r.top,
                     r.right - r.left,
                     r.bottom - r.top,
                     SWP_NOZORDER | SWP_NOACTIVATE);
    }//if
} /*scaleControl*/

static void scaleWindow (HWND hwnd)
{
    RECT r;

    GetWindowRect(hwnd, &r);    //screen coords
    SetWindowPos(hwnd, NULL,
                 0, 0,
                 (r.right - r.left)*xTo/xFrom,
                 (r.bottom - r.top)*yTo/yFrom,
                 SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE);

} /*scaleWindow*/

/*****************************************************************************/

// Helper function used by *.8bf import conversion...

#ifndef STANDALONE

static char *putCstring(char *p, const char label[12], const char field[256])
{
    int i;

    p += wsprintf(p, "%12.12s\"", label);
    for (i = 0; i < 256 && field[i] != 0; i++) {
        switch (field[i]) {
        case '\a':
            *p++ = '\\';
            *p++ = 'a';
            break;
        case '\b':
            *p++ = '\\';
            *p++ = 'b';
            break;
        case '\f':
            *p++ = '\\';
            *p++ = 'f';
            break;
        case '\n':
            p += wsprintf(p, "\\n\"\r\n            \"");
            break;
        case '\r':
            *p++ = '\\';
            *p++ = 'r';
            break;
        case '\t':
            *p++ = '\\';
            *p++ = 't';
            break;
        case '\v':
            *p++ = '\\';
            *p++ = 'v';
            break;
        case '\\':
            *p++ = '\\';
            *p++ = '\\';
            break;
#if 0
        case '\'':
            *p++ = '\\';
            *p++ = '\'';
            break;
#endif
        case '\"':
            *p++ = '\\';
            *p++ = '\"';
            break;
        default:
            *p++ = field[i];
            break;
        }//switch
    }//for
    p += wsprintf(p, "\"\r\n");
    return p;
} /*putCstring*/

#endif //ifndef STANDALONE

/*****************************************************************************/
#ifndef STANDALONE

int CheckForFileSave(void) {
    char szMessage[512];
    int iResult;

    // Returns IDCANCEL if user wants to cancel the current operation,
    // or if the Save failed in some way (including the case in which
    // the user explicitly cancels the Save).

    if (gParams->fNeedSourceSave) {
        // Get the message string...
        // (e.g., "The filter source code has been modified.\n\n"
        //        "Do you want to save it first?"
        LoadString(hDllInstance, IDS_SAVESRC, szMessage, sizeof szMessage);
        iResult = msgBox(MB_ICONWARNING|MB_YESNOCANCEL, "FilterMeister", szMessage);
        switch (iResult)
        {
        case IDYES:
        case IDOK:
            // Try to do a Save first...
            SendMessage(MyHDlg, WM_COMMAND, IDC_SAVE, 0);
            // cancel if save fails (in which case fNeedSourceSave is
            // still TRUE)...
            if (gParams->fNeedSourceSave)
                return IDCANCEL;
            else
                return IDNO;
            break;
        case IDCANCEL:
        case IDABORT:
            return IDCANCEL;
            break;
        case IDNO:
        default:
            break;
        }//switch
    }
    //No source save needed or wanted.
    return IDNO;
} /*CheckForFileSave*/

#endif //ifndef STANDALONE

/************************* FMCODE resource routines **************************************/

#ifdef FULL

void EncryptFmCode(char *p, int iAlgorithm, int iKey)
{
    CB_SECTION_HEADER *pHeader = (CB_SECTION_HEADER *)p;
    DWORD *pDW;
    int i;
    int iChecksum = 0;

    //note that we rely on the fact that all sections
    //are padded out to a multiple of 4 bytes to allow
    //us to encrypt a dword at a time

    //compute and store CRC???

    pHeader->iEncrypt = iAlgorithm;
    pHeader->iKey = iKey;

    switch (iAlgorithm) {
    case 1:
        // Algorithm 1: simple XOR
        p += sizeof(CB_SECTION_HEADER);
        pDW = (DWORD *)p;
        for (i = pHeader->nBytes; i > 0; i -= 4) {
            iChecksum += *pDW;
            *pDW++ ^= iKey;
        }//for
        break;
    case 2:
        // Algorithm 2: XOR against mutating key
        p += sizeof(CB_SECTION_HEADER);
        pDW = (DWORD *)p;
        for (i = pHeader->nBytes; i > 0; i -= 4) {
            iChecksum += *pDW;
            *pDW++ ^= iKey;
            //mutate the key
            iKey += 0x31459acf;
        }//for
        break;
	case 0:
    default:
        //default is no encryption, no checksum
        break;
    } //switch
    pHeader->iChecksum = iChecksum;
} /*EncryptFmCode*/

#endif //FULL


#ifdef STANDALONE

static BOOL DecryptFmCode(CB_SECTION_HEADER *pHeader)
{
    char *p = (char *)pHeader;
    DWORD *pDW;
    int i;
    int iAlgorithm;
    int iKey;
    int iChecksum = 0;

    //note that we rely on the fact that all sections
    //are padded out to a multiple of 4 bytes to allow
    //us to decrypt a dword at a time

    iAlgorithm = pHeader->iEncrypt;
    iKey = pHeader->iKey;

    //some simple credibility checks...
    if (pHeader->nBytes < 4 || pHeader->nBytes > 80000000) { //increase??? from 8MB to approx 80MB?
#ifdef VERBOSE
        ErrorOk("Implausible FMCODE section header size (%d)", pHeader->nBytes);
#endif
        return FALSE; //implausible -- exit quickly
    }

#if 0
    Info("Decrypting CB section, kind = %d", pHeader->iKind);
#endif

    switch (iAlgorithm) {
    case 1:
        // Algorithm 1: simple XOR
        p += sizeof(CB_SECTION_HEADER);
        pDW = (DWORD *)p;
        for (i = pHeader->nBytes; i > 0; i -= 4) {
            *pDW ^= iKey;
            iChecksum += *pDW++;
        }//for
        break;
    case 2:
        // Algorithm 2: XOR against mutating key
        p += sizeof(CB_SECTION_HEADER);
        pDW = (DWORD *)p;
        for (i = pHeader->nBytes; i > 0; i -= 4) {
            *pDW ^= iKey;
            iChecksum += *pDW++;
            //mutate the key
            iKey += 0x31459acf;
        }//for
        break;
	case 0:
    default:
        //default is no encryption, no checksum
        break;
    } //switch

#if 0
    Info("Finished decrypting CB section, kind = %d", pHeader->iKind);
#endif

    //check the checksum
    return (iChecksum == pHeader->iChecksum);
} /*DecryptFmCode*/


int LoadFmCode(void) {
    HRSRC hRes;         // handle/ptr. to res. info.
    HRSRC hResLoad;     // handle to loaded resource  
    char *pFmCodeBase;  // pointer to base of FMCODE resource data 
    char *pFmCode;      // pointer into FMCODE resource data 
	int FmCodeSubID;	// Resource subtype for FMCODE
    char *pTemp;        // temporary pointer to resources
    int cb;             // Size of resource (bytes)
    int nBytes;         // Number of code bytes in section
    int nSections = 0;  // Number of sections loaded
    int padCount;       // Number of pad bytes at end of section
    CB_SECTION_HEADER *pHeader;
    //CB *pCB;
	

    // Locate the FMCODE 132 or 164 resource. 
#ifdef _WIN64
	FmCodeSubID = 164;
#else //Win32
	FmCodeSubID = 132;
#endif
    hRes = FindResource(hDllInstance, MAKEINTRESOURCE(FmCodeSubID), "FMCODE"); 
    if (hRes == NULL) { 
        ErrorOk("Could not find FMCODE %d resource.", FmCodeSubID);
        return 0;
    } 

    // Check for reasonable size of resource.
    if ((cb=SizeofResource(hDllInstance, hRes)) < 100)
    {
        ErrorOk("FMCODE too short: %d bytes", cb);
        return 0;
    }

    // Load the FMCODE resource into global memory. 
    hResLoad = LoadResource(hDllInstance, hRes); 
    if (hResLoad == NULL) { 
        ErrorOk("Could not load FMCODE %d resource.", FmCodeSubID);
        return 0;
    } 

    // Lock the FMCODE resource into global memory. 
    pTemp = LockResource(hResLoad); // UNLOCK LATER??? No- UnlockResource() is a No-op!!!
    if (pTemp == NULL) { 
        ErrorOk("Could not lock FMCODE resource.");
        return 0;
    }

    // Make a read/write copy of FMCODE in case it's stored in readonly memory
    // (otherwise, we can't decrypt it!)	//WHY NOT COPY INTO gParams buffer & decrypt there???
    pFmCode = pFmCodeBase = malloc(cb);
    if (!pFmCode) {
        ErrorOk("Could not allocate %d bytes for FMCODE resource.", cb);
        return 0;
    }
    memcpy(pFmCode, pTemp, cb);

    //Parse and distribute the FMCODE payload...
    if (((DWORD)pFmCode & 3) != 0) {
        ErrorOk("FMCODE is not dword-aligned.");
        free(pFmCodeBase);
        return 0;
    }
    while (cb >= sizeof(CB_SECTION_HEADER)) {
        pHeader = (CB_SECTION_HEADER *)pFmCode;
		if (!DecryptFmCode(pHeader)) {
            // Decryption/checksum failed
            ErrorOk("FMCODE is corrupt");
            free(pFmCodeBase);
            return 0;
        }
        pFmCode += sizeof(CB_SECTION_HEADER);
        cb -= sizeof(CB_SECTION_HEADER);
        if (pHeader->iKind == CBK_LITPOOL) {
            // Load up the literal pool
			memcpy(&gFmcp->ndlits, pFmCode, sizeof(gFmcp->ndlits));		//X64 OKAY???
            if (gFmcp->ndlits > MAX_DLITS) {
                ErrorOk("FMCODE literal pool too large: %d entries", gFmcp->ndlits);
                gFmcp->ndlits = 0;
                free(pFmCodeBase);
                return 0;
            }
            pFmCode += sizeof(gFmcp->ndlits);
            cb -= sizeof(gFmcp->ndlits);
            nBytes = pHeader->nBytes - sizeof(gFmcp->ndlits);
            if (nBytes > cb) {
                ErrorOk("FMCODE too short for literal pool:\n"
                        "need %d bytes, but only have %d bytes",
                        nBytes, cb);
                free(pFmCodeBase);
                return 0;
            }
            if (nBytes > sizeof(gParams->dlitPool)) {
                ErrorOk("Code section too long for literal poool: %d bytes",
                        nBytes);
                free(pFmCodeBase);
                return 0;
            }
            if (nBytes != (int)(gFmcp->ndlits * sizeof(gFmcp->dlitPool[0]))) {
                ErrorOk("FMCODE literal pool count is wrong: %d", gFmcp->ndlits);
                free(pFmCodeBase);
                return 0;
            }
            memcpy(gFmcp->dlitPool, pFmCode, nBytes);
        }
        else if (pHeader->iKind == CBK_INITFMC) {
            // Load up initial values for FM context record
			nBytes = pHeader->nBytes;
            if (nBytes != (int)(offsetof(FMcontext, use_CMOV) -			//X64 OKAY???
                                offsetof(FMcontext, need_iuv) +
                                sizeof(gFmcp->use_CMOV)))
            {
                ErrorOk("FMCODE initial value size is wrong");
                free(pFmCodeBase);
                return 0;
            }
            memcpy(&gFmcp->need_iuv, pFmCode, nBytes);
            // Synchronize global (parameter) flags to FM context flags...
            gParams->flags &= ~(AFH_USE_MMX | AFH_USE_CMOV | AFH_USE_SSE | AFH_IS_TILEABLE);
            if (gFmcp->use_MMX)     gParams->flags |= AFH_USE_MMX;
            if (gFmcp->use_CMOV)    gParams->flags |= AFH_USE_CMOV;
            if (gFmcp->use_SSE)		gParams->flags |= AFH_USE_SSE;
            if (gFmcp->isTileable)  gParams->flags |= AFH_IS_TILEABLE;
        }
        else {
            // Must be a code buffer...
            BYTE *pBuffer;
            int maxBytes;
            pFmCode += sizeof(CB);
            cb -= sizeof(CB);
            nBytes = pHeader->nBytes - sizeof(CB);
            if (nBytes > cb) {
                ErrorOk("FMCODE too short for section %d:\n"
                        "need %d bytes, but only have %d bytes",
                        pHeader->iKind, nBytes, cb);
                free(pFmCodeBase);
                return 0;
            }
            switch (pHeader->iKind) {
            case CBK_FOREVERYPIXEL:
				pBuffer = gParams->codeBuffer;
                maxBytes = sizeof(gParams->codeBuffer);
                break;
            case CBK_INITCODE:
                pBuffer = gParams->CB_initCode;
                maxBytes = sizeof(gParams->CB_initCode);
                break;
            case CBK_DIVCHECK:
                pBuffer = gParams->CB_divCheck;
                maxBytes = sizeof(gParams->CB_divCheck);
                break;
            case CBK_ONZERODIVIDE:
                pBuffer = gParams->CB_onZeroDivide;
                maxBytes = sizeof(gParams->CB_onZeroDivide);
                break;
            case CBK_ONDIVIDEOVERFLOW:
                pBuffer = gParams->CB_onDivideOverflow;
                maxBytes = sizeof(gParams->CB_onDivideOverflow);
                break;
            case CBK_ONCTL:
                pBuffer = gParams->CB_onCtl;
                maxBytes = sizeof(gParams->CB_onCtl);
                break;
            case CBK_ONFILTERSTART:
                pBuffer = gParams->CB_onFilterStart;
                maxBytes = sizeof(gParams->CB_onFilterStart);
                break;
            case CBK_FOREVERYTILE:
                pBuffer = gParams->CB_forEveryTile;
                maxBytes = sizeof(gParams->CB_forEveryTile);
                break;
            case CBK_ONFILTEREND:
                pBuffer = gParams->CB_onFilterEnd;
                maxBytes = sizeof(gParams->CB_onFilterEnd);
                break;
            case CBK_USERFUNCTIONS:
                pBuffer = gParams->CB_UserFunctions;
                maxBytes = sizeof(gParams->CB_UserFunctions);
                break;
            default:
                ErrorOk("Unknown section kind in FMCODE: %d", pHeader->iKind);
                free(pFmCodeBase);
                return 0;
            }//switch
            if (nBytes > maxBytes) {
                ErrorOk("Code section %d too long for buffer: %d bytes",
                        pHeader->iKind, nBytes);
                free(pFmCodeBase);
                return 0;
            }
            memcpy(pBuffer, pFmCode, nBytes);
        }
        pFmCode += nBytes;
        cb -= nBytes;
        //pad to next dword boundary...
        padCount = -(int)pFmCode & 3;
        pFmCode += padCount;
        cb -= padCount;
        nSections++;
    }//while
	
    if (cb != 0) {
        // Due to padding???
        Warn("Leftover bytes at end of FMCODE: %d", cb);
    }

    //delete the temporary read/write copy of FMCODE
    free(pFmCodeBase);

#ifdef VERBOSE
    Info("Loaded FMCODE %d resource: %d sections", FmCodeSubID, nSections);
#endif

    return 1;
} /*LoadFmCode*/

#endif //#ifdef STANDALONE

/*****************************************************************************/

// Search PATH and FM_PATH for specified file name.
// Returns pointer to full path name if found, else
// returns pointer to original file name.
// CAUTION: Full path name is stored in a static buffer!

const char *FmSearchFile(const char *szFileName)
{
    static char pathbuffer[_MAX_PATH+1];

    /* Search for file in PATH environment variable: */
    _searchenv(szFileName, "PATH", pathbuffer);
    if (*pathbuffer != '\0') {
        //found it in PATH search list...
#if 0
        Info("Found file in PATH:\n%s", pathbuffer);
#endif
        return pathbuffer;
    }
    else {
        /* Search for file in FM_PATH environment variable: */
        _searchenv(szFileName, "FM_PATH", pathbuffer);
        if (*pathbuffer != '\0') {
            //found it in FM_PATH search list...
#if 0
        Info("Found file in FM_PATH:\n%s", pathbuffer);
#endif
            return pathbuffer;
        }
    }
    // If not found in PATH or FM_PATH, then just return
    // the file name verbatim...
#if 0
    Info("Not found in PATH or FM_PATH:\n%s", szFileName);
#endif
    return szFileName;
} /*FmSearchFile*/


/*****************************************************************************/

#ifndef STANDALONE

// Get text from the Source Editor (if it is active), regardless of whether
// it is implemented as a simple Edit control or a Rich Edit control.
//
// Returns TRUE if successful, else FALSE.

BOOL FmGetEditorText(void)
{
    size_t textsize;

    if (hEditorDlg) {
        // Source editor is active...
#ifdef USE_EDIT_CONTROL
        // DANGER!!! Win32 bug? GetDlgItemText sign-extends the 16-bit char
        // count to 32-bits, so if > 32767, it goes negative!!!
        // NOTE: Max # chars returned is 65535 (and the nMaxCount param is masked
        // by 0xffff), so need to use something else for Rich Edit controls if we
        // want max source code size > 64K!!!
        textsize = GetDlgItemText(hEditorDlg, IDC_EDIT2, gParams->szSrcCode,
                                    sizeof(gParams->szSrcCode)) & 0xffff;
#elif 0
        ////// THIS METHOD CAUSES FLASHING AND OTHER BAD SIDE EFFECTS /////
        // GetWIndowText won't return more than 64K chars, so we
        // need to use EM_GETSELTEXT to retrieve > 64K chars from
        // a Rich Edit control.
        CHARRANGE saveSel;
        static const CHARRANGE allSel = {0, -1};
        //save current selection
        SendMessage(hEdit2, EM_EXGETSEL, 0, (LPARAM)&saveSel);
        //select all source
        SendMessage(hEdit2, EM_EXSETSEL, 0, (LPARAM)&allSel);
        //WARNING!!! EM_GETSELTEXT doesn't check for buffer overflow!!!!
        //retrieve selected text
        textsize = SendMessage(hEdit2, EM_GETSELTEXT, 0, (LPARAM)gParams->szSrcCode);
        //restore saved selection
        SendMessage(hEdit2, EM_EXSETSEL, 0, (LPARAM)&saveSel);
#else
        ///// METHOD USING EM_GETTEXTRANGE ////////
        TEXTRANGE tr;
        tr.chrg.cpMin = 0;
        tr.chrg.cpMax = MAX_SOURCE_CODE_SIZE;   //select all text
        tr.lpstrText = gParams->szSrcCode;   //Check for overflow!!!!
        textsize = SendMessage(hEdit2, EM_GETTEXTRANGE, 0, (LPARAM)&tr);
#endif //USE_EDIT_CONTROL
        if (textsize >= sizeof(gParams->szSrcCode)-1)
        {   // Probably had to truncate to fit in gParams->szSrcCode...
#if 0
            //debugging...
            Info("FmGetEditorText: text size = %Iu", textsize);
#endif
#if 0
            ShowAlert(IDS_MAXTEXT);
#else   
            ErrorOk("Too much source code (more than %d characters).",
                    MAX_SOURCE_CODE_SIZE);
#endif
            // Set focus back to Source Code Edit Text Item
            SetFocus (hEdit2);
            return FALSE;
        }
    }
    // Got valid source code buffer...
    return TRUE;
} /*FmGetEditorText*/

#endif //ifndef STANDALONE

/*****************************************************************************/
// Filter dialog subclass procedure

#ifndef STANDALONE

static WNDPROC wpOrigFilterProc;

static LRESULT WINAPI FilterSubclassProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{ 
  
    if (wMsg == WM_KEYDOWN && wParam == VK_F12) {
        // Toggle the editor.
		
        PostMessage(MyHDlg, WM_COMMAND, IDC_EDIT, 0);
        return 0;   //processed
    }

    return CallWindowProc(wpOrigFilterProc, hwnd, wMsg, wParam, lParam);

} /*FilterSubclassProc*/

static WNDPROC wpOrigEditCtlProc;
    
static LRESULT WINAPI EditCtlSubclassProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{ 
    if (wMsg == WM_KEYDOWN && wParam == VK_F12) {
        // Toggle the editor.

        PostMessage(MyHDlg, WM_COMMAND, IDC_EDIT, 0);
        return 0;   //processed

    }

    return CallWindowProc(wpOrigEditCtlProc, hwnd, wMsg, wParam, lParam);

} /*EditCtlSubclassProc*/

#endif //STANDLAONE


LRESULT CALLBACK KeyHookProc( int code, WPARAM wParam, LPARAM lParam){
	
	if (code==HC_ACTION){
		
		//if (wParam==WM_KEYDOWN){
		if (!(lParam & 2147483648)){ //Check if bit 31 is not set -> key down
		
			//if (gParams->fValidCompile){ //Only after FM code was compiled otherwise memory error
				//if (gParams->gDialogEvent & 4){ //Only if keydown event is set for the dialog
					//if (GetForegroundWindow() == MyHDlg){ //not for editor window
						
						if (!(lParam & 1073741824)){ //Do not call repeatedly when key is pressed down

							//KBDLLHOOKSTRUCT  *pKeyBoard = (KBDLLHOOKSTRUCT *)lParam;
							//fm_triggerEvent((int)pKeyBoard->vkCode, FME_KEYDOWN, 0);
							
							fm_triggerEvent((int)wParam, FME_KEYDOWN, 0);
						}

					//}
				//}
			//}

		} else { //WM_KEYUP

			fm_triggerEvent((int)wParam, FME_KEYUP, 0);

		}

	}
	
	return CallNextHookEx ( KeyHook, code, wParam, lParam);

}
 




    
/*****************************************************************************/
#if 0
            // for debugging
            LONG bu;
            int buX;
            int buY;
#endif

static OPENFILENAME ofn;
static char szFileName[_MAX_PATH + 1];
char szTitleName[_MAX_FNAME + _MAX_EXT + 1];
FILETIME szFileTime;


#ifndef STANDALONE
//load source code -> cut out from below!
int openSourceFile (){

                         FILE  *file;
                         int    iLength;
                         PSTR   pstrBuffer, pstrBuffer2, p1, p2, p3;
                         int nBareCR, nDoubleCR;
                         time_t now = time(NULL);
/*                         char string[1024];
                         char string2[1024];
                         char string3[1024];
*/
                        
                         if (strlen (szFileName)==0){
                                return false;
                         }

/*
                         //Check for preprocessor instructions
                         if (!(file = fopen (szFileName, "r"))) {
                              msgBox(MB_ICONERROR, "FilterMeister I/O Error",
                                     "Could not open file\r\n%s", szFileName);
                              //FM_is_busy--;
                              return false;//break;
                         }

                         //fgets (string,15,file);
                         //fscanf(file, "%s\n", &string);
                         fgets (string,1023,file);
					     memset (string+strlen(string)-1, 0 ,1);
                         if ( strcmp(string,"//PREPROCESSOR")== 0){
                                //fgets (string,4,file); //Read end of line
                                //fscanf(file, "%s\n", &string);
                                //fscanf(file, "%s\n", &string2);
                                //fscanf(file, "%s\n", &string3);
                                //fgets (string,1023,file); //file path to processed file
                                //fgets (string2,1023,file); //preprocessor path
                                //fgets (string3,1023,file); //preprocessor command line options

                                fgets (string,1023,file); //file path to processed file
					            memset (string+strlen(string)-1, 0 ,1);
                                fgets (string2,1023,file);//preprocessor path
					            memset (string2+strlen(string2)-1, 0 ,1);
                                fgets (string3,1023,file);//preprocessor command line options
					            memset (string3+strlen(string3)-1, 0 ,1);
                                
                                if ((int)ShellExecute(gFmcp->hDlg, "open",
                                    (char *)string2+2,
                                    (char *)string3+2,
                                    NULL,//(char *)fm_getFilterInstallDir(),
                                    SW_SHOWNORMAL) > 32){
                                        
                                        strcpy (szFileName, string+2);
                                        fclose (file);

                                } else {
                                    msgBox(MB_ICONERROR, "FilterMeister Error",
                                     "Could not run preprocessor\r\n%s", string2+2);
                                    return false;
                                }
                                
                         }
*/

                         //Open File
                         if (!(file = fopen (szFileName, "rb"))) {
                              msgBox(MB_ICONERROR, "FilterMeister I/O Error",
                                     "Could not open file\r\n%s", szFileName);
                              //FM_is_busy--;
                              return false;//break;
                         }

                         // Determine file length...
                         fseek (file, 0, SEEK_END);
                         iLength = ftell (file);
                         rewind (file);

                         if (!(pstrBuffer = (PSTR) malloc (iLength + 1)))
                         {
                            fclose (file);
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not allocate %d bytes of memory.", iLength + 1);
                            //FM_is_busy--;
                            return false;//break;
                         }

                         fread (pstrBuffer, 1, iLength, file);
                         //check for correct read length???
                         
                         fclose (file);
                         pstrBuffer[iLength] = '\0';


                         { //Grab file date 
                            HANDLE fHandle = CreateFile(szFileName, GENERIC_READ | GENERIC_WRITE, 
                                    FILE_SHARE_READ | FILE_SHARE_DELETE,
                                    NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL);
                            GetFileTime (fHandle, NULL,NULL, &szFileTime);
                            CloseHandle (fHandle);
                         }


                         //Perform any necessary import file conversions

#define FUDGE 2048
                         if (iLength > 4 && _strnicmp(pstrBuffer, "%RGB", 4) == 0) {
                             // It's a Filter Factory (*.afs) file...

                             // Quick and dirty algorithm -- fails if two consecutive \r's

                             // Conversion will convert 9 0x0D's to 0x0D/0x0A pairs,
                             // and 4 0x0D/0x0D pairs to 0x0D/0x0A/0x0D/0x0A quadruples,
                             // so the result buffer must be 9 + 4*2 bytes longer than
                             // the input buffer.
                             char szYear[5];
                             char *pVal[8] = {NULL};
                             BOOL bSuppressThisCtl;
                             BOOL bExtendedAfs = FALSE;
                             char *zap1_from, *zap1_to;
                             char *zap2_from, *zap2_to;

                             if (!(pstrBuffer2 = (PSTR) malloc (iLength + 9 + 4*2 + 1 + FUDGE)))
                             {
                                msgBox(MB_ICONERROR, "FilterMeister System Error",
                                       "Could not allocate %d bytes of memory.",
                                       iLength + 9 + 4*2 + 1 + FUDGE);
                                free(pstrBuffer);  //free the input buffer
                                //FM_is_busy--;
                                return false;//break;
                             }

                             nBareCR = 0;
                             nDoubleCR = 0;
                             p1 = pstrBuffer;
                             p2 = pstrBuffer2;
                             p2 += wsprintf(p2,
                                "%%FFP\r\n\r\n"
                                            );
                             zap1_from = p2;
                             p2 += sprintf(p2,  //can't use wsprintf -- it doesn't support '*'???
                                "Category   :\"FilterMeister\"\r\n"
                                "Title      :\"%.*s\"\r\n"
                                "Copyright  :\"%s\"\r\n"
                                "Author     :\"unknown\"\r\n",
                                (ofn.nFileExtension - ofn.nFileOffset > 1) ?
                                (ofn.nFileExtension - ofn.nFileOffset - 1) : _MAX_PATH,
                                &szFileName[ofn.nFileOffset],
                                (strftime(szYear, sizeof(szYear), "%Y", localtime(&now)), szYear)
                                            );
                             zap1_to = p2;
                             p2 += wsprintf(p2,
                                "Filename   :\"%s\"\r\n"
                                "Description:\"Code converted by FilterMeister from Filter Factory source \"\r\n"
                                "            \"file %s \"\r\n"
                                "            \"on %.24s.\"\r\n"
                                "Version    :\"",
                                szTitleName,
                                szFileName,
                                ctime(&now)
                                            );
                             while (*p1) {
                                 if (p1[0] == 0x0D) {
                                     if (nBareCR < 9) {
                                         //convert first 9 bare CR's to CR/LF's.
                                         if (nBareCR == 0) {
                                             p2 += wsprintf(p2,
                                                 "\"\r\n\r\n"
                                                 );
                                             zap2_from = p2;
                                             p2 += wsprintf(p2,
                                                 "// Control definitions:\r\n"
                                                 "// Edit control labels as desired, and\r\n"
                                                 "// delete any unwanted controls.\r\n"
                                                 );
                                         }
                                         *p2++ = *p1++;
                                         *p2++ = 0x0A;
                                         if (nBareCR < 8) {
                                             p2 += wsprintf(p2,
                                                 "Ctl(%d):\"Control &%d:\",val=",
                                                 nBareCR, nBareCR);
                                             pVal[nBareCR] = p1;    //save pointer to val(n)
                                         }
                                         else {
                                             p2 += wsprintf(p2, "\r\n");
                                             zap2_to = p2;
                                             p2 += wsprintf(p2, "R:\r\n");
                                         }
                                         nBareCR++;
                                     }
                                     else if (p1[1] == 0x0D) {
                                         //convert next 4 CR/CR pairs to CR/LF/CR/LF quads,
                                         //then convert remaining CR/CR pairs to CR/LF pairs
                                         nDoubleCR++;
                                         switch (nDoubleCR) {
                                         case 5:
                                         case 6:
                                         case 7:
                                         case 8:
                                             *p2++ = '\"';
                                             break;
                                         case 9:
                                         case 10:
                                         case 11:
                                         case 12:
                                         case 13:
                                         case 14:
                                         case 15:
                                         case 16:
                                             if (!bSuppressThisCtl) {
                                                 p2 += wsprintf(p2, "\",val=");
                                                 if (p3=pVal[nDoubleCR - 9]) {
                                                     //fill in the val(n) field
                                                     while (*p3 >= '0' && *p3 <= '9') {
                                                         *p2++ = *p3++;
                                                     }
                                                 }
                                                 else {
                                                     //default the value to 0
                                                     *p2++ = '0';
                                                 }
                                             }
                                             break;
                                         }//switch
                                         *p2++ = *p1++;
                                         *p2++ = 0x0A;
                                         if (nDoubleCR <= 4 || nDoubleCR == 8) {
                                             *p2++ = *p1++;
                                             *p2++ = 0x0A;
                                         }
                                         else {
                                             p1++;
                                         }
                                         if (nDoubleCR < 4) {
                                             p2 += wsprintf(p2,
                                                 "%c:\r\n",
                                                 "RGBA"[nDoubleCR]);
                                         }
                                         else if (*p1 && nDoubleCR == 4) {
                                             p2 += wsprintf(p2, "Category   :\"");
                                             bExtendedAfs = TRUE;
                                         }
                                         else if (*p1 && nDoubleCR == 5) {
                                             p2 += wsprintf(p2, "Title      :\"");
                                         }
                                         else if (*p1 && nDoubleCR == 6) {
                                             p2 += wsprintf(p2, "Author     :\"");
                                         }
                                         else if (*p1 && nDoubleCR == 7) {
                                             p2 += wsprintf(p2, "Copyright  :\"");
                                         }
                                         else if (p1[0] >= '1' && p1[0] <= '8' &&
                                                  p1[1] == ':' && p1[2] == ' ') {
                                             if (p1[3] == 0x0D) {
                                                 //If label field is null, suppress
                                                 //this control...
                                                 bSuppressThisCtl = TRUE;
                                             }
                                             else {
                                                 bSuppressThisCtl = FALSE;
                                                 p2 += wsprintf(p2, "Ctl(%c):\"", p1[0] - 1);
                                             }
                                             p1 += 3;
                                         }
                                     }
                                     else {
                                         //delete bare CR's beyond the first 9
                                         //(FF inserts bare CR's in formulas to
                                         //break up lines into < 64 chars)
                                         p1++;
                                     }
                                 }
                                 else if (p1[0] == '\\' && p1[1] == 'r') {
                                     //convert embedded "\\r" to CR/LF pair
                                     *p2++ = 0X0D;
                                     *p2++ = 0X0A;
                                     p1 += 2;
                                 }
                                 else {
                                     //just copy the char
                                     *p2++ = *p1++;
                                 }
                             } //while
                             *p2 = 0;   //terminating null
                             //Zap non-extended keys if extended info was present...
                             if (bExtendedAfs) {
                                 //Must do deletions from high addresses to low addresses...
                                 strcpy(zap2_from, zap2_to);    //delete non-extended ctl's
                                 strcpy(zap1_from, zap1_to);    //delete synthesized Category.
                                                                //Title, Copyright, Author
                             }
                             free(pstrBuffer);  //free the input buffer
                             pstrBuffer = pstrBuffer2;  //output buffer becomes input buffer
                         } //if AFS file

                         else if (iLength > 100 && _strnicmp(pstrBuffer, "MZ", 2) == 0) {
                             // It might be a Filter Factory Plug-in (*.8bf) file...
                             typedef struct {   //structure of FF PARM resource
                                 int cbSize;    //size of this structure
                                 int nVersion;  //0=original FF, 1=standalone filter
                                 int val[8];    //initial values of controls
                                 int popDialog; //true if need to pop a parameter dialog
                                 int unknown1;
                                 int unknown2;
                                 int unknown3;
                                 int map_used[4];   //true if map(n) is used
                                 int ctl_used[8];   //true if ctl(n) is used
                                 char category[252];    //Category name
                                 // Michael Johannhanwahr's protect flag...
                                 int iProtected;        // == 1 means protected
                                 char title[256];       //Filter title
                                 char copyright[256];   //Copyright info
                                 char author[256];      //Filter author(s)
                                 char map[4][256];      //4 map labels
                                 char ctl[8][256];      //8 control labels
                                 char formula[4][1024]; //4 channel formulas
                             } PARM_T;
                             // Following code is cribbed from the Win32 SDK section
                             // on "updating Resources".  NOTE that there is an error
                             // in the example: LockResource(hRes) should be
                             // LockResource(hResLoad)!!!
                             HMODULE hMod;       // handle to .8bf file 
                             HRSRC hRes;         // handle/ptr. to res. info. in hMod
                             HRSRC hResLoad;     // handle to loaded resource  
                             PARM_T *pParm;      // pointer to PARM resource data 
                             int cb;             // Size of resource (bytes)
                             int i;
                             
                             // Load the .8bf file.
#if 0
                             hMod = LoadLibrary(szFileName); 
#else
                             hMod = LoadLibraryEx(szFileName, NULL, LOAD_LIBRARY_AS_DATAFILE); 
#endif
 
                             if (hMod == NULL) { 
                                 msgBox(MB_ICONERROR, "FilterMeister Error",
                                        "Could not load plug-in file\r\n%s"
                                        "\r\n\r\n"
                                        "Maybe a DLL is missing?", szFileName);
                                 free(pstrBuffer);  //free the input buffer
                                 //FM_is_busy--;
                                 return false;//break;
                             } 
 
                             // Locate the PARM 16 resource in the .8bf file. 
                             hRes = FindResource(hMod, MAKEINTRESOURCE(16), "PARM"); 
 
                             if (hRes == NULL) { 
                                 msgBox(MB_ICONERROR, "Not a Filter Factory Plug-in File",
                                     "Not a Filter Factory plug-in file:\r\n\r\n%s" 
                                     "\r\n\r\n"
                                     "(Plug-in doesn't have a PARM resource.)", 
                                     szFileName);
                                 FreeLibrary(hMod);
                                 free(pstrBuffer);  //free the input buffer
                                 //FM_is_busy--;
                                 return false;//break;
                             } 

                             // Check for reasonable size of resource.
                             if ((cb=SizeofResource(hMod, hRes)) < sizeof(*pParm))
                             {
                                  msgBox(MB_ICONERROR, "Not a Filter Factory Plug-in File",
                                     "Not a Filter Factory plug-in file:\r\n\r\n%s"
                                     "\r\n\r\n"
                                     "(Size of PARM resources is %d bytes.)", 
                                     szFileName, cb);
                                  FreeLibrary(hMod);
                                  free(pstrBuffer);  //free the input buffer
                                  //FM_is_busy--;
                                  return false;//break;
                             }
 
                             // Load the PARM resource into global memory. 
                             hResLoad = LoadResource(hMod, hRes); 
 
                             if (hResLoad == NULL) { 
                                 msgBox(MB_ICONERROR, "FilterMeister Error",
                                     "Could not load PARM resource from file\r\n%s", 
                                     szFileName);
                                 FreeLibrary(hMod);
                                 free(pstrBuffer);  //free the input buffer
                                 //FM_is_busy--;
                                 return false;//break;
                             } 
 
                             // Lock the PARM resource into global memory. 
                             pParm = LockResource(hResLoad); 
 
                             if (pParm == NULL) { 
                                 msgBox(MB_ICONERROR, "FilterMeister Error",
                                     "Could not lock PARM resource.");
                                 FreeLibrary(hMod);
                                 free(pstrBuffer);  //free the input buffer
                                 //FM_is_busy--;
                                 return false;//break;
                             }

                             // Check embedded size of PARM structure.
                             if (pParm->cbSize != sizeof(*pParm)) {
                                  msgBox(MB_ICONERROR, "Not a Filter Factory Plug-in File",
                                     "Not a Filter Factory plug-in file:\r\n\r\n%s"
                                     "\r\n\r\n"
                                     "(Size of PARM structure is %d bytes.)", 
                                     szFileName, pParm->cbSize);
                                  FreeLibrary(hMod);
                                  free(pstrBuffer);  //free the input buffer
                                  //FM_is_busy--;
                                  return false;//break;
                             }

                             // Check that this is a standalone filter, not
                             // the original Filter Factory.
                             if (pParm->nVersion == 0) {
                                  msgBox(MB_ICONERROR, "Not a Filter Factory Plug-in File",
                                     "Not a Filter Factory standalone plug-in:\r\n\r\n%s"
                                     "\r\n\r\n"
                                     "(This is the original Filter Factory!)", 
                                     szFileName, pParm->cbSize);
                                  FreeLibrary(hMod);
                                  free(pstrBuffer);  //free the input buffer
                                  //FM_is_busy--;
                                  return false;//break;
                             }
 
                             // Check Michael Johannhanwahr's Protect flag --
                             // Plugin Manager 2.0 sets this flag to prevent
                             // the casual user from seeing the filter's source
                             // code, so we should respect it also.
                             if (pParm->iProtected != 0) {
                                  msgBox(MB_ICONERROR, "Protected Filter Factory Plug-in",
                                      "Sorry, the source code for this filter is protected\r\n"
                                      "and the filter cannot be imported:\r\n\r\n"
                                      "%s",
                                     szFileName, pParm->cbSize);
                                  FreeLibrary(hMod);
                                  free(pstrBuffer);  //free the input buffer
                                  //FM_is_busy--;
                                  return false;//break;
                             }
 
#if 0
                             { int i;
                             Info("Successfully loaded PARM, size = %d.", cb);
                             Info("pParm = %8.8x", pParm);
                             Info("Parm cbSize = %d", pParm->cbSize);
                             Info("Parm version = %d", pParm->nVersion);
                             Info("Category:\r\n%.252s", pParm->category);
                             Info("Michael's protect flag = %d", pParm->iProtected);
                             Info("Title:\r\n%.256s", pParm->title);
                             Info("Copyright:\r\n%.256s", pParm->copyright);
                             Info("Author:\r\n%.256s", pParm->author);
                             for (i = 0; i < 4; i++) {
                                 if (pParm->map_used[i]) {
                                     Info("map(%d) = %.256s", i, pParm->map[i]);
                                 }
                             }
                             for (i = 0; i < 8; i++) {
                                 if (pParm->ctl_used[i]) {
                                     Info("ctl(%d) = %.256s", i, pParm->ctl[i]);
                                     Info("val(%d) = %d", i, pParm->val[i]);
                                 }
                             }
                             Info("R:\r\n%.512s", pParm->formula[0]);
                             Info("G:\r\n%.512s", pParm->formula[1]);
                             Info("B:\r\n%.512s", pParm->formula[2]);
                             Info("A:\r\n%.512s", pParm->formula[3]);
                             FM_is_busy--;
                             break;
                             }
#else
                             p2 = pstrBuffer;   //More than big enough!
                             p2 += wsprintf(p2, "%%FFP\r\n\r\n");

                             p2 = putCstring(p2, "Category   :", pParm->category);
                             p2 = putCstring(p2, "Title      :", pParm->title);
                             p2 = putCstring(p2, "Copyright  :", pParm->copyright);
                             p2 = putCstring(p2, "Author     :", pParm->author);

                             p2 += wsprintf(p2,
                                "Filename   :\"%s\"\r\n"
                                "Description:\"Code converted by FilterMeister from Filter Factory Plug-in \"\r\n"
                                "            \"file %s \"\r\n"
                                "            \"on %.24s.\"\r\n"
                                "Version    :\"1.0\"\r\n",
                                szTitleName,
                                szFileName,
                                ctime(&now)
                                    );

                             p2 += wsprintf(p2,
                                 "\r\n"
                                 "// Control definitions:\r\n\r\n"
                                 );
                             for (i = 0; i < 4; i++) {
                                 if (pParm->map_used[i]) {
                                     p2 += wsprintf(p2,
                                         "map(%d):\"%.256s\"\r\n",
                                         i, pParm->map[i]);
                                 }
                             }
                             for (i = 0; i < 8; i++) {
                                 if (pParm->ctl_used[i]) {
                                     p2 += wsprintf(p2,
                                         "ctl(%d):\"%.256s\",val=%d\r\n",
                                         i, pParm->ctl[i], pParm->val[i]);
                                 }
                             }
                             p2 += wsprintf(p2,
                                 "\r\nR:%.1024s\r\n", pParm->formula[0]);
                             p2 += wsprintf(p2,
                                 "\r\nG:%.1024s\r\n", pParm->formula[1]);
                             p2 += wsprintf(p2,
                                 "\r\nB:%.1024s\r\n", pParm->formula[2]);
                             p2 += wsprintf(p2,
                                 "\r\nA:%.1024s\r\n", pParm->formula[3]);
#endif
                             // Clean up. 
                             if (!FreeLibrary(hMod)) { 
                                 msgBox(MB_ICONERROR, "FilterMeister Error",
                                     "Could not free plug-in DLL.");
                                 free(pstrBuffer);  //free the input buffer
                                 FM_is_busy--;
                                 return false;//break;
                             } 
                         } //if .8bf file

                         gParams->fValidParse = FALSE; //invalidate any previous parse
                         gParams->fValidCompile = FALSE; //invalidate any previous compile
                         gParams->fGotSourceCode = TRUE;
                         gParams->fNeedSourceSave = FALSE;

#ifndef USE_EDIT_CONTROL
                         //Disable EN_CHANGE notification messages from the Rich Edit control.
                         SendMessage(hEdit2, EM_SETEVENTMASK, 0,
                                     SendMessage(hEdit2, EM_GETEVENTMASK, 0, 0) & ~ENM_CHANGE);
#endif

                         //SetDlgItemText(hEditorDlg?hEditorDlg:hDlg, IDC_EDIT2, pstrBuffer);
                         SetDlgItemText(hEditorDlg?hEditorDlg:gFmcp->hDlg, IDC_EDIT2, pstrBuffer);

#ifndef USE_EDIT_CONTROL
                         //Re-enable EN_CHANGE notification messages from the Rich Edit control.
                         SendMessage(hEdit2, EM_SETEVENTMASK, 0,
                                     SendMessage(hEdit2, EM_GETEVENTMASK, 0, 0) | ENM_CHANGE);
#endif
                         //SHOULD WARN if source code size > MAX...
                         {
                             size_t textsize = strlen(pstrBuffer);
                             if (textsize >= sizeof(gParams->szSrcCode)-1)
                             {   // Probably will have to truncate to fit in gParams->szSrcCode...
#if 0
                                 //debugging...
                                 Info("IDC_LOAD: text size = %Iu", textsize);
#endif
#if 0
                                 ShowAlert(IDS_MAXTEXT);
#else   
                                 ErrorOk("Too much source code (more than %d characters).",
                                         MAX_SOURCE_CODE_SIZE);
#endif
                             }
                         }

                         free (pstrBuffer);

                         // Set Focus to Source Code Edit Text Item
                         //SetFocus (GetDlgItem (hEditorDlg?hEditorDlg:hDlg, IDC_EDIT2));
                         SetFocus (GetDlgItem (hEditorDlg?hEditorDlg:gFmcp->hDlg, IDC_EDIT2));
                         
                         
                         return true;

}
#endif //ifndef STANDALONE




INT_PTR WINAPI FilterDialogProc(HWND hDlg, UINT wMsg, WPARAM wParam, LPARAM lParam)       // Win32 Change
{
#if !defined(STATIC_FILTER) && !defined(STANDALONE)
    long  lSize;
    int   nItem;
    char  szInput[32];
#endif
#if defined(USE_BITMAP) || defined(FML)
    // See BDLG32 (msdn sample 4544)
    static HBITMAP hBmpDialog;  // Handle of bitmap to display as dialog background.
    static HBITMAP hBmpOld;     // Handle of old bitmap to restore.
    static BITMAP  bmInfo;      // Bitmap information structure.
    static HDC     hdcMem;      // Compatible Memory DC for dialog.
    static POINT   pt;          // Position for upper left corner of bitmap.
    static SIZE    size;        // Size (width and height) of dialog client rectangle.
#endif
    char  buff[128];
    char  buff2[128];
    char *p;
    HWND  hCtrl;
    
    //Move up to top level
    //static HCURSOR hcHand;
    
    
    int idd;    // WIN32 Change
    int iCtl;
    static int iCtl2;
    int iCmd;
    long   prevScaleFactor = scaleFactor;
    static POINT ptPrev, ptNew;

    //Moved to main level of AfhFMUI.c, Harald Heim. Dec 14, 2002
    //static POINT imageOffset = {0, 0};


#ifndef NOOPTIMIZE_QUICK_TOGGLE
    static LONG fMouseMoved = 0;
#endif
    static GPtr globals = NULL;         /* needs to be static */
    static int iCurrentMouseoverCtl = -1;   //should re-init whenever controls change
                                            //e.g. after a recompile
#if 0
    static HACCEL hAccel;
#endif

    switch  (wMsg) {

        case  WM_INITDIALOG:

            /* set up globals   */
            globals = (GPtr) lParam;
            gFmcp->hDlg = MyHDlg = hDlg;

            // Create the Tool Tip control first!
#if 1
            {//scope
                INITCOMMONCONTROLSEX iccex;
                BOOL status;

                iccex.dwSize = sizeof(INITCOMMONCONTROLSEX);
                iccex.dwICC = ICC_WIN95_CLASSES;  //all Win95 classes
                status = InitCommonControlsEx(&iccex);
#if 0
                Info("InitCommonControlsEx status = %x", status);
#endif
            }//end scope
#else
            // Should link dynamically, so we only load comctl32.dll
            // if necessary!
            InitCommonControls();
#endif

            // For older versions of the SDK include files...
#ifndef TTS_BALLOON

#define TTS_NOANIMATE           0x10
#define TTS_NOFADE              0x20
#define TTS_BALLOON             0x40
#define TTS_CLOSE               0x80

//flags
#define TTF_PARSELINKS          0x1000


// ToolTip Icons (Set with TTM_SETTITLE)
#define TTI_NONE                0
#define TTI_INFO                1
#define TTI_WARNING             2
#define TTI_ERROR               3

#define TTM_SETTITLEA           (WM_USER + 32)  // wParam = TTI_*, lParam = char* szTitle
#define TTM_SETTITLEW           (WM_USER + 33)  // wParam = TTI_*, lParam = wchar* szTitle

#ifdef UNICODE
#define TTM_SETTITLE            TTM_SETTITLEW
#else
#define TTM_SETTITLE            TTM_SETTITLEA
#endif

#endif  //old include file


            ghTT = CreateWindowEx(  
                0,         // extended styles -- WS_EX_TOPMOST ???
                TOOLTIPS_CLASS,   // predefined class 
                NULL,       // text 
				WS_POPUP | TTS_ALWAYSTIP | TTS_NOPREFIX | WS_BORDER,
                CW_USEDEFAULT,         // starting x position 
                CW_USEDEFAULT,         // starting y position 
                CW_USEDEFAULT,        // button width 
                CW_USEDEFAULT,        // button height 
                hDlg,       // parent window 
                NULL,
                hDllInstance,
                NULL);      // pointer not needed 
#if 0
            Info("Tool Tip control created, ghTT = %x,\n"
                 "                          hDlg = %x,\n"
                 "                         hDllInstance = %x",
                 ghTT, hDlg, hDllInstance);
#endif

#if 1
            // Set max tooltip width...
            SendMessage(ghTT, TTM_SETMAXTIPWIDTH, 0, 300);
#endif

#if 1
            // Set tooltip background color...
            //SendMessage(ghTT, TTM_SETTIPBKCOLOR, (COLORREF)0xcccccc, 0);
            // Set tooltip text color...
            //SendMessage(ghTT, TTM_SETTIPTEXTCOLOR, RGB(255,0,0), 0);
#endif

#if 0
            // Set a title and icon for balloon tips...
            SendMessage(ghTT, TTM_SETTITLE, TTI_INFO, (LPARAM)"FilterMeister says:");
#endif

#if 1
            // Set tooltip delay times...
            {
                int iTime = 5000;   // initial show time in milliseconds
#if 0
                // TTDT_AUTOMATIC sets autopop = initial*10, reshow = initial/5...
                SendMessage(ghTT, TTM_SETDELAYTIME, TTDT_AUTOMATIC, MAKELONG(iTime, 0));
#elif 1
                // TTDT_AUTOMATIC with initial < 0 sets all values to default...
                SendMessage(ghTT, TTM_SETDELAYTIME, TTDT_AUTOMATIC, MAKELONG(-1, 0));
#else
                // Set each delay parameter indidually...
                SendMessage(ghTT, TTM_SETDELAYTIME, TTDT_INITIAL, MAKELONG(8000, 0));
                SendMessage(ghTT, TTM_SETDELAYTIME, TTDT_AUTOPOP, MAKELONG(10000, 0));
                SendMessage(ghTT, TTM_SETDELAYTIME, TTDT_RESHOW, MAKELONG(3000, 0));
#endif
            }
#endif

            // Per MSDN, we must explicitly define a ToolTip
            // control as topmost...
#if 0
            // but it screws up z-pos of common dialogs such as Load/Save!!!
            SetWindowPos(ghTT, HWND_TOPMOST, 0, 0, 0, 0,
                         SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
#endif

            // Save or restore dialog styles...
            if (gParams->gInitialInvocation) {
                //On first invocation, save initial dialog styles.
                gParams->gDialogInitialStyle =
                gParams->gDialogStyle = GetWindowLong(hDlg, GWL_STYLE);
                gParams->gDialogInitialStyleEx =
                gParams->gDialogStyleEx = GetWindowLong(hDlg, GWL_EXSTYLE);
            }
            else {
				//On subsequent invocations, restore styles in effect
                //when dialog last exited (but turn off VISIBLE bit for now).
#if 0
                Info("Prev style = %x\r\nCurrent style = %x",
                        gParams->gDialogStyle,
                        GetWindowLong(hDlg, GWL_STYLE));
                Info("Prev styleEx = %x\r\nCurrent styleEx = %x",
                        gParams->gDialogStyleEx,
                        GetWindowLong(hDlg, GWL_EXSTYLE));
#endif
                SetWindowLong(hDlg, GWL_STYLE, gParams->gDialogStyle & ~WS_VISIBLE);
                SetWindowLong(hDlg, GWL_EXSTYLE, gParams->gDialogStyleEx);
            }

            // Restore dialog position and size, and predefined controls...
            if (gParams->gInitialInvocation) {
                //Initial invocation...
                fm_setDialogPos(FALSE, -1, 0, 485,178); //388, 142);

                //Now define the default controls:
                //???? Could just pre-fill the ctls array with all the
                //???? right parameters, and let "Initialize user controls"
                //???? do all the dirty work?  But would also have to do
                //???? that in resetAllCtls...
                createPredefinedCtls();
            }
            else {
                //On subsequent invocations, use coordinates that were
                //saved at previous exit...
                SetWindowPos(hDlg, NULL,
                             gParams->gDialogRect.left,
                             gParams->gDialogRect.top,
                             gParams->gDialogRect.right - gParams->gDialogRect.left,
                             gParams->gDialogRect.bottom - gParams->gDialogRect.top,
                             SWP_NOZORDER | SWP_NOACTIVATE);
				
				// Restore/refresh dialog theme...
				//Info ("%d",gParams->gDialogTheme);
				//fm_setDialogTheme((int) gParams->gDialogTheme);
				
                //Avoid that the XP theme is reactivated on second invocation
				//setDefaultWindowTheme(GetDlgItem(hDlg, IDC_PROGRESS1));
                //setDefaultWindowTheme(GetDlgItem(hDlg, IDC_BUTTON1));
                //setDefaultWindowTheme(GetDlgItem(hDlg, IDC_BUTTON2));
                //setDefaultWindowTheme(GetDlgItem(hDlg, IDC_PROXY_GROUPBOX));
            }

			
			fm_setDialogTheme((int) gParams->gDialogTheme);



#if 0
            //testing enhanced metafile static control....
            {//scope
                HENHMETAFILE hImage = NULL;
                LONG lRes;
                char *lpszMetaFile = "e:\\msdev\\projects\\afhfm04\\buttrfly.wmf";
                //char *lpszMetaFile = "e:\\msdev\\projects\\afhfm04\\Lightblb.wmf";
                //char *lpszMetaFile = "e:\\msdev\\projects\\afhfm04\\siren.emf";
                //char *lpszMetaFile = "e:\\msdev\\projects\\afhfm04\\wizrd322.WMF";
                //char *lpszMetaFile = "e:\\msdev\\projects\\afhfm04\\wizrd32p.WMF";
                //char *lpszMetaFile = "e:\\msdev\\projects\\afhfm04\\FMGardener.emf";
                //char *lpszMetaFile = "e:\\msdev\\projects\\afhfm04\\FMGardenerFlatSS.emf";
                //char *lpszMetaFile = "e:\\shell3\\public_html\\flood\\clipart\\siren.emf";
#if 0
                // Read from *.emf file
                hImage = GetEnhMetaFile(lpszMetaFile);
#elif 0
                // Read from *.wmf file
                // (Doesn't work on files with initial 22-byte Aldus Placeable header.)
                HMETAFILE hMf;
                int cb;             // Size of resource (bytes)
                BYTE *lpvData;

                hMf = GetMetaFile(lpszMetaFile);
                if (!hMf) {
                    ErrorOk("Could not get WMF %s", lpszMetaFile);
                }
                else {
                    //later...
                    //DeleteMetaFile(hMf);
                    //Get size of metafile data
                    cb = GetMetaFileBitsEx(hMf, 0, NULL);
                    if (!cb) {
                        DeleteMetaFile(hMf);
                    }
                    else {
                        lpvData = malloc(cb);
                        if (!lpvData) {
                            DeleteMetaFile(hMf);
                            ErrorOk("Could not allocate %d bytes for WMF file", cb);
                        }
                        else {
                            //Now get the actual data bits...
                            if (!(cb = GetMetaFileBitsEx(hMf, cb, lpvData))) {
                                DeleteMetaFile(hMf);
                                free(lpvData);
                                ErrorOk("GetMetaFileBitsEx failed");
                            }
                            else {
                                //Info("WMF 1st dword = %x", *(DWORD *)lpvData);
                                hImage = SetWinMetaFileBits(cb, lpvData, NULL /*hdcRef*/, NULL /*lpmfp*/);
                                //Okay to delete wmf data here???
                                free(lpvData);
                            }
                        }
                    }
                }
#elif 0
                // Read from *.wmf (Plain Windows or Aldus Placeable) file
                FILE *file;
                int cb;             // Size of resource (bytes)
                BYTE *lpvData;

                if (!(file = fopen(lpszMetaFile, "rb"))) {
                    ErrorOk("Could not open WMF file\n%s", lpszMetaFile);
                }
                else {
                    // Determine file length...
                    fseek(file, 0, SEEK_END);
                    cb = ftell(file);
                    rewind(file);

                    if (!(lpvData = malloc(cb)))
                    {
                        fclose(file);
                        ErrorOk("Could not allocate %d bytes for WMF file", cb);
                    }
                    else {
                        //Now get the actual data bits...
                        cb = fread(lpvData, 1, cb, file);
                        //check for correct read length???
                        if (cb < 100) {
                            //implausible file size
                            ErrorOk("WMF file is too short (%d bytes)", cb);
                        }
                        else {
                            //Info("WMF 1st dword = %x", *(DWORD *)lpvData);
                            if (*(DWORD *)lpvData == 0x9AC6CDD7) {
                                //file has an Aldus placeable header
                                //--skip over it...
                                lpvData += 22;
                                cb -= 22;
                                //Info("WMF adjusted 1st dword = %x", *(DWORD *)lpvData);
                            }
                            if (((LPMETAHEADER)lpvData)->mtSize * 2 != (size_t)cb) {
                                ErrorOk("WMF file is wrong size (%d instead of %d)",
                                    cb, ((LPMETAHEADER)lpvData)->mtSize * 2);
                            }
                            else {
                                hImage = SetWinMetaFileBits(cb, lpvData,
                                                            NULL /*hdcRef*/,
                                                            NULL /*lpmfp*/);
                            }
                        }

                        free(lpvData);
                    }//if malloc

                    fclose(file);
                }//if fopen
#else
                // Read from resource
                // Could also use LoadIMage (  IMAGE_ENHMETAFILE... ) ???
                HRSRC hRes;         // handle/ptr. to res. info.
                HRSRC hResLoad;     // handle to loaded resource  
                char *pFmMf;        // pointer to FMMETAFILE resource data 
                int cb;             // Size of resource (bytes)

                // Locate the FMMETAFILE resource. 
                hRes = FindResource(hDllInstance, MAKEINTRESOURCE(IDR_FMMETAFILE1), "FMMETAFILE"); 
                //Info("hRes = %x", hRes);
                if (hRes == NULL) { 
                    ErrorOk("Could not find FMMETAFILE resource.");
                } 
                else {
                    // Get size of resource.
                    cb = SizeofResource(hDllInstance, hRes);
                    if (cb < 100) {
                        //Implausible size
                        ErrorOk("FMMETAFILE resource is too short (%d bytes)", cb);
                    }
                    else {
                        // Load the FMMETAFILE resource into global memory. 
                        hResLoad = LoadResource(hDllInstance, hRes); 
                        //Info("hResLoad = %x", hResLoad);
                        if (hResLoad == NULL) { 
                            ErrorOk("Could not load FMMETAFILE resource.");
                        } 
                        else {
                            // Lock the FMMETAFILE resource into global memory. 
                            pFmMf = LockResource(hResLoad); 
                            //Info("pFmMf = %x", pFmMf);
                            if (pFmMf == NULL) { 
                                ErrorOk("Could not lock FMMETAFILE resource.");
                            }
                            else {
                                // Determine what kind of metafile..,
                                if (((LPENHMETAHEADER)pFmMf)->iType == EMR_HEADER &&
                                    ((LPENHMETAHEADER)pFmMf)->dSignature == ENHMETA_SIGNATURE) {
                                    //It's an enhanced metafile
                                    hImage = SetEnhMetaFileBits(cb, pFmMf);
                                }
                                else {
                                    //Probably a Windows metafile
                                    if (*(DWORD *)pFmMf == 0x9AC6CDD7) {
                                        //metafile has an Aldus placeable header
                                        //--skip over it...
                                        pFmMf += 22;
                                        cb -= 22;
                                        //Info("WMF adjusted 1st dword = %x", *(DWORD *)pFmMf);
                                    }
                                    if (((LPMETAHEADER)pFmMf)->mtSize * 2 != (size_t)cb) {
                                        ErrorOk("WMF resource is wrong size (%d instead of %d)",
                                            cb, ((LPMETAHEADER)pFmMf)->mtSize * 2);
                                    }
                                    else {
                                        hImage = SetWinMetaFileBits(cb, pFmMf,
                                                                    NULL /*hdcRef*/,
                                                                    NULL /*lpmfp*/);
                                    }
                                }
                            }
                        }
                    }
                }
#endif
                if (!hImage) {
                    ErrorOk("Could not get metafile %s", lpszMetaFile);
                }
                else {
                    //Info("hImage = %x", hImage);
                    //Info("EMF 1st dword = %x", *(DWORD *)hImage);
                    lRes = SendDlgItemMessage(hDlg, IDC_EMF,
                        STM_SETIMAGE, IMAGE_ENHMETAFILE, (LPARAM)hImage);
                    //DeleteEnhMetaFile(lRes);  //delete previous image???
                    //DeleteEnhMetaFile(hImage);    //do this at end of dialog
                    //Info("STM_SETIMAGE returned %d", lRes);
                }
            }//scope
#endif

#ifndef STANDALONE
            fm_setCtlText(CTL_EDIT, 0); //update EDIT control text
#endif //ifndef STANDALONE

#ifdef STANDALONE
            //disable and hide the edit control (if it exists)
            EnableWindow(GetDlgItem(hDlg, IDX_EDIT), 0);
            ShowWindow(GetDlgItem(hDlg, IDX_EDIT), SW_HIDE);
#endif

#ifdef USE_BITMAP
#ifndef STRETCH_BITMAP
            // If the system font is not "small" (100%), then the dialog coordinates
            // will have been automatically stretched.  Since we aren't stretching
            // the bitmap, we need to "unstretch" all the dialog box coordinates
            // to bring everything back into registration with the bitmap.
            {//scope
            RECT r;
            int i;

#if 0
            //for debugging...
            #define SMALL_BUX 8     // Small system font width averages 8 pixels.
            #define SMALL_BUY 16    // Small system font height is 16 pixels.
            bu = GetDialogBaseUnits();  //Useless: actual scaling depends on dialog font!
            buX = LOWORD(bu);
            buY = HIWORD(bu);
#endif

            // Figure out what scaling was actually used by checking
            // the actual height and width of the dialog box against
            // what we thought it should be.
            GetWindowRect(hDlg, &r);    //Screen coords okay, only need width and height.
            xFrom = r.right - r.left;
            yFrom = r.bottom - r.top;
            //Here's what they should be if small fonts were in effect..
#if defined(FM_LIFTCOV)
            xTo = 579;
            yTo = 419;
#else
            //xTo = 459;
            //yTo = 380;
            xTo = 734;
            yTo = 314;
#endif

            // Scale the dialog box.
            scaleWindow(hDlg);

            // Scale the individual controls...
            scaleControl(hDlg, IDC_PROXY);
            scaleControl(hDlg, IDC_BUTTON1);
            scaleControl(hDlg, IDC_BUTTON2);
            scaleControl(hDlg, IDX_OK);
            scaleControl(hDlg, IDX_CANCEL);
            
            scaleControl(hDlg, IDC_PROGRESS1);
            scaleControl(hDlg, IDC_PROXY_GROUPBOX);

#ifdef ALLOW_DRAGOUT
            scaleControl(hDlg, IDC_DRAGOUT);
#endif
            scaleControl(hDlg, 498);    // Zoom factor
            scaleControl(hDlg, 499);    // Host name
            scaleControl(hDlg, 500);    // Filter title
#if defined(FM_LIFTCOV)
            scaleControl(hDlg, 501);    // Filter case
#endif

            for (i = 0; i < N_CTLS; i++)
            {   //scale the controls
                if (gParams->ctl[i].inuse)
                {
                    switch (gParams->ctl[i].class) {
                    case CC_SCROLLBAR:
                        scaleControl(hDlg, IDC_CTLBASE+i);    //scrollbar i
                        scaleControl(hDlg, IDC_BUDDY1+i);     //scrollbar i value
                        scaleControl(hDlg, IDC_BUDDY2+i);     //scrollbar i label
                        break;
                    default:
                        scaleControl(hDlg, IDC_CTLBASE+i);    //control i
                        break;
                    } //switch
                }
            }
            }//scope
#endif  //!STRETCH_BITMAP

            // Create clipping region for the bitmap.
            {//scope
                HRGN hRgn1, hRgn2;
                RECT rect;
                int nWidth, nHeight, r, L, T, R, B, tbh;

#ifndef NO_TITLE_BAR
                tbh = 0;
#else
                tbh = -16;
#endif

#if defined(FM_LIFTCOV)
                // Clipping region for LIFTCOV...
                GetWindowRect(hDlg, &rect); // returns screen coords
                nWidth = rect.right - rect.left;
                nHeight = rect.bottom - rect.top;
                // Main rect
                hRgn1 = CreateRectRgn(L=7, T=25+tbh, R=nWidth-8, B=nHeight-6);
                // Round the u/r corner
                r = 20;
                R++;    //shift just a hair to the right
                hRgn2 = CreateRectRgn(R-r, T, R, T+r);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_DIFF);
                hRgn2 = CreateEllipticRgn(R-2*r, T, R, T+2*r);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_OR);
                // Round the l/l corner
                r = 20;
                hRgn2 = CreateRectRgn(L, B-r, L+r, B);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_DIFF);
                hRgn2 = CreateEllipticRgn(L+1, B-2*r, L+2*r, B+1);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_OR);
#else
                // Clipping region for RADSINAL...
                GetWindowRect(hDlg, &rect); // returns screen coords
                nWidth = rect.right - rect.left;
                nHeight = rect.bottom - rect.top;
                // Main rect
                hRgn1 = CreateRectRgn(15, T=31+tbh, R=nWidth-42, B=nHeight-44);
                // Round the u/r corner
                r = 5;
                hRgn2 = CreateRectRgn(R-r, T, R, T+r);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_DIFF);
                hRgn2 = CreateEllipticRgn(R-2*r, T, R, T+2*r);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_OR);
                // Notch for OK button
                hRgn2 = CreateRectRgn(L=nWidth-111, T=nHeight-114, nWidth, nHeight);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_DIFF);
                // Round the u/l corner of the notch
                r = 15;
                hRgn2 = CreateRectRgn(L, T, L+r, T+r);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_OR);
                hRgn2 = CreateEllipticRgn(L, T, L+2*r, T+2*r);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_DIFF);
                // Round the u/r corner of the notch
                r = 5;
                hRgn2 = CreateRectRgn(R-r, T-r, R, T);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_DIFF);
                hRgn2 = CreateEllipticRgn(R-2*r, T-2*r, R, T);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_OR);
                // Round the l/l corner of the notch
                r = 5;
                hRgn2 = CreateRectRgn(L-r, B-r, L, B);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_DIFF);
                hRgn2 = CreateEllipticRgn(L-2*r, B-2*r, L, B);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_OR);
                // OK button
                hRgn2 = CreateEllipticRgn(nWidth-106, nHeight-112, nWidth-9, nHeight-15);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_OR);
                // Round the l/r corner of the main rect (minus the cutout)
                r = 5;
                L = 73;
                hRgn2 = CreateRectRgn(L-2, B-r, L+r, B);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_DIFF);
                hRgn2 = CreateEllipticRgn(L, B-2*r, L+2*r, B);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_OR);
                // Lower left cutout
                hRgn2 = CreateEllipticRgn(-40, nHeight-110, 78, nHeight-15);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_DIFF);
                // Round the u/l corner of the main rect (minus the cutout)
                r = 5;
                L = 15;
                T = 78+tbh;
                hRgn2 = CreateRectRgn(L, T-2, L+r, T+r);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_DIFF);
                hRgn2 = CreateEllipticRgn(L, T, L+2*r, T+2*r);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_OR);
                // Upper left cutout
                hRgn2 = CreateEllipticRgn(-200, -25+tbh, 210, 78+tbh);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_DIFF);
#endif //FM_LIFTCOV

#ifndef NO_TITLE_BAR
                // Title bar for dragging...
                hRgn2 = CreateRectRgn(15, 4, nWidth-42, 20);
                CombineRgn(hRgn1, hRgn1, hRgn2, RGN_OR);
#endif
                SetWindowRgn(
                        hDlg,   // handle to window whose window region is to be set
                        hRgn1,  // handle to region 
                        TRUE    // window redraw flag 
                           );
                //delete the regions with DeleteObj()???
            }//end scope
#elif defined(STATIC_FILTER)
            // try a rounded corner look?
            {//scope
                HRGN hRgn;
                RECT rect;
                int nWidth, nHeight;

                GetWindowRect(hDlg, &rect); // returns screen coords
                nWidth = rect.right - rect.left;
#if defined(FM_LIFTCOV)
                nHeight = (rect.bottom - rect.top)*92/100;  // Use top 92%
#else
                nHeight = (rect.bottom - rect.top)*55/100;  // Use top 55%
#endif
                hRgn = CreateRoundRectRgn(
                    4,      // x-coordinate of the region's upper-left corner 
                    4,      // y-coordinate of the region's upper-left corner 
                    nWidth - 4, // x-coordinate of the region's lower-right corner 
                    nHeight,    // y-coordinate of the region's lower-right corner 
                    40,     // height of ellipse for rounded corners 
                    50      // width of ellipse for rounded corners 
                    );
                SetWindowRgn(
                        hDlg,   // handle to window whose window region is to be set
                        hRgn,   // handle to region 
                        TRUE    // window redraw flag 
                           );
            }//end scope
#elif 0
            // try a non-rectangular look?
            {//scope
                HRGN hRgn;
                hRgn = CreateEllipticRgn(
                    0,  // x-coordinate of the upper-left corner of the bounding rectangle 
                    0,  // y-coordinate of the upper-left corner of the bounding rectangle 
                    400,    // x-coordinate of the lower-right corner of the bounding rectangle  
                    300     // y-coordinate of the lower-right corner of the bounding rectangle  
                   );   
                SetWindowRgn(
                        hDlg,   // handle to window whose window region is to be set
                        hRgn,   // handle to region 
                        TRUE    // window redraw flag 
                           );
            }//end scope
#endif //STATIC_FILTER

#if defined(USE_BITMAP)
            // See BDLG32 (msdn sample 4544)
            {   // Setup bitmap for display as dialog background.
                RECT    rectClient;
                HDC     hDC;

                // Load bitmap resource -- remember to call DeleteObject when done.
                // (interface.bmp should actually be 457x374 pixels)
                hBmpDialog = LoadBitmap(hDllInstance, (LPCTSTR) IDB_BITMAP2);
                if (!hBmpDialog)
                    ;   //failed

                // Get bitmap information.
                GetObject(hBmpDialog, sizeof(BITMAP), &bmInfo);

                GetClientRect(hDlg, &rectClient);
                size.cx = rectClient.right;     // zero-based
                size.cy = rectClient.bottom;    // zero-based
                pt.x = rectClient.left;
                pt.y = rectClient.top;

                // Create a compatible memory DC using the dialog's DC.
                hDC = GetDC(hDlg);  // release later!
                hdcMem = CreateCompatibleDC(hDC);
                if (!hdcMem)
                        ;   //failed

                // Select the bitmap into the memory DC.
                hBmpOld = SelectObject(hdcMem, hBmpDialog);
                SetMapMode(hdcMem, GetMapMode(hDC));    //needed???

                // Release the client window DC.
                ReleaseDC(hDlg, hDC);
            }
#endif //USE_BITMAP

            // Restore any dialog background image.
            //Need to make this pointer relative to gParams????
            gDialogImage = gParams->gDialogImage_save;

            // Restore any dialog clipping region.
            if (gParams->gDialogRegion != NULL) {
                SetWindowRgn(
                    hDlg,   // handle to window whose window region is to be set
                    gParams->gDialogRegion, // handle to region 
                    FALSE   // window redraw flag 
                       );
            }

            hcHand = LoadCursor(hDllInstance, MAKEINTRESOURCE(IDC_CURSOR1));
			cursorResource = IDC_CURSOR1;

            // Do this first, to compute scaleFactor...
            GetProxyItemRect (hDlg, &gProxyRect); 
            SetupFilterRecordForProxy (globals, &scaleFactor, scaleFactor, &imageOffset);

            // But don't run the filter until all other params are computed...
            //DoFilterRect (globals);

            // Set focus to OK button.
            SetFocus (GetDlgItem (hDlg, IDX_OK));

            // Give the zoom buttons a more readable font...
            // (But then the focus rectangles overwrite the font????)
            SendDlgItemMessage(hDlg, IDC_BUTTON1, WM_SETFONT,
                               (WPARAM) GetStockObject(SYSTEM_FIXED_FONT),
                               FALSE /*redraw flag*/);
            SendDlgItemMessage(hDlg, IDC_BUTTON2, WM_SETFONT,
                               (WPARAM) GetStockObject(SYSTEM_FIXED_FONT),
                               FALSE /*redraw flag*/);

            // Enable/disable zoom buttons...
            EnableWindow(GetDlgItem(hDlg, IDC_BUTTON1), scaleFactor > 1);
            EnableWindow(GetDlgItem(hDlg, IDC_BUTTON2), scaleFactor < MAX_SCALE_FACTOR);

            // Display Filter Case...
            // Should calculate this even if no filter dialog???
            strcpy( buff2, "" );
            switch (gStuff->filterCase)
            {
            case 0:
                strcat( buff2, "Host is not Photoshop 3.0 compatible");
                break;
            case filterCaseUnsupported:
                strcat( buff2, "Unsupported filter case" );
                break;
            case filterCaseFlatImageNoSelection:
                strcat( buff2, "Flat image, no selection" );
                break;
            case filterCaseFlatImageWithSelection:
                strcat( buff2, "Flat image with selection" );
                break;
            case filterCaseFloatingSelection:
                strcat( buff2, "Floating selection" );
                break;
            case filterCaseEditableTransparencyNoSelection:
                strcat( buff2, "Editable transparency, no selection" );
                break;
            case filterCaseEditableTransparencyWithSelection:
                strcat( buff2, "Editable transparency with selection" );
                break;
            case filterCaseProtectedTransparencyNoSelection:
                strcat( buff2, "Protected transparency, no selection" );
                break;
            case filterCaseProtectedTransparencyWithSelection:
                strcat( buff2, "Protected transparency with selection" );
                break;
            default:
                strcat( buff2, "Unknown filter case" );
                break;
            } /*switch*/

            strcpy(gFmcp->filterCaseText, buff2);

            // Set title bar text
            {
                char szTitle[MAX_LABEL_SIZE+1];
                // E.g., "AFH FilterMeister Beta 0.4.0"
                LoadString(hDllInstance, 2, szTitle, sizeof szTitle);
                if (gParams->gInitialInvocation) {
                    //set initial default filter title
                    //better to do this elsewhere???
                    strcpy(gFmcp->filterTitleText, szTitle);
                }
            }

            // Show host
            // Should calculate this even if no filter dialog???
            switch (gStuff->hostSig)
            {
            case 0:     //Ouch! Anyone else sets hostSIg to 0?
                p = "Macromedia Fireworks";
                break;
            case '8BIM':
                p = "Adobe Photoshop";
                break;
            case 'ART3':
                p = "Adobe Illustrator 3";
                break;
            case 'ARTZ':
                p = "Adobe Illustrator 88";
                break;
            case 'ARTY':
                p = "Adobe Illustrator 1.1";
                break;
            case 'ART5':
                p = "Adobe Illustrator 6";  // [sic?]
                break;
            case 'ARPI':
                p = "Adobe Illustrator 7";
                break;
            case 'ALD5':
                p = "PageMaker 5";
                break;
            case 'ALD6':
                p = "PageMaker 6";
                break;
            case 'AD65':
                p = "PageMaker 6.5";
                break;
            case 'FXTC':
                p = "Adobe After Effects";
                break;
            case 'PHUT':
                p = "Adobe PhotoDeluxe";
                break;
#if 0
            case 'PSAd':
                p = "Adobe ImageReady";	//(tm)
                break;
#else
            // The "Photoshop Adapter" is used by several Adobe
            // products, including AI 8.
            case 'PSAd':
                p = "Adobe Photoshop Adapter";
                break;
#endif
            case 'PMAN':
                p = "STOIK Picture Man";        //(tm) or (R)???
                break;
            case 'PrMr':
            case '8B)M':    //Adobe Premiere 5.0 Tryout ???
                p = "Adobe Premiere";
                break;
            case 'Cdrw':
                p = "CorelDRAW";
                break;
            case 'ARAX':    //They got this one backwards???
                p = "CorelXARA";
                break;
            case 'BABL':
                p = "Equilibrium DeBabelizer";
                break;
            case 'PSPI':
                p = "Jasc Paint Shop Pro";           //(tm) or (R)???
                break;
            case 'PSP8':
                p = "Jasc Paint Shop Pro 8";           //(tm) or (R)???
                break;
            case 'PSP9':
                p = "Jasc Paint Shop Pro 9";           //(tm) or (R)???
                break;
            case 'DAD6':
                p = "ACDSystems Canvas";           //(tm) or (R)???
                break;
            case 'PHLI':
                p = "PhotoLine";                //(tm)??? Computerinsel GmbH
                break;
            case 'PLU':     // [sic] Only 3 chars!
                p = "Ulead PhotoImpact";
                break;
            case 'PPUB':
                p = "Micrografx Picture Publisher";
                break;
            case 'FSX3':    // supposed to be this...
            case '3XSF':    // but they got it backwards???
                p = "Fractal Design Painter";   //now MetaCreations???
                break;
            case 'FSDA':
                p = "Fractal Design Dabbler";   //now MetaCreations Art Dabbler???
                break;
            case 'UP20':
                p = "Megalux Ultimate Paint 2.0";
                break;
            case '    ':    // Yechh!  I suppose they forgot to set this!
                p = "UTHSCSA ImageTool";         //(tm) or (R)???
                break;
            default:
                strcpy(buff, "Unknown Host (\"");
                strncat(buff, (const char *)&gStuff->hostSig, 4);
                strcat(buff, "\")");
                p = buff;
                break;
            } /*switch*/

            strcpy(gFmcp->filterHostText, p);

//Commented out by Harald Heim, March 9, 2003
#if 0
            SetDlgItemText(hDlg, 499, p);
#endif


            // Initialize user controls
            if (!gParams->gInitialInvocation)       //does this break standalones???
              for (iCtl = 0;  iCtl < N_CTLS;  iCtl++)
              {
                if (gParams->ctl[iCtl].inuse)
                {
#if 1
                    //createCtl clobbers these, so save them for later
                    int  val       = gParams->ctl[iCtl].val;
                    int  initVal   = gParams->ctl[iCtl].initVal;
                    int  minval    = gParams->ctl[iCtl].minval;
                    int  maxval    = gParams->ctl[iCtl].maxval;
                    int  lineSize  = gParams->ctl[iCtl].lineSize;
                    int  pageSize  = gParams->ctl[iCtl].pageSize;
					int  thumbSize  = gParams->ctl[iCtl].thumbSize;
                    int  imageType = gParams->ctl[iCtl].imageType;
                    int  divisor   = gParams->ctl[iCtl].divisor;
                    int  ticFreq   = gParams->ctl[iCtl].tb.ticFreq;
                    CTLACTION action = gParams->ctl[iCtl].action;
                    COLORREF textColor = gParams->ctl[iCtl].textColor;
                    COLORREF bkColor = gParams->ctl[iCtl].bkColor;
                    char tooltip[MAX_TOOLTIP_SIZE+1];
                    char image[_MAX_PATH+1];
					int noeditborder=false;
					int anchor = gParams->ctl[iCtl].anchor;
					int tabControl = gParams->ctl[iCtl].tabControl;
					int tabSheet = gParams->ctl[iCtl].tabSheet;
                    int scripted = gParams->ctl[iCtl].scripted;

                    strcpy(tooltip, gParams->ctl[iCtl].tooltip);
                    strcpy(image, gParams->ctl[iCtl].image);

					if (!(gParams->ctl[iCtl].buddy1Style & WS_BORDER)) noeditborder=true;
					
                    gParams->ctl[iCtl].inuse = FALSE;  //so fm_createCtl can work ???
                                                        //without trying to delete
                                                        //stale handles, etc. ???
                    fm_createCtl(iCtl,
                                 gParams->ctl[iCtl].class,         //class
                                 (INT_PTR)gParams->ctl[iCtl].label,//text
                                 gParams->ctl[iCtl].xPos,          //x
                                 gParams->ctl[iCtl].yPos,          //y
                                 gParams->ctl[iCtl].width,         //w
                                 gParams->ctl[iCtl].height,        //h
                                 gParams->ctl[iCtl].style,         //styles
                                 gParams->ctl[iCtl].styleEx,       //extended styles
                                 gParams->ctl[iCtl].properties,    //properties
                                 gParams->ctl[iCtl].enabled);      //visible/enabled

                    //restore clobbered attributes
                    gParams->ctl[iCtl].initVal = initVal;
					fm_setCtlTheme(iCtl, gParams->ctl[iCtl].theme); // Overwrites the background color if theme enabled
					fm_setCtlTab(iCtl, tabControl, tabSheet);		// Overwrites the background color if theme enabled
                    fm_setCtlRange(iCtl, minval, maxval);
                    fm_setCtlLineSize(iCtl, lineSize);
                    fm_setCtlPageSize(iCtl, pageSize);
					fm_setCtlThumbSize(iCtl, thumbSize);
                    fm_setCtlAction(iCtl, action);
                    fm_setCtlVal(iCtl, val);
                    fm_setCtlFontColor(iCtl, textColor);
                    fm_setCtlColor(iCtl, bkColor);
                    fm_setCtlToolTip(iCtl, (INT_PTR)tooltip, 0);
                    fm_setCtlImage(iCtl, (INT_PTR)image, imageType);
                    fm_setCtlDivisor(iCtl, divisor);
					fm_setCtlAnchor(iCtl, anchor);
                    fm_setCtlScripting(iCtl, scripted);

					//Set buddy styles
					if (gParams->ctl[iCtl].class == CC_STANDARD || gParams->ctl[iCtl].class == CC_SLIDER){
						fm_setCtlBuddyStyle (iCtl,0,gParams->ctl[iCtl].buddy2Style);
						fm_setCtlBuddyStyle (iCtl,1,gParams->ctl[iCtl].buddy1Style);
						fm_setCtlBuddyStyleEx (iCtl,0,gParams->ctl[iCtl].buddy2StyleEx);
						fm_setCtlBuddyStyleEx (iCtl,1,gParams->ctl[iCtl].buddy1StyleEx);
						if (noeditborder) fm_clearCtlBuddyStyle (iCtl,1,WS_BORDER);
					}

                    switch (gParams->ctl[iCtl].class) {
                    case CC_TRACKBAR:
					case CC_SLIDER:
                        fm_setCtlTicFreq(iCtl, ticFreq);
                        break;
                    default:
                        break;
                    } //switch class

#else
                    hCtrl = GetDlgItem(hDlg, IDC_CTLBASE+iCtl);
                    if (gParams->ctl[iCtl].minval > gParams->ctl[iCtl].maxval) {
                        //range is inverted...
                        SetScrollRange(hCtrl, SB_CTL,
                                       gParams->ctl[iCtl].maxval,
                                       gParams->ctl[iCtl].minval,
                                       FALSE /*no redraw*/);
                        SetScrollPos(hCtrl, SB_CTL,
                                     gParams->ctl[iCtl].minval + gParams->ctl[iCtl].maxval -
                                        gParams->ctl[iCtl].val,
                                     TRUE /*redraw*/);
                    }
                    else {
                        //normal range (min <= max)
                        SetScrollRange(hCtrl, SB_CTL,
                                       gParams->ctl[iCtl].minval,
                                       gParams->ctl[iCtl].maxval,
                                       FALSE /*no redraw*/);
                        SetScrollPos(hCtrl, SB_CTL,
                                     gParams->ctl[iCtl].val,
                                     TRUE /*redraw*/);
                    }
                    SetDlgItemInt(hDlg, IDC_BUDDY1+iCtl,
                                 gParams->ctl[iCtl].val,
                                 TRUE /*signed*/);

                    //set control text
#if 0
                    fm_fsetCtlText(iCtl, (int)gParams->ctl[iCtl].label);
#else
                    //avoid unnecessary redrawing...
                    SetDlgItemText(hDlg, IDC_BUDDY2+iCtl, gParams->ctl[iCtl].label);
#endif

                    //set control enable state
#if 0
                    enableCtl(iCtl);	//MWVDL 2008-07-18
#else
                    {
                        int level = gParams->ctl[iCtl].enabled;
                        EnableWindow(GetDlgItem(hDlg, IDC_CTLBASE+iCtl), (level & 2) != 0);
                        // enable or disable the control edit field
                        EnableWindow(GetDlgItem(hDlg, IDC_BUDDY1+iCtl), (level & 2) != 0);
                        // enable or disable the control text label (no, looks crappy)
                        ////EnableWindow(GetDlgItem(hDlg, IDC_BUDDY2+iCtl), (level & 2) != 0);
                        // show or hide the control
                        ShowWindow(GetDlgItem(hDlg, IDC_CTLBASE+iCtl), (level & 1)?SW_SHOW:SW_HIDE);
                        // show or hide the control edit field
                        ShowWindow(GetDlgItem(hDlg, IDC_BUDDY1+iCtl), (level & 1)?SW_SHOW:SW_HIDE);
                        // show or hide the control text label
                        ShowWindow(GetDlgItem(hDlg, IDC_BUDDY2+iCtl), (level & 1)?SW_SHOW:SW_HIDE);
                    }
#endif
#endif
                } else if (iCtl == CTL_PREVIEW || iCtl == CTL_FRAME || iCtl == CTL_ZOOM || iCtl == CTL_PROGRESS) {

					
					//if (iCtl == CTL_PREVIEW) Info ("%d\n%d\n%d\n%d\n",gParams->ctl[iCtl].xPos,
					//	gParams->ctl[iCtl].yPos,gParams->ctl[iCtl].width,gParams->ctl[iCtl].height );
					
					fm_setCtlPos (iCtl, gParams->ctl[iCtl].xPos,
						gParams->ctl[iCtl].yPos,
						gParams->ctl[iCtl].width,
                        gParams->ctl[iCtl].height);

					enableCtl(iCtl); //MWVDL 2008-07-18

				}
              } //for iCtl





// ------------------------------------------------------------------------------------------------
// ------------------------------------------------------------------------------------------------
// ------------------------------------------------------------------------------------------------
// ------------------------------------------------------------------------------------------------



#if 0
            // play around with creating controls dynamically....

            //create a static owner-draw control

            {
                HWND hwndButton;
                HFONT hDlgFont;
                /* Convert dialog units to pixels */
                RECT r;
                r.left = 352;   //x
                r.top = 97;    //y
                r.right = 32;  //w
                r.bottom = 43;  //h
                MapDialogRect(hDlg, &r);

                hwndButton = CreateWindowEx(  
                    0L,         // extended styles
                    "STATIC",   // predefined class 
                    "Custom",       // button text 
#if 1
                    WS_VISIBLE | WS_CHILD | SS_OWNERDRAW | SS_NOTIFY ,  // styles 
#else
                    WS_VISIBLE | WS_CHILD | SS_OWNERDRAW | SS_NOTIFY ,  // styles 
                    WS_VISIBLE | WS_CHILD | SS_ETCHEDFRAME | SS_NOTIFY ,  // styles 
                    WS_VISIBLE | WS_CHILD | SS_WHITERECT | SS_NOTIFY ,  // styles 
#endif
 
                    // Size and position values are given explicitly, because 
                    // the CW_USEDEFAULT constant gives zero values for buttons. 
                    r.left,         // starting x position 
                    r.top,         // starting y position 
                    r.right,        // button width 
                    r.bottom,        // button height 
                    hDlg,       // parent window 
                    (HMENU) 3986,       // button ID
#if 1
                    hDllInstance,
#else
                    (HINSTANCE) GetWindowLong(hDlg, GWL_HINSTANCE), 
#endif
                    NULL);      // pointer not needed 

                /// Set control font to same as dialog font
                if ((hDlgFont = (HFONT)SendMessage(hDlg, WM_GETFONT, 0, 0L)) != NULL)
                {
                     SendMessage(hwndButton, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                }
            }
#endif

#if 0
            // play around with creating controls dynamically....

            //create a pushbutton

            {
                HWND hwndButton;
                HFONT hDlgFont;
                /* Convert dialog units to pixels */
                RECT r;
                r.left = 180;   //x
                r.top = 126;    //y
                r.right = 60;  //w
                r.bottom = 14;  //h
                MapDialogRect(hDlg, &r);

                hwndButton = CreateWindowEx(  
                    0L,         // extended styles
                    "BUTTON",   // predefined class 
                    "Custom",       // button text 
#if 1
                    WS_VISIBLE | WS_CHILD | BS_AUTO3STATE | BS_LEFTTEXT | BS_RIGHT | WS_TABSTOP,  // styles 
#else
                    WS_VISIBLE | WS_CHILD | BS_PUSHBUTTON | WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | BS_PUSHBUTTON | BS_FLAT | WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | BS_AUTOCHECKBOX | WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | BS_AUTOCHECKBOX | BS_PUSHLIKE | WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | BS_AUTO3STATE | BS_PUSHLIKE | WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | BS_AUTO3STATE | BS_FLAT | WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | BS_AUTORADIOBUTTON | WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | BS_AUTORADIOBUTTON | BS_PUSHLIKE |WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | BS_GROUPBOX | WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | BS_GROUPBOX | BS_FLAT | WS_TABSTOP,  // styles 
#endif
 
                    // Size and position values are given explicitly, because 
                    // the CW_USEDEFAULT constant gives zero values for buttons. 
                    r.left,         // starting x position 
                    r.top,         // starting y position 
                    r.right,        // button width 
                    r.bottom,        // button height 
                    hDlg,       // parent window 
                    (HMENU) 3987,       // button ID
#if 1
                    hDllInstance,
#else
                    (HINSTANCE) GetWindowLong(hDlg, GWL_HINSTANCE), 
#endif
                    NULL);      // pointer not needed 

                /// Set control font to same as dialog font
                if ((hDlgFont = (HFONT)SendMessage(hDlg, WM_GETFONT, 0, 0L)) != NULL)
                {
                     SendMessage(hwndButton, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                }
            }
#endif

#if 0
            // play around with creating controls dynamically....

            //create a combobox

            {
                HWND hwndButton;
                HFONT hDlgFont;
                /* Convert dialog units to pixels */
                RECT r;
                r.left = 180;   //x
                r.top = 60;    //y
                r.right = 60;  //w
                r.bottom = 50;  //h
                MapDialogRect(hDlg, &r);

                hwndButton = CreateWindowEx(  
                    ES_WANTRETURN /*doesn't work?*/,         // extended styles
                    "COMBOBOX",   // predefined class 
                    "Custom",       // button text 
#if 1
                    WS_VISIBLE | WS_CHILD | CBS_DROPDOWNLIST | CBS_SORT | WS_VSCROLL | WS_TABSTOP,  // styles 
#else
                    WS_VISIBLE | WS_CHILD | CBS_DROPDOWN | CBS_SORT | CBS_AUTOHSCROLL | WS_VSCROLL | WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | CBS_DROPDOWN | WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | CBS_SIMPLE | WS_TABSTOP,  // styles 
#endif
 
                    // Size and position values are given explicitly, because 
                    // the CW_USEDEFAULT constant gives zero values for buttons. 
                    r.left,         // starting x position 
                    r.top,         // starting y position 
                    r.right,        // button width 
                    r.bottom,        // button height 
                    hDlg,       // parent window 
                    (HMENU) 3988,       // button ID
                    hDllInstance,
                    NULL);      // pointer not needed 

                // Set control to use the Extended UI
                SendMessage(hwndButton, CB_SETEXTENDEDUI, TRUE, 0);

                /// Set control font to same as dialog font
                if ((hDlgFont = (HFONT)SendMessage(hDlg, WM_GETFONT, 0, 0L)) != NULL)
                {
                     SendMessage(hwndButton, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                }

                // Fill combobox with a directory listing
                {
                    char szPathSpec[128] = "c:\\adobe\\photoshop\\plugins\\afh\\*.*";

                    DlgDirListComboBox(
                        hDlg,   // handle to dialog box with combobox  
                        szPathSpec, // pointer to path or filename string  
                        3988,   // identifier of combobox 
                        0,  // identifier of static control 
                        0   // file attributes to display 
                       );
                }

                //Set current selection to 2nd item in list
                SendMessage(hwndButton, CB_SETCURSEL, 1, 0);
            }
#endif

#if 0
            // play around with creating controls dynamically....

            //create a listbox

            {
                HWND hwndButton;
                HFONT hDlgFont;
                /* Convert dialog units to pixels */
                RECT r;
                r.left = 180;   //x
                r.top = 10;    //y
                r.right = 60;  //w
                r.bottom = 50;  //h
                MapDialogRect(hDlg, &r);

                hwndButton = CreateWindowEx(  
                    ES_WANTRETURN /*doesn't work?*/,         // extended styles
                    "LISTBOX",   // predefined class 
                    "Custom",       // button text 
#if 1
                    WS_VISIBLE | WS_CHILD | LBS_STANDARD | WS_TABSTOP,  // styles 
#else
                    WS_VISIBLE | WS_CHILD | LBS_STANDARD | LBS_EXTENDEDSEL | WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | LBS_NOTIFY | WS_VSCROLL | WS_TABSTOP,  // styles 
#endif
 
                    // Size and position values are given explicitly, because 
                    // the CW_USEDEFAULT constant gives zero values for buttons. 
                    r.left,         // starting x position 
                    r.top,         // starting y position 
                    r.right,        // button width 
                    r.bottom,        // button height 
                    hDlg,       // parent window 
                    (HMENU) 3989,       // button ID
                    hDllInstance,
                    NULL);      // pointer not needed 

                /// Set control font to same as dialog font
                if ((hDlgFont = (HFONT)SendMessage(hDlg, WM_GETFONT, 0, 0L)) != NULL)
                {
                     SendMessage(hwndButton, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                }

                // Fill listbox with a directory listing
                {
                    char szPathSpec[128] = "e:\\msdev\\projects\\afhfm04\\*.ffp";

                    DlgDirList(
                        hDlg,   // handle to dialog box with list box  
                        szPathSpec, // pointer to path or filename string  
                        3989,   // identifier of list box 
                        0,  // identifier of static control 
                        0   // file attributes to display 
                       );
                }

                //Set current selection (LB_SETCURSEL only works for single
                //selection list boxes - use LB_SELITEMRANGEEX for multiple
                //selection list boxes)
                SendMessage(hwndButton, LB_SETCURSEL, 3, 0);
            }
#endif

#if 0
            // play around with creating controls dynamically....

            //create a scrollbar

            {
                HWND hCtl;
                /* Convert dialog units to pixels */
                RECT r;
#if 0
                //vertical
                r.left = 270;   //x
                r.top = 10;    //y
                r.right = 8;  //w
                r.bottom = 90;  //h
#else
                //horizontal
                r.left = 270;   //x
                r.top = 105;    //y
                r.right = 90;  //w
                r.bottom = 8;  //h
#endif
                MapDialogRect(hDlg, &r);

                hCtl = CreateWindowEx(  
                    0,         // extended styles
                    "SCROLLBAR",   // predefined class 
                    "Custom",       // button text 
#if 1
                    WS_VISIBLE | WS_CHILD | SBS_HORZ | WS_TABSTOP,  // styles 
#else
                    WS_VISIBLE | WS_CHILD | SBS_VERT | WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | WS_TABSTOP,  // styles 
#endif
 
                    // Size and position values are given explicitly, because 
                    // the CW_USEDEFAULT constant gives zero values for buttons. 
                    r.left,         // starting x position 
                    r.top,         // starting y position 
                    r.right,        // button width 
                    r.bottom,        // button height 
                    hDlg,       // parent window 
                    (HMENU) 3992,       // button ID
                    hDllInstance,
                    NULL);      // pointer not needed 

                SetScrollRange(hCtl, SB_CTL, 0, 20, FALSE /*redraw*/);
                SetScrollPos(hCtl, SB_CTL, 5, TRUE /*redraw*/);

#if 0
                /// Set control font to same as dialog font
                {
                    HFONT hDlgFont;
                    if ((hDlgFont = (HFONT)SendMessage(hDlg, WM_GETFONT, 0, 0L)) != NULL)
                    {
                         SendMessage(hCtl, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                    }
                }
#endif
            }
#endif

#if 0
            // play around with creating controls dynamically....

            //create a trackbar

            {
//#include <commctrl.h>
                HWND hCtl;
                /* Convert dialog units to pixels */
                RECT r;

                // Should link dynamically, so only load comctl32.dll
                // if necessary!
                InitCommonControls();
#if 0
                //vertical
                r.left = 270;   //x
                r.top = 10;    //y
                r.right = 8;  //w
                r.bottom = 90;  //h
#else
                //horizontal
                r.left = 270;   //x
                r.top = 70;    //y
                r.right = 90;  //w
                r.bottom = 20;  //h
#endif
                MapDialogRect(hDlg, &r);

                hCtl = CreateWindowEx(  
                    0,         // extended styles
                    TRACKBAR_CLASS,   // predefined class 
                    "Custom",       // button text 
#if 1
                    WS_VISIBLE | WS_CHILD | TBS_BOTTOM | TBS_TOOLTIPS | TBS_ENABLESELRANGE | WS_TABSTOP,  // styles 
#else
                    WS_VISIBLE | WS_CHILD | TBS_BOTTOM | TBS_FIXEDLENGTH | TBS_ENABLESELRANGE | WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | TBS_BOTTOM | TBS_ENABLESELRANGE | WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | TBS_BOTTOM | TBS_NOTHUMB | TBS_ENABLESELRANGE | WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | TBS_BOTTOM | TBS_AUTOTICKS | TBS_ENABLESELRANGE | WS_TABSTOP,  // styles 
                    WS_VISIBLE | WS_CHILD | WS_TABSTOP,  // styles 
#endif
 
                    // Size and position values are given explicitly, because 
                    // the CW_USEDEFAULT constant gives zero values for buttons. 
                    r.left,         // starting x position 
                    r.top,         // starting y position 
                    r.right,        // button width 
                    r.bottom,        // button height 
                    hDlg,       // parent window 
                    (HMENU) 3990,       // button ID
                    hDllInstance,
                    NULL);      // pointer not needed 

                SendMessage(hCtl, TBM_SETRANGE, TRUE /*redraw*/, MAKELONG(0, 255));
                SendMessage(hCtl, TBM_SETPAGESIZE, 0 /*unused*/, 5);
                SendMessage(hCtl, TBM_SETSEL, TRUE /*redraw*/, MAKELONG(40, 75));
                SendMessage(hCtl, TBM_SETPOS, TRUE /*set*/, 50);
#if 1
                //TBM_SETTHUMBLENGTH requires TBS_FIXEDLENGTH???
                //sets height of pointer????
                SendMessage(hCtl, TBM_SETTHUMBLENGTH, 10 /*pixels*/, 0); //no effect??
#endif

#if 1
                {
                    // Create a buddy edit control for the trackbar....

                    HWND hEdit = CreateWindowEx(  
                        0,         // extended styles
                        "EDIT",   // predefined class 
                        "Custom",       // button text 
                        WS_VISIBLE | WS_CHILD | WS_BORDER | WS_TABSTOP,  // styles 
                        // Size and position values are given explicitly, because 
                        // the CW_USEDEFAULT constant gives zero values for buttons. 
                        0,         // starting x position 
                        0,         // starting y position 
                        r.bottom*2,        // control width 
                        r.bottom/2,        // control height 
                        hDlg,       // parent window 
                        (HMENU) 3997,       // control ID
                        hDllInstance,
                        NULL);      // pointer not needed 

                    /// Set edit buddy font to same as dialog font
                    {
                        HFONT hDlgFont;
                        if ((hDlgFont = (HFONT)SendMessage(hDlg, WM_GETFONT, 0, 0L)) != NULL)
                        {
                             SendMessage(hEdit, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                        }
                    }
                    SendMessage(hCtl, TBM_SETBUDDY, FALSE /*right/bottom*/,
                                (LPARAM) hEdit);
                }//scope
#endif

            }
#endif

#if 0
            // play around with creating controls dynamically....

            //create an up-down control

            {
//#include <commctrl.h>
                HWND hCtl;
                /* Convert dialog units to pixels */
                RECT r;

                // Should link dynamically, so only load comctl32.dll
                // if necessary!
                InitCommonControls();
#if 1
                //vertical
                r.left = 180;   //x
                r.top = 100;    //y
                r.right = 10;  //w
                r.bottom = 20;  //h
#else
                //horizontal
                r.left = 270;   //x
                r.top = 70;    //y
                r.right = 90;  //w
                r.bottom = 20;  //h
#endif
                MapDialogRect(hDlg, &r);

                hCtl = CreateWindowEx(  
                    0,         // extended styles
                    UPDOWN_CLASS,   // predefined class 
                    "Custom",       // button text 
#if 1
                    WS_VISIBLE | WS_CHILD | UDS_ARROWKEYS | WS_TABSTOP,  // styles 
#else
#endif
 
                    // Size and position values are given explicitly, because 
                    // the CW_USEDEFAULT constant gives zero values for buttons. 
                    r.left,         // starting x position 
                    r.top,         // starting y position 
                    r.right,        // button width 
                    r.bottom,        // button height 
                    hDlg,       // parent window 
                    (HMENU) 3993,       // button ID
                    hDllInstance,
                    NULL);      // pointer not needed 

#if 0
                SendMessage(hCtl, TBM_SETTHUMBLENGTH, 40 /*pixels*/, 0); //no effect??
#endif

#if 0
                /// Set control font to same as dialog font
                {
                    HFONT hDlgFont;
                    if ((hDlgFont = (HFONT)SendMessage(hDlg, WM_GETFONT, 0, 0L)) != NULL)
                    {
                         SendMessage(hCtl, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                    }
                }
#endif
            }
#endif

#if 0
            // play around with creating controls dynamically....

            //create a spinner control

            {
//#include <commctrl.h>
                HWND hCtl;
                HWND hSpinner;
                /* Convert dialog units to pixels */
                RECT r;

                // Should link dynamically, so only load comctl32.dll
                // if necessary!
                InitCommonControls();
#if 1
                //vertical
                r.left = 270;   //x
                r.top = 30;    //y
                r.right = 80;  //w
                r.bottom = 12;  //h
#else
                //horizontal
                r.left = 270;   //x
                r.top = 70;    //y
                r.right = 90;  //w
                r.bottom = 20;  //h
#endif
                MapDialogRect(hDlg, &r);

                //First create the buddy edit control...

                hCtl = CreateWindowEx(  
                    WS_EX_CLIENTEDGE,         // extended styles
                    "EDIT",   // predefined class 
                    "Custom",       // button text 
#if 1
                    WS_VISIBLE | WS_CHILD | WS_BORDER | WS_TABSTOP,  // styles 
#else
#endif
 
                    // Size and position values are given explicitly, because 
                    // the CW_USEDEFAULT constant gives zero values for buttons. 
                    r.left,         // starting x position 
                    r.top,         // starting y position 
                    r.right,        // button width 
                    r.bottom,        // button height 
                    hDlg,       // parent window 
                    (HMENU) 3994,       // control ID
                    hDllInstance,
                    NULL);      // pointer not needed 

#if 1
                hSpinner = CreateUpDownControl(
                    WS_VISIBLE | WS_CHILD | WS_BORDER | WS_TABSTOP |
#if 0
                    UDS_ALIGNLEFT | UDS_SETBUDDYINT | UDS_ARROWKEYS,    
#else
                    UDS_ALIGNRIGHT | UDS_SETBUDDYINT | UDS_ARROWKEYS,   
#endif
                    r.left,         // starting x position 
                    r.top,         // starting y position 
                    r.right,        // button width 
                    r.bottom,        // button height 
                    hDlg,       // parent window 
                    3995,       // control ID
                    hDllInstance,
                    hCtl,       //edit buddy
                    2000,   
                    0,  
                    40  
                   );
#endif

#if 0
                SendMessage(hCtl, TBM_SETTHUMBLENGTH, 40 /*pixels*/, 0); //no effect??
#endif

#if 1
                /// Set control font to same as dialog font
                {
                    HFONT hDlgFont;
                    if ((hDlgFont = (HFONT)SendMessage(hDlg, WM_GETFONT, 0, 0L)) != NULL)
                    {
                         SendMessage(hCtl, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                    }
                }
#endif
            }
#endif

#if 0
            // play around with creating controls dynamically....

            //create a progress bar control

            {
//#include <commctrl.h>
                HWND hCtl;
                /* Convert dialog units to pixels */
                RECT r;

#if 1
                INITCOMMONCONTROLSEX iccex;

                iccex.dwSize = sizeof(INITCOMMONCONTROLSEX);
                iccex.dwICC = ICC_PROGRESS_CLASS;
                InitCommonControlsEx(&iccex);
#else
                // Should link dynamically, so only load comctl32.dll
                // if necessary!
                InitCommonControls();
#endif
#if 1
                //vertical
                r.left = 180;   //x
                r.top = 10;    //y
                r.right = 10;  //w
                r.bottom = 100;  //h
#else
                //horizontal
                r.left = 270;   //x
                r.top = 70;    //y
                r.right = 90;  //w
                r.bottom = 10;  //h
#endif
                MapDialogRect(hDlg, &r);

                hCtl = CreateWindowEx(  
                    0,         // extended styles
                    PROGRESS_CLASS,   // predefined class 
                    "Custom",       // button text 
#if 1
                    WS_VISIBLE | WS_CHILD | PBS_VERTICAL | WS_BORDER,  // styles 
#else
                    WS_VISIBLE | WS_CHILD | PBS_SMOOTH | PBS_VERTICAL,  // styles 
                    WS_VISIBLE | WS_CHILD | PBS_SMOOTH,  // styles 
                    WS_VISIBLE | WS_CHILD,  // styles 
#endif
 
                    // Size and position values are given explicitly, because 
                    // the CW_USEDEFAULT constant gives zero values for buttons. 
                    r.left,         // starting x position 
                    r.top,         // starting y position 
                    r.right,        // button width 
                    r.bottom,        // button height 
                    hDlg,       // parent window 
                    (HMENU) 3996,       // button ID
                    hDllInstance,
                    NULL);      // pointer not needed 

#if 1
                SendMessage(hCtl, PBM_SETPOS, 50, 0);
                SendMessage(hCtl, PBM_STEPIT, 0, 0);

#define PBM_SETBARCOLOR (WM_USER + 9)
                SendMessage(hCtl, PBM_SETBARCOLOR, 0, RGB(255,0,0));  //IE 4.71

#define PBM_SETBKCOLOR 0x2001
                SendMessage(hCtl, PBM_SETBKCOLOR, 0, RGB(0,100,40));  //IE 4.71
#endif

#if 0
                /// Set control font to same as dialog font
                {
                    HFONT hDlgFont;
                    if ((hDlgFont = (HFONT)SendMessage(hDlg, WM_GETFONT, 0, 0L)) != NULL)
                    {
                         SendMessage(hCtl, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                    }
                }
#endif
            }
#endif

#if 1
            // play around with creating controls dynamically....

            //create a tooltip control

            {
//#include <commctrl.h>
                /* Convert dialog units to pixels */
#if 0
                TOOLINFO ti;


                // Should link dynamically, so only load comctl32.dll
                // if necessary!
                InitCommonControls();

                ghTT = CreateWindowEx(  
                    0,         // extended styles
                    TOOLTIPS_CLASS,   // predefined class 
                    "Custom",       // button text 
#if 1
                    WS_VISIBLE | WS_CHILD| WS_POPUP | TTS_ALWAYSTIP,  // styles 
#else
                    TTS_ALWAYSTIP | TTS_NOPREFIX,  // styles 
#endif
 
                    // Size and position values are given explicitly, because 
                    // the CW_USEDEFAULT constant gives zero values for buttons. 
                    CW_USEDEFAULT,         // starting x position 
                    CW_USEDEFAULT,         // starting y position 
                    CW_USEDEFAULT,        // button width 
                    CW_USEDEFAULT,        // button height 
                    hDlg,       // parent window 
                    (HMENU) 3991,       // button ID
                    hDllInstance,
                    NULL);      // pointer not needed 
#endif

#if 0
                //register a tooltip for a control
                ti.cbSize = sizeof(TOOLINFO);
                ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS;
                ti.hwnd = hDlg;
                ti.uId = (UINT) GetDlgItem(hDlg, 3987);
                //ti.rect = xxx;
                ti.hinst = hDllInstance;
                ti.lpszText = "This is a 3-state checkbox control.";
                SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);

                //register a tooltip for a control
                ti.cbSize = sizeof(TOOLINFO);
                ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS;
                ti.hwnd = hDlg;
                ti.uId = (UINT) GetDlgItem(hDlg, 3988);
                //ti.rect = xxx;
                ti.hinst = hDllInstance;
                ti.lpszText = "This is a drop-down combobox control.";
                SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);

                //register a tooltip for a control
                ti.cbSize = sizeof(TOOLINFO);
                ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS;
                ti.hwnd = hDlg;
                ti.uId = (UINT) GetDlgItem(hDlg, 3989);
                //ti.rect = xxx;
                ti.hinst = hDllInstance;
                ti.lpszText = "This is a listbox control.";
                SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);

                //register a tooltip for a control
                ti.cbSize = sizeof(TOOLINFO);
                ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS;
                ti.hwnd = hDlg;
                ti.uId = (UINT) GetDlgItem(hDlg, 3990);
                //ti.rect = xxx;
                ti.hinst = hDllInstance;
                ti.lpszText = "This is a trackbar control.";
                SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);

                //register a tooltip for a control
                ti.cbSize = sizeof(TOOLINFO);
                ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS;
                ti.hwnd = hDlg;
                ti.uId = (UINT) GetDlgItem(hDlg, 3992);
                //ti.rect = xxx;
                ti.hinst = hDllInstance;
                ti.lpszText = "This is a scrollbar control.";
                SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);

                //register a tooltip for a control
                ti.cbSize = sizeof(TOOLINFO);
                ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS;
                ti.hwnd = hDlg;
                ti.uId = (UINT) GetDlgItem(hDlg, 3993);
                //ti.rect = xxx;
                ti.hinst = hDllInstance;
                ti.lpszText = "This is an up-down control.";
                SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);

                //register a tooltip for a control
                ti.cbSize = sizeof(TOOLINFO);
                ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS;
                ti.hwnd = hDlg;
                ti.uId = (UINT) GetDlgItem(hDlg, 3995);
                //ti.rect = xxx;
                ti.hinst = hDllInstance;
                ti.lpszText = "This is a spinner control.";
                SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);

                //register a tooltip for a control
                ti.cbSize = sizeof(TOOLINFO);
                ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS;
                ti.hwnd = hDlg;
                ti.uId = (UINT) GetDlgItem(hDlg, 3996);
                //ti.rect = xxx;
                ti.hinst = hDllInstance;
                ti.lpszText = "This is a progress bar control.";
                SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);

                //register a tooltip for a control
                ti.cbSize = sizeof(TOOLINFO);
                ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS;
                ti.hwnd = hDlg;
                ti.uId = (UINT) GetDlgItem(hDlg, 499);
                //ti.rect = xxx;
                ti.hinst = hDllInstance;
                ti.lpszText = "Host application";
                SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);
#endif


//Not really needed
#if 0
                {
                    //register a tooltip for the PREVIEW control
                //DOESN'T work if proxy static text field is invisible!!!
                // So use a rect within the dialog client area and
                // relay mouse messages to the tooltip control...

                    TOOLINFO ti;
                ti.cbSize = sizeof(TOOLINFO);
                    ti.uFlags = 0;                      // or TTF_SUBCLASS???
                ti.hwnd = hDlg;
                    ti.uId = (UINT) CTL_PREVIEW;         // our id for this control
                ti.rect.left = gProxyRect.left;
                ti.rect.top = gProxyRect.top;
                ti.rect.right = gProxyRect.right;
                ti.rect.bottom = gProxyRect.bottom;
                ti.hinst = hDllInstance;
                
                //Edited by Harald Heim, January 12, 2003
                    ti.lpszText = "Left-drag to pan, or <A>right click</A> to choose a point";
                    //ti.lpszText = "Preview";
                
                SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);
                }
#endif

#if 0
                {
                    //register a tooltip for the ZOOM-IN button
                    TOOLINFO ti;
                ti.cbSize = sizeof(TOOLINFO);
                    ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS | TTF_CENTERTIP;
                ti.hwnd = hDlg;
                ti.uId = (UINT) GetDlgItem(hDlg, IDC_BUTTON1);
                //ti.rect = xxx;
                ti.hinst = hDllInstance;
                ti.lpszText = "Zoom in";
                SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);

                    //register a tooltip for the ZOOM-OUT button
                ti.cbSize = sizeof(TOOLINFO);
                    ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS | TTF_RTLREADING;
                ti.hwnd = hDlg;
                ti.uId = (UINT) GetDlgItem(hDlg, IDC_BUTTON2);
                //ti.rect = xxx;
                ti.hinst = hDllInstance;
                ti.lpszText = "Zoom out";
                SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);
                }
#endif

#if 0
                {
                    //register a tooltip for the PROGRESS control
                    TOOLINFO ti;
                ti.cbSize = sizeof(TOOLINFO);
                ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS;
                ti.hwnd = hDlg;
                ti.uId = (UINT) GetDlgItem(hDlg, IDC_PROGRESS1);
                //ti.rect = xxx;
                ti.hinst = hDllInstance;
                ti.lpszText = "Progress indicator";
                SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);
                }
#endif

#if 0
                {
                    //register a tooltip for the ABOUT control
                    TOOLINFO ti;
                ti.cbSize = sizeof(TOOLINFO);
                ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS;
                ti.hwnd = hDlg;
                ti.uId = (UINT) GetDlgItem(hDlg, 3986);
                //ti.rect = xxx;
                ti.hinst = hDllInstance;
                ti.lpszText = "About this filter";
                SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);
                }
#endif

#if 0
                /// Set control font to same as dialog font
                {
                    HFONT hDlgFont;
                    if ((hDlgFont = (HFONT)SendMessage(hDlg, WM_GETFONT, 0, 0L)) != NULL)
                    {
                         SendMessage(ghTT, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                    }
                }
#endif
            }
#endif



// ------------------------------------------------------------------------------------------------
// ------------------------------------------------------------------------------------------------
// ------------------------------------------------------------------------------------------------
// ------------------------------------------------------------------------------------------------




            // Update Filter parameter block display
            ShowFilterParameterBlock(globals, hDlg);

#ifdef STANDALONE
            if (gParams->gInitialInvocation) {
                // Load code from FMCODE resource
                gParams->fValidCompile = FALSE;	//pessimistic
				if (LoadFmCode()) {
                    // Execute the initialization code...
                    gParams->fValidCompile = TRUE;
                    if (setjmp(Mde_error_exit_jb)) {
                        //Someone longjmp'ed back to us...
                        gParams->fValidCompile = FALSE;
                    }
                    else {
                        //execute init code
#if 0
						INT_PTR retval;
						BYTE *p = gParams->CB_initCode;
						Info("About to call CB_initCode from WM_INITDIALOG in FilterDialogProc\n"
							"gParams->CB_initCode = %p\n"
							"gFmcp = %p\n"
							"gParams->fValidCompile = %d.\n",
							gParams->CB_initCode, gFmcp, gParams->fValidCompile);
						Info("CB_initCode code bytes at %p:\n"
							 "%2.2x %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n"
							 "%2.2x %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n"
							 "%2.2x %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n"
							 "%2.2x %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n"
							 , p
							 , p[ 0], p[ 1], p[ 2], p[ 3], p[ 4], p[ 5], p[ 6], p[ 7]
							 , p[ 8], p[ 9], p[10], p[11], p[12], p[13], p[14], p[15]
							 , p[16], p[17], p[18], p[19], p[20], p[21], p[22], p[23]
							 , p[24], p[25], p[26], p[27], p[28], p[29], p[30], p[31]
							 );
						retval =
#endif
                        X86_call_fmf1(gParams->CB_initCode, (INT_PTR)gFmcp);
#if 0
						Info("Returned from calling CB_initCode\n"
							 "retval = %d.\n"
							 "gParams->fValidCompile = %d.\n",
							 retval,
							 gParams->fValidCompile);
#endif
                    }
                }
            }
#endif


#ifdef SCRIPTABLE
            //Write the script params to the controls
            if (ScriptParamRead){
                int n;
                for (n=0;n<256;n++) {
                    if (ScriptParamUsed[n] && gParams->ctl[n].inuse && gParams->ctl[n].scripted){
                        fm_setCtlVal(n, ScriptParam[n]);
                    }
                }
            }
#endif


            // Restore/refresh dialog title...
            fm_setDialogText((INT_PTR) gParams->gDialogTitle);

			setDefaultWindowTheme(GetDlgItem(hDlg, IDC_PROGRESS1));
            setDefaultWindowTheme(GetDlgItem(hDlg, IDC_BUTTON1));
            setDefaultWindowTheme(GetDlgItem(hDlg, IDC_BUTTON2));
            setDefaultWindowTheme(GetDlgItem(hDlg, IDC_PROXY_GROUPBOX));


			// Start the hit-test timer running...
            KillTimer(hDlg, 2); //Cancel any previous hit-test timer.
            SetTimer(hDlg, 2, 20, NULL);   //previous Period = 250 msec.

			//LLKeyHook = SetWindowsHookEx(WH_KEYBOARD_LL,LLKeyHookProc, (HINSTANCE) GetWindowLong(hDlg, GWL_HINSTANCE),(DWORD)NULL);
			//KeyHook = SetWindowsHookEx(WH_KEYBOARD,KeyHookProc, (HINSTANCE) GetWindowLong(hDlg, GWL_HINSTANCE),(DWORD)NULL);
			

            // Finally, compute the proxy image now that all params are set...
            // (But skip it if AdvanceState() returned an error.)

            if (gResult == noErr) {
                // Set a timer to trigger a delayed proxy update,
                // giving the message loop a chance to paint the dialog
                // window first...
                KillTimer(hDlg, 1); //Cancel any previous trigger.
                if (!SetTimer(hDlg, 1, 100, NULL)) {
                    // If SetTimer() fails, compute the proxy immediately.
                    if (!fDragging) {
                        DoFilterRect (globals);
                    }
                }
            } //noErr

#if 0
            //Get Accelerators for Filter dialog
            hAccel = LoadAccelerators(hDllInstance, MAKEINTRESOURCE(IDR_ACCELERATOR_EDIT));
#endif

#ifndef STANDALONE
            // Always start the editor...
            gParams->flags |= AFH_EDITMODE;

            if (gParams->flags & AFH_EDITMODE) {
                // Start the Editor dialog.
                StartEditor();
            }

            // Subclass the filter dialog...
            wpOrigFilterProc = (WNDPROC) SetWindowLongPtr(hDlg, 
                                                       GWLP_WNDPROC,
                                                       (LONG_PTR)FilterSubclassProc);
            // Also subclass the Edit control...
            wpOrigEditCtlProc = (WNDPROC) SetWindowLongPtr(GetDlgItem(hDlg, IDX_EDIT), 
                                                       GWLP_WNDPROC,
                                                       (LONG_PTR)EditCtlSubclassProc);
#endif //STANDALONE



            //Added by Harald Heim, March 23, 2003
#ifndef STANDALONE
            //Display Welcome Message
            {
                
                int iValue=0;

                RQD(ROK(HKEY_CURRENT_USER, "Software\\AFH Systems\\FilterMeister"), "welcome", (DWORD *) &iValue);


                if ( !iValue )
                {   // Should implement this as a STRING resource???
                    Info ("Welcome to FilterMeister!\n\n"

                          "To get started please open one of the .ffp files from "
						  "the 'sourcecode' subfolder by presssing the 'Open' button. "
						  "Pressing the 'Compile' button executes that code.\n\n"

                          "Press the 'Help' button to get more information about "
						  "FilterMeister and its programming language.\n\n"

                          "We wish you a pleasant and productive time with FilterMeister!\n\n"

                          "The FilterMeister Development Team");
                    RSD(RCK(HKEY_CURRENT_USER, "Software\\AFH Systems\\FilterMeister"), "welcome", 1); 
            
                    //Display Editor at startup
                    ////SendMessage(hDlg, WM_COMMAND, IDC_EDIT, 0);
                }
               
            }
#endif

			/*if (gParams->gInitialInvocation) {
				//set current directory to filter install directory
				_chdir(fm_getFilterInstallDir());
			}*/

			//previous = false if initial invocation
			if (gParams->gDialogEvent & 1) 
				fm_triggerEvent(0, FME_INIT, (int)!gParams->gInitialInvocation);

			
            return  FALSE; // FALSE says we set the focus to a control.

#ifndef STANDALONE
        case WM_ACTIVATE:

            if (SendMessage(GetDlgItem(hEditorDlg, IDC_AUTOLOAD), BM_GETCHECK, 0, 0) == BST_CHECKED){
                    int fActive = LOWORD(wParam);           // activation flag 
                    //fMinimized = (BOOL) HIWORD(wParam); // minimized flag 
                    HWND hwndPrevious = (HWND) lParam;       // window handle 
                    //HWND activeWindow = GetActiveWindow();

                    //if ( LOWORD(wParam) == WA_CLICKACTIVE && (hwndPrevious !=hEditorDlg) && (hwndPrevious !=hDlg)){ 
                    if ( LOWORD(wParam) > 0 && (hwndPrevious !=hEditorDlg) && (hwndPrevious !=hDlg)){ 
                        //Reload and compile  
                        FILETIME dFileTime;
                        HANDLE fHandle = CreateFile(szFileName, GENERIC_READ | GENERIC_WRITE, 
                                    FILE_SHARE_READ | FILE_SHARE_DELETE,
                                    NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL);
                        GetFileTime (fHandle, NULL,NULL, &dFileTime);
                        CloseHandle (fHandle);
                            
                        if (szFileTime.dwLowDateTime != dFileTime.dwLowDateTime || 
                        szFileTime.dwHighDateTime != dFileTime.dwHighDateTime){
                            openSourceFile();
                            SendMessage(hDlg, WM_COMMAND, IDC_GO, 0);
                        }
                    }
            }
    
            return 0;

#endif

        case WM_ERASEBKGND:
            if (gDialogImage)
            // See BDLG32 (msdn sample 4544)
            {   // Setup bitmap file for display as dialog background.
                HDC     hDC;

                // First try loading the bitmap from an embedded BITMAP resource...
                hBmpDialog = LoadBitmap(hDllInstance, gDialogImage);

                //if (hBmpDialog) Info("Loaded bitmap from resource: %s", gDialogImage);

                // If it's not embedded, try searching for it as an
                // external bitmap file...
                if (!hBmpDialog) {
                    // Load bitmap from file...
                    hBmpDialog = LoadImage(hDllInstance,
                                           FmSearchFile(gDialogImage),
                                           IMAGE_BITMAP,
                                           0, 0,
                                           LR_LOADFROMFILE);
                }
                if (!hBmpDialog) {
                    //failed
#ifdef STANDALONE
                    //For standalone version, we just quietly ignore the
                    //failure to find the image file, and fall through to
                    //paint the background with whatever color or gradient
                    //is in effect.
                    gDialogImage = NULL;
                    gParams->gDialogImage_save = NULL;//don't retry on subsequent invocations
#else
                    gDialogImage = NULL;//first, to prevent reentry
                    ErrorOk("Could not load background image from file\n%s",
                            gParams->gDialogImage_save);
                    gParams->gDialogImage_save = NULL;//don't retry on subsequent invocations
                    return TRUE;    //otherwise background paints too much????
#endif
                }
                else {
                    // Get bitmap information.
                    GetObject(hBmpDialog, sizeof(BITMAP), &bmInfo);

                    // Create a compatible memory DC using the dialog's DC.
                    hDC = GetDC(hDlg);  // release later!
                    hdcMem = CreateCompatibleDC(hDC);
                    if (!hdcMem) {
                        //failed
                        // Release the client window DC.
                        ReleaseDC(hDlg, hDC);
                        DeleteObject(hBmpDialog);
                        hBmpDialog = NULL;
                        gParams->gDialogImage_save = gDialogImage = NULL;//first, to prevent reentry
                        ErrorOk("Could not create memory DC for background image");
                        return TRUE;    //otherwise background paints too much????
                    }
                    else {
                        // Select the bitmap into the memory DC.
                        hBmpOld = SelectObject(hdcMem, hBmpDialog);
                        SetMapMode(hdcMem, GetMapMode(hDC));    //needed???

                        // Set gDialogImage to null so we don't open this file again
                        // during this invocation.
                        gDialogImage = NULL;
                    }
                    // Release the client window DC.
                    ReleaseDC(hDlg, hDC);
                }
            }
            if (hBmpDialog && gParams->gDialogImage_save)
            {   // Display a bitmap as the dialog background.
                // MUST use the DC passed in wParam, not the DC of hDlg --
                // The clipping/update regions are different!!!
                HDC hDC = (HDC) wParam; // handle of device context

                if (gParams->gDialogImageMode == DIM_STRETCHED)
                //mode=stretched
                {   // Compress/stretch the bitmap as necessary...
                    // See BDLG32 (msdn sample 4544)
                    RECT rectClient;
                    POINT pt;
                    int oldMode;

                    GetClientRect(hDlg, &rectClient);
                    size.cx = rectClient.right;     // zero-based
                    size.cy = rectClient.bottom;    // zero-based
                    pt.x = rectClient.left;
                    pt.y = rectClient.top;
                    oldMode = SetStretchBltMode(hDC, gParams->gDialogImageStretchMode);
                    StretchBlt(hDC, pt.x, pt.y, size.cx, size.cy,
                               hdcMem, 0, 0, bmInfo.bmWidth, bmInfo.bmHeight,
                               SRCCOPY);
                    SetStretchBltMode(hDC, oldMode);
                }
                else if (gParams->gDialogImageMode == DIM_TILED)
                //mode=tiled
                {   // Use multiple straight BitBlts... (a la Prosise)
                    int x, y;
                    POINT ptSize, ptOrg, ptTotalSize;
                    RECT rectClient;

                    GetClientRect(hDlg, &rectClient);
                    size.cx = rectClient.right;     // zero-based
                    size.cy = rectClient.bottom;    // zero-based
                    ptSize.x = bmInfo.bmWidth;
                    ptSize.y = bmInfo.bmHeight;
                    DPtoLP(hDC, &ptSize, 1);
                    //need to convert size.cx, size.cy to LP??
                    ptTotalSize.x = size.cx;
                    ptTotalSize.y = size.cy;
                    DPtoLP(hDC, &ptTotalSize, 1);
                    ptOrg.x = 0;
                    ptOrg.y = 0;
                    DPtoLP(hdcMem, &ptOrg, 1);
                    for (y = 0; y < ptTotalSize.y; y += ptSize.y) {
                        for (x = 0; x < ptTotalSize.x; x += ptSize.x) {
                            BitBlt(hDC, x, y, ptSize.x, ptSize.y,
                                   hdcMem, ptOrg.x, ptOrg.y, SRCCOPY);
                        }
                    }
                }
                else
                //mode=exact
                {   // Use straight BitBlt... (a la Petzold)
                    POINT   ptSize, ptOrg;
                    ptSize.x = bmInfo.bmWidth;
                    ptSize.y = bmInfo.bmHeight;
                    DPtoLP(hDC, &ptSize, 1);
                    ptOrg.x = 0;
                    ptOrg.y = 0;
                    DPtoLP(hdcMem, &ptOrg, 1);
                    BitBlt(hDC, 0, 0, ptSize.x, ptSize.y,
                           hdcMem, ptOrg.x, ptOrg.y, SRCCOPY);
                }
            }
            else
            {   // Do a gradient fill a la Prosise BitmapDemo.cpp...
                HBRUSH hBrush[64];
                int i;
                int nWidth;
                int nHeight;
                RECT rect;
                // MUST use the DC passed in wParam, not the DC of hDlg --
                // The clipping/update regions are different!!!
                HDC hDC = (HDC) wParam; // handle of device context

                //COLORREF color1 = RGB(255,0,0);
                //COLORREF color2 = RGB(0,255,0);
                //COLORREF color1 =  GetSysColor(COLOR_BTNFACE);//RGB(0xD4,0xD0,0xC8);  //yellow grey
                //COLORREF color2 =  GetSysColor(COLOR_BTNFACE);//RGB(0xD4,0xD0,0xC8);   //yellow grey

                COLORREF color1 = gParams->gDialogGradientColor1;
                COLORREF color2 = gParams->gDialogGradientColor2;
                int r1 = GetRValue(color1);
                int g1 = GetGValue(color1);
                int b1 = GetBValue(color1);
                int r2 = GetRValue(color2);
                int g2 = GetGValue(color2);
                int b2 = GetBValue(color2);

                GetClientRect(hDlg, &rect);
                nWidth = rect.right - rect.left;
                nHeight = rect.bottom - rect.top;

#ifndef NOTRACE
                Trace (globals, "WM_ERASEBKGND\n"
                                " nWidth = %d\n"
                                " nHeight = %d\n"
                                " hDC = %#x",
                                nWidth,
                                nHeight,
                                hDC);
#endif

                if (color1 == color2) {
                    // optimize this if color1 == color2
                    hBrush[0] = CreateSolidBrush(color1);
                    FillRect(hDC, &rect, hBrush[0]);
                    DeleteObject(hBrush[0]);
                    //SetTimer(hDlg, 1002, 100, NULL);
                }
                else
                {
                    // create a 64-step linear gradient
                    for (i=0; i < 64; i++)
                    {
                        hBrush[i] = CreateSolidBrush(
                            RGB( (r1*(63-i)+r2*i)/63, (g1*(63-i)+g2*i)/63, (b1*(63-i)+b2*i)/63 )
                            );
                    }

                    if (gParams->gDialogGradientDirection == 0) {
                        // vertical gradient
#if 0
                        for (i=0; i < nHeight; i++) {
                            SetRect(&rect, 0, i, nWidth, i+1);
                            FillRect(hDC, &rect, hBrush[(i*64)/nHeight]);
                        }
#else
                        for (i=0; i < 64; i++) {
                            SetRect(&rect, 0, i*nHeight/64, nWidth, (i+1)*nHeight/64);
                            FillRect(hDC, &rect, hBrush[i]);
                        }
#endif
                    }
                    else {
                        // horizontal gradient
#if 0
                        for (i=0; i < nWidth; i++) {
                            SetRect(&rect, i, 0, i+1, nHeight);
                            FillRect(hDC, &rect, hBrush[(i*64)/nWidth]);
                        }
#else
                        for (i=0; i < 64; i++) {
                            SetRect(&rect, i*nWidth/64, 0, (i+1)*nWidth/64, nHeight);
                            FillRect(hDC, &rect, hBrush[i]);
                        }
#endif
                    }

                    for (i=0; i < 64; i++)
                        DeleteObject(hBrush[i]);
                } /* color1 != color2 */
                
            }

			//Apply Preview Color
			if (gParams->ctl[CTL_PREVIEW].bkColor != -1){
				//Only if image does not fill the preview
				if ( (gFmcp->x_start == 0 && gFmcp->x_end == gFmcp->X) || (gFmcp->y_start == 0 && gFmcp->y_end == gFmcp->Y) ){
					HBRUSH hBrush;
					RECT rect;
					HDC hDC = (HDC) wParam;

					GetWindowRect( GetDlgItem(hDlg, 101), &rect );
					ScreenToClient (hDlg, (LPPOINT)&rect);
					ScreenToClient (hDlg, (LPPOINT)&(rect.right));
					hBrush = CreateSolidBrush(gParams->ctl[CTL_PREVIEW].bkColor);
					FillRect(hDC, &rect, hBrush);
					DeleteObject(hBrush);
				}
			}
			

#if 1
            return TRUE;   // No more background painting needed.
#else
            // Seems to make things worse...
            return SetDlgMsgResult(hDlg, wMsg, TRUE);   // No more background painting needed.
#endif

        case WM_DRAWITEM:
            {//scope
                LPDRAWITEMSTRUCT pdis = (LPDRAWITEMSTRUCT)lParam;
                if (pdis->CtlID >= IDC_CTLBASE && pdis->CtlID < IDC_CTLBASE+N_CTLS) {
                    int n = pdis->CtlID - IDC_CTLBASE;
                    
					//if (n==CTL_PREVIEW) Info ("%d",gParams->ctl[n].inuse,!(gParams->ctl[n].enabled != 3));

                    if (gParams->ctl[n].inuse && (gParams->ctl[n].properties & CTP_DRAWITEM) && !(gParams->ctl[n].enabled != 3) ){
                    
                        PerformControlAction(
                                n,                      //control index
                                FME_DRAWITEM,           //FM event code
                                0,                      //previous value
                                WM_DRAWITEM,            //message type
                                0,                      //notification code
                                FALSE,                  //fActionable
                                TRUE,0);                  //fProcessed
                    }
                    
                    //It's a user-defined control
                    if (gParams->ctl[n].class != CC_IMAGE) {
                        return FALSE;       //for now
                    }
#if 1
                    //Info("WM_DRAWITEM: ctl = %d", n);
                    //InvalidateRect(pdis->hwndItem, NULL, TRUE); //Infinite loop!
#endif
                    if (gParams->ctl[n].image[0] == '\0') {
                        //no image
                        return FALSE;       //not handled
                    }
                    if (pdis->itemAction & ODA_DRAWENTIRE) {
                        //stretch a bitmap

                        if (gParams->ctl[n].im.hbm == 0) {
                            //first time through, load bitmap into a memory DC
                            //first try to load from RT_BITMAP resource...
                            gParams->ctl[n].im.hbm = LoadImage(hDllInstance,
                                                               gParams->ctl[n].image,
                                                               IMAGE_BITMAP,
                                                               0, 0,
                                                               LR_DEFAULTCOLOR);
                            if (gParams->ctl[n].im.hbm == 0) {
                                //if not in resources, load from file...
                                gParams->ctl[n].im.hbm = LoadImage(hDllInstance,
                                                                   FmSearchFile(gParams->ctl[n].image),
                                                                   IMAGE_BITMAP,
                                                                   0, 0,
                                                                   LR_LOADFROMFILE);
                            }
                            if (gParams->ctl[n].im.hbm == 0) {
                                gParams->ctl[n].im.hbm = (HBITMAP) 1; //1-shot stopper
#ifndef STANDALONE
                                ErrorOk("Can\'t load bitmap image:\n%s",
                                        gParams->ctl[n].image);
#endif
                                return FALSE;   //failed
                            }
                            else {
                                COLORREF crTransparent;

                                // Get bitmap information.
                                GetObject(gParams->ctl[n].im.hbm, sizeof(BITMAP),
                                          &gParams->ctl[n].im.bmInfo);

                                // Create a compatible memory DC.
                                gParams->ctl[n].im.hdcMem = CreateCompatibleDC(pdis->hDC);
                                if (!gParams->ctl[n].im.hdcMem) {
                                        return FALSE;   //failed
                                }
                                SetMapMode(gParams->ctl[n].im.hdcMem, GetMapMode(pdis->hDC));

                                // Select the bitmap into the memory DC.
                                gParams->ctl[n].im.hbmOld = SelectObject(gParams->ctl[n].im.hdcMem,
                                                                         gParams->ctl[n].im.hbm);

                                // Choose the transparent color to be the
                                // color of the pixel at (0,0)
                                crTransparent = GetPixel(gParams->ctl[n].im.hdcMem, 0, 0);

                                // Set the background color of the image
                                // to the transparent color chosen above
                                SetBkColor(gParams->ctl[n].im.hdcMem, crTransparent);

                                // Create another memory DC for the AND mask
                                gParams->ctl[n].im.hdcAnd = CreateCompatibleDC(pdis->hDC);
                                if (!gParams->ctl[n].im.hdcAnd) {
                                        return FALSE;   //failed
                                }
                                SetMapMode(gParams->ctl[n].im.hdcAnd, GetMapMode(pdis->hDC));

                                // Create a monochrome bitmap
                                gParams->ctl[n].im.hbmAnd =
                                    CreateBitmap(gParams->ctl[n].im.bmInfo.bmWidth,
                                                 gParams->ctl[n].im.bmInfo.bmHeight,
                                                 1, 1, NULL);

                                // Select the monochrome bitmap into the AND DC
                                gParams->ctl[n].im.hbmOldAnd =
                                    SelectObject(gParams->ctl[n].im.hdcAnd,
                                                 gParams->ctl[n].im.hbmAnd);

                                // Copy the image into the monochrome DC, setting
                                // all background-colored pixels to 1, and all
                                // other pixels to 0
                                BitBlt(gParams->ctl[n].im.hdcAnd,
                                       0, 0,
                                       gParams->ctl[n].im.bmInfo.bmWidth,
                                       gParams->ctl[n].im.bmInfo.bmHeight,
                                       gParams->ctl[n].im.hdcMem,
                                       0, 0,
                                       SRCCOPY);

                                // Set transparent bits in the image to black
                                BitBlt(gParams->ctl[n].im.hdcMem,
                                       0, 0,
                                       gParams->ctl[n].im.bmInfo.bmWidth,
                                       gParams->ctl[n].im.bmInfo.bmHeight,
                                       gParams->ctl[n].im.hdcAnd,
                                       0, 0,
                                       0x220326); //dest<-dest AND NOT src
                            }
                        }
                        if ((DWORD_PTR)gParams->ctl[n].im.hbm > 1) {
                            int oldMode = SetStretchBltMode(pdis->hDC, COLORONCOLOR);
                            //Set foreground color to black and background color to
                            //white in the display DC to get desired monochrome
                            //conversion for the AND mask...
                            COLORREF oldTextColor = SetTextColor(pdis->hDC, RGB(0,0,0));
                            COLORREF oldBkColor = SetBkColor(pdis->hDC, RGB(255,255,255));
                            StretchBlt(pdis->hDC,
                                       pdis->rcItem.left,
                                       pdis->rcItem.top,
                                       pdis->rcItem.right - pdis->rcItem.left,
                                       pdis->rcItem.bottom - pdis->rcItem.top,
                                       gParams->ctl[n].im.hdcAnd,
                                       0, 0, 
                                       gParams->ctl[n].im.bmInfo.bmWidth, 
                                       gParams->ctl[n].im.bmInfo.bmHeight,
                                       SRCAND);
                            //Restore the display DC's previous foreground and background
                            //colors...
                            SetTextColor(pdis->hDC, oldTextColor);
                            SetBkColor(pdis->hDC, oldBkColor);
#if 1
                            StretchBlt(pdis->hDC,
                                       pdis->rcItem.left,
                                       pdis->rcItem.top,
                                       pdis->rcItem.right - pdis->rcItem.left,
                                       pdis->rcItem.bottom - pdis->rcItem.top,
                                       gParams->ctl[n].im.hdcMem,
                                       0, 0, 
                                       gParams->ctl[n].im.bmInfo.bmWidth, 
                                       gParams->ctl[n].im.bmInfo.bmHeight,
#if 1
                                       SRCPAINT);   //or SRCINVERT
#else
                                       SRCINVERT);   //or SRCINVERT
#endif
#endif
                            SetStretchBltMode(pdis->hDC, oldMode);
                        }
                    } //if ODA_DRAWENTIRE
                    if (pdis->itemAction & ODA_SELECT) {
                    }
                    if (pdis->itemAction & ODA_FOCUS) {
                    }
                    return TRUE;    //handled
                }//if user control
            }//scope
            return FALSE;   //not handled

        case WM_CTLCOLORSTATIC:
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORSCROLLBAR:
#if 0
            // Yes, this controls nonreadonly edit box colors;
            // Readonly edit boxes seem to use WM_CTLCOLORSTATIC instead.
        case WM_CTLCOLOREDIT:
#endif

            {
                HDC hdc = (HDC) wParam;   // handle of display context
                HWND hwnd = (HWND) lParam; // handle of static control
                int iCtl;

                iCtl = GetDlgCtrlID(hwnd);    // ID of control
				
				if (// iCtl >= IDC_BUDDY2 && iCtl < IDC_BUDDY2+N_CTLS   // Scrollbar labels
                    //|| iCtl >= 600 && iCtl < 600+N_CHECKBOXES   // Checkboxes--NFG, not 600!
                    // || iCtl == IDC_SLIDER1          // trackbar -- NFG!!!
                       (iCtl >= 803 && iCtl <= 899) // Static text but NOT Group boxes
                       //(iCtl >= 800 && iCtl <= 899) // Static text and [NFG]Group boxes
                    || (iCtl >= IDC_TRACE && iCtl <= IDC_USE_SSE)			//AARRGH!!
                    || (iCtl >= IDC_DISASSEMB && iCtl <= IDC_USE_CMOV)
                    || (iCtl == IDC_USE_INLINES)
                    //|| iCtl == 410                  // Slider1 label
                    || iCtl == 499                  // Host Name
                    || iCtl == 500                  // Filter Title
                    || iCtl == 501                  // Filter Case
                    || iCtl == 3987                 // Custom control above
                    || iCtl == 3988                 // Custom control above
                    || iCtl == 3989                 // Custom control above
                    // || iCtl == 3990                 // NFG for trackbar!!
#if 0
                    || iCtl == IDC_EMF
                    //|| iCtl == IDC_ICON1
#endif
#ifdef ALLOW_DRAGOUT
                    || iCtl == IDC_DRAGOUT
#endif
#ifdef USE_BITMAP
                    || iCtl == IDX_OK || iCtl == IDX_CANCEL
#ifdef FM_LIFTCOV
                    || iCtl == IDC_BUTTON1 || iCtl == IDC_BUTTON2   // Zoom buttons
#endif
#endif
                    )
                {   // We want these controls to be transparent...
                    // Note that these controls are NOT updated dynamically
                    // (which doesn't work right, since the backgrounds don't
                    // get erased properly).
                    SetBkMode(hdc, TRANSPARENT);
#if defined(xxxxx)
                    SetTextColor(hdc, RGB(  0,  0,  0));    //black text
#elif defined(FM_LIFTCOV)
                    SetTextColor(hdc, RGB(255,255,  0));    //yellow text
#else
                    SetTextColor(hdc, RGB(255,255,255));    //white text
#endif
                    //should really use SetDlgMsgResult...
                    return (INT_PTR) (HBRUSH) GetStockObject(HOLLOW_BRUSH); // create statically?
                }
                else if (iCtl >= IDC_CTLBASE && iCtl < IDC_CTLBASE+N_CTLS) {
                    // user control...
                    int n = iCtl - IDC_CTLBASE;
                    if (!gParams->ctl[n].inuse) return FALSE; //not processed
                    if (gParams->ctl[n].bkColor == -1) {
                        //transparent background
                        if (gParams->ctl[n].class == CC_STANDARD ||
							gParams->ctl[n].class == CC_SLIDER ||
                            gParams->ctl[n].class == CC_SCROLLBAR) {
                            //scrollbars are NFG with transparent body,
                            //so bail out to the Windows default handling
                            //in this case...
                            return FALSE;   //not processed
                        }
                        //else, go ahead and set transparent background mode
                        SetBkMode(hdc, TRANSPARENT);
                    }
                    else {
                        //solid background color
                        SetBkMode(hdc, OPAQUE);
                        SetBkColor(hdc, gParams->ctl[n].bkColor);
                    }
                    SetTextColor(hdc, gParams->ctl[n].textColor);
                    return (INT_PTR)gParams->ctl[n].hBrush;         //background brush
                }
                else if (iCtl >= IDC_BUDDY2 && iCtl < IDC_BUDDY2+N_CTLS) {
                    // static text buddy for a scrollbar...
                    int n = iCtl - IDC_BUDDY2;
                    if (!gParams->ctl[n].inuse) return FALSE; //not processed
                    //for now, we always want static text buddies to have
                    //a transparent background
                    SetBkMode(hdc, TRANSPARENT);
                    SetTextColor(hdc, gParams->ctl[n].textColor);
                    return (INT_PTR)GetStockObject(HOLLOW_BRUSH);         //background brush
                }
                else if (iCtl >= 800 && iCtl <= 802) { //Advanced mode group boxes
                    SetBkColor(hdc, RGB(128,128,128));  //gray background
                    SetTextColor(hdc, RGB(255,255,255));    //white text
                    //should really use SetDlgMsgResult...
                    return (INT_PTR) (HBRUSH) GetStockObject(GRAY_BRUSH);
                }
                else if (iCtl != IDC_PROXY             // Could this be the proxy update bug???
                     //&&  iCtl != IDC_SLIDER1           // NFG otherwise
                     //&&  iCtl != IDC_ICON1
                     &&  iCtl != 3990                  // NFG otherwise for trackbar
                     )
                {
                    SetBkColor(hdc, RGB(255,255,255));  //white background
                    SetTextColor(hdc, RGB(0,0,100));    //dark blue text
                    //should really use SetDlgMsgResult...
                    return (INT_PTR) (HBRUSH) GetStockObject(WHITE_BRUSH); // create statically?
                }
                else return FALSE;  // Not processed.
            }
            break;

        case WM_TIMER:
            switch (wParam) {
            case 1:
                // Cancel the proxy-delay timer (make it a one-shot).
                KillTimer(hDlg, 1);

                if (gResult == noErr) {
                    //If we are dragging, the proxy will have already been
                    //updated.
                    if (!fDragging) {
                        // Do Filtering operation
                        DoFilterRect (globals);
                        // Invalidate Proxy Item (and repaint background if scale factor
                        // increased; i.e., if proxy image may have shrunk)...
                        {
                            static long originalScaleFactor = -1; //force initial update
                            UpdateProxy(hDlg, scaleFactor > originalScaleFactor);
                            originalScaleFactor = scaleFactor;
                        }
                    }//if not dragging
                }//if noErr
                break;

            case 2:
 
                if (GetForegroundWindow() == MyHDlg){ //Only make it work if FM dialog is active

                    //check for cursor in a MOUSEOVER or MOUSEMOVE control
                    RECT rcCtl;
                    int i;
                    int hitTest;
                    POINT pt;
                    BOOL RetVal;
                    int enableState;

                    //Added by Harald Heim, Jun 6,2002
                    //int NoInvisible;

                    GetCursorPos(&pt);
                    for (i = 0; i < N_CTLS; ++i) {

                         //Check if control is enabled
                        if (gParams->ctl[i].tabControl>=0){
                            enableState = (int)fm_ctlEnabledAs (i); //Control is on a tab sheet
                        } else {
                           enableState = gParams->ctl[i].enabled;
                        }

						//optimized a bit by Alex Hunter 9/17/2013
                        if (i==CTL_PREVIEW ||
							(gParams->ctl[i].inuse && (gParams->ctl[i].properties & (CTP_MOUSEMOVE|CTP_MOUSEOVER)) && !(enableState != 3 && MouseOverWhenInvisible == 0)) ) {
                        //Changed by Harald Heim, Marc 24,2003
                        //if (i==CTL_PREVIEW || (gParams->ctl[i].inuse && ((gParams->ctl[i].properties & CTP_MOUSEMOVE) || (gParams->ctl[i].properties & CTP_MOUSEOVER)) && !(enableState != 3 && MouseOverWhenInvisible == 0)) ) {
                        //if (i==CTL_PREVIEW || (gParams->ctl[i].inuse && ((gParams->ctl[i].properties & CTP_MOUSEMOVE) || (gParams->ctl[i].properties & CTP_MOUSEOVER)) && !(gParams->ctl[i].enabled != 3 && MouseOverWhenInvisible == 0)) ) {
                        //if (gParams->ctl[i].inuse && gParams->ctl[i].properties & CTP_MOUSEOVER) {
                            

                            if (i==CTL_PREVIEW) {
                                RetVal = GetWindowRect(GetDlgItem(hDlg, IDC_PROXY), &rcCtl); 
                            } else {
                                RetVal = GetWindowRect(gParams->ctl[i].hCtl, &rcCtl);
                            }

                            //Mouse event only for tab buttons
                            if (gParams->ctl[i].class == CC_TAB) {
                                RECT rect;
                                TabCtrl_GetItemRect (gParams->ctl[i].hCtl, TabCtrl_GetItemCount(gParams->ctl[i].hCtl)-1, &rect);
                                rcCtl.right = rcCtl.left + rect.right-1;
                                rcCtl.bottom = rcCtl.top + rect.bottom-1;

                                //hitTest = pt.x >= rcCtl.left && pt.x < rcCtl.right &&
                                //          pt.y >= rcCtl.top  && pt.y < rcCtl.bottom;
                           
                            
                            } /*
                            
                              //Include labels and edit box
                              else if (gParams->ctl[i].class == CC_STANDARD || gParams->ctl[i].class == CC_SLIDER){
                                RECT rect;
                                
                                hitTest = pt.x >= rcCtl.left && pt.x < rcCtl.right &&
                                          pt.y >= rcCtl.top  && pt.y < rcCtl.bottom;

                                if (!hitTest && GetWindowRect(gParams->ctl[i].hBuddy1, &rect)) {
                                    ScreenToClient (hDlg, (LPPOINT)&rect);
                                    ScreenToClient (hDlg, (LPPOINT)&(rect.right));
                                    Info ("%d - %d",rect.left,pt.x);
                                    hitTest = pt.x >= rect.left && pt.x < rect.right &&
                                              pt.y >= rect.top  && pt.y < rect.bottom;
                                }
                                if (!hitTest && GetWindowRect(gParams->ctl[i].hBuddy2, &rect)) {
                                    ScreenToClient (hDlg, (LPPOINT)&rect);
                                    ScreenToClient (hDlg, (LPPOINT)&(rect.right));
                                    hitTest = pt.x >= rect.left && pt.x < rect.right &&
                                              pt.y >= rect.top  && pt.y < rect.bottom;
                                }                                

                            } else { //Normal controls */
                                
                                hitTest = pt.x >= rcCtl.left && pt.x < rcCtl.right &&
                                          pt.y >= rcCtl.top  && pt.y < rcCtl.bottom;
                            //}



                            //this control wants mouseover or mousemove notifications
                            if (RetVal) {
                            //if (GetWindowRect(gParams->ctl[i].hCtl, &rcCtl)) {

                                if (hitTest) {
                                    //it's a hit!
                                    
                                     if (gParams->ctl[i].properties & CTP_MOUSEMOVE){
                                        //this control wants mousemove notifications

                                        PerformControlAction(
                                            i,                      //control index
                                            FME_MOUSEMOVE,          //FM event code
                                            0,                      //previous value
                                            wMsg,                   //message type
                                            0,                      //notification code
                                            FALSE,                  //fActionable
                                            TRUE,0);                  //fProcessed

                                    } 
                                    
                                     if (gParams->ctl[i].properties & CTP_MOUSEOVER){ //i==CTL_PREVIEW || 
                                    
                                        if (i != iCurrentMouseoverCtl) {
                                            //MOUSEOUT to iCurrentMouseoverCtl if >= 0
                                            if (iCurrentMouseoverCtl >= 0) {
                                                PerformControlAction(
                                                    iCurrentMouseoverCtl,   //control index
                                                    FME_MOUSEOUT,           //FM event code
                                                    0,                      //previous value
                                                    wMsg,                   //message type
                                                    0,                      //notification code
                                                    FALSE,                  //fActionable
                                                    TRUE,0);                  //fProcessed
                                            }
                                            
                                           
                                            //MOUSEOVER to i
                                            iCurrentMouseoverCtl = i;
                                            PerformControlAction(
                                                iCurrentMouseoverCtl,   //control index
                                                FME_MOUSEOVER,          //FM event code
                                                0,                      //previous value
                                                wMsg,                   //message type
                                                0,                      //notification code
                                                FALSE,                  //fActionable
                                                TRUE,0);                  //fProcessed

                                        }
                                    }
                                    break;
                                }
                            }
                        }
                    } //for i
                    if (i >= N_CTLS) {  //dodgy!!!
                        //MOUSEOUT to iCurrentMouseoverCtl if >= 0
                        if (iCurrentMouseoverCtl >= 0) {
                            PerformControlAction(
                                iCurrentMouseoverCtl,   //control index
                                FME_MOUSEOUT,           //FM event code
                                0,                      //previous value
                                wMsg,                   //message type
                                0,                      //notification code
                                FALSE,                  //fActionable
                                TRUE,0);                  //fProcessed
                            iCurrentMouseoverCtl = -1;
                        }
                    }
                }
                break;

            //Added by Harald Heim, Dec 13,2002
            case 3:
            case 4:
            case 5:
            case 6:
            case 7:
            case 8:
            case 9:
            case 10:
            case 11:
            case 12:

                //Trigger Timer Event
                PerformControlAction(
                    (int)(wParam-3),		//(pseudo-)control index -- for setting FM global var 'n', 
											// but 'ctlClass' will be garbage!
                    FME_TIMER,				//FM event code
                    0,                      //previous value
                    wMsg,                   //message type
                    0,                      //notification code
                    FALSE,                  //fActionable
                    TRUE,0);                  //fProcessed

                break;
             
            case 1000: //MouseWheel Timer
                KillTimer(hDlg, 1000);
                DoAction(gParams->ctl[iCtl2].action);
                break;

            case 1001: //Preview Update Timer -> max. 50 times per second
                KillTimer(hDlg, 1001);
                DoAction(gParams->ctl[iCtl2].action);
                break;

            //case 1002: //WM_PAINT -> Redraw Proxy
            //    KillTimer(hDlg, 1002);              
            //    break;

            default:
                return FALSE;   //not handled

            }//switch
            return TRUE;    //handled

        case WM_DESTROY:
            // Kill any leftover timers
            KillTimer(hDlg, 1);
            KillTimer(hDlg, 2);

            KillTimer(hDlg, 1000);
            KillTimer(hDlg, 1001);

            {   // Restore old dialog bitmap.
                if (hBmpOld && hdcMem) {
                    SelectObject(hdcMem, hBmpOld);
                }
                // Need to DeleteObject() the bitmap that was loaded.
                if (hBmpDialog) {
                    if (!DeleteObject(hBmpDialog)) {
#ifndef STANDALONE
                        ErrorOk("DeleteObject #5 failed: %8.8x", hBmpDialog);
#endif
                    }
                }
                // Delete the memory DC.
                if (hdcMem) {
                    DeleteDC(hdcMem);
                }
            }

#ifndef STANDALONE
#define CHECK(what) \
            /* Info(#what ", control %d", n); */ \
            if (!(what)) {\
                ErrorOk("CHECK failed at line %d", __LINE__);\
            }
#else
#define CHECK(what) (what)
#endif

            //Delete any acquired resources in user controls...
            {//scope
                int n;
                for (n = 0;  n < N_CTLS;  n++)
                {
                    if (gParams->ctl[n].inuse) {
                        //what about hCtl, hBuddy1, hBuddy2????
                        if (gParams->ctl[n].hBrush) {
                            CHECK(DeleteObject(gParams->ctl[n].hBrush));
                            gParams->ctl[n].hBrush = NULL;
                        }
#if 0
                        //NO! Fonts are shared among controls,
                        //and NT (but not Win9x) generates a check
                        //and vanishes Photoshop!
                        if (gParams->ctl[n].hFont) {
                            CHECK(DeleteObject(gParams->ctl[n].hFont));
                            gParams->ctl[n].hFont = NULL;
                        }
#endif
                        if (gParams->ctl[n].class == CC_IMAGE) {
                            //select old bitmaps back into DC's, then
                            //delete all image bitmaps and DC's.

                            if (gParams->ctl[n].im.hbmOld) {
                                SelectObject(gParams->ctl[n].im.hdcMem, gParams->ctl[n].im.hbmOld);
                                gParams->ctl[n].im.hbmOld = 0;
                            }
                            if (gParams->ctl[n].im.hbm) {
                                CHECK(DeleteObject(gParams->ctl[n].im.hbm));
                                gParams->ctl[n].im.hbm = 0;
                            }
                            if (gParams->ctl[n].im.hdcMem) {
                                CHECK(DeleteDC(gParams->ctl[n].im.hdcMem));
                                gParams->ctl[n].im.hdcMem = 0;
                            }

                            if (gParams->ctl[n].im.hbmOldAnd) {
                                SelectObject(gParams->ctl[n].im.hdcAnd, gParams->ctl[n].im.hbmOldAnd);
                                gParams->ctl[n].im.hbmOldAnd = 0;
                            }
                            if (gParams->ctl[n].im.hbmAnd) {
                                CHECK(DeleteObject(gParams->ctl[n].im.hbmAnd));
                                gParams->ctl[n].im.hbmAnd = 0;
                            }
                            if (gParams->ctl[n].im.hdcAnd) {
                                CHECK(DeleteDC(gParams->ctl[n].im.hdcAnd));
                                gParams->ctl[n].im.hdcAnd = 0;
                            }
                        }//if CC_IMAGE
                        else if (gParams->ctl[n].class == CC_METAFILE) {
                            //set no image, and retrieve previous image
                            HENHMETAFILE hPrevImage;
                            hPrevImage = (HENHMETAFILE)SendMessage(gParams->ctl[n].hCtl,
                                                                   STM_SETIMAGE,
                                                                   IMAGE_ENHMETAFILE,
                                                                   (LPARAM)NULL);
                            //delete previous image, if any
                            if (hPrevImage) {
                                CHECK(DeleteEnhMetaFile(hPrevImage));
                            }
                        }//if CC_METAFILE
                        else if (gParams->ctl[n].class == CC_BITMAP) {
                            //set no image, and retrieve previous bitmap
                            HBITMAP hPrevImage;
                            hPrevImage = (HBITMAP)SendMessage(gParams->ctl[n].hCtl,
                                                              STM_SETIMAGE,
                                                              IMAGE_BITMAP,
                                                              (LPARAM)NULL);
                            //delete previous bitmap, if any
                            if (hPrevImage) {
                                CHECK(DeleteObject(hPrevImage));
                            }
                        }//if CC_BITMAP
                    }//if inuse
                } //for n
            }//scope

#ifndef STANDALONE
            // Remove the subclass from this dialog. 
            SetWindowLongPtr(hDlg, GWLP_WNDPROC, (LONG_PTR)wpOrigFilterProc);
#if 0
            //// NO!!!! Window may change with each compile!!!!
            // Remove the subclass from the Edit control. 
            SetWindowLongPtr(GetDlgItem(hDlg, IDX_EDIT), GWLP_WNDPROC, (LONG_PTR)wpOrigEditCtlProc);
#endif
#endif //STANDALONE

            // Delete the proxy RGB buffer, if any.
            if (pRGBbuffer) {
                free(pRGBbuffer);
                pRGBbuffer = NULL;
            }

            return FALSE;   // Not processed.


        case WM_NCHITTEST:
#if defined(USE_BITMAP) || defined(FML)
            {
                POINT   pt;
                UINT    nHitTest = HTCLIENT;

                pt.x = LOWORD(lParam);
                pt.y = HIWORD(lParam);
                
                
                //Test if Size grip
                if (sizeGrip){
                    nHitTest = (UINT)DefWindowProc(hDlg, wMsg, wParam, lParam);
                    if( nHitTest == HTCLIENT ) 
                    { 
                            RECT rc; 
                            GetWindowRect(hDlg, &rc ); 
                            rc.left = rc.right - GetSystemMetrics(SM_CXHSCROLL); 
                            rc.top = rc.bottom - GetSystemMetrics(SM_CYVSCROLL); 
                        
                            if( PtInRect(&rc, pt) ){ 
                                nHitTest = HTBOTTOMRIGHT;
                                return SetDlgMsgResult(hDlg, wMsg, nHitTest);
                            }
                    }
                }


                ScreenToClient(hDlg, &pt);
                if (pt.x >= gProxyRect.left && pt.x < gProxyRect.right &&
                    pt.y >= gProxyRect.top  && pt.y < gProxyRect.bottom)
                {   // It's in our proxy preview window...
                    nHitTest = HTCLIENT;
                }
#ifdef FML
                else if (gParams->gDialogDragMode == 2)  //Drag=None
                {
                    // Call default window proc to perform hit test...
                    nHitTest = (UINT)DefWindowProc(hDlg, wMsg, wParam, lParam);
                    // Change title bar hit to client hit to prevent dragging...
                    if (nHitTest == HTCAPTION)
                        nHitTest = HTCLIENT;
                }
#endif

                else if (gParams->gDialogDragMode > 2)  //Drag=Any Control
                {
                    nHitTest = (UINT)DefWindowProc(hDlg, wMsg, wParam, lParam);

                    if (nHitTest == HTCLIENT){
                        
                        RECT rcCtl;

                        GetWindowRect(gParams->ctl[gParams->gDialogDragMode].hCtl, &rcCtl);
                        ScreenToClient (hDlg, (LPPOINT)&rcCtl);
                        ScreenToClient (hDlg, (LPPOINT)&(rcCtl.right));

                        if (pt.x >= rcCtl.left && pt.x < rcCtl.right &&
                            pt.y >= rcCtl.top  && pt.y < rcCtl.bottom) 
                        {
                        
                                nHitTest = HTCAPTION;

                        }
                    }
                    
                }  
                else
                {
                    

#ifndef NO_TITLE_BAR
                    // Call default window proc to perform hit test...
                    nHitTest = (UINT)DefWindowProc(hDlg, wMsg, wParam, lParam);
                    // Change client hit to title bar hit to force dragging...
                    if (nHitTest == HTCLIENT){

#endif
#ifdef FML
                      if (gParams->gDialogDragMode == 1)   //Drag=Background
#endif

                          if (GetSystemMetrics(SM_SWAPBUTTON)){ //Left Handed Mouse setting
                                if (GetAsyncKeyState(VK_RBUTTON) & 0x8000)
                                    nHitTest = HTCAPTION;
                          } else {//Standard Right Handed Mouse setting
                                if (GetAsyncKeyState(VK_LBUTTON) & 0x8000)
                                    nHitTest = HTCAPTION;
                          }
                    }

                }
                return SetDlgMsgResult(hDlg, wMsg, nHitTest);
            }
#endif //USE_BITMAP||FML
            return FALSE;   // Not processed -- pass to DefWindowProc().

        case WM_SIZE:
            
            if (fPaint) //make sure that it isn't executed before the dialog appears, otherwise Memory Error
            {
                int fwSizeType = (int)wParam;	// resizing flag 
                int nWidth = LOWORD(lParam);	// width of client area 
                int nHeight = HIWORD(lParam);	// height of client area 

				int dialogWidth = (int)fm_PixelsToHDBUs(nWidth);
				int dialogHeight = (int)fm_PixelsToVDBUs(nHeight);
				
				fScaling = true;
                if(!scaleLock) fm_scaleCtls(dialogWidth, dialogHeight);
				fScaling = false;

                fSizing=true;

                PerformControlAction(
                            fwSizeType,             //control index
                            FME_SIZE,               //FM event code
                            0,                      //previous value
                            wMsg,                   //message type
                            0,                      //notification code
                            FALSE,                  //fActionable
                            TRUE,0);                  //fProcessed


                //When maximizing or restoring window there is no WM_EXITSIZEMOVE event, so we have to do this now
                //Because there is no way to distinguish a restore, we have to use fMaximized
                if (fwSizeType == SIZE_MAXIMIZED || fMaximized){
                    
                    if (fwSizeType == SIZE_MAXIMIZED) 
                           fMaximized = TRUE;
                    else 
                           fMaximized = FALSE;

                    //Make sure that the preview is displayed correctly
                    GetProxyItemRect (hDlg, &gProxyRect); 
                    SetupFilterRecordForProxy (globals, &scaleFactor, scaleFactor, &imageOffset);

                    PerformControlAction(
                                0,                      //control index
                                FME_EXITSIZE,           //FM event code
                                0,                      //previous value
                                WM_EXITSIZEMOVE,        //message type
                                0,                      //notification code
                                FALSE,                  //fActionable
                                TRUE,0);                  //fProcessed

					
					RedrawWindow(MyHDlg,NULL,NULL,RDW_ERASE|RDW_INVALIDATE|RDW_ERASENOW|RDW_UPDATENOW);
					DoFilterRect (MyGlobals);
					UpdateProxy(MyHDlg, FALSE);

                }

            
            }    
            return FALSE;

/*        case WM_ENTERSIZEMOVE:

            if (fPaint) //make sure that it isn't executed before the dialog appears, otherwise Memory Error
            {
                PerformControlAction(
                            0,                      //control index
                            FME_ENTERSIZE,               //FM event code
                            0,                      //previous value
                            wMsg,                   //message type
                            0,                      //notification code
                            FALSE,                  //fActionable
                            TRUE,0);                  //fProcessed
            }
            return FALSE;
*/

        case WM_GETMINMAXINFO:
            {
                LPMINMAXINFO lpmmi = (LPMINMAXINFO) lParam; // address of structure 
 
                if (MaxMinInfo.ptMinTrackSize.x >0 && MaxMinInfo.ptMinTrackSize.y >0)
                    lpmmi->ptMinTrackSize = (&MaxMinInfo)->ptMinTrackSize;
                if (MaxMinInfo.ptMaxTrackSize.x >0 && MaxMinInfo.ptMaxTrackSize.y >0)
                    lpmmi->ptMaxTrackSize = (&MaxMinInfo)->ptMaxTrackSize;

            }
            return FALSE;

                    
        case WM_EXITSIZEMOVE:
            
            
            if (fSizing){
                
                fSizing = FALSE;

                //Make sure that the preview is displayed correctly
                GetProxyItemRect (hDlg, &gProxyRect); 
                SetupFilterRecordForProxy (globals, &scaleFactor, scaleFactor, &imageOffset);

                PerformControlAction(
                            0,                      //control index
                            FME_EXITSIZE,           //FM event code
                            0,                      //previous value
                            wMsg,                   //message type
                            0,                      //notification code
                            FALSE,                  //fActionable
                            TRUE,0);                  //fProcessed
                 
                //To avoid artifacts by size grip
                if (sizeGrip){
                    RedrawWindow(
                        hDlg,   // handle of entire dialog window
                        NULL,   // address of structure with update rectangle, NULL=>entire client area
                        NULL,   // handle of update region
                        RDW_ERASE|RDW_INVALIDATE|RDW_ERASENOW|RDW_UPDATENOW     // array of redraw flags
                    );
                }
            
				
				RedrawWindow(MyHDlg,NULL,NULL,RDW_ERASE|RDW_INVALIDATE|RDW_ERASENOW|RDW_UPDATENOW);
				DoFilterRect (MyGlobals);
				UpdateProxy(MyHDlg, FALSE);

            }



            return FALSE;

        case WM_PAINT:
            fPaint = true;
            ShowOutputBuffer(globals, hDlg, fDragging);     // Draw Proxy Item
            
            //Draw Resize handle
            if (sizeGrip){
                RECT rc;
                static RECT rcOld;
                HDC  hDC = GetDC(hDlg);

                GetClientRect(hDlg, &rc ); 
                rc.left = rc.right - GetSystemMetrics(SM_CXHSCROLL); 
                rc.top = rc.bottom - GetSystemMetrics(SM_CYVSCROLL);

                if (rcOld.right - rcOld.left > 0 && !(EqualRect(&rcOld, &rc)==1) ) {
                    //Delete previous grip
                    HBRUSH hBr = GetSysColorBrush(COLOR_BTNFACE);
                    FillRect (hDC, &rcOld, hBr);
                    DeleteObject (hBr);
                }

                DrawFrameControl(hDC, &rc, DFC_SCROLL, DFCS_SCROLLSIZEGRIP );

                ReleaseDC(hDlg, hDC);

                rcOld = rc;
            }

            
            return TRUE;    // SDK samples had FALSE, but that causes update
                            // problems when full-dragging a window across
                            // the proxy... does the "fix" introduce any new
                            // problems? 
            break;

        case WM_MOUSEMOVE:
            {
                int x = (SHORT) LOWORD(lParam);
                int y = (SHORT) HIWORD(lParam);

#if 1
                {
                    // relay this mouse message to the tooltip control...
                    MSG msg; 

                    msg.lParam = lParam; 
                    msg.wParam = wParam; 
                    msg.message = wMsg; 
                    msg.hwnd = hDlg;
#if 1   //not needed???
                    msg.time = 0;
                    msg.pt.x = 40;
                    msg.pt.y = 40;
#endif
                    SendMessage(ghTT, TTM_RELAYEVENT, 0, (LPARAM) &msg); 
                }
#endif
                if (fDragging)
                {
                    POINT tempOffset;

                    SetCursor(LoadCursor(NULL, IDC_SIZEALL));
                    ptNew.x = x;
                    ptNew.y = y;
#ifndef NOOPTIMIZE_QUICK_TOGGLE
                    //Only update if image can be dragged in preview
                    if ( gFmcp->x_start != 0 || gFmcp->y_start != 0 || gFmcp->x_end != gFmcp->X || gFmcp->y_end != gFmcp->Y ){
                        fMouseMoved |= (ptNew.x - ptPrev.x) | (ptNew.y - ptPrev.y);
                    }
#endif
                    //Only update if image can be dragged in preview
                    if (gFmcp->x_start != 0 || gFmcp->y_start != 0 || gFmcp->x_end != gFmcp->X || gFmcp->y_end != gFmcp->Y ){
                        imageOffset.x -= (ptNew.x - ptPrev.x) * scaleFactor * gFmcp->scrollFactor;// / gFmcp->enlargeFactor;
                        imageOffset.y -= (ptNew.y - ptPrev.y) * scaleFactor * gFmcp->scrollFactor;// / gFmcp->enlargeFactor;
                    }
					ptPrev = ptNew;
					GetProxyItemRect (hDlg, &gProxyRect);
                    tempOffset = imageOffset; // don't modify imageOffset while dragging
                    

                    if ( ((gParams->flags & AFH_ZERO_OUTPUT_DRAG) == AFH_ZERO_OUTPUT_DRAG) && gFmcp->x_start == 0 && gFmcp->y_start == 0 && gFmcp->x_end == gFmcp->X && gFmcp->y_end == gFmcp->Y ){
                        
                         //Temporarily deactivate Zero drag to avoid redraw problem
                         gParams->flags &= ~AFH_ZERO_OUTPUT_DRAG;

                         SetupFilterRecordForProxy (globals, &scaleFactor, scaleFactor, &tempOffset);
                               
                         //Activate Zero drag again
                         gParams->flags |= AFH_ZERO_OUTPUT_DRAG;

                    }else {

                        SetupFilterRecordForProxy (globals, &scaleFactor, scaleFactor, &tempOffset);
                    }



                    if (gResult == noErr) {
                        // Run the filter only if want to drag the filtered output
                        // rather than the raw input...
                        if (gParams->flags & AFH_DRAG_FILTERED_OUTPUT)
                        {
                            DoFilterRect (globals);
                        }
                        UpdateProxy(hDlg, FALSE);
                    }//noErr
                    SetCursor(LoadCursor(NULL, IDC_SIZEALL)); //again, in case WAIT cursor appeared
                    //MOUSEOUT to iCurrentMouseoverCtl if >= 0
                    if (iCurrentMouseoverCtl >= 0) {
                        PerformControlAction(
                            iCurrentMouseoverCtl,   //control index
                            FME_MOUSEOUT,           //FM event code
                            0,                      //previous value
                            wMsg,                   //message type
                            0,                      //notification code
                            FALSE,                  //fActionable
                            TRUE,0);                  //fProcessed
                        iCurrentMouseoverCtl = -1;
                    }
                    return TRUE; // ??
                }

                //Proxy Box
                else if (x >= gProxyRect.left && x < gProxyRect.right &&
                         y >= gProxyRect.top && y < gProxyRect.bottom)
                {
                    
                    //if (gParams->ctl[CTL_PREVIEW].inuse){

                        SetCursor(hcHand);
                
                        /*   // Why did Harald comment this out????

                             //Found the reason for commenting this out! This code sends a mouseout 
                             //to the preview itself which causes some problems in my FFP code. 
                             //iCurrentMouseoverCtl must be set to the preview control 
                             //at this point already.

                        //MOUSEOUT to iCurrentMouseoverCtl if >= 0
                        if (iCurrentMouseoverCtl >= 0) {
                            PerformControlAction(
                                iCurrentMouseoverCtl,   //control index
                                FME_MOUSEOUT,           //FM event code
                                0,                      //previous value
                                wMsg,                   //message type
                                0,                      //notification code
                                FALSE,                  //fActionable
                                TRUE,0);                  //fProcessed
                            iCurrentMouseoverCtl = -1;
                        }
                        */
                    
                        //Added by Harald Heim, Dec 13, 2002
                        //Send MouseMove for Proxy
                    
                        if (gParams->ctl[CTL_PREVIEW].properties & CTP_MOUSEMOVE){
                            PerformControlAction(
                                CTL_PREVIEW,   //control index
                                FME_MOUSEMOVE,           //FM event code
                                0,                      //previous value
                                wMsg,                   //message type
                                0,                      //notification code
                                FALSE,                  //fActionable
                                TRUE,0);                  //fProcessed
                        }
                    //}


#if 1
                    return TRUE; //??
#else
                    return FALSE; //so underlying controls can see mouse and pass to tooltips???
#endif
                }
                else
                {
                    //check for hit in a MOUSEOVER control
                    RECT rcCtl;
                    int i;
                    int D = 0;
                    int hitTest;
                    int enableState;

                    for (i = 0; i < N_CTLS; ++i) {
                        
                        //Check if control is enabled
                        if (gParams->ctl[i].tabControl>=0){
                            enableState = (int)fm_ctlEnabledAs (i); //Control is on a tab sheet
                        } else {
                           enableState = gParams->ctl[i].enabled;
                        }


                        if (gParams->ctl[i].inuse && ((gParams->ctl[i].properties & CTP_MOUSEMOVE) || (gParams->ctl[i].properties & CTP_MOUSEOVER)) && enableState == 3) {
                        //if (gParams->ctl[i].inuse && (gParams->ctl[i].properties & CTP_MOUSEOVER) && !(gParams->ctl[i].enabled != 3 && MouseOverWhenInvisible == 0)) {
                        //if (gParams->ctl[i].inuse && gParams->ctl[i].enabled == 3 && gParams->ctl[i].properties & CTP_MOUSEOVER) {
                        //if (gParams->ctl[i].inuse && gParams->ctl[i].properties & CTP_MOUSEOVER) {

                            //this control wants mouseover notifications
                            if (GetWindowRect(gParams->ctl[i].hCtl, &rcCtl)) {
                                ScreenToClient (hDlg, (LPPOINT)&rcCtl);
                                ScreenToClient (hDlg, (LPPOINT)&(rcCtl.right));


                                //Mouse event only for tab buttons
                                if (gParams->ctl[i].class == CC_TAB) {
                                    RECT rect;

                                    TabCtrl_GetItemRect (gParams->ctl[i].hCtl, TabCtrl_GetItemCount(gParams->ctl[i].hCtl)-1, &rect);
                                    rcCtl.right = rcCtl.left + rect.right-1;
                                    rcCtl.bottom = rcCtl.top + rect.bottom-1;

                                }
                                    
                                            
                                hitTest = x >= rcCtl.left-D && x < rcCtl.right+D &&
                                          y >= rcCtl.top-D  && y < rcCtl.bottom+D;
                                

                                if (hitTest) {
                                    //it's a hit!
                                    
                                    if (gParams->ctl[i].properties & CTP_MOUSEMOVE){
                                        //this control wants mousemove notifications

                                        PerformControlAction(
                                            i,                      //control index
                                            FME_MOUSEMOVE,          //FM event code
                                            0,                      //previous value
                                            wMsg,                   //message type
                                            0,                      //notification code
                                            FALSE,                  //fActionable
                                            TRUE,0);                  //fProcessed

                                    } 
                                    
                                    if (gParams->ctl[i].properties & CTP_MOUSEOVER){
                                    
                                        if (i != iCurrentMouseoverCtl) {
                                            //MOUSEOUT to iCurrentMouseoverCtl if >= 0
                                            if (iCurrentMouseoverCtl >= 0) {
                                                PerformControlAction(
                                                    iCurrentMouseoverCtl,   //control index
                                                    FME_MOUSEOUT,           //FM event code
                                                    0,                      //previous value
                                                    wMsg,                   //message type
                                                    0,                      //notification code
                                                    FALSE,                  //fActionable
                                                    TRUE,0);                  //fProcessed
                                            }
                                                
                                            //MOUSEOVER to i
                                            iCurrentMouseoverCtl = i;
                                            PerformControlAction(
                                                iCurrentMouseoverCtl,   //control index
                                                FME_MOUSEOVER,          //FM event code
                                                0,                      //previous value
                                                wMsg,                   //message type
                                                0,                      //notification code
                                                FALSE,                  //fActionable
                                                TRUE,0);                  //fProcessed
                                        
                                        }
                                        
                                    }
                                    
                                    break;

                                }
                            }
                            
                        }
                    } //for i

                    if (i >= N_CTLS) {  //dodgy!!!
                        //MOUSEOUT to iCurrentMouseoverCtl if >= 0
                        if (iCurrentMouseoverCtl >= 0) {
                            PerformControlAction(
                                iCurrentMouseoverCtl,   //control index
                                FME_MOUSEOUT,           //FM event code
                                0,                      //previous value
                                wMsg,                   //message type
                                0,                      //notification code
                                FALSE,                  //fActionable
                                TRUE,0);                  //fProcessed
                            iCurrentMouseoverCtl = -1;
                        }
                    }
                }
            }
            return FALSE;
            break;

			case WM_NOTIFY:
				
				iCmd = COMMANDCMD(wParam, lParam);

				hCtrl = ((LPNMHDR)lParam)->hwndFrom;
                iCtl = GetWindowLong(hCtrl, GWL_ID);
                if (iCtl < IDC_CTLBASE || iCtl >= IDC_CTLBASE + N_CTLS)
                    return FALSE;
                iCtl -= IDC_CTLBASE;
                
                switch (((LPNMHDR)lParam)->code)
				{
					case TCN_SELCHANGE:
					{ 
						int val = TabCtrl_GetCurSel(hCtrl);
						int i, oldvalue;
						
						if (val!=-1) {
							
                            oldvalue = gParams->ctl[iCtl].val;

                            fm_setCtlVal (iCtl, val);

							for (i = 0; i < N_CTLS; ++i) {
								if (gParams->ctl[i].inuse && gParams->ctl[i].tabControl == iCtl &&  gParams->ctl[i].tabSheet != -1){
									enableCtl(i);	//MWVDL 2008-07-18
								}	
							}

							return PerformControlAction(
								iCtl,  //control index
								FME_CLICKED,            //FM event code
								oldvalue, //previous value
								wMsg,               //message type
								iCmd,               //notification code
								FALSE,
								TRUE,0);
						}

					} 
				}
            
				return TRUE;

			
			case WM_LBUTTONDOWN:
            {
                int x = (SHORT) LOWORD(lParam);
                int y = (SHORT) HIWORD(lParam);

#if 1
                {
                    // relay this mouse message to the tooltip control...
                    MSG msg; 

                    msg.lParam = lParam; 
                    msg.wParam = wParam; 
                    msg.message = wMsg; 
                    msg.hwnd = hDlg; 
                    SendMessage(ghTT, TTM_RELAYEVENT, 0, (LPARAM) &msg); 
                }
#endif
                if (x >= gProxyRect.left && x < gProxyRect.right &&
                    y >= gProxyRect.top && y < gProxyRect.bottom)
                {
                    
                    //if (gParams->ctl[CTL_PREVIEW].inuse){
                        //Merged by Harald Heim, May 20, 2003
                        if (ClickDrag == 0)
                        {
                    
                            SetCapture(hDlg);
                            SetCursor(LoadCursor(NULL, IDC_SIZEALL));
                            ptPrev.x = x;
                            ptPrev.y = y;
    #ifndef NOOPTIMIZE_QUICK_TOGGLE
                            fMouseMoved = 0;
    #endif
                            fDragging = TRUE;

                            if (gParams->ctl[CTL_PREVIEW].properties & CTP_PREVIEWDRAG){
                                PerformControlAction(
                                    CTL_PREVIEW,   //control index
                                    FME_PREVIEWDRAG,        //FM event code
                                    0,                      //previous value
                                    wMsg,                   //message type
                                    0,                      //notification code
                                    FALSE,                  //fActionable
                                    TRUE,0);                  //fProcessed
                            }
                        
                            gParams->flags |= AFH_DRAGGING_PROXY;
                            UpdateProxy(hDlg, FALSE);  // To display raw input.
                            return TRUE; //??
                        }
                        else
                        {
                            //send FME_LEFTCLICKED_DOWN event for Proxy
                            SetCursor(hcHand);
                            PerformControlAction(
                                CTL_PREVIEW,   //control index
                                FME_LEFTCLICKED_DOWN,   //FM event code
                                0,                      //previous value
                                wMsg,                   //message type
                                0,                      //notification code
                                FALSE,                  //fActionable
                                TRUE,0);                  //fProcessed
                        }
                    //}

                }
                
                /*else
                {
                    //check for hit
                    RECT rcCtl;
                    int i;
                    int D = 0;
                    for (i = 0; i < N_CTLS; ++i) {
 
                        if (gParams->ctl[i].inuse  && gParams->ctl[i].enabled == 3) { //&& gParams->ctl[i].class == CC_OWNERDRAW
                        //if (gParams->ctl[i].inuse && ((gParams->ctl[i].properties & CTP_MOUSEMOVE) || (gParams->ctl[i].properties & CTP_MOUSEOVER)) && gParams->ctl[i].enabled == 3) {
                        
                            //this control needs notifications
                            if (GetWindowRect(gParams->ctl[i].hCtl, &rcCtl)) {
                                ScreenToClient (hDlg, (LPPOINT)&rcCtl);
                                ScreenToClient (hDlg, (LPPOINT)&(rcCtl.right));
                                if (x >= rcCtl.left-D && x < rcCtl.right+D &&
                                    y >= rcCtl.top-D  && y < rcCtl.bottom+D) {
                                    //it's a hit!
                                    
                                    //if (gParams->ctl[i].properties & CTP_MOUSEMOVE){
                                        //this control wants mousemove notifications

                                        PerformControlAction(
                                            i,                      //control index
                                            FME_LEFTCLICKED_DOWN,   //FM event code
                                            0,                      //previous value
                                            wMsg,                   //message type
                                            0,                      //notification code
                                            FALSE,                  //fActionable
                                            TRUE,0);                  //fProcessed

                                    //} 
                                    
                                    break;

                                }
                            }
                            
                        }
                    } //for i
                }*/


            }

            return FALSE;
            break;


        case WM_LBUTTONUP:
#if 1
            {
                // relay this mouse message to the tooltip control...
                MSG msg; 

                msg.lParam = lParam; 
                msg.wParam = wParam; 
                msg.message = wMsg; 
                msg.hwnd = hDlg; 
                SendMessage(ghTT, TTM_RELAYEVENT, 0, (LPARAM) &msg); 
            }
#endif

            {
                int x = (SHORT) LOWORD(lParam);
                int y = (SHORT) HIWORD(lParam);

                //if (gParams->ctl[CTL_PREVIEW].inuse){
                    if (fDragging)
                    {
            
                        if (x >= gProxyRect.left && x < gProxyRect.right &&
                            y >= gProxyRect.top && y < gProxyRect.bottom)
                        {
                            SetCursor(hcHand);
                        }
                        else
                        {
                            //SetCursor(LoadCursor(NULL, IDC_ARROW));  //???
                        }
                        ReleaseCapture();
						fDragging = FALSE;
                        gParams->flags &= ~AFH_DRAGGING_PROXY;
                        ptNew.x = x;
                        ptNew.y = y;
        #ifndef NOOPTIMIZE_QUICK_TOGGLE
                        //Only update if image can be dragged in preview
                        if (gFmcp->x_start != 0 || gFmcp->y_start != 0 || gFmcp->x_end != gFmcp->X || gFmcp->y_end != gFmcp->Y ){
                            fMouseMoved |= (ptNew.x - ptPrev.x) | (ptNew.y - ptPrev.y);
                        }
        #endif
                        //Only update if image can be dragged in preview
                        if (gFmcp->x_start != 0 || gFmcp->y_start != 0 || gFmcp->x_end != gFmcp->X || gFmcp->y_end != gFmcp->Y ){
                            imageOffset.x -= (ptNew.x - ptPrev.x) * scaleFactor * gFmcp->scrollFactor;// / gFmcp->enlargeFactor;
                            imageOffset.y -= (ptNew.y - ptPrev.y) * scaleFactor * gFmcp->scrollFactor;// / gFmcp->enlargeFactor;
                        }
                        GetProxyItemRect (hDlg, &gProxyRect); 


                        if ( ((gParams->flags & AFH_ZERO_OUTPUT_DRAG) == AFH_ZERO_OUTPUT_DRAG) && gFmcp->x_start == 0 && gFmcp->y_start == 0 && gFmcp->x_end == gFmcp->X && gFmcp->y_end == gFmcp->Y ){
                        
                             //Temporarily deactivate Zero drag to avoid redraw problem
                             gParams->flags &= ~AFH_ZERO_OUTPUT_DRAG;

                             SetupFilterRecordForProxy (globals, &scaleFactor, scaleFactor, &imageOffset);
                               
                             //Activate Zero drag again
                             gParams->flags |= AFH_ZERO_OUTPUT_DRAG;

                        }else {
                            SetupFilterRecordForProxy (globals, &scaleFactor, scaleFactor, &imageOffset);
                        }


                        if (gResult == noErr) {
                            // Do filter and update proxy...
        #ifndef NOOPTIMIZE_QUICK_TOGGLE
                            //if mouse didn't move since LBUTTONDOWN, no need to run filter...
                            //gives a quick before/after preview toggle

                            if (fMouseMoved) {

                                VRect scaledFilterRect;

                                //Setup x_start etc.
								gFmcp->filterRect32 = GetFilterRect(); //needed??
								scaledFilterRect = gFmcp->filterRect32;
                                scaleRect32(&scaledFilterRect, 1, gFmcp->scaleFactor);
								gFmcp->inRect32 = GetInRect();	//needeed??
                                gFmcp->x_start = gFmcp->inRect32.left - scaledFilterRect.left;
                                gFmcp->y_start = gFmcp->inRect32.top - scaledFilterRect.top;
                                gFmcp->x_end = gFmcp->x_start + gFmcp->columns;
                                gFmcp->y_end = gFmcp->y_start + gFmcp->rows;
                            
                                if (gParams->ctl[CTL_PREVIEW].properties & CTP_PREVIEWDRAG){
                                    PerformControlAction(
                                        CTL_PREVIEW,   //control index
                                        FME_PREVIEWDRAG,        //FM event code
                                        1,                      //previous value
                                        wMsg,                   //message type
                                        0,                      //notification code
                                        FALSE,                  //fActionable
                                        TRUE,0);                  //fProcessed
                                }
    
                                DoFilterRect (globals);
                                fMouseMoved = 0;
                            }
        #else
                            DoFilterRect (globals);
        #endif
                        
                            UpdateProxy(hDlg, FALSE);
                        }//noErr
                        return TRUE;
                    }

                    //Info ("WM_LBUTTONUP\nfDragging: %d\nx: %d, y: %d\n%d,%d - %d,%d",fDragging,x,y,gProxyRect.left,gProxyRect.top,gProxyRect.right,gProxyRect.bottom);

                    //Added by Harald Heim, Dec 13, 2002
                    if (x >= gProxyRect.left && x < gProxyRect.right &&
                        y >= gProxyRect.top  && y < gProxyRect.bottom)
					{
                            //Info ("Should send FME_LEFTCLICKED_UP now");
                            
                            //send FME_LEFTCLICKED_UP event for Proxy
                            //SetCursor(hcHand);
                            PerformControlAction(
                                CTL_PREVIEW,   //control index
                                FME_LEFTCLICKED_UP,           //FM event code
                                0,                      //previous value
                                wMsg,                   //message type
                                0,                      //notification code
                                FALSE,                  //fActionable
                                TRUE,                   //fProcessed
                                1);                     //override busy logic

                    }

                //}//gParams->ctl[CTL_PREVIEW].inuse
			}

            return FALSE;
            break;


        case WM_CONTEXTMENU:

            {
                int x = (SHORT) LOWORD(lParam);
                int y = (SHORT) HIWORD(lParam);

                RECT rcCtl;
                int i;

                for (i = 0; i < N_CTLS; ++i) {
                    
                    if (gParams->ctl[i].inuse && (gParams->ctl[i].properties & CTP_CONTEXTMENU) && !(gParams->ctl[i].enabled != 3) ){
                    
                        if (GetWindowRect(gParams->ctl[i].hCtl, &rcCtl)) {
                            
                            if (x >= rcCtl.left && x < rcCtl.right &&
                                y >= rcCtl.top  && y < rcCtl.bottom) {
                                //it's a hit!
                                
                                PerformControlAction(
                                    i,   //control index
                                    FME_CONTEXTMENU,        //FM event code
                                    0,                      //previous value
                                    wMsg,                   //message type
                                    0,                      //notification code
                                    FALSE,                  //fActionable
                                    TRUE,0);                  //fProcessed

                                break;
                            }
                        }
                    }
                } //for i

                return false;
            }
            

#if 1
        case WM_RBUTTONDOWN:
            
            //Added by Harald Heim, Aug 9, 2002
            {
                // relay this mouse message to the tooltip control...
                MSG msg; 

                msg.lParam = lParam; 
                msg.wParam = wParam; 
                msg.message = wMsg; 
                msg.hwnd = hDlg; 
                SendMessage(ghTT, TTM_RELAYEVENT, 0, (LPARAM) &msg); 
            }

            {
                int x = (SHORT) LOWORD(lParam);
                int y = (SHORT) HIWORD(lParam);


                if (x >= gProxyRect.left && x < gProxyRect.right &&
                    y >= gProxyRect.top && y < gProxyRect.bottom) // && gParams->ctl[CTL_PREVIEW].inuse
                {

                    //Merged by Harald Heim, May 20, 2003
                    if (ClickDrag == 1)
                    {
                        SetCapture(hDlg);
                        SetCursor(LoadCursor(NULL, IDC_SIZEALL));
                        ptPrev.x = x;
                        ptPrev.y = y;
#ifndef NOOPTIMIZE_QUICK_TOGGLE
                        fMouseMoved = 0;
#endif
                        fDragging = TRUE;

                        if (gParams->ctl[CTL_PREVIEW].properties & CTP_PREVIEWDRAG){
                            PerformControlAction(
                                CTL_PREVIEW,   //control index
                                FME_PREVIEWDRAG,        //FM event code
                                0,                      //previous value
                                wMsg,                   //message type
                                0,                      //notification code
                                FALSE,                  //fActionable
                                TRUE,0);                  //fProcessed
                        }
                        
                        gParams->flags |= AFH_DRAGGING_PROXY;
                        UpdateProxy(hDlg, FALSE);  // To display raw input.
                        return TRUE; //??
                    } 
                    else 
                    {
                        //send FME_RIGHTCLICKED_DOWN event for Proxy
                        //SetCursor(hcHand);
                        PerformControlAction(
                            CTL_PREVIEW,   //control index
                            FME_RIGHTCLICKED_DOWN,           //FM event code
                            0,                      //previous value
                            wMsg,                   //message type
                            0,                      //notification code
                            FALSE,                  //fActionable
                            TRUE,0);                  //fProcessed
                      
                    }

                } else {

                    //Right mouse button on dialog background
                    PerformControlAction(
                        CTL_BACKGROUND,   //control index
                        FME_CONTEXTMENU,           //FM event code
                        0,                      //previous value
                        wMsg,                   //message type
                        0,                      //notification code
                        FALSE,                  //fActionable
                        TRUE,0);                  //fProcessed

                }   
            }

            return FALSE;
            break;


        case WM_RBUTTONUP:

            //Added by Harald Heim, Dec 13, 2002
            {
                // relay this mouse message to the tooltip control...
                MSG msg; 

                msg.lParam = lParam; 
                msg.wParam = wParam; 
                msg.message = wMsg; 
                msg.hwnd = hDlg; 
                SendMessage(ghTT, TTM_RELAYEVENT, 0, (LPARAM) &msg); 
            }

            {
                int x = (SHORT) LOWORD(lParam);
                int y = (SHORT) HIWORD(lParam);

                //if (gParams->ctl[CTL_PREVIEW].inuse){
                    if (fDragging)
                    {
                        if (x >= gProxyRect.left && x < gProxyRect.right &&
                            y >= gProxyRect.top && y < gProxyRect.bottom)
                        {
                            SetCursor(hcHand);
                        }
                        else
                        {
                            //SetCursor(LoadCursor(NULL, IDC_ARROW));  //???
                        }
                        ReleaseCapture();
                        fDragging = FALSE;
                        gParams->flags &= ~AFH_DRAGGING_PROXY;
                        ptNew.x = x;
                        ptNew.y = y;
        #ifndef NOOPTIMIZE_QUICK_TOGGLE
                        //Only update if image can be dragged in preview
                        if (gFmcp->x_start != 0 || gFmcp->y_start != 0 || gFmcp->x_end != gFmcp->X || gFmcp->y_end != gFmcp->Y){
                            fMouseMoved |= (ptNew.x - ptPrev.x) | (ptNew.y - ptPrev.y);
                        }
        #endif
                        //Only update if image can be dragged in preview
                        if (gFmcp->x_start != 0 || gFmcp->y_start != 0 || gFmcp->x_end != gFmcp->X || gFmcp->y_end != gFmcp->Y ){
                            imageOffset.x -= (ptNew.x - ptPrev.x) * scaleFactor * gFmcp->scrollFactor;// / gFmcp->enlargeFactor;
                            imageOffset.y -= (ptNew.y - ptPrev.y) * scaleFactor * gFmcp->scrollFactor;// / gFmcp->enlargeFactor;
                        }
                        GetProxyItemRect (hDlg, &gProxyRect);


                        if ( ((gParams->flags & AFH_ZERO_OUTPUT_DRAG) == AFH_ZERO_OUTPUT_DRAG) && gFmcp->x_start == 0 && gFmcp->y_start == 0 && gFmcp->x_end == gFmcp->X && gFmcp->y_end == gFmcp->Y ){
                        
                             //Temporarily deactivate Zero drag to avoid redraw problem
                             gParams->flags &= ~AFH_ZERO_OUTPUT_DRAG;

                             SetupFilterRecordForProxy (globals, &scaleFactor, scaleFactor, &imageOffset);
                               
                             //Activate Zero drag again
                             gParams->flags |= AFH_ZERO_OUTPUT_DRAG;

                        }else {

                            SetupFilterRecordForProxy (globals, &scaleFactor, scaleFactor, &imageOffset);
                        }


                        if (gResult == noErr) {
                            // Do filter and update proxy...
        #ifndef NOOPTIMIZE_QUICK_TOGGLE
                            //if mouse didn't move since LBUTTONDOWN, no need to run filter...
                            //gives a quick before/after preview toggle

                            if (fMouseMoved) {
                            
                                VRect scaledFilterRect;

                                //Setup x_start etc.
								gFmcp->filterRect32 = GetFilterRect();	//needed??
								scaledFilterRect = gFmcp->filterRect32;
                                scaleRect32(&scaledFilterRect, 1, gFmcp->scaleFactor);
								gFmcp->inRect32 = GetInRect();
                                gFmcp->x_start = gFmcp->inRect32.left - scaledFilterRect.left;
                                gFmcp->y_start = gFmcp->inRect32.top - scaledFilterRect.top;
                                gFmcp->x_end = gFmcp->x_start + gFmcp->columns;
                                gFmcp->y_end = gFmcp->y_start + gFmcp->rows;
                            
                                
                                if (gParams->ctl[CTL_PREVIEW].properties & CTP_PREVIEWDRAG){
                                    PerformControlAction(
                                        CTL_PREVIEW,   //control index
                                        FME_PREVIEWDRAG,        //FM event code
                                        1,                      //previous value
                                        wMsg,                   //message type
                                        0,                      //notification code
                                        FALSE,                  //fActionable
                                        TRUE,0);                  //fProcessed
                                }
                          
    
                                DoFilterRect (globals);
                                fMouseMoved = 0;
                            }
        #else
                            DoFilterRect (globals);
        #endif
                            UpdateProxy(hDlg, FALSE);
                        }//noErr
                        return TRUE;
                    }



                    if (x >= gProxyRect.left && x < gProxyRect.right &&
                        y >= gProxyRect.top && y < gProxyRect.bottom)
                    {
                        //send FME_RIGHTCLICKED_UP event for Proxy
                        //SetCursor(hcHand);
                        PerformControlAction(
                            CTL_PREVIEW,   //control index
                            FME_RIGHTCLICKED_UP,           //FM event code
                            0,                      //previous value
                            wMsg,                   //message type
                            0,                      //notification code
                            FALSE,                  //fActionable
                            TRUE,1);                  //fProcessed

                    }

                //}//gParams->ctl[CTL_PREVIEW].inuse
            }

            return FALSE;
            break;
            //End of Added by Harald Heim, Dec 13, 2002



        case WM_MBUTTONDOWN:
        case WM_MBUTTONUP:
            {
                // relay this mouse message to the tooltip control...
                MSG msg; 

                msg.lParam = lParam; 
                msg.wParam = wParam; 
                msg.message = wMsg; 
                msg.hwnd = hDlg; 
                SendMessage(ghTT, TTM_RELAYEVENT, 0, (LPARAM) &msg); 
            }
            return FALSE;
            break;
#endif

#if 0
        // Doesn't work -- Dlg eats all key codes???
        case WM_KEYDOWN:          
			

            

			if (wParam == VK_F12) {
                // Toggle the editor.
                SendMessage(hDlg, WM_COMMAND, IDC_EDIT, 0);
                return 0;   //processed
            }
            
            

            return 1;   //not processed
            break;
#endif

        case 0x20A://WM_MOUSEWHEEL:
            {
                int zDelta = (short) HIWORD(wParam);
                int oldvalue;

                int i;
                HWND focusHWND = GetFocus();

                for (i = 0; i < N_CTLS; ++i) {

                    if (gParams->ctl[i].inuse && !(gParams->ctl[i].enabled != 3) ){
                    
                        if (gParams->ctl[i].hCtl == focusHWND) {

                            if (gParams->ctl[i].class == CC_STANDARD ||
                                gParams->ctl[i].class == CC_SCROLLBAR ||
								gParams->ctl[i].class == CC_SLIDER){
                    
                                oldvalue = gParams->ctl[i].val;
        
                                if (zDelta<0)
                                    gParams->ctl[i].val += gParams->ctl[i].pageSize;
                                else
                                    gParams->ctl[i].val -= gParams->ctl[i].pageSize;

    
                                fm_setCtlVal(i, gParams->ctl[i].val);

                                PerformControlAction(
                                                i,                   //control index
                                                FME_CHANGED,			//FM event code
                                                oldvalue,               //previous value
                                                wMsg,                   //message type
                                                0,                      //notification code
                                                FALSE,                  //fActionable
                                                TRUE,0);                  //fProcessed  
                            
                                //Perform the default action for this control.
                            
                                KillTimer(hDlg, 1000); //Cancel any previous trigger.
                                iCtl2 = i;

                                if (GetKeyState(VK_SHIFT) < 0){
                                    if (!SetTimer(hDlg, 1001, 20, NULL)) //Update max. every 20 ms
                                        DoAction(gParams->ctl[i].action);
                                } else if (!SetTimer(hDlg, 1000, 500, NULL)) {
                                    DoAction(gParams->ctl[i].action);
                                }
                                
                            }
                        } 
                   }

                }
            }
            return 0;   //processed


        case WM_HSCROLL:
        case WM_VSCROLL:

            //handles both scrollbar and trackbar messages
            //(Note that TB_PAGEDOWN == SB_PAGEDOWN, etc.)

            { //Start scope
                int oldvalue; 

                hCtrl = (HWND) lParam;
                iCtl = GetWindowLong(hCtrl, GWL_ID);
                if (iCtl < IDC_CTLBASE || iCtl >= IDC_CTLBASE + N_CTLS)
                    return FALSE;
                iCtl -= IDC_CTLBASE;
                idd = COMMANDID (wParam);
    #if 0
                if (gParams->ctl[iCtl].class == CC_TRACKBAR) {
                    Info("TB notification %d", idd);
                }
    #endif

                oldvalue = gParams->ctl[iCtl].val;

                switch (idd)
                {
                case SB_PAGEDOWN:
                    gParams->ctl[iCtl].val += gParams->ctl[iCtl].pageSize;
                    break;
                case SB_LINEDOWN:
                    gParams->ctl[iCtl].val += gParams->ctl[iCtl].lineSize;
                    break;
                case SB_PAGEUP:
                    gParams->ctl[iCtl].val -= gParams->ctl[iCtl].pageSize;
                    break;
                case SB_LINEUP:
                    gParams->ctl[iCtl].val -= gParams->ctl[iCtl].lineSize;
                    break;
                case SB_TOP:
                    gParams->ctl[iCtl].val = gParams->ctl[iCtl].minval;
                    break;
                case SB_BOTTOM:
                    gParams->ctl[iCtl].val = gParams->ctl[iCtl].maxval;
                    break;
                case SB_THUMBTRACK:
                case SB_THUMBPOSITION:
                    
					if (gParams->ctl[iCtl].class == CC_STANDARD || gParams->ctl[iCtl].class == CC_SCROLLBAR){
						
						SCROLLINFO si;
						ZeroMemory(&si, sizeof(si));
						si.cbSize = sizeof(si);
						si.fMask = SIF_TRACKPOS;

						if (GetScrollInfo(hCtrl, SB_CTL, &si) )
							gParams->ctl[iCtl].val = si.nTrackPos; //get 32bit scroll bar value
						else
							gParams->ctl[iCtl].val = (short)HIWORD(wParam); //get 16bit scroll bar value

					} else { //Trackbar and Slider

						gParams->ctl[iCtl].val = (int)SendMessage(hCtrl, TBM_GETPOS, 0, 0); //get 32bit track bar value
					}
					
                    if (gParams->ctl[iCtl].class == CC_STANDARD ||
						gParams->ctl[iCtl].class == CC_SLIDER ||
                        gParams->ctl[iCtl].class == CC_SCROLLBAR)
                    {
                        if (gParams->ctl[iCtl].minval > gParams->ctl[iCtl].maxval) {
                            //range is inverted...
                            gParams->ctl[iCtl].val = gParams->ctl[iCtl].minval +
                                                     gParams->ctl[iCtl].maxval -
                                                     gParams->ctl[iCtl].val;
                        }
                    } //if
                    break;
                default:
                    return FALSE;
                } /*switch*/
                
                if (oldvalue != gParams->ctl[iCtl].val)
                {
                    
					if ( (idd== SB_THUMBPOSITION || idd == SB_THUMBTRACK) && gParams->ctl[iCtl].thumbSize > 0){

						int range = abs(gParams->ctl[iCtl].maxval - gParams->ctl[iCtl].minval);
						gParams->ctl[iCtl].val = gParams->ctl[iCtl].val * range / (range - gParams->ctl[iCtl].thumbSize-1);

					}					
					
					if ( (idd== SB_THUMBPOSITION || idd == SB_THUMBTRACK) && gParams->ctl[iCtl].gamma != 100){
						
                        int temp = gParams->ctl[iCtl].gamma;
                        double r = (gParams->ctl[iCtl].gamma/100.0); //Inverse Gamma needs to be applied as a correction

                        if (gParams->ctl[iCtl].minval >= 0){
                            double absRange = (double) abs (gParams->ctl[iCtl].maxval - gParams->ctl[iCtl].minval) - 1.0;
                            gParams->ctl[iCtl].val = gParams->ctl[iCtl].minval + (int)(pow( (gParams->ctl[iCtl].val - gParams->ctl[iCtl].minval) / absRange, r) * absRange + 0.5);
                        } else {
                            if (gParams->ctl[iCtl].val >= 0){
                                double absRange = (double) gParams->ctl[iCtl].maxval - 1.0;
                                gParams->ctl[iCtl].val = (int)(pow( gParams->ctl[iCtl].val / absRange, r) * absRange + 0.5);
                    
                            } else {
                                double absRange = (double) abs (gParams->ctl[iCtl].minval) - 1.0;
                                gParams->ctl[iCtl].val = - (int)(pow( abs(gParams->ctl[iCtl].val) / absRange, r) * absRange + 0.5);
                            }
                        }

                        //if (oldvalue != gParams->ctl[iCtl].val){ //Avoid jumping when only clicking on scrollbar knob
                            //Avoid double gamma calculation
                            //gParams->ctl[iCtl].gamma = 100;
                            fm_setCtlVal(iCtl, gParams->ctl[iCtl].val);
                            //gParams->ctl[iCtl].gamma = temp;
                        //}
					
                    } else {
                        //let fm_setCtlVal() do all the work...
                        fm_setCtlVal(iCtl, gParams->ctl[iCtl].val);
                    }
                  
                
                    //Added by Harald Heim, May 22, 2003
                    PerformControlAction(
                                iCtl,                   //control index
                                FME_CHANGED,       		//FM event code
                                oldvalue,               //previous value
                                wMsg,                   //message type
                                0,                      //notification code
                                FALSE,                  //fActionable
                                TRUE,0);                  //fProcessed                                        
                    

                } //if CC_STANDARD || CC_SCROLLBAR

            }//scope


            // Perform default action for THUMBTRACK events only if TRACK property is set
            // (or sense is reversed by SHIFT key).
            if (idd != SB_THUMBTRACK ||
                (((gParams->ctl[iCtl].properties & CTP_TRACK) != 0) ^
                 (GetKeyState(VK_SHIFT) < 0))
               )
            {
                KillTimer(hDlg, 1001); //Cancel any previous trigger.
                iCtl2 = iCtl;
                //Perform the default action for this control.
                if (!SetTimer(hDlg, 1001, 20, NULL)){ //Update max. every 20 ms
                    DoAction(gParams->ctl[iCtl].action);
                }
            }
            return TRUE;

        case WM_COMMAND:
            idd = COMMANDID (wParam);              // WIN32 Change
            iCmd = COMMANDCMD(wParam, lParam);


            if (idd >= IDC_BUDDY2 && idd < IDC_BUDDY2 + N_CTLS) {
                
                FME fmeCode;

                if (iCmd == BN_CLICKED) {
                    fmeCode = FME_CLICKED;
                } else if (iCmd == STN_DBLCLK) {
                    fmeCode = FME_DBLCLK;
                    //Set Default Value
                    if (gParams->ctl[idd - IDC_BUDDY2].defval != 0xffffffff){	//or -1 for X64???
                        fm_setCtlVal(idd - IDC_BUDDY2, gParams->ctl[idd - IDC_BUDDY2].defval);
                    }
                }

                if (iCmd == BN_CLICKED || iCmd == STN_DBLCLK) {
                    return PerformControlAction(
                        idd - IDC_BUDDY2,   //control index
                        fmeCode,            //FM event code
                        0,                  //previous value
                        wMsg,               //message type
                        0,                  //notification code
                        FALSE,
                        TRUE,0);
                }
            }


            if (idd >= IDC_BUDDY1 && idd < IDC_BUDDY1 + N_CTLS) {
                static BOOL changed = FALSE;
                // NUMEDIT buddy control for a scrollbar
                if (iCmd == EN_CHANGE) {
                    changed = TRUE;
                    
                    if (EditBoxUpdate != 0){

                    //if (gParams->ctl[idd - IDC_BUDDY1].properties & CTP_MOUSEOVER) {
                    
                        BOOL fTranslated;
                        int iVal;
#if 1
                        //Changed by Harald Heim, May 1, 2003
                        if (gParams->ctl[idd - IDC_BUDDY1].divisor == 1) { // Set Integer Value

                            iVal = (int)GetDlgItemInt(hDlg, idd, &fTranslated, TRUE /*signed*/);

                        } else 
#endif
                        { // Set Double Value
                        
                            char  szInput[32];
                            double dVal;
                            char *pInvalid;

                            fTranslated = GetDlgItemText (hDlg, idd, szInput, 32);
                            // WRONG!  strtod() is locale-dependent!!!
                            dVal = strtod( szInput, &pInvalid);     //should check errno, *pInvalid==0 ???
#if 0
                            iVal = (int) (gParams->ctl[idd - IDC_BUDDY1].divisor * dVal);   //should this round???
#else
                            iVal = (int) (gParams->ctl[idd - IDC_BUDDY1].divisor * dVal + (dVal<0.0?-0.5:0.5));
#endif
                        }

                        gParams->ctl[idd - IDC_BUDDY1].val = iVal;

                        /*PerformControlAction(
                                idd - IDC_BUDDY1,   //control index
                                FME_EDITCHANGED,        //FM event code
                                iVal,                      //previous value
                                wMsg,                   //message type
                                0,                      //notification code
                                FALSE,                  //fActionable
                                TRUE,0);                  //fProcessed
                        */

                        if (!fTranslated) {
                            //bad syntax ... force an illegal value so setCtlVal()      //or just set to 0???
                            //will fix it up...
                            iVal = -9845783;  //black magic
                            fTranslated = TRUE;
                        }
                        if (fTranslated) {
                        
                            //Added by Harald Heim, May 22, 2003
                            int oldvalue = gParams->ctl[idd - IDC_BUDDY1].val;

                            //set value of associated control
                            //(calling setCtlVal() may be dangerous overkill,
                            //possibly even resulting in an infinite loop???)
                            fm_setCtlVal(idd - IDC_BUDDY1, iVal);


                        
                            //Added by Harald Heim, May 22, 2003
                            PerformControlAction(
                                idd - IDC_BUDDY1,       //control index
                                FME_CHANGED,//FME_VALUECHANGED,       //FM event code
                                oldvalue,               //previous value
                                wMsg,                   //message type
                                0,                      //notification code
                                FALSE,                  //fActionable
                                TRUE,0);                  //fProcessed                                        


                            if (!fDragging) {
                                if (EditBoxUpdate==1){
                                    //Update Preview immediately
                                    DoAction(gParams->ctl[idd - IDC_BUDDY1].action);

                                } else {
                                    //Delayed update via timer
                                    iCtl2 = idd - IDC_BUDDY1;//Store control number for timer
                                    if (!SetTimer(hDlg, 1001, EditBoxUpdate, NULL)) //Update max. every 20 ms
                                        DoAction(gParams->ctl[idd - IDC_BUDDY1].action);
                                }
                            }

                        }
                        changed = FALSE;

                    }

                }
                else if (iCmd == EN_SETFOCUS) {
                    changed = FALSE;
                    
                    if (gParams->ctl[idd - IDC_BUDDY1].properties & CTP_MOUSEOVER) {
                        
                        PerformControlAction(
                                idd - IDC_BUDDY1,   //control index
                                FME_SETFOCUS,			//FM event code
                                0,                      //previous value
                                wMsg,                   //message type
                                0,                      //notification code
                                FALSE,                  //fActionable
                                TRUE,0);                  //fProcessed
                    }

                }
                else if (iCmd == EN_KILLFOCUS ||
                         iCmd == EN_MAXTEXT) {
                    BOOL fTranslated;
                    int iVal;

                    if (iCmd == EN_MAXTEXT) {
                        //presumably because user hit RETURN
                        changed = TRUE;
                    }

                    if (iCmd == EN_KILLFOCUS && (gParams->ctl[idd - IDC_BUDDY1].properties & CTP_MOUSEOVER)) {
                        
                        PerformControlAction(
                                idd - IDC_BUDDY1,   //control index
                                FME_KILLFOCUS,          //FM event code
                                0,                      //previous value
                                wMsg,                   //message type
                                0,                      //notification code
                                FALSE,                  //fActionable
                                TRUE,0);                  //fProcessed
                    }

                    //update only if the NUMEDIT control actually changed!
                    //(e.g., keep track of EN_CHANGE notices).
                    //Otherwise, switching to another app when a NUMEDIT
                    //has the focus will cause an unwarranted filter run...
                    if (!changed) return TRUE; //???


#if 1
                    //Changed by Harald Heim, May 1, 2003
                    if (gParams->ctl[idd - IDC_BUDDY1].divisor == 1) { // Set Integer Value

                        iVal = (int)GetDlgItemInt(hDlg, idd, &fTranslated, TRUE /*signed*/);

                    } else 
#endif
                    { // Set Double Value
                        
                        char  szInput[32];
                        double dVal;
                        char *pInvalid;

                        fTranslated = GetDlgItemText (hDlg, idd, szInput, 32);
                        // WRONG!  strtod() is locale-dependent!!!
                        dVal = strtod( szInput, &pInvalid);     //should check errno, *pInvalid==0 ???
#if 0
                        iVal = (int) (gParams->ctl[idd - IDC_BUDDY1].divisor * dVal);   //should this round???
#else
                        iVal = (int) (gParams->ctl[idd - IDC_BUDDY1].divisor * dVal + (dVal<0.0?-0.5:0.5));
#endif
                    }

                    //Info("iVal = %d\nfTranslated = %d", iVal, fTranslated);

                    if (!fTranslated) {
                        //bad syntax ... force an illegal value so setCtlVal()      //or just set to 0???
                        //will fix it up...
                        iVal = -9845783;  //black magic
                        fTranslated = TRUE;
                    }
                    if (fTranslated) {
                        
                        //Added by Harald Heim, May 22, 2003
                        int oldvalue = gParams->ctl[idd - IDC_BUDDY1].val;

                        //set value of associated control
                        //(calling setCtlVal() may be dangerous overkill,
                        //possibly even resulting in an infinite loop???)
                        fm_setCtlVal(idd - IDC_BUDDY1, iVal);
                        
                        //Added by Harald Heim, May 22, 2003
                        PerformControlAction(
                            idd - IDC_BUDDY1,       //control index
                            FME_CHANGED,       		//FM event code
                            oldvalue,               //previous value
                            wMsg,                   //message type
                            0,                      //notification code
                            FALSE,                  //fActionable
                            TRUE,0);                  //fProcessed                                        


                        if (!fDragging) {
                            // Update Filter parameter block display
                            ShowFilterParameterBlock(globals, hDlg);
                            if (gResult == noErr) {
                                // Do Filtering operation
                                DoFilterRect (globals);
                                // Invalidate Proxy Item
                                UpdateProxy(hDlg, FALSE);
                            }
                        }
                    }
                    changed = FALSE;
#if 0
                    return SetDlgMsgResult(hDlg, wMsg, TRUE);
#else
                    return TRUE; //so EN_MAXTEXT won't ding the bell -- ineffective???
#endif
                }
#if 0
                else {
                    //unexpected notification????
                    static int counter = 0;
                    static int wmsg_save = 0;
                    static int wparam_save = 0;
                    static int lparam_save = 0;
                    static int idd_save = 0;
                    static int cmd_save = 0;
                    wmsg_save = wMsg;
                    wparam_save = wParam;
                    lparam_save = lParam;
                    idd_save = idd;
                    cmd_save = iCmd;
                    counter++;
                    return TRUE;
                }
#endif
                return TRUE;   //???
            }

            if (idd >= IDC_CTLBASE && idd < IDC_CTLBASE + N_CTLS) {
                int iVal = 0;

                //Added by Harald Heim
                int oldvalue=0;

                BOOL fProcessed = TRUE;
                BOOL fActionable = FALSE;
                FME fmeCode = FME_UNKNOWN;
                // It's a user control...
                switch (gParams->ctl[idd - IDC_CTLBASE].class) {
                case CC_CHECKBOX:
                case CC_RADIOBUTTON:
                case CC_PUSHBUTTON:
                case CC_OWNERDRAW:
                case CC_STATICTEXT:
                case CC_FRAME:
                case CC_RECT:
                case CC_IMAGE:
                case CC_BITMAP:
                case CC_ICON:
                case CC_METAFILE:
                case CC_GROUPBOX:
                    // NOTE that BN_CLICKED == STN_CLICKED!!!
                    if (iCmd == BN_CLICKED) {
                        //From a button or static class control...
                        static int reentry_depth = 0;
                        static int reentry_count = 0;

                        //check for reentry...
                        if (reentry_depth++ > 0) {
                            //reentered!!! Log the fact, and exit quickly...
                            reentry_count++;
                            reentry_depth--;
#if 0
                            return FALSE;   //do default processing anyway??
#else
                            return TRUE;    //no default processing either
#endif
                        }

                        fmeCode = FME_CLICKED;

                        switch (gParams->ctl[idd - IDC_CTLBASE].class) {
                        case CC_CHECKBOX:
                        case CC_RADIOBUTTON:
                            iVal = (int)SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0);
                            switch (gParams->ctl[idd - IDC_CTLBASE].style & BS_TYPEMASK) {
                            case BS_CHECKBOX:
                            case BS_AUTOCHECKBOX:
                            case BS_3STATE:
                            case BS_AUTO3STATE:
                            case BS_AUTORADIOBUTTON:
                                
                                //Added by Harald Heim, May 22, 2003
                                oldvalue = gParams->ctl[idd - IDC_CTLBASE].val;
                        
                                fm_setCtlVal(idd - IDC_CTLBASE, iVal);
                                break;
                            case BS_RADIOBUTTON:
                                //setCtlVal causes reentry????
                                //Need to update pre_ctl[] and ctl[].val????
                                break;
                            }//switch button subtype
                            break;
                        case CC_PUSHBUTTON:
                        case CC_OWNERDRAW:
                        case CC_STATICTEXT:
                        case CC_FRAME:
                        case CC_RECT:
                        case CC_IMAGE:
                        case CC_BITMAP:
                        case CC_ICON:
                        case CC_METAFILE:
                        case CC_GROUPBOX:
                            //these classes don't maintain their own value...
                            iVal = 1;
                            fm_setCtlVal(idd - IDC_CTLBASE, iVal);
                            break;
                        default:
                            break;
                        }//switch class

                         reentry_depth--; //release reentry lock

                    }//BN_CLICKED
                    else {
                        switch (gParams->ctl[idd - IDC_CTLBASE].class) {
                        case CC_CHECKBOX:
                        case CC_RADIOBUTTON:
                        case CC_PUSHBUTTON:
                        case CC_GROUPBOX:
                        case CC_OWNERDRAW:
                            //Button-style controls...
                            if (iCmd == BN_DBLCLK) {
                                fProcessed = FALSE;     //not processed
                                fActionable = FALSE;    //for now
                                fmeCode = FME_DBLCLK;
                            }
                            /*else if (iCmd == BN_PUSHED) {
                                fProcessed = FALSE;     //not processed
                                fActionable = FALSE;    //for now
                                fmeCode = FME_LEFTCLICKED_DOWN;
                            }
							else if (iCmd == BN_UNPUSHED) {
                                fProcessed = FALSE;     //not processed
                                fActionable = FALSE;    //for now
                                fmeCode = FME_LEFTCLICKED_UP;
                            }*/
							else {
                                fProcessed = FALSE;     //not processed
                                fActionable = FALSE;    //for now
                                fmeCode = FME_UNKNOWN;
                            }
                            break;
                        case CC_STATICTEXT:
                        case CC_FRAME:
                        case CC_RECT:
                        case CC_IMAGE:
                        case CC_BITMAP:
                        case CC_ICON:
                        case CC_METAFILE:
                            //Static-style controls...
                            if (iCmd == STN_DBLCLK) {
                                fProcessed = FALSE;     //not processed
                                fActionable = FALSE;    //for now
                                fmeCode = FME_DBLCLK;
                            }
                            else {
                                fProcessed = FALSE;     //not processed
                                fActionable = FALSE;    //for now
                                fmeCode = FME_UNKNOWN;
                            }
                            break;
                        }//switch
                    }//if !BN_CLICKED
                    break;

                case CC_LISTBOX:
                    if (iCmd == LBN_SELCHANGE) {
                        
                        //Added by Harald Heim, May 22, 2003
                        oldvalue = gParams->ctl[idd - IDC_CTLBASE].val;
                        
                        iVal = (int)SendMessage(COMMANDWND(wParam), LB_GETCURSEL, 0, 0);
                        fm_setCtlVal(idd - IDC_CTLBASE, iVal);
                        fmeCode = FME_CLICKED;  //for now
                    }
                    else if (iCmd == LBN_DBLCLK) {
                        //Requires LBS_NOTIFY
                        fProcessed = FALSE;     //not processed
                        fActionable = FALSE;    //for now
                        fmeCode = FME_DBLCLK;
                    }
                    else {
                        fProcessed = FALSE;   //not processed
                    }
                    break;

                case CC_COMBOBOX:
#if defined(_DEBUG) && 0
                    {   //debug combobox events
                        char sz[256];
                        static int iSeq = 0;
                        iSeq++;
                        sprintf(sz, "CB event %.4d: Ctl = %d, Cmd = %d (%s)\n",
                            iSeq,
                            idd - IDC_CTLBASE,
                            iCmd,
                            iCmd==CBN_ERRSPACE      ?"CBN_ERRSPACE":
                            iCmd==CBN_SELCHANGE     ?"CBN_SELCHANGE":
                            iCmd==CBN_DBLCLK        ?"CBN_DBLCLK":
                            iCmd==CBN_SETFOCUS      ?"CBN_SETFOCUS":
                            iCmd==CBN_KILLFOCUS     ?"CBN_KILLFOCUS":
                            iCmd==CBN_EDITCHANGE    ?"CBN_EDITCHANGE":
                            iCmd==CBN_EDITUPDATE    ?"CBN_EDITUPDATE":
                            iCmd==CBN_DROPDOWN      ?"CBN_DROPDOWN":
                            iCmd==CBN_CLOSEUP       ?"CBN_CLOSEUP":
                            iCmd==CBN_SELENDOK      ?"CBN_SELENDOK":
                            iCmd==CBN_SELENDCANCEL  ?"CBN_SELENDCANCEL":
                            "???"
                            );
                        OutputDebugString(sz);
                    }
#endif
                    if (iCmd == CBN_SELENDOK /*||
                        iCmd == CBN_SELCHANGE*/) {
                        
                        oldvalue = gParams->ctl[idd - IDC_CTLBASE].val;
                        
                        iVal = (int)SendMessage(COMMANDWND(wParam), CB_GETCURSEL, 0, 0);
                        fm_setCtlVal(idd - IDC_CTLBASE, iVal);
                        fmeCode = FME_CLICKED;  //for now
                    }
                    else if (iCmd == CBN_DBLCLK) {
                        //Can only occur for CBS_SIMPLE...
                        fProcessed = FALSE;     //not processed
                        fActionable = FALSE;    //for now
                        fmeCode = FME_DBLCLK;
                    }
                    else if (iCmd == CBN_DROPDOWN) {
                        fProcessed = FALSE;     //not processed
                        fActionable = FALSE;    //for now
                        fmeCode = FME_COMBO_DROPDOWN;
                    }
                    else if (iCmd == CBN_CLOSEUP) {
                        fProcessed = FALSE;     //not processed
                        fActionable = FALSE;    //for now
                        fmeCode = FME_COMBO_CLOSEUP;
                    }
                    else {
                        fProcessed = FALSE;   //not processed
                    }
                    break;

				case CC_EDIT:
					
					
					if (iCmd == EN_CHANGE) {
						
						GetDlgItemText(hDlg, idd, gParams->ctl[idd - IDC_CTLBASE].label,1024);
						
						oldvalue = (int)strlen(gParams->ctl[idd - IDC_CTLBASE].label);
						fmeCode = FME_CHANGED;//FME_VALUECHANGED;
					
					} else if (iCmd == EN_SETFOCUS) {

						//fmeCode = FME_CLICKED;
						fmeCode = FME_SETFOCUS;//FME_SETEDITFOCUS;
						
					} else if (iCmd == EN_KILLFOCUS) {

						fmeCode = FME_KILLFOCUS;//FME_KILLEDITFOCUS;
					
					} else if (iCmd == EN_MAXTEXT){

						fmeCode = FME_CHANGED;//FME_VALUECHANGED;
					}

					fProcessed = FALSE;     //not processed
                    fActionable = FALSE;    //for now

					break;

                default:
                    fProcessed = FALSE;   //not processed
                }//switch class


#ifdef SCRIPTABLE
                //Set script values before CTL_OK event, so that the user can still modify them
                if (idd - IDC_CTLBASE == CTL_OK && fmeCode == FME_CLICKED){
                        int i;
                        for (i=0;i<256;i++) {
                           ScriptParam[i] = gFmcp->pre_ctl[i]; //gParams->ctl[i].val
                           ScriptParamUsed[i] = gParams->ctl[i].inuse && gParams->ctl[i].scripted;
                        }
                }
#endif


                fActionable = fProcessed;   //for now
                return PerformControlAction(
                    idd - IDC_CTLBASE,  //control index
                    fmeCode,            //FM event code
                    oldvalue,           //previous value
                    wMsg,               //message type
                    iCmd,               //notification code
                    fActionable,
                    fProcessed,0);
            }//user control

            switch  (idd) {
#if 0
            case 3988:  // test combo box
                if (iCmd == CBN_SELENDOK /*||
                    iCmd == CBN_SELCHANGE*/) {
                    char s[256];
                    int index = SendMessage(COMMANDWND(wParam), CB_GETCURSEL, 0, 0);
                    SendMessage(COMMANDWND(wParam), CB_GETLBTEXT, index, (LPARAM)s);
                    Info("Combo box cmd=%d\nindex = %d\nstring = '%s'", iCmd, index, s);
                }
                break;

            case 3989:  // test list box
                if (iCmd == LBN_SELCHANGE /*||
                    iCmd == LBN_DBLCLK*/) {
                    char s[256];
                    int index = SendMessage(COMMANDWND(wParam), LB_GETCURSEL, 0, 0);
                    SendMessage(COMMANDWND(wParam), LB_GETTEXT, index, (LPARAM)s);
                    Info("List box cmd=%d\nindex = %d\nstring = '%s'", iCmd, index, s);
                }
                break;
#endif

#if !defined(BETA) && !defined(STANDALONE)


              case  100:
                    if (iCmd == EN_CHANGE)   // WIN32 Change
                    {          
              
                        // user changed the bandwidth edit text
                        char *pInvalid;

                        GetDlgItemText(hDlg, 100, szInput, 32);
                        lSize = strtol(szInput, &pInvalid, 10);
                        if  ( ! *pInvalid && lSize >= 1 && lSize <= 1000)
                        {
                            gParams->bandWidth = lSize;

                            //Do we really want to update the Proxy???
                            //The Proxy update is unaffected by changes
                            //in bandwidth!!!
                            if (!fDragging) {
                                // Update Filter parameter block display
                                ShowFilterParameterBlock(globals, hDlg);
                                if (gResult == noErr) {
                                    // Do Filtering operation
                                    DoFilterRect (globals);
                                    // Invalidate Proxy Item
                                    UpdateProxy(hDlg, FALSE);
                                }//noErr
                            }
                        }
                    }
                    break;


              case IDC_TRACE:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_TRACE;
                    else
                        gParams->flags &= ~AFH_TRACE;
                    break;

              case IDC_MMX:
                    //ignore BN_SETFOCUS, BN_KILLFOCUS...
                    if (iCmd == BN_CLICKED) {
                        if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                            gParams->flags |= AFH_USE_MMX;
                        else
                            gParams->flags &= ~AFH_USE_MMX;
                        gFmcp->use_MMX = (gParams->flags & AFH_USE_MMX) != 0;
                        // Invalidate any generated code to force re-codegen
                        gParams->fValidCompile = FALSE;
                    }
                    break;

              case IDC_ADVSTATE:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_USE_ADVANCESTATE;
                    else
                        gParams->flags &= ~AFH_USE_ADVANCESTATE;
                    break;

              case IDC_HOSTPAD:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_USE_HOST_PADDING;
                    else
                        gParams->flags &= ~AFH_USE_HOST_PADDING;
                    break;

              case IDC_OPTZOOM:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_INIT_OPTIMAL_ZOOM;
                    else
                        gParams->flags &= ~AFH_INIT_OPTIMAL_ZOOM;
                    break;

              case IDC_BIGGULP:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_PROXY_BIG_GULP;
                    else
                        gParams->flags &= ~AFH_PROXY_BIG_GULP;
                    break;

              case IDC_ZDRAG:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                    {
                        gParams->flags |= AFH_ZERO_OUTPUT_DRAG;
                        // and clear DRAG_FILTERED_OUTPUT...
                        SendMessage(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_DRAGOUT), BM_SETCHECK, 0, 0);
                        gParams->flags &= ~AFH_DRAG_FILTERED_OUTPUT;
                    }
                    else
                        gParams->flags &= ~AFH_ZERO_OUTPUT_DRAG;
                    break;

              case IDC_FLUSHA:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_FLUSH_ADVANCESTATE;
                    else
                        gParams->flags &= ~AFH_FLUSH_ADVANCESTATE;
                    break;

#if 0	//no longer used
              case IDC_OPTBEV:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_OPTIMIZE_BEVEL;
                    else
                        gParams->flags &= ~AFH_OPTIMIZE_BEVEL;
                    break;
#endif

              case IDC_DISASSEMB:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_DISASSEMBLE;
                    else
                        gParams->flags &= ~AFH_DISASSEMBLE;
                    break;

              case IDC_DUMPSDL:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_DUMP_SDL;
                    else
                        gParams->flags &= ~AFH_DUMP_SDL;
                    break;

              case IDC_PSEUDOCODE:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_PSEUDOCODE;
                    else
                        gParams->flags &= ~AFH_PSEUDOCODE;
                    break;

              case IDC_USE_CMOV:
                    //ignore BN_SETFOCUS, BN_KILLFOCUS...
                    if (iCmd == BN_CLICKED) {
                        if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                            gParams->flags |= AFH_USE_CMOV;
                        else
                            gParams->flags &= ~AFH_USE_CMOV;
                        gFmcp->use_CMOV = (gParams->flags & AFH_USE_CMOV) != 0;
                        // Invalidate any generated code to force re-codegen
                        gParams->fValidCompile = FALSE;
                    }
                    break;

              case IDC_USE_SSE:
                    //ignore BN_SETFOCUS, BN_KILLFOCUS...
                    if (iCmd == BN_CLICKED) {
                        if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                            gParams->flags |= AFH_USE_SSE;
                        else
                            gParams->flags &= ~AFH_USE_SSE;
                        gFmcp->use_SSE = (gParams->flags & AFH_USE_SSE) != 0;
                        // Invalidate any generated code to force re-codegen
                        gParams->fValidCompile = FALSE;
                    }
                    break;

              case IDC_USE_INLINES:
                    //ignore BN_SETFOCUS, BN_KILLFOCUS...
                    if (iCmd == BN_CLICKED) {
                        if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                            gParams->flags |= AFH_USE_INLINES;
                        else
                            gParams->flags &= ~AFH_USE_INLINES;
                        // Invalidate parse and codegen to force a re-parse,
                        // since in-lining is determined at parse time.
                        gParams->fValidParse = FALSE;
                        gParams->fValidCompile = FALSE;
                    }
                    break;

              case IDC_RGB:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_SUPPORT_RGB;
                    else
                        gParams->flags &= ~AFH_SUPPORT_RGB;
                    break;

              case IDC_GRAY:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_SUPPORT_GRAY;
                    else
                        gParams->flags &= ~AFH_SUPPORT_GRAY;
                    break;

              case IDC_CMYK:  
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_SUPPORT_CMYK;
                    else
                        gParams->flags &= ~AFH_SUPPORT_CMYK;
                    break;

              case IDC_LAB:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_SUPPORT_LAB;
                    else
                        gParams->flags &= ~AFH_SUPPORT_LAB;
                    break;

              case IDC_INDEXED:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_SUPPORT_INDEXED;
                    else
                        gParams->flags &= ~AFH_SUPPORT_INDEXED;
                    break;

              case IDC_BITMAP:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_SUPPORT_BITMAP;
                    else
                        gParams->flags &= ~AFH_SUPPORT_BITMAP;
                    break;

              case IDC_DUOTONE:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_SUPPORT_DUOTONE;
                    else
                        gParams->flags &= ~AFH_SUPPORT_DUOTONE;
                    break;

              case IDC_MULTI:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                        gParams->flags |= AFH_SUPPORT_MULTI;
                    else
                        gParams->flags &= ~AFH_SUPPORT_MULTI;
                    break;

#if !defined(STATIC_FILTER) || defined(ALLOW_DRAGOUT)
              // Process "Drag Filtered Output" check box...
              case IDC_DRAGOUT:
                    if (SendMessage(COMMANDWND(wParam), BM_GETCHECK, 0, 0))
                    {
                        gParams->flags |= AFH_DRAG_FILTERED_OUTPUT;
                        // and clear ZERO_OUTPUT_DRAG...
#if !defined(STATIC_FILTER)
                        SendMessage(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_ZDRAG), BM_SETCHECK, 0, 0);
#endif
                        gParams->flags &= ~AFH_ZERO_OUTPUT_DRAG;
                    }
                    else
                        gParams->flags &= ~AFH_DRAG_FILTERED_OUTPUT;
                    break;
#endif

#endif //if !defined(BETA) && !defined(STANDALONE)

#ifndef STATIC_FILTER
#ifndef STANDALONE
              case IDC_PARSE:
                  {   // Hit the "Parse" button.
                      // Temporary KLUDGE: use temporary szCode buffer
                      // until fix Mdp_syntaxErr so it doesn't modify
                      // the source code buffer!!!!
                      static char szCode[MAX_SOURCE_CODE_SIZE+1];

                      gParams->fValidParse = FALSE; //pessimistic
                      gParams->fValidCompile = FALSE;

                      // Copy latest source code to buffer if necessary...
                      if (!FmGetEditorText()) break;
                      assert (sizeof gParams->szSrcCode <= sizeof szCode);
                      strcpy(szCode, gParams->szSrcCode);	//NEW: Pre-process from gParams to szCode!!!

                      // Got valid source code buffer...
                      {   // parse the code...
                          int iss;
                          unsigned int parser_flags = 0;
#if 0
                          //debugging...
                          Info("Compiling %d chars...", strlen(szCode));
#endif
						  // Reset the filterTitleText, filterCategoryText and
						  // filterFilenameText strings, which may then get set
						  // by the parser... (Or could we init these strings
						  // to their default values here?)
						  gParams->filterTitleText[0] = 0;
						  gParams->filterCategoryText[0] = 0;
						  gParams->filterFilenameText[0] = 0;

                          if (gParams->flags & AFH_USE_INLINES) {
                              parser_flags |= MDP_USE_INLINES;
                          }
                          // Clear any previous text from output window...
                          SetDlgItemText(hEditorDlg?hEditorDlg:hDlg, IDC_EDIT1, "");
                          iss = Mdp_parse(szCode, parser_flags, gParams->SDL, NSDL);
                          gParams->fValidParse = iss & 1;
                          if (gParams->fValidParse) {
                              //success -- dump the SDL
#ifdef FULL
                              if ((gParams->flags & (AFH_DUMP_SDL|AFH_EDITMODE|AFH_ADVANCED)) ==
                                                    (AFH_DUMP_SDL|AFH_EDITMODE|AFH_ADVANCED)) {
                                  Mdp_dumpSDL(szCode);
                              }
#endif
                          }
                          else if (iss == MD386__SYNTAX) {
                              // Error...
                              // Warning: Mdp_syntaxErr modifies contents of source buffer!!!
                              char *lpToken = Mdp_syntaxErr();
                              // Highlight the error "token" in the source edit window...
#ifdef USE_EDIT_CONTROL
                              Edit_SetSel(hEdit2, lpToken-szCode, (lpToken-szCode)+strlen(lpToken));
#else
                              {
                                 CHARRANGE ichCharRange;
                                 ichCharRange.cpMin = (LONG)(lpToken-szCode);
                                 ichCharRange.cpMax = ichCharRange.cpMin + (LONG)strlen(lpToken);
                                 SendMessage(hEdit2, EM_EXSETSEL, 0, (LPARAM)&ichCharRange);
                              }
#endif //USE_EDIT_CONTROL
                              // ...and scroll it into view if necessary!
                              Edit_ScrollCaret(hEdit2);
                              msgBox(MB_ICONERROR, "FilterMeister Syntax Error",
                                    "Error while looking for %s\nnear \"%s\"",
                                    MDP_looking_for, *lpToken ? lpToken : "end_of_source");
                          }
                      }//scope
                      // Set focus back to Source Code Edit Text Item
                      SetFocus (GetDlgItem (hEditorDlg?hEditorDlg:hDlg, IDC_EDIT2));
                  }//scope
                  break;

              case IDC_COMPILE:
                  {   // Hit the "Codegen" button.
                      if (!gParams->fValidParse) {
                          // Need to (re-)parse before codegen...
                          SendMessage(hDlg, WM_COMMAND, IDC_PARSE, 0);
                      }

                      if (gParams->fValidParse)
                      {   // Evaluate the SDL and dump the output from the Estack...
                          // Clear any previous text from output window...
                          SetDlgItemText(hEditorDlg?hEditorDlg:hDlg, IDC_EDIT1, "");
                          gParams->fValidCompile = TRUE; //assume the compile will succeed...
                          Mdc_dump_estack(gParams->SDL);
                          //Enable or disable the Bandwidth Edit field, depending
                          //on whether the algorithm is tileable or not...
                          EnableWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, 100), gFmcp->isTileable);
                          
                          //Added by Harald Heim, March 21, 2003
                          // Reset FM dialog to its default size.
                          fm_setDialogPos(0, -2, -2, 485 , 178);
                      }
#if 0
                      else
                      {   // No valid SDL to evaluate.
                          
                          // Reset FM dialog to its default size.
                          fm_setDialogPos(0, -2, -2, 485 , 178);

                          MessageBeep(MB_ICONHAND);
                      }
#endif
                      // Set focus back to Source Code Edit Text Item
                      SetFocus (GetDlgItem (hEditorDlg?hEditorDlg:hDlg, IDC_EDIT2));
                  }
                  break;
#endif //ifndef STANDALONE

              case IDC_INIT:
                  {   // Hit the "Initialize" button.
#ifdef STANDALONE
                      gParams->fValidCompile = TRUE;
#else
                      if (!gParams->fValidCompile) {
                          // Need to (re-)codegen before init...
                          SendMessage(hDlg, WM_COMMAND, IDC_COMPILE, 0);
                      }
#endif
                      if (gParams->fValidCompile)
                      {   // Execute the initialization code
                         
						  //Clean up threads
						  fm_terminateThread(0);

						  if (strcmp(szFileName,"")==0){
							//set current directory to filter install directory
							_chdir(fm_getFilterInstallDir());
							//Info ("%s",fm_getFilterInstallDir());
						  } else {
							static char tempFolder[_MAX_PATH + 1];
							char * pch;
							if (pch=strrchr(szFileName, 0x5C)){		//Use ASCII CHar!!
								memset (pch,0,1);
								strcpy(tempFolder, szFileName);
								memset (pch,0x5C,1);
								_chdir(tempFolder);
								//Info ("%s",tempFolder);
							} else {
								_chdir(fm_getFilterInstallDir());
								//Info ("%s",fm_getFilterInstallDir());
							}
						  }

						  //Reset Dialog Events
						  gParams->gDialogEvent = 0;

                          //Reset dialog restrictions
                          MaxMinInfo.ptMinTrackSize.x = 0;
                          MaxMinInfo.ptMinTrackSize.y = 0;
                          MaxMinInfo.ptMaxTrackSize.x = 0;
                          MaxMinInfo.ptMaxTrackSize.y = 0;
						  
						  fm_resetAllCtls();

#ifndef STANDALONE
                          //Create a new UniqueID!!!
                          {
                             int i;

                             srand((unsigned)time(NULL));
                             for (i=0;i<36;i++){
                                gParams->filterUniqueID[i] = 48 + (rand() % 10); //Values from 0 to 9
                             }
                             gParams->filterUniqueID[36] = 0; 
                         }
                         strcpy(gFmcp->filterUniqueID, gParams->filterUniqueID);
#endif

#ifdef STANDALONE
                          if (setjmp(Mde_error_exit_jb)) {
                              //Someone longjmp'ed back to us...
                              //Force another re-compile/init
                              gParams->fValidCompile = FALSE;
                          }
                          else {
                              //execute init code
							  ///Info("About to call CB_initCode from IDC_INIT in FilterDialogProc");
                              X86_call_fmf1(gParams->CB_initCode, (INT_PTR)gFmcp);
                          }
#else
                          Mdc_execute_init_code();
#endif


#ifndef STANDALONE
                    #ifdef SCRIPTABLE
                            //Write the script params to the controls
                            if (ScriptParamRead){
                                int n;
                                for (n=0;n<256;n++) {
                                    if (ScriptParamUsed[n] && gParams->ctl[n].inuse && gParams->ctl[n].scripted){
                                        fm_setCtlVal(n, ScriptParam[n]);
                                    }
                                }
                            }
                    #endif
#endif                          
                          
                          // Update dialog title (re-interpolate)...
                          fm_setDialogText((INT_PTR) gParams->gDialogTitle);

						  setDefaultWindowTheme(GetDlgItem(hDlg, IDC_PROGRESS1));
						  setDefaultWindowTheme(GetDlgItem(hDlg, IDC_BUTTON1));
						  setDefaultWindowTheme(GetDlgItem(hDlg, IDC_BUTTON2));
						  setDefaultWindowTheme(GetDlgItem(hDlg, IDC_PROXY_GROUPBOX));

#ifndef STANDALONE
						  if (gParams->gDialogEvent & 1)
							  fm_triggerEvent(0, FME_INIT, (int)!gParams->gInitialInvocation);
#endif
                          
						  // Invalidate Proxy Item
                          UpdateProxy(hDlg, FALSE);
                      }
#if 0
                      else
                      {   // No valid code to execute.
                          MessageBeep(MB_ICONHAND);
                      }
#endif

#ifndef STANDALONE
                      // Set focus back to Source Code Edit Text Item
                      SetFocus (GetDlgItem (hEditorDlg?hEditorDlg:hDlg, IDC_EDIT2));
#endif
                  }
                  break;

              case IDC_GO:
                  {   // Hit the "Go!" button.
                      if (!gParams->fValidCompile) {
                          // Need to (re-)codegen and init before go...
                          SendMessage(hDlg, WM_COMMAND, IDC_INIT, 0);
                      }

                      if (gParams->fValidCompile)
                      {   // Run the filter and update the proxy
                          if (!fDragging) {
                              // Update Filter parameter block display
                              ShowFilterParameterBlock(globals, hDlg);
                              if (gResult == noErr) {
                                  //Race condition?  If we just did an IDC_INIT,
                                  //the message loop may not have had time to finish
                                  //repainting the interface, so we need to introduce
                                  //a small delay...
#if 0
                                  // Do Filtering operation
                                  DoFilterRect (globals);
                                  // Invalidate Proxy Item
                                  UpdateProxy(hDlg, FALSE);
#else
                                  // Set a timer to trigger a delayed proxy update,
                                  // giving the message loop a chance to paint the dialog
                                  // window first...
                                  KillTimer(hDlg, 1); //Cancel any previous trigger.
                                  if (!SetTimer(hDlg, 1, 10, NULL)) {
                                      // If SetTimer() fails, update the proxy immediately.
                                      // Do Filtering operation
                                      DoFilterRect (globals);
                                      // Invalidate Proxy Item
                                      UpdateProxy(hDlg, FALSE);
                                  }
                              }//noErr
                          }
                      }
#if 0
                      else
                      {   // No valid code to execute.
                          MessageBeep(MB_ICONHAND);
                      }
#endif
#endif

#ifndef STANDALONE
                      // Set focus back to Source Code Edit Text Item
                      //SetFocus (GetDlgItem (hEditorDlg?hEditorDlg:hDlg, IDC_EDIT2));
                      SetFocus(hDlg);
#endif
                  }
                  break;

#ifndef STANDALONE
              case  IDC_EDIT2:          // user changed the source code edit text
                    if (iCmd == EN_CHANGE)   // WIN32 Change
                    {
#if 0
                        //debugging...
                        Info("IDC_EDIT2: EN_CHANGE");
#endif
                        gParams->fValidParse = FALSE; //invalidate any previous parse
                        gParams->fValidCompile = FALSE; //invalidate any previous compile
                        gParams->fGotSourceCode = TRUE;
                        gParams->fNeedSourceSave = TRUE;
                        return FALSE;   // not handled
                    }
                    else if (iCmd == EN_ERRSPACE ||
                             iCmd == EN_MAXTEXT )
                    {   //too much text!
                        gParams->fValidParse = FALSE; //invalidate any previous parse
                        gParams->fValidCompile = FALSE; //invalidate any previous compile
                        gParams->fGotSourceCode = TRUE; //even though it's NFG
                        gParams->fNeedSourceSave = TRUE;
#if 0
                        //debugging...
                        Info("iCmd = %x", iCmd);
#endif
#if 0
                        ShowAlert(IDS_MAXTEXT);
#else
                        ErrorOk("Too much source code (more than %d characters).",
                                MAX_SOURCE_CODE_SIZE);
#endif
                        // Set focus back to Source Code Edit Text Item
                        SetFocus(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_EDIT2));
                    }
                    else
                        return FALSE;   //not handled
                    break;
#endif //ifndef STANDALONE

#ifndef STANDALONE
			  case IDC_X64GEN:			// User pushed "X64" or "X32" button
				  //Toggle the X64GEN flag
				  gParams->flags ^= AFH_X64GEN;
				  if (X64GEN) {
                      SetDlgItemText(hEditorDlg?hEditorDlg:hDlg, IDC_X64GEN, "&X64");
				  }
				  else {
                      SetDlgItemText(hEditorDlg?hEditorDlg:hDlg, IDC_X64GEN, "&X32");
				  }
				  // Invalidate parse and codegen to force a re-parse,
                  // since toggling X64 may set different macro defs.
                  gParams->fValidParse = FALSE;
                  gParams->fValidCompile = FALSE;
				  break;

              case IDC_EDIT:            // User pushed "Edit >>>" button
                  //Toggle BOTH flags (AFH_ADVANCED gets toggled again below)...
                  gParams->flags ^= AFH_EDITMODE | AFH_ADVANCED;
                  {
                      fm_setCtlToolTip(CTL_EDIT,
                                 (gParams->flags & AFH_EDITMODE) ?
                                 (INT_PTR)"Close source edit window" :
                                 (INT_PTR)"Edit filter source code",
                                 0);
                  }
                  // Fall through...

              case IDC_ADVANCED:            // User pushed "Advanced >>>" button
                  {
                      gParams->flags ^= AFH_ADVANCED;
                      fm_setCtlText(CTL_EDIT, 0); //update EDIT control text

                      if (gParams->flags & AFH_EDITMODE) {
                          // Open the editor dialog...
                          StartEditor();
                      }
                      else {
#if 0
                          // Contract dialog window to hide editing controls
                          fm_setDialogPos(FALSE, -1, 0, 485,178); //388, 142);
#else
                          // Close editor dialog if it's active
                          if (hEditorDlg) {
                              LPARAM iss;
                              //Save the source code in our filter parameter block...
#if 0
                              // Should complain if too much text????
                              GetDlgItemText(hEditorDlg, IDC_EDIT2, gParams->szSrcCode,
                                             sizeof(gParams->szSrcCode));
#else
                              FmGetEditorText();
#endif
                              //Send a CANCEL command to the editor window...
                              iss = SendMessage(hEditorDlg, WM_COMMAND, IDCANCEL, 0);
                              hEditorDlg = 0;
                          }
#endif
                      }

                      if (!(gParams->flags & AFH_EDITMODE)) {
                          //nothing
                      }
#ifdef FULL
                      else if (gParams->flags & AFH_ADVANCED) {
                          // Expand dialog window to show advanced editing controls
                          //SetDlgItemText(hDlg, IDX_EDIT, "&Edit <<<");
#if 0
                          SetDlgItemText(hEditorDlg?hEditorDlg:hDlg, IDC_ADVANCED, "&Advanced <<<");
#else
                          SetDlgItemText(hEditorDlg?hEditorDlg:hDlg, IDC_ADVANCED, "&<<<");
#endif
                          SetDlgItemText(hEditorDlg?hEditorDlg:hDlg, IDC_GO, "Pre&view");
#if 0
                          fm_setDialogPos(FALSE, -1, 0, 671,305); //532, 407);
#else
                          {
                              /* Convert dialog units to pixels */
                              RECT rcDialog;
                              RECT rcClient;
                              RECT r;
                              r.left = 0;
                              r.top = 0;
                              r.right = 671;//532;
                              r.bottom = 305; //273;
                              MapDialogRect(hEditorDlg, &r);
                              // Need to add SM_CYCAPTION, SM_CXDLGFRAME, etc.
                              // Add the difference between dialog window size and dialog client
                              // size, so it automatically adjusts for caption, border, etc.
                              GetWindowRect(hEditorDlg, &rcDialog);
                              GetClientRect(hEditorDlg, &rcClient);
                              r.right += (rcDialog.right-rcDialog.left) - (rcClient.right-rcClient.left);
                              r.bottom += (rcDialog.bottom-rcDialog.top) - (rcClient.bottom-rcClient.top);
                              SetWindowPos(hEditorDlg, NULL,
                                           0, 0, r.right, r.bottom,
                                           SWP_NOZORDER | SWP_NOACTIVATE | SWP_NOMOVE);
                          }
#endif
                      }
#endif //FULL
                      else {
                          // Contract dialog window to hide advanced editing controls
#ifdef FULL
#if 0
                          SetDlgItemText(hEditorDlg?hEditorDlg:hDlg, IDC_ADVANCED, "&Advanced >>>");
#else
                          SetDlgItemText(hEditorDlg?hEditorDlg:hDlg, IDC_ADVANCED, "&>>>");
#endif
                          SetDlgItemText(hEditorDlg?hEditorDlg:hDlg, IDC_GO, "&Compile");
#endif
#if 0
                          fm_setDialogPos(FALSE, -1, 0, 485,178);  //388, 311
#else
                          {
                              /* Convert dialog units to pixels */
                              RECT rcDialog;
                              RECT rcClient;
                              RECT r;
                              r.left = 0;
                              r.top = 0;
                              r.right = 482; //388;
                              r.bottom = 207;//173;
                              MapDialogRect(hEditorDlg, &r);
                              // Need to add SM_CYCAPTION, SM_CXDLGFRAME, etc.
                              // Add the difference between dialog window size and dialog client
                              // size, so it automatically adjusts for caption, border, etc.
                              GetWindowRect(hEditorDlg, &rcDialog);
                              GetClientRect(hEditorDlg, &rcClient);
                              r.right += (rcDialog.right-rcDialog.left) - (rcClient.right-rcClient.left);
                              r.bottom += (rcDialog.bottom-rcDialog.top) - (rcClient.bottom-rcClient.top);
                              SetWindowPos(hEditorDlg, NULL,
                                           0, 0, r.right, r.bottom,
                                           SWP_NOZORDER | SWP_NOACTIVATE | SWP_NOMOVE);
                          }
#endif
                      }
                      //Enable/disable Edit mode buttons...
                      EnableWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_LOAD), (gParams->flags & AFH_EDITMODE) != 0);
                      EnableWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_SAVE), (gParams->flags & AFH_EDITMODE) != 0);
                      EnableWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_X64GEN), (gParams->flags & AFH_EDITMODE) != 0);
                      EnableWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_FONT), (gParams->flags & AFH_EDITMODE) != 0);
                      EnableWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_GO), (gParams->flags & AFH_EDITMODE) != 0);
                      EnableWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_MAKE), (gParams->flags & AFH_EDITMODE) != 0);
                      EnableWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_HELP1), (gParams->flags & AFH_EDITMODE) != 0);
                      EnableWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_HELP2), (gParams->flags & AFH_EDITMODE) != 0);
#ifdef FULL
                      EnableWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_ADVANCED), (gParams->flags & AFH_EDITMODE) != 0);
                      //Show/Hide Advanced mode buttons...
#if 1 //IDC_PREPROCESS is disabled for now
					  EnableWindow(GetDlgItem(hDlg, IDC_PREPROCESS), 0);
#else
                      EnableWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_PREPROCESS), (gParams->flags & AFH_ADVANCED) != 0);
#endif
                      EnableWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_PARSE), (gParams->flags & AFH_ADVANCED) != 0);
                      EnableWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_COMPILE), (gParams->flags & AFH_ADVANCED) != 0);
                      EnableWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_INIT), (gParams->flags & AFH_ADVANCED) != 0);
                      ShowWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_PREPROCESS), gParams->flags & AFH_ADVANCED?SW_SHOW:SW_HIDE);
                      ShowWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_PARSE), gParams->flags & AFH_ADVANCED?SW_SHOW:SW_HIDE);
                      ShowWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_COMPILE), gParams->flags & AFH_ADVANCED?SW_SHOW:SW_HIDE);
                      ShowWindow(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_INIT), gParams->flags & AFH_ADVANCED?SW_SHOW:SW_HIDE);
#endif
                  }

                  // Update Filter parameter block display
                  ShowFilterParameterBlock(globals, hDlg);

                  if (gParams->flags & AFH_EDITMODE) {
                      // Set Focus to Source Code Edit Text Item
                      SetFocus (GetDlgItem (hEditorDlg?hEditorDlg:hDlg, IDC_EDIT2));
                  }

                  // Should also repaint the background if it's a gradient (or image?)!
                  break;
#endif //ifndef STANDALONE

#ifndef STANDALONE
                case IDC_LOAD:
                  FM_is_busy++;
                  
                  if (GetAsyncKeyState (VK_SHIFT) <=-32767){
                    //Reload and compile  
                    openSourceFile();
                    SendMessage(hDlg, WM_COMMAND, IDC_GO, 0);
                  
                  } else {
                     //Code stolen from Petzold Chapter 11...
                     static char szFilter[] = 
                      "All Filter Source Files (*.fml;*.ffp;*.afs;*.8bf;*.ffw)\0"
                                               "*.fml;*.ffp;*.afs;*.8bf;*.ffw\0"    //1
                      "FilterMeister Source Files (*.fml)\0*.fml\0"                 //2
                      "Filter Factory Plus Source Files (*.ffp)\0*.ffp\0"           //3
                      "Filter Factory Source Files (*.afs;*.8bf)\0*.afs;*.8bf\0"    //4
                      "Filter Factory Wizard Source Files (*.ffw)\0*.ffw\0"         //5
                      "Text Files (*.txt)\0*.txt\0"                                 //6
                      "All Files (*.*)\0*.*\0"                                      //7
                      ;

                     if (CheckForFileSave() == IDCANCEL) {
                         FM_is_busy--;
                         break;
                     }

                     ofn.lStructSize       = sizeof (OPENFILENAME) ;
#if 0
                     ofn.hwndOwner         = hEditorDlg?hEditorDlg:hDlg ;   //causes hang
#else
                     ofn.hwndOwner         = hDlg ;
#endif
                     ofn.hInstance         = NULL ;
                     ofn.lpstrFilter       = szFilter ;
                     ofn.lpstrCustomFilter = NULL ;
                     ofn.nMaxCustFilter    = 0 ;
                     //ofn.nFilterIndex      = 0 ;
                     ofn.nMaxFile          = _MAX_PATH ;
                     ofn.nMaxFileTitle     = _MAX_FNAME + _MAX_EXT ;
                     ofn.lpstrInitialDir   = NULL ;
                     ofn.lpstrTitle        = "Open Filter Source File" ;
                     ofn.nFileOffset       = 0 ;
                     ofn.nFileExtension    = 0 ;
                     ofn.lpstrDefExt       = "ffp" ;
                     ofn.lCustData         = 0L ;
                     ofn.lpfnHook          = NULL ;
                     ofn.lpTemplateName    = NULL ;
                     ofn.lpstrFile         = szFileName ;
                     ofn.lpstrFileTitle    = szTitleName ;
                     ofn.Flags             = OFN_HIDEREADONLY /*| OFN_CREATEPROMPT*/ ;

                     if (GetOpenFileName (&ofn)) {
                            openSourceFile();
                     }
                     
                  }
                  FM_is_busy--;
                  break;

              case IDC_SAVE:
                  FM_is_busy++;
                  {
                     //Code stolen from Petzold Chapter 11...
                     static OPENFILENAME sfn;
                     static char szFilter[] = 
                      "FilterMeister Source File (*.fml)\0*.fml\0"          //1
                      "Filter Factory Plus Source File (*.ffp)\0*.ffp\0"    //2
                      "Filter Factory Source File (*.afs)\0*.afs\0"         //3
                      "Filter Factory Wizard Source File (*.ffw)\0*.ffw\0"  //4
                      "Text File (*.txt)\0*.txt\0"                          //5
                      ;

                     sfn.lStructSize       = sizeof (OPENFILENAME) ;
#if 0
                     sfn.hwndOwner         = hEditorDlg?hEditorDlg:hDlg ;   //causes hang
#else
                     sfn.hwndOwner         = hDlg ;
#endif
                     sfn.hInstance         = NULL ;
                     sfn.lpstrFilter       = szFilter ;
                     sfn.lpstrCustomFilter = NULL ;
                     sfn.nMaxCustFilter    = 0 ;
                     sfn.nFilterIndex      = 0 ;
                     sfn.nMaxFile          = _MAX_PATH ;
                     sfn.nMaxFileTitle     = _MAX_FNAME + _MAX_EXT ;
                     sfn.lpstrInitialDir   = NULL ;
                     sfn.lpstrTitle        = NULL ;
                     sfn.nFileOffset       = 0 ;
                     sfn.nFileExtension    = 0 ;
                     sfn.lpstrDefExt       = "ffp" ;
                     sfn.lCustData         = 0L ;
                     sfn.lpfnHook          = NULL ;
                     sfn.lpTemplateName    = NULL ;
                     sfn.lpstrFile         = szFileName ;
                     sfn.lpstrFileTitle    = szTitleName ;
                     sfn.Flags             = OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT ;

                     //Set default save file type...
#if 0
                     if (ofn.nFilterIndex <= 1 ||   // undefined or Any filter source
                         ofn.nFilterIndex == 3 ||   // *.afs
                         ofn.nFilterIndex == 6) {   // *.*
                         // Default file type is ffp (and we can't save .afs
                         // files, so use the default save type in this case also).
                         sfn.nFilterIndex = 1;  // ffp save index
                     }
                     else {
                         // Map load file type to same save file type...
                         sfn.nFilterIndex = ofn.nFilterIndex - 1;
                     }
#else
                     if (_stricmp(&szFileName[ofn.nFileExtension], "afs") == 0 ||
                         _stricmp(&szFileName[ofn.nFileExtension], "8bf") == 0) {
                         // Change extension to .ffp...
                         strcpy(&szFileName[ofn.nFileExtension], "ffp");
                         //Achh, no.  Don't modify ofn.nFilterIndex -- it messes
                         //up a subsequent Open.
                         ofn.nFilterIndex = 3;  // *.ffp
                         //Whoops!!! If save is canceled, should set ext back
                         //in case user wants to reload it!!!!
                     }
                     switch (ofn.nFilterIndex) {
                     case 2:        // *.fml
                     case 3:        // *.ffp
                     case 5:        // *.ffw
                     case 6:        // *.txt
                         sfn.nFilterIndex = ofn.nFilterIndex - 1;
                         break;
                     case 4:        // *.afs or *.8bf
                         // Can't save *.afs/*.8bf files, so truncate the extension
                         // and use the default save file type...
                         if (ofn.nFileExtension > 0 &&
                             szFileName[ofn.nFileExtension] == '.') {
                             // lop off the '.' before the extension
                             szFileName[ofn.nFileExtension-1] = '\0';
                             //Ouch -- the SaveAs control adds the '.' back in???
                             // I guess we should just hard code an extension
                             // change of .afs.8bf to .ffp.
                         }
                         else {
                             szFileName[ofn.nFileExtension] = '\0';
                         }
                         // Fall through to default...
                     default:
						 if (gParams->fValidCompile){
							 if (strstr(gParams->szSrcCode,"%fml")!=0 || strstr (gParams->szSrcCode,"%FML")!=0){//ACK!! REDO THIS!!!
							 	sfn.nFilterIndex = 1;      // *.fml
								 if (strcmp (&szFileName[ofn.nFileExtension-1],".fml")!=0) 
									szFileName[ofn.nFileExtension-1] = '\0';
							 } else {
								 sfn.nFilterIndex = 2;      // *.ffp
								 if (strcmp (&szFileName[ofn.nFileExtension-1],".ffp")!=0)
									szFileName[ofn.nFileExtension-1] = '\0';
							}
						 } else {
							if (strcmp (&szFileName[ofn.nFileExtension-1],".fml")==0)
								 sfn.nFilterIndex = 1;      // *.fml
							else
								 sfn.nFilterIndex = 2;      // *.ffp
						 }

                         break;
                     } //switch
#endif

                     if (GetSaveFileName(&sfn)) {
                         FILE  *file;
                         size_t iLength;
#if 0
                         PSTR   pstrBuffer;
#endif

                         if (!(file = fopen (szFileName, "wb"))) {
                              msgBox(MB_ICONERROR, "FilterMeister I/O Error",
                                     "Could not create or overwrite file\r\n%s", szFileName);
                              FM_is_busy--;
                              break;
                         }

                         //WRONG if !hEditorDlg, in which case the source resides in gParams!!!
                         // Is this correct for Rich Edit controls with > 64K text????
#if 0
                         iLength = GetWindowTextLength(GetDlgItem(hEditorDlg?hEditorDlg:hDlg, IDC_EDIT2));

                         if (!(pstrBuffer = (PSTR) malloc (iLength + 1)))
                         {
                            fclose (file);
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not allocate %d bytes of memory.", iLength + 1);
                            FM_is_busy--;
                            break;
                         }

                         //WRONG if !hEditorDlg, in which case the source resides in gParams!!!
                         //Do this a la IDC_PARSE!!!
                         GetDlgItemText(hEditorDlg?hEditorDlg:hDlg, IDC_EDIT2, pstrBuffer, iLength + 1);

                         if ((int)fwrite(pstrBuffer, 1, iLength, file) != iLength) {
                             fclose(file);
                             free(pstrBuffer);
                             msgBox(MB_ICONERROR, "FilterMeister I/O Error",
                                    "Could not write file %s", szTitleName);
                             FM_is_busy--;
                             break;
                         }

                         free(pstrBuffer);
#else
                         if (!FmGetEditorText()) {
                             //too much text!
#if 0
                             //Should save anyway, even if source is too big to compile??
                             //(in which case, issue a warning but save anyway???)
                             fclose(file);
                             FM_is_busy--;
                             break;
#endif
                         }

                         iLength = strlen(gParams->szSrcCode);

                         if ((int)fwrite(gParams->szSrcCode, 1, iLength, file) != iLength) {
                             fclose(file);
                             msgBox(MB_ICONERROR, "FilterMeister I/O Error",
                                    "Could not write file %s", szTitleName);
                             FM_is_busy--;
                             break;
                         }
#endif

                         if (fclose(file)) {
                             msgBox(MB_ICONERROR, "FilterMeister I/O Error",
                                    "Could not close file %s", szTitleName);
                             FM_is_busy--;
                             break;
                         }

                         gParams->fNeedSourceSave = FALSE;

                         { //Grab file date 
                            HANDLE fHandle = CreateFile(szFileName, GENERIC_READ | GENERIC_WRITE, 
                                    FILE_SHARE_READ | FILE_SHARE_DELETE,
                                    NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL);
                            GetFileTime (fHandle, NULL,NULL, &szFileTime);
                            CloseHandle (fHandle);
                         }

                     } //if GetSaveFileName
                  }
                  FM_is_busy--;
                  break;
#endif //ifndef STANDALONE

#if 0
              case IDC_FIND:
                  {
                      Info("Find is not yet implemented");
                  }
                  break;
#endif

              //Added by Harald Heim, March 22, 2003
              case IDC_HELP1:    //(IDC_HELP conflicts with winuser.h)
                  {

                      char szPath[256];
                      int i;
                      DWORD ret = GetModuleFileName(hDllInstance, szPath, sizeof(szPath));
                      //Info("Filter filename = %s (size = %d)", szPath, ret);
                      for (i = ret - 1; i >= 0 && szPath[i] != '\\'; --i)
                          ;
                      szPath[i+1] = '\0';
                      //Info("Filter path = \"%s\"", szPath);

                      ret = (DWORD)(DWORD_PTR)ShellExecute(hDlg, "open", "help.htm", NULL, szPath, SW_SHOWNORMAL);
                      if (ret <= 32) {
                          ErrorOk("Help file can't be found");
                      }

                  }
                  break;



              case IDC_HELP2:
                  {
#if 0
                      Info("Help is not yet available.");
#elif 0
                      // Create a modeless dialog box
                    HWND hEditorDlg;
                    hEditorDlg = CreateDialogParam(hDllInstance,
                                       (LPSTR)"EDITOR",
                #if 0
                                       NULL,        //no owner
                #else
                                       (HWND)MyHDlg,
                #endif
                                       (FARPROC)AuthorizeDialogProc,
                                       (LPARAM)0);
                    //Show the dialog
                    ShowWindow(hEditorDlg, SW_SHOW);
#elif 0 //!defined(STANDALONE)
                      // Test dialog box in separate thread...
                      StartEditor();
#elif 1
    #ifndef STANDALONE
                      FM_is_busy++;
    #endif
                      DoAbout(globals, filterSelectorParameters);
    #ifndef STANDALONE
                      FM_is_busy--;
    #endif
#elif 1
                      char szPath[256];
                      int i;
                      DWORD ret = GetModuleFileName(hDllInstance, szPath, sizeof(szPath));
                      Info("Filter filename = %s (size = %d)", szPath, ret);
                      for (i = ret - 1; i >= 0 && szPath[i] != '\\'; --i)
                          ;
                      szPath[i+1] = '\0';
                      Info("Filter path = \"%s\"", szPath);
#elif 1
                      DWORD ret = (DWORD)ShellExecute(hDlg, "open", "help.html", NULL, NULL, SW_SHOWNORMAL);
                      if (ret <= 32) {
                          ErrorOk("ShellExecute failed: %d", ret);
                      }
#else
                      //For now, pop up the "About" box.
                      char message[256];
                      char caption[60];

                      LoadString(hDllInstance, 1, message, sizeof(message));
                      LoadString(hDllInstance, 2, caption, sizeof(caption));
                      msgBox(MB_APPLMODAL|MB_ICONINFORMATION|MB_OK, caption, message);
                      //MessageBox((HWND) (platform->hwnd),message,caption,MB_APPLMODAL|MB_ICONINFORMATION|MB_OK);
#endif
                  }
                  break;

#endif //ifndef STATIC_FILTER

#if 0
              case IDC_EMF:
              case 3986:
                      DoAbout(globals, filterSelectorParameters);
                      break;
#endif

              case IDC_BUTTON2: // -
				#ifdef HIGHZOOM 
					if (scaleFactor==1 && gFmcp->enlargeFactor>1)
						gFmcp->enlargeFactor -= 2;
					else
				#endif
						scaleFactor += 2;
                    // Fall through...
              case IDC_BUTTON1: // +
				#ifdef HIGHZOOM   
					if (scaleFactor==1 && gFmcp->enlargeFactor<16)
						gFmcp->enlargeFactor++;
					else
				#endif
						scaleFactor--;

                    if (GetKeyState(VK_SHIFT) < 0) {
                        //SHIFT modifies meaning of +/-
                        if (scaleFactor < prevScaleFactor) {
                            // Shift + means zoom 100 %
                            scaleFactor = 1;
							gFmcp->enlargeFactor = 1;
							
                        }
                        else {
							#ifdef HIGHZOOM 
							if (scaleFactor == 1 && gFmcp->enlargeFactor == 15){
								gFmcp->enlargeFactor = 1;
							} else if (scaleFactor == 1){
								gFmcp->enlargeFactor = 16;
							} else {
							#endif
								// Shift - means optimal zoom
								scaleFactor = -888; //magic # for SetupFilterRecordForProxy.
								gFmcp->enlargeFactor = 1;
							#ifdef HIGHZOOM 
							}
							#endif
                        }
                    }
                    else if (GetKeyState(VK_CONTROL) < 0) {
                        //CONTROL modifies meaning of +/-
                        if (scaleFactor < prevScaleFactor) {
                            // Ctrl + means zoom 100 %
                            scaleFactor = 1;
							gFmcp->enlargeFactor = 1;
                        }
                        else {
                            #ifdef HIGHZOOM 
							if (scaleFactor == 1 && gFmcp->enlargeFactor == 15){
								gFmcp->enlargeFactor = 1;
							} else if (scaleFactor == 1){
								gFmcp->enlargeFactor = 16;
							} else {
							#endif
								// Ctrl + means zoom to minimum size
								scaleFactor = MAX_SCALE_FACTOR;	// should be limited???
								gFmcp->enlargeFactor = 1;
							#ifdef HIGHZOOM
							}
							#endif
                        }
                    }
#if 0
                    else if (scaleFactor == 13 || scaleFactor == 15)
                    {
                        // Skip scalefactor 13 or 15...
                        if (scaleFactor > prevScaleFactor)
                            scaleFactor++;  //zooming out
                        else
                            scaleFactor--;  //zooming in
                    }
#else
					else {
                        // Skip undesirable scale factors...
						if (scaleFactor > prevScaleFactor) {
                            //zooming out
							scaleFactor = roundUpScaleFactor(scaleFactor);
						}
						else {
                            //zooming in
							scaleFactor = roundDownScaleFactor(scaleFactor);
						}
					}
#endif
                    GetProxyItemRect (hDlg, &gProxyRect); 
                    //Should really do this in WM_TIMER, once final scale factor is set???
                    SetupFilterRecordForProxy (globals, &scaleFactor, prevScaleFactor, &imageOffset);

					#ifdef HIGHZOOM
						EnableWindow(GetDlgItem(hDlg, IDC_BUTTON1), scaleFactor > 1 || gFmcp->enlargeFactor < 16 );
					#else
						EnableWindow(GetDlgItem(hDlg, IDC_BUTTON1), scaleFactor > 1);
					#endif
                    EnableWindow(GetDlgItem(hDlg, IDC_BUTTON2), scaleFactor < MAX_SCALE_FACTOR);
                    // Should switch input focus to other button when a button
                    // gets disabled ... otherwise, can't TAB away from disabled
                    // button using keyboard only.
                    #ifdef HIGHZOOM
						if (scaleFactor == 1 && gFmcp->enlargeFactor==16) SetFocus(GetDlgItem(hDlg, IDC_BUTTON2));
					#else
						if (scaleFactor == 1) SetFocus(GetDlgItem(hDlg, IDC_BUTTON2));
					#endif
                    if (scaleFactor == MAX_SCALE_FACTOR) SetFocus(GetDlgItem(hDlg, IDC_BUTTON1));

                    // Update Filter parameter block display
                    ShowFilterParameterBlock(globals, hDlg);

                    if (gResult == noErr) {
                        // Set a timer to trigger a delayed proxy update,
                        // giving the user a chance to click several times
                        // on the zoom button first...
                        KillTimer(hDlg, 1); //Cancel any previous trigger.
                        if (!SetTimer(hDlg, 1, 500, NULL)) {
                            // If SetTimer() fails, update the proxy immediately.
                            if (!fDragging) {
                                // Do Filtering operation
                                DoFilterRect (globals);
                                // Invalidate Proxy Item (and repaint background if scale factor
                                // increased; i.e., if proxy image may have shrunk)...
                                // (To eliminate MORE flashing, only repaint BG if shrunk, and
                                // view width or height is less than Proxy width or height.)
                                UpdateProxy(hDlg, scaleFactor > prevScaleFactor);
                            }
                        }

                    }//noErr

                
                    break;

              case  IDOK:
#ifndef STANDALONE


                    if (hEditorDlg) {
                        // Read and validate value of bandWidth...
                        char *pInvalid;

                        GetDlgItemText(hEditorDlg, nItem = 100, szInput, 32);
                        lSize = strtol(szInput, &pInvalid, 10);
                        if  ( ! *pInvalid && lSize >= 1 && lSize <= 1000)
                            gParams->bandWidth = lSize;
                        else {
                            if  (*pInvalid)
                                lSize  = 1000;

                            if  (lSize < 1)
                                SetDlgItemInt(hEditorDlg, nItem, 1, TRUE ); //signed
                            else
                                SetDlgItemInt(hEditorDlg, nItem, 1000, TRUE );//signed

                            SendDlgItemMessage(hEditorDlg, nItem, EM_SETSEL, 0, (LPARAM)-1);   // WIN32 Change
                            SetFocus(GetDlgItem(hEditorDlg, nItem));
                            MessageBeep (MB_ICONEXCLAMATION);
                            break;
                        }
                    }


                    // If source needs to be recompiled, ask user and recompile
                    // if requested...
                    // (Keep this message in sync with the same message in
                    // DoFilterRect().  But note that here we use yes/no/cancel,
                    // not just yes/no...)
                    if (!gParams->fValidCompile && gParams->fGotSourceCode) {
                        char szMessage[512];
                        // Get the message string...
                        // (e.g., "The compiled code is not up-to-date.\n\n"
                        //        "Would you like to compile it first?"
                        LoadString(hDllInstance, IDS_RECOMPILE, szMessage, sizeof szMessage);
                        switch (YesNoCancel(szMessage))
                        {
                        case IDYES:
                        case IDOK:
                            // (re-)codegen and init before applying filter...
                            SendMessage(hDlg, WM_COMMAND, IDC_INIT, 0);
                            // cancel if compilation fails...
                            if (!gParams->fValidCompile) return FALSE;
                            break;  //proceed after successful re-compile
                        case IDCANCEL:
                        case IDABORT:
                            return FALSE;   //not handled????
                            break;
                        case IDNO:
                        default:
                            break;  //proceed without re-compiling
                        }//switch
                    }
#endif //ifndef STANDALONE
                    // fall through...
              case  IDCANCEL:
	
					//Trigger cancel event on ESC key or x titlebar button, but not on OK or Cancel button
					if (idd == IDCANCEL && cancelEventPerformed==0 && gParams->fValidCompile){
						fm_triggerEvent(CTL_CANCEL, FME_CLICKED, 0);
						//cancelEventPerformed=1; //Do not set it to 1
					}

#ifndef STANDALONE
                    if (CheckForFileSave() == IDCANCEL) break;
#endif

                    if (fDragging)
                    {
                        ReleaseCapture();
                        //SetCursor(LoadCursor(NULL, IDC_ARROW));  //???
                        fDragging = FALSE;
                        gParams->flags &= ~AFH_DRAGGING_PROXY;
                    }
#ifndef STATIC_FILTER
#ifndef STANDALONE
#if 0
                    if (hEditorDlg) {
                        //Save the source code in our filter parameter block...
                        GetDlgItemText(hEditorDlg, IDC_EDIT2, gParams->szSrcCode,
                                       sizeof(gParams->szSrcCode));
                        //Should complain if too much text ????
                    }
#else
                    //Save the source code in our filter parameter block...
                    FmGetEditorText();
#endif
                    //gParams->fValidCompile = FALSE; //for now!!!
#endif //ifndef STANDALONE
#endif //ifndef STATIC_FILTER

                    // Preserve values in FM context record across invocations...
                    gParams->need_iuv = gFmcp->need_iuv;
                    gParams->need_d = gFmcp->need_d;
                    gParams->need_m = gFmcp->need_m;
                    gParams->need_tbuf = gFmcp->need_tbuf;
                    gParams->need_t2buf = gFmcp->need_t2buf;
					gParams->need_t3buf = gFmcp->need_t3buf;
					gParams->need_t4buf = gFmcp->need_t4buf;
                    gParams->needPremiere = gFmcp->needPremiere;
                    gParams->needPadding = gFmcp->needPadding;
                    gParams->bandWidth = gFmcp->bandWidth;
#ifndef DONT_SAVE_GLOBALS
                    /* Save the following predeclared global variables */
                    //// NO!!! Should be done AFTER the filter has run on the host image!!!!!
                    gParams->globvars = gFmcp->globvars;
#endif //DONT_SAVE_GLOBALS
                    // !!!Too bad C doesn't allow direct array assignments, e.g.:
                    ///// gParams->filterAuthorText = gFmcp->filterAuthorText;
                    memcpy(gParams->filterAuthorText, gFmcp->filterAuthorText, sizeof(gParams->filterAuthorText));
                    memcpy(gParams->filterCategoryText, gFmcp->filterCategoryText, sizeof(gParams->filterCategoryText));
                    memcpy(gParams->filterTitleText, gFmcp->filterTitleText, sizeof(gParams->filterTitleText));
                    memcpy(gParams->filterFilenameText, gFmcp->filterFilenameText, sizeof(gParams->filterFilenameText));
                    memcpy(gParams->filterCopyrightText, gFmcp->filterCopyrightText, sizeof(gParams->filterCopyrightText));
                    memcpy(gParams->filterDescriptionText, gFmcp->filterDescriptionText, sizeof(gParams->filterDescriptionText));
                    memcpy(gParams->filterOrganizationText, gFmcp->filterOrganizationText, sizeof(gParams->filterOrganizationText));
                    memcpy(gParams->filterURLText, gFmcp->filterURLText, sizeof(gParams->filterURLText));
                    memcpy(gParams->filterVersionText, gFmcp->filterVersionText, sizeof(gParams->filterVersionText));
                    memcpy(gParams->filterUniqueID, gFmcp->filterUniqueID, sizeof(gParams->filterUniqueID));
                    gParams->ndlits = gFmcp->ndlits;
                    memcpy(gParams->dlitPool, gFmcp->dlitPool, sizeof(gParams->dlitPool));

#ifndef STANDALONE
                    //save CB's across invocations, for use by Make...
                    //Must not have any "relocatable" pointers in the CB's !!!
                    //(pointers into the param block itself are okay???)
                    //Maybe it's best to always force a recompile when reinvoked,
                    //instead of trying to preserve code across invocations?
                    //Otherwise, may need to attempt to relocate the code pointers
                    //in case the address of gParams changed???  Or store the code
                    //buffer pointers as self-relative to gParams?
                    gParams->cbCodeBuffer = cbCodeBuffer;
                    gParams->cbInitCode = cbInitCode;
                    gParams->cbDivCheck = cbDivCheck;
                    gParams->cbZeroDivide = cbZeroDivide;
                    gParams->cbDivideOverflow = cbDivideOverflow;
                    gParams->cbOnCtl = cbOnCtl;
                    gParams->cbOnFilterStart = cbOnFilterStart;
                    gParams->cbForEveryTile = cbForEveryTile;
                    gParams->cbOnFilterEnd = cbOnFilterEnd;
                    gParams->cbUserFunctions = cbUserFunctions;
#endif

                    //save current position and size of dialog
                    GetWindowRect(hDlg, &gParams->gDialogRect);

                    //save current dialog styles
                    gParams->gDialogStyle = GetWindowLong(hDlg, GWL_STYLE);
                    gParams->gDialogStyleEx = GetWindowLong(hDlg, GWL_EXSTYLE);
                    gParams->gInitialInvocation = FALSE;

                    //save current dialog clipping region, if any
                    // create a dummy region first...
                    gParams->gDialogRegion = CreateRectRgn(0, 0, 0, 0);
                    if (GetWindowRgn(hDlg, gParams->gDialogRegion) == ERROR) {
                        //no clipping region
                        DeleteObject(gParams->gDialogRegion);
                        gParams->gDialogRegion = NULL;
                    }

#ifndef STANDALONE
                    // Close editor dialog if it's active
                    if (hEditorDlg) {
                        LRESULT iss;
                        //From the Platform SDK docs:
                        // A thread cannot use DestroyWindow to destroy a window
                        // created by a different thread. 
#if 0
                        DestroyWindow(hEditorDlg);
#else
                        iss = SendMessage(hEditorDlg, WM_COMMAND, IDCANCEL, 0);
#endif
                        hEditorDlg = 0;
                    }
#endif //STANDALONE

                    gFmcp->hDlg = NULL;    //So msgBox et al. will use global handle.
                    MyHDlg = NULL;          //ditto
                    EndDialog(hDlg, idd);          // WIN32 change
                    break;

              default:
                    return FALSE;   // not handled
            }
            break;

        default:
            return  FALSE;  // not handled
    }

    return  TRUE;   // handled; do not call defProc
} /*FilterDialogProc*/

/*****************************************************************************/




//function added by Harald Heim on Apr 2, 2002
int setZoom (HWND hDlg, int n, int m)
{
    
    static GPtr globals = NULL;         /* needs to be static */    // EH????
    long   prevScaleFactor = scaleFactor;
	//int prevEnlargeFactor = gFmcp->enlargeFactor;
    //////static POINT imageOffset = {0, 0};      // why local ???

    scaleFactor = n;
	gFmcp->enlargeFactor = m;

    if (scaleFactor != prevScaleFactor){ //&& gFmcp->enlargeFactor != prevEnlargeFactor
    
        GetProxyItemRect (hDlg, &gProxyRect); 
        SetupFilterRecordForProxy (globals, &scaleFactor, prevScaleFactor, &imageOffset);

		#ifdef HIGHZOOM
			EnableWindow(GetDlgItem(hDlg, IDC_BUTTON1), scaleFactor > 1 || gFmcp->enlargeFactor < 16 );
		#else
			EnableWindow(GetDlgItem(hDlg, IDC_BUTTON1), scaleFactor > 1);
		#endif
        EnableWindow(GetDlgItem(hDlg, IDC_BUTTON2), scaleFactor < MAX_SCALE_FACTOR);
        #ifdef HIGHZOOM
			if (scaleFactor == 1 && gFmcp->enlargeFactor==16) SetFocus(GetDlgItem(hDlg, IDC_BUTTON2));
		#else
			if (scaleFactor == 1) SetFocus(GetDlgItem(hDlg, IDC_BUTTON2));
		#endif
        if (scaleFactor == MAX_SCALE_FACTOR) SetFocus(GetDlgItem(hDlg, IDC_BUTTON1));

        // Update Filter parameter block display to update the proxy scale factor...
        ShowFilterParameterBlock(globals, hDlg);

        if (!fDragging) {
            DoFilterRect (globals);
            UpdateProxy(hDlg, scaleFactor > prevScaleFactor);
        }
        
        return true;
    }

    return false;
}



//function added by Harald Heim on Dec 14, 2002
INT_PTR scrollPreview (HWND hDlg, INT_PTR mode, INT_PTR offsetX, INT_PTR offsetY)
{

    GPtr globals = NULL;        // EH???

    if (mode==0){
        imageOffset.x += (int)offsetX * scaleFactor;
        imageOffset.y += (int)offsetY * scaleFactor;
    } else if (mode==1){
        imageOffset.x = (int)offsetX * scaleFactor;
        imageOffset.y = (int)offsetY * scaleFactor;
    }

    GetProxyItemRect (hDlg, &gProxyRect);
    SetupFilterRecordForProxy (globals, &scaleFactor, scaleFactor, &imageOffset ); //&tempOffset

    if (gResult == noErr) {
        // Run the filter only if want to drag the filtered output
        // rather than the raw input...
        if (gParams->flags & AFH_DRAG_FILTERED_OUTPUT)
        {
            DoFilterRect (globals);
        }
        UpdateProxy(hDlg, FALSE);
    }//noErr

    return 1;
}



//function added by Harald Heim on January, 2003
INT_PTR setPreviewCursor (INT_PTR res)
{
    HCURSOR hcHand_old = hcHand;

	cursorResource = res;

    // How about loading a named resource????

    //Standard Windows Cursors
    if (res >= 32512)
        hcHand = LoadCursor(NULL,  MAKEINTRESOURCE(res));
    //Load from Resource
    else
        hcHand = LoadCursor(hDllInstance, MAKEINTRESOURCE(res));

    if (hcHand != NULL){
            DestroyCursor(hcHand_old);
            SetCursor(hcHand);
    }


    return (INT_PTR)hcHand;

}

INT_PTR getPreviewCursor (void)
{

	return cursorResource;

}


//#ifndef STANDALONE  -- who is calling this in STANDALONE mode???


void FMprint(const char *buff)
{
#if 0
	//we shouldn't be called if no Edit1 control...
    assert(hEdit1);
#else
	//but someone is calling us from a standalone!!
	if (!hEdit1) return; //get out quick!!
#endif

#if 1	//extra overhead okay??
    // May not be necessary to set the selection to the end of
    // text if that's where it defaults to anyway...
	{//scope
		int ichEnd = Edit_GetTextLength(hEdit1);
		Edit_SetSel(hEdit1, ichEnd, ichEnd);
	}
#endif
    Edit_ReplaceSel(hEdit1, buff);
} /*FMprint*/

//#endif //ifndef STANDALONE

/*****
Similarly, by default the TAB key moves the input focus to the next control in the tabbing sequence. 
However, it might be useful for the user to be able to enter TAB characters in an edit control. By 
default, multiple-line edit controls accept CTRL+TAB to enter TAB characters in an edit control.
The following sections describe four methods that can be used to implement alternative behavior for 
the ENTER and TAB keys. It should be noted that some users may object if the ENTER and TAB keys no 
longer act in the default fashion.

Use ES_WANTRETURN in Windows version 3.1

Windows version 3.1 supplies a new edit control style, ES_WANTRETURN, which causes an ENTER key to 
advance to a new line and not be passed to the dialog manager. ES_WANTRETURN does not affect TAB keys.

Subclass the edit control

An edit control subclass procedure could respond to the WM_GETDLGCODE message with DLGC_WANTALLKEYS 
to receive all keyboard input. This causes the edit control to be sent all key input, including ENTER 
and TAB keys. The following subclass procedure could be used:
  
LONG FAR PASCAL AllKeyFunc(HWND hWnd, WORD Message, WORD wParam,
   LONG lParam)
{
   //
   // This subclass function makes pressing an ENTER key
   // create a new line when the focus is in an edit control.
   //

   if ( Message == WM_GETDLGCODE )
      return DLGC_WANTALLKEYS;

   return CallWindowProc(lpfnOldClassProc, hWnd, Message, wParam,
      lParam);
}

  ****/

/****
FOLLOWING SHOWS HOW TO SUBCLASS:

Here is how to make an edit control read-only in Windows version 3.0:
  
// Declare the subclass function.
long FAR PASCAL SubEditWndProc(HWND hWnd, unsigned wMessage,
                WORD wParam, LONG lParam);

//
// Create an edit control.
//
hWndEdit=CreateWindow("EDIT",
                  "Edit Control Line Numbers",
                  WS_OVERLAPPEDWINDOW,
                  35, 35, 400, 150,
                  NULL, NULL, hInstance, NULL);
//
// Subclass the edit control.
//
static FARPROC     lpSubClassProc;

lpSubClassProc=MakeProcInstance((FARPROC)SubEditWndProc,hInstance);
     SetWindowLong(hWndEdit, GWL_WNDPROC, (LONG)lpSubClassProc);
.
.
.

//
// Read-only subclass function for an edit control.
//
long FAR PASCAL SubEditWndProc(HWND hWnd, unsigned wMessage, WORD
   wParam, LONG lParam)
{
   switch ( wMessage ) {
      case WM_CUT:
      case WM_KEYDOWN:
      case WM_PASTE:
      case WM_CHAR:
         // Do nothing here.
      break;

      default:
         CallWindowProc (lpPrevWndFunc, hWnd, wMessage, wParam,
                         lParam );
   }
}

****/

/*****************************************************************************/
void PromptUserForInput (GPtr globals)
{

    INT_PTR    nResult;
	//HWND hDlg;
	//MSG msg;

    PlatformData *platform;

    platform = ((FilterRecordPtr) gStuff)->platformData;

    /* Query the user for parameters. */

	nResult = DialogBoxParam(
	//hDlg = CreateDialogParam(
		               hDllInstance,
                       (LPSTR)"FILTERPARAM",
                       (HWND)platform->hwnd,
                       (DLGPROC)FilterDialogProc,
                        (LPARAM)globals);

	
	/*
	ShowWindow(hDlg, SW_SHOW);

	while (GetMessage(&msg, NULL, 0, 0)) {
	    //if (IsDialogMessage(hDlg, &msg)) {
        //    continue;
        //}
		TranslateMessage(&msg);
        DispatchMessage(&msg);
    }//while

	if (hDlg) {
        DestroyWindow(hDlg);
        hDlg = 0;
    }
	*/

    if  (nResult == IDCANCEL || nResult == -1)
        gResult = 1; // don't continue

}   

/*****************************************************************************/
void PromptUserForInputNoProxy (GPtr globals)
{

    INT_PTR    nResult;
    PlatformData *platform;

    platform = ((FilterRecordPtr) gStuff)->platformData;

    /* Query the user for parameters. */

    nResult = DialogBoxParam(
		               hDllInstance,
                       (LPSTR)"FILTERPARAM",
                       (HWND)platform->hwnd,
                       (DLGPROC)FilterDialogProc,
                        (LPARAM)globals);
}   

/****************************************************************************/
/* Example for ShowAlert() function which takes a string ID as parameter    */
/* and displays a message box                                               */
/****************************************************************************/

short ShowAlert (short stringID)
{
    char szMessage[256];
    char szTitle[128];

    LoadString(hDllInstance, stringID, szMessage, sizeof szMessage);
    LoadString(hDllInstance, 2, szTitle, sizeof szTitle);
    return  MessageBox(MyHDlg, szMessage, szTitle, MB_OK | MB_ICONHAND | MB_TASKMODAL);

}

/*****************************************************************************/

#ifndef NOTRACE
void Trace (GPtr globals, const char *szFormat, ...)
{
    char szBuffer[1024];
    char *pArgs;
    int nResult=0; //Added to fix compiler error
//#if 0
    if ((gParams->flags & (AFH_TRACE|AFH_DRAGGING_PROXY)) == AFH_TRACE)
//#endif
    {   // Trace if Trace flag set but not while dragging proxy...

    if  (nResult == IDCANCEL || nResult == -1)
        gResult = 1; // don't continue

        pArgs = (char *) &szFormat + sizeof(szFormat);
        wvsprintf(szBuffer, szFormat, pArgs);
        MessageBox(MyHDlg, szBuffer, "Filter trace", MB_OK | MB_TASKMODAL);
    }
} /*Trace*/
#endif



/*****************************************************************************/

// 21Feb2014:
// _MT mods (for statically linking the CRT) kindly provided by Harald Heim.

/* Initialization and termination code for window's dlls. */
// Every 32-Bit DLL has an entry point DLLInit (It is actually called DllMain()...)
 
#if defined(_MT) && !defined(_DLL)
    //Needed for static linking of the C Runtime
    BOOL WINAPI _CRT_INIT(HANDLE, DWORD, LPVOID);
#endif
 
BOOL APIENTRY DLLInit(HANDLE hInstance, DWORD fdwReason, LPVOID lpReserved)
{
 
    if (fdwReason == DLL_PROCESS_ATTACH)
        hDllInstance = hInstance;
   
#if defined(_MT) && !defined(_DLL)
    //Needed for static linking of the C Runtime
    if (fdwReason == DLL_PROCESS_ATTACH || fdwReason == DLL_THREAD_ATTACH || fdwReason == DLL_PROCESS_DETACH || fdwReason == DLL_THREAD_DETACH){
        if (!_CRT_INIT(hInstance, fdwReason, lpReserved)) return(FALSE);
    }
#endif
 
    return TRUE;   // Indicate that the DLL was initialized successfully.
}
 
/*****************************************************************************/
